/* This file was generated by build/generator/gen_base.py */

static struct {
  int errcode;
  const char *errname;
} svn__errno[] = {
  { 0, "success" }
};

static struct {
  int errcode;
  const char *errname;
} svn__apr_errno[] = {
  { 0, "APR_SUCCESS" },
  { 10000, "SOCBASEERR" },
  { 10001, "SOCEPERM" },
  { 10003, "SOCESRCH" },
  { 10004, "SOCEINTR" },
  { 10006, "SOCENXIO" },
  { 10009, "SOCEBADF" },
  { 10013, "SOCEACCES" },
  { 10014, "SOCEFAULT" },
  { 10022, "SOCEINVAL" },
  { 10024, "SOCEMFILE" },
  { 10032, "SOCEPIPE" },
  { 10035, "SOCEWOULDBLOCK" },
  { 10036, "SOCEINPROGRESS" },
  { 10037, "SOCEALREADY" },
  { 10038, "SOCENOTSOCK" },
  { 10039, "SOCEDESTADDRREQ" },
  { 10040, "SOCEMSGSIZE" },
  { 10041, "SOCEPROTOTYPE" },
  { 10042, "SOCENOPROTOOPT" },
  { 10043, "SOCEPROTONOSUPPORT" },
  { 10044, "SOCESOCKTNOSUPPORT" },
  { 10045, "SOCEOPNOTSUPP" },
  { 10046, "SOCEPFNOSUPPORT" },
  { 10047, "SOCEAFNOSUPPORT" },
  { 10048, "SOCEADDRINUSE" },
  { 10049, "SOCEADDRNOTAVAIL" },
  { 10050, "SOCENETDOWN" },
  { 10051, "SOCENETUNREACH" },
  { 10052, "SOCENETRESET" },
  { 10053, "SOCECONNABORTED" },
  { 10054, "SOCECONNRESET" },
  { 10055, "SOCENOBUFS" },
  { 10056, "SOCEISCONN" },
  { 10057, "SOCENOTCONN" },
  { 10058, "SOCESHUTDOWN" },
  { 10059, "SOCETOOMANYREFS" },
  { 10060, "SOCETIMEDOUT" },
  { 10061, "SOCECONNREFUSED" },
  { 10062, "SOCELOOP" },
  { 10063, "SOCENAMETOOLONG" },
  { 10064, "SOCEHOSTDOWN" },
  { 10065, "SOCEHOSTUNREACH" },
  { 10066, "SOCENOTEMPTY" },
  { 20000, "APR_OS_START_ERROR" },
  { 20000, "APR_UTIL_ERRSPACE_SIZE" },
  { 20001, "APR_ENOSTAT" },
  { 20002, "APR_ENOPOOL" },
  { 20004, "APR_EBADDATE" },
  { 20005, "APR_EINVALSOCK" },
  { 20006, "APR_ENOPROC" },
  { 20007, "APR_ENOTIME" },
  { 20008, "APR_ENODIR" },
  { 20009, "APR_ENOLOCK" },
  { 20010, "APR_ENOPOLL" },
  { 20011, "APR_ENOSOCKET" },
  { 20012, "APR_ENOTHREAD" },
  { 20013, "APR_ENOTHDKEY" },
  { 20014, "APR_EGENERAL" },
  { 20015, "APR_ENOSHMAVAIL" },
  { 20016, "APR_EBADIP" },
  { 20017, "APR_EBADMASK" },
  { 20019, "APR_EDSOOPEN" },
  { 20020, "APR_EABSOLUTE" },
  { 20021, "APR_ERELATIVE" },
  { 20022, "APR_EINCOMPLETE" },
  { 20023, "APR_EABOVEROOT" },
  { 20024, "APR_EBADPATH" },
  { 20025, "APR_EPATHWILD" },
  { 20026, "APR_ESYMNOTFOUND" },
  { 20027, "APR_EPROC_UNKNOWN" },
  { 20028, "APR_ENOTENOUGHENTROPY" },
  { 50000, "APR_OS_ERRSPACE_SIZE" },
  { 70000, "APR_OS_START_STATUS" },
  { 70001, "APR_INCHILD" },
  { 70002, "APR_INPARENT" },
  { 70003, "APR_DETACH" },
  { 70004, "APR_NOTDETACH" },
  { 70005, "APR_CHILD_DONE" },
  { 70006, "APR_CHILD_NOTDONE" },
  { 70007, "APR_TIMEUP" },
  { 70008, "APR_INCOMPLETE" },
  { 70012, "APR_BADCH" },
  { 70013, "APR_BADARG" },
  { 70014, "APR_EOF" },
  { 70015, "APR_NOTFOUND" },
  { 70019, "APR_ANONYMOUS" },
  { 70020, "APR_FILEBASED" },
  { 70021, "APR_KEYBASED" },
  { 70022, "APR_EINIT" },
  { 70023, "APR_ENOTIMPL" },
  { 70024, "APR_EMISMATCH" },
  { 70025, "APR_EBUSY" },
  { 100000, "APR_UTIL_START_STATUS" },
  { 100001, "APR_ENOKEY" },
  { 100002, "APR_ENOIV" },
  { 100003, "APR_EKEYTYPE" },
  { 100004, "APR_ENOSPACE" },
  { 100005, "APR_ECRYPT" },
  { 100006, "APR_EPADDING" },
  { 100007, "APR_EKEYLENGTH" },
  { 100008, "APR_ENOCIPHER" },
  { 100009, "APR_ENODIGEST" },
  { 100010, "APR_ENOENGINE" },
  { 100011, "APR_EINITENGINE" },
  { 100012, "APR_EREINIT" },
  { 120000, "APR_OS_START_USEERR" },
  { 120000, "APR_OS_START_USERERR" },
  { 620000, "APR_OS_START_CANONERR" },
  { 620001, "APR_EACCES" },
  { 620002, "APR_EEXIST" },
  { 620003, "APR_ENAMETOOLONG" },
  { 620004, "APR_ENOENT" },
  { 620005, "APR_ENOTDIR" },
  { 620006, "APR_ENOSPC" },
  { 620007, "APR_ENOMEM" },
  { 620008, "APR_EMFILE" },
  { 620009, "APR_ENFILE" },
  { 620010, "APR_EBADF" },
  { 620011, "APR_EINVAL" },
  { 620012, "APR_ESPIPE" },
  { 620013, "APR_EAGAIN" },
  { 620014, "APR_EINTR" },
  { 620015, "APR_ENOTSOCK" },
  { 620016, "APR_ECONNREFUSED" },
  { 620017, "APR_EINPROGRESS" },
  { 620018, "APR_ECONNABORTED" },
  { 620019, "APR_ECONNRESET" },
  { 620020, "APR_ETIMEDOUT" },
  { 620021, "APR_EHOSTUNREACH" },
  { 620022, "APR_ENETUNREACH" },
  { 620023, "APR_EFTYPE" },
  { 620024, "APR_EPIPE" },
  { 620025, "APR_EXDEV" },
  { 620026, "APR_ENOTEMPTY" },
  { 620027, "APR_EAFNOSUPPORT" },
  { 670000, "APR_OS_START_EAIERR" },
  { 720000, "APR_OS_START_SYSERR" }
};
