/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 3.0.12
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */


#ifndef SWIGRUBY
#define SWIGRUBY
#endif


/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13



#include <ruby.h>

/* Ruby 1.9.1 has a "memoisation optimisation" when compiling with GCC which
 * breaks using rb_intern as an lvalue, as SWIG does.  We work around this
 * issue for now by disabling this.
 * https://sourceforge.net/tracker/?func=detail&aid=2859614&group_id=1645&atid=101645
 */
#ifdef rb_intern
# undef rb_intern
#endif

/* Remove global macros defined in Ruby's win32.h */
#ifdef write
# undef write
#endif
#ifdef read
# undef read
#endif
#ifdef bind
# undef bind
#endif
#ifdef close
# undef close
#endif
#ifdef connect
# undef connect
#endif


/* Ruby 1.7 defines NUM2LL(), LL2NUM() and ULL2NUM() macros */
#ifndef NUM2LL
#define NUM2LL(x) NUM2LONG((x))
#endif
#ifndef LL2NUM
#define LL2NUM(x) INT2NUM((long) (x))
#endif
#ifndef ULL2NUM
#define ULL2NUM(x) UINT2NUM((unsigned long) (x))
#endif

/* Ruby 1.7 doesn't (yet) define NUM2ULL() */
#ifndef NUM2ULL
#ifdef HAVE_LONG_LONG
#define NUM2ULL(x) rb_num2ull((x))
#else
#define NUM2ULL(x) NUM2ULONG(x)
#endif
#endif

/* RSTRING_LEN, etc are new in Ruby 1.9, but ->ptr and ->len no longer work */
/* Define these for older versions so we can just write code the new way */
#ifndef RSTRING_LEN
# define RSTRING_LEN(x) RSTRING(x)->len
#endif
#ifndef RSTRING_PTR
# define RSTRING_PTR(x) RSTRING(x)->ptr
#endif
#ifndef RSTRING_END
# define RSTRING_END(x) (RSTRING_PTR(x) + RSTRING_LEN(x))
#endif
#ifndef RARRAY_LEN
# define RARRAY_LEN(x) RARRAY(x)->len
#endif
#ifndef RARRAY_PTR
# define RARRAY_PTR(x) RARRAY(x)->ptr
#endif
#ifndef RFLOAT_VALUE
# define RFLOAT_VALUE(x) RFLOAT(x)->value
#endif
#ifndef DOUBLE2NUM
# define DOUBLE2NUM(x) rb_float_new(x)
#endif
#ifndef RHASH_TBL
# define RHASH_TBL(x) (RHASH(x)->tbl)
#endif
#ifndef RHASH_ITER_LEV
# define RHASH_ITER_LEV(x) (RHASH(x)->iter_lev)
#endif
#ifndef RHASH_IFNONE
# define RHASH_IFNONE(x) (RHASH(x)->ifnone)
#endif
#ifndef RHASH_SIZE
# define RHASH_SIZE(x) (RHASH(x)->tbl->num_entries)
#endif
#ifndef RHASH_EMPTY_P
# define RHASH_EMPTY_P(x) (RHASH_SIZE(x) == 0)
#endif
#ifndef RSTRUCT_LEN
# define RSTRUCT_LEN(x) RSTRUCT(x)->len
#endif
#ifndef RSTRUCT_PTR
# define RSTRUCT_PTR(x) RSTRUCT(x)->ptr
#endif



/*
 * Need to be very careful about how these macros are defined, especially
 * when compiling C++ code or C code with an ANSI C compiler.
 *
 * VALUEFUNC(f) is a macro used to typecast a C function that implements
 * a Ruby method so that it can be passed as an argument to API functions
 * like rb_define_method() and rb_define_singleton_method().
 *
 * VOIDFUNC(f) is a macro used to typecast a C function that implements
 * either the "mark" or "free" stuff for a Ruby Data object, so that it
 * can be passed as an argument to API functions like Data_Wrap_Struct()
 * and Data_Make_Struct().
 */
 
#ifdef __cplusplus
#  ifndef RUBY_METHOD_FUNC /* These definitions should work for Ruby 1.4.6 */
#    define PROTECTFUNC(f) ((VALUE (*)()) f)
#    define VALUEFUNC(f) ((VALUE (*)()) f)
#    define VOIDFUNC(f)  ((void (*)()) f)
#  else
#    ifndef ANYARGS /* These definitions should work for Ruby 1.6 */
#      define PROTECTFUNC(f) ((VALUE (*)()) f)
#      define VALUEFUNC(f) ((VALUE (*)()) f)
#      define VOIDFUNC(f)  ((RUBY_DATA_FUNC) f)
#    else /* These definitions should work for Ruby 1.7+ */
#      define PROTECTFUNC(f) ((VALUE (*)(VALUE)) f)
#      define VALUEFUNC(f) ((VALUE (*)(ANYARGS)) f)
#      define VOIDFUNC(f)  ((RUBY_DATA_FUNC) f)
#    endif
#  endif
#else
#  define VALUEFUNC(f) (f)
#  define VOIDFUNC(f) (f)
#endif

/* Don't use for expressions have side effect */
#ifndef RB_STRING_VALUE
#define RB_STRING_VALUE(s) (TYPE(s) == T_STRING ? (s) : (*(volatile VALUE *)&(s) = rb_str_to_str(s)))
#endif
#ifndef StringValue
#define StringValue(s) RB_STRING_VALUE(s)
#endif
#ifndef StringValuePtr
#define StringValuePtr(s) RSTRING_PTR(RB_STRING_VALUE(s))
#endif
#ifndef StringValueLen
#define StringValueLen(s) RSTRING_LEN(RB_STRING_VALUE(s))
#endif
#ifndef SafeStringValue
#define SafeStringValue(v) do {\
    StringValue(v);\
    rb_check_safe_str(v);\
} while (0)
#endif

#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
#define rb_define_alloc_func(klass, func) rb_define_singleton_method((klass), "new", VALUEFUNC((func)), -1)
#define rb_undef_alloc_func(klass) rb_undef_method(CLASS_OF((klass)), "new")
#endif

static VALUE _mSWIG = Qnil;

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */


/* Define some additional error types */
#define SWIG_ObjectPreviouslyDeletedError  -100


/* Define custom exceptions for errors that do not map to existing Ruby
   exceptions.  Note this only works for C++ since a global cannot be
   initialized by a function in C.  For C, fallback to rb_eRuntimeError.*/

SWIGINTERN VALUE 
getNullReferenceError(void) {
  static int init = 0;
  static VALUE rb_eNullReferenceError ;
  if (!init) {
    init = 1;
    rb_eNullReferenceError = rb_define_class("NullReferenceError", rb_eRuntimeError);
  }
  return rb_eNullReferenceError;
} 

SWIGINTERN VALUE 
getObjectPreviouslyDeletedError(void) {
  static int init = 0;
  static VALUE rb_eObjectPreviouslyDeleted ;
  if (!init) {
    init = 1;
    rb_eObjectPreviouslyDeleted = rb_define_class("ObjectPreviouslyDeleted", rb_eRuntimeError);
  }
  return rb_eObjectPreviouslyDeleted;
} 


SWIGINTERN VALUE
SWIG_Ruby_ErrorType(int SWIG_code) {
  VALUE type;
  switch (SWIG_code) {
  case SWIG_MemoryError:
    type = rb_eNoMemError;
    break;
  case SWIG_IOError:
    type = rb_eIOError;
    break;
  case SWIG_RuntimeError:
    type = rb_eRuntimeError;
    break;
  case SWIG_IndexError:
    type = rb_eIndexError;
    break;
  case SWIG_TypeError:
    type = rb_eTypeError;
    break;
  case SWIG_DivisionByZero:
    type = rb_eZeroDivError;
    break;
  case SWIG_OverflowError:
    type = rb_eRangeError;
    break;
  case SWIG_SyntaxError:
    type = rb_eSyntaxError;
    break;
  case SWIG_ValueError:
    type = rb_eArgError;
    break;
  case SWIG_SystemError:
    type = rb_eFatal;
    break;
  case SWIG_AttributeError:
    type = rb_eRuntimeError;
    break;
  case SWIG_NullReferenceError:
    type = getNullReferenceError();
    break;
  case SWIG_ObjectPreviouslyDeletedError:
    type = getObjectPreviouslyDeletedError();
    break;
  case SWIG_UnknownError:
    type = rb_eRuntimeError;
    break;
  default:
    type = rb_eRuntimeError;
  }
  return type;
}


/* This function is called when a user inputs a wrong argument to
   a method.
 */
SWIGINTERN 
const char* Ruby_Format_TypeError( const char* msg,
				   const char* type, 
				   const char* name, 
				   const int argn,
				   VALUE input )
{
  char buf[128];
  VALUE str;
  VALUE asStr;
  if ( msg && *msg )
    {
      str = rb_str_new2(msg);
    }
  else
    {
      str = rb_str_new(NULL, 0);
    }

  str = rb_str_cat2( str, "Expected argument " );
  sprintf( buf, "%d of type ", argn-1 );
  str = rb_str_cat2( str, buf );
  str = rb_str_cat2( str, type );
  str = rb_str_cat2( str, ", but got " );
  str = rb_str_cat2( str, rb_obj_classname(input) );
  str = rb_str_cat2( str, " " );
  asStr = rb_inspect(input);
  if ( RSTRING_LEN(asStr) > 30 )
    {
      str = rb_str_cat( str, StringValuePtr(asStr), 30 );
      str = rb_str_cat2( str, "..." );
    }
  else
    {
      str = rb_str_append( str, asStr );
    }

  if ( name )
    {
      str = rb_str_cat2( str, "\n\tin SWIG method '" );
      str = rb_str_cat2( str, name );
      str = rb_str_cat2( str, "'" );
    }

  return StringValuePtr( str );
}

/* This function is called when an overloaded method fails */
SWIGINTERN 
void Ruby_Format_OverloadedError(
				 const int argc,
				 const int maxargs,
				 const char* method, 
				 const char* prototypes 
				 )
{
  const char* msg = "Wrong # of arguments";
  if ( argc <= maxargs ) msg = "Wrong arguments";
  rb_raise(rb_eArgError,"%s for overloaded method '%s'.\n"  
	   "Possible C/C++ prototypes are:\n%s",
	   msg, method, prototypes);
}

/* -----------------------------------------------------------------------------
 * rubytracking.swg
 *
 * This file contains support for tracking mappings from 
 * Ruby objects to C++ objects.  This functionality is needed
 * to implement mark functions for Ruby's mark and sweep
 * garbage collector.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(ST_DATA_T_DEFINED)
/* Needs to be explicitly included for Ruby 1.8 and earlier */
#include <st.h>
#endif

/* Ruby 1.8 actually assumes the first case. */
#if SIZEOF_VOIDP == SIZEOF_LONG
#  define SWIG2NUM(v) LONG2NUM((unsigned long)v)
#  define NUM2SWIG(x) (unsigned long)NUM2LONG(x)
#elif SIZEOF_VOIDP == SIZEOF_LONG_LONG
#  define SWIG2NUM(v) LL2NUM((unsigned long long)v)
#  define NUM2SWIG(x) (unsigned long long)NUM2LL(x)
#else
#  error sizeof(void*) is not the same as long or long long
#endif

/* Global hash table to store Trackings from C/C++
   structs to Ruby Objects. 
*/
static st_table* swig_ruby_trackings = NULL;

static VALUE swig_ruby_trackings_count(ANYARGS) {
  return SWIG2NUM(swig_ruby_trackings->num_entries);
}


/* Setup a hash table to store Trackings */
SWIGRUNTIME void SWIG_RubyInitializeTrackings(void) {
  /* Create a hash table to store Trackings from C++
     objects to Ruby objects. */

  /* Try to see if some other .so has already created a 
     tracking hash table, which we keep hidden in an instance var
     in the SWIG module.
     This is done to allow multiple DSOs to share the same
     tracking table.
  */
  VALUE trackings_value = Qnil;
  /* change the variable name so that we can mix modules
     compiled with older SWIG's - this used to be called "@__trackings__" */
  ID trackings_id = rb_intern( "@__safetrackings__" );
  VALUE verbose = rb_gv_get("VERBOSE");
  rb_gv_set("VERBOSE", Qfalse);
  trackings_value = rb_ivar_get( _mSWIG, trackings_id );
  rb_gv_set("VERBOSE", verbose);

  /* The trick here is that we have to store the hash table
  pointer in a Ruby variable. We do not want Ruby's GC to
  treat this pointer as a Ruby object, so we convert it to
  a Ruby numeric value. */
  if (trackings_value == Qnil) {
    /* No, it hasn't.  Create one ourselves */
    swig_ruby_trackings = st_init_numtable();
    rb_ivar_set( _mSWIG, trackings_id, SWIG2NUM(swig_ruby_trackings) );
  } else {
    swig_ruby_trackings = (st_table*)NUM2SWIG(trackings_value);
  }

  rb_define_virtual_variable("SWIG_TRACKINGS_COUNT", swig_ruby_trackings_count, NULL);
}

/* Add a Tracking from a C/C++ struct to a Ruby object */
SWIGRUNTIME void SWIG_RubyAddTracking(void* ptr, VALUE object) {
  /* Store the mapping to the global hash table. */
  st_insert(swig_ruby_trackings, (st_data_t)ptr, object);
}

/* Get the Ruby object that owns the specified C/C++ struct */
SWIGRUNTIME VALUE SWIG_RubyInstanceFor(void* ptr) {
  /* Now lookup the value stored in the global hash table */
  VALUE value;

  if (st_lookup(swig_ruby_trackings, (st_data_t)ptr, &value)) {
    return value;
  } else {
    return Qnil;
  }
}

/* Remove a Tracking from a C/C++ struct to a Ruby object.  It
   is very important to remove objects once they are destroyed
   since the same memory address may be reused later to create
   a new object. */
SWIGRUNTIME void SWIG_RubyRemoveTracking(void* ptr) {
  /* Delete the object from the hash table */
  st_delete(swig_ruby_trackings, (st_data_t *)&ptr, NULL);
}

/* This is a helper method that unlinks a Ruby object from its
   underlying C++ object.  This is needed if the lifetime of the
   Ruby object is longer than the C++ object */
SWIGRUNTIME void SWIG_RubyUnlinkObjects(void* ptr) {
  VALUE object = SWIG_RubyInstanceFor(ptr);

  if (object != Qnil) {
    if (TYPE(object) != T_DATA)
      abort();
    DATA_PTR(object) = 0;
  }
}

/* This is a helper method that iterates over all the trackings
   passing the C++ object pointer and its related Ruby object
   to the passed callback function. */

/* Proxy method to abstract the internal trackings datatype */
static int swig_ruby_internal_iterate_callback(void* ptr, VALUE obj, void(*meth)(void* ptr, VALUE obj)) {
  (*meth)(ptr, obj);
  return ST_CONTINUE;
}

SWIGRUNTIME void SWIG_RubyIterateTrackings( void(*meth)(void* ptr, VALUE obj) ) {
  st_foreach(swig_ruby_trackings, (int (*)(ANYARGS))&swig_ruby_internal_iterate_callback, (st_data_t)meth);
}

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Ruby API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

SWIGINTERN VALUE
SWIG_Ruby_AppendOutput(VALUE target, VALUE o) {
  if (NIL_P(target)) {
    target = o;
  } else {
    if (TYPE(target) != T_ARRAY) {
      VALUE o2 = target;
      target = rb_ary_new();
      rb_ary_push(target, o2);
    }
    rb_ary_push(target, o);
  }
  return target;
}

/* For ruby1.8.4 and earlier. */
#ifndef RUBY_INIT_STACK
   RUBY_EXTERN void Init_stack(VALUE* addr);
#  define RUBY_INIT_STACK \
   VALUE variable_in_this_stack_frame; \
   Init_stack(&variable_in_this_stack_frame);
#endif


#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * rubyrun.swg
 *
 * This file contains the runtime support for Ruby modules
 * and includes code for managing global variables and pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* For backward compatibility only */
#define SWIG_POINTER_EXCEPTION  0

/* for raw pointers */
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Ruby_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Ruby_ConvertPtrAndOwn(obj, pptr, type, flags, own)
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Ruby_NewPointerObj(ptr, type, flags)
#define SWIG_AcquirePtr(ptr, own)                       SWIG_Ruby_AcquirePtr(ptr, own)
#define swig_owntype                                    swig_ruby_owntype

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Ruby_ConvertPacked(obj, ptr, sz, ty, flags)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Ruby_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_ConvertPtr(obj, pptr, type, 0)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_NewPointerObj(ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Ruby_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Ruby_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Ruby_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer) 		SWIG_Ruby_SetModule(pointer)


/* Error manipulation */

#define SWIG_ErrorType(code)                            SWIG_Ruby_ErrorType(code)               
#define SWIG_Error(code, msg)            		rb_raise(SWIG_Ruby_ErrorType(code), "%s", msg)
#define SWIG_fail                        		goto fail				 


/* Ruby-specific SWIG API */

#define SWIG_InitRuntime()                              SWIG_Ruby_InitRuntime()              
#define SWIG_define_class(ty)                        	SWIG_Ruby_define_class(ty)
#define SWIG_NewClassInstance(value, ty)             	SWIG_Ruby_NewClassInstance(value, ty)
#define SWIG_MangleStr(value)                        	SWIG_Ruby_MangleStr(value)		  
#define SWIG_CheckConvert(value, ty)                 	SWIG_Ruby_CheckConvert(value, ty)	  

#include "assert.h"

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  VALUE klass;
  VALUE mImpl;
  void  (*mark)(void *);
  void  (*destroy)(void *);
  int trackObjects;
} swig_class;


/* Global pointer used to keep some internal SWIG stuff */
static VALUE _cSWIG_Pointer = Qnil;
static VALUE swig_runtime_data_type_pointer = Qnil;

/* Global IDs used to keep some internal SWIG stuff */
static ID swig_arity_id = 0;
static ID swig_call_id  = 0;

/*
  If your swig extension is to be run within an embedded ruby and has
  director callbacks, you should set -DRUBY_EMBEDDED during compilation.  
  This will reset ruby's stack frame on each entry point from the main 
  program the first time a virtual director function is invoked (in a 
  non-recursive way).
  If this is not done, you run the risk of Ruby trashing the stack.
*/

#ifdef RUBY_EMBEDDED

#  define SWIG_INIT_STACK                            \
      if ( !swig_virtual_calls ) { RUBY_INIT_STACK } \
      ++swig_virtual_calls;
#  define SWIG_RELEASE_STACK --swig_virtual_calls;
#  define Ruby_DirectorTypeMismatchException(x) \
          rb_raise( rb_eTypeError, "%s", x ); return c_result;

      static unsigned int swig_virtual_calls = 0;

#else  /* normal non-embedded extension */

#  define SWIG_INIT_STACK
#  define SWIG_RELEASE_STACK
#  define Ruby_DirectorTypeMismatchException(x) \
          throw Swig::DirectorTypeMismatchException( x );

#endif  /* RUBY_EMBEDDED */


SWIGRUNTIME VALUE 
getExceptionClass(void) {
  static int init = 0;
  static VALUE rubyExceptionClass ;
  if (!init) {
    init = 1;
    rubyExceptionClass = rb_const_get(_mSWIG, rb_intern("Exception"));
  }
  return rubyExceptionClass;
} 

/* This code checks to see if the Ruby object being raised as part
   of an exception inherits from the Ruby class Exception.  If so,
   the object is simply returned.  If not, then a new Ruby exception
   object is created and that will be returned to Ruby.*/
SWIGRUNTIME VALUE
SWIG_Ruby_ExceptionType(swig_type_info *desc, VALUE obj) {
  VALUE exceptionClass = getExceptionClass();
  if (rb_obj_is_kind_of(obj, exceptionClass)) {
    return obj;
  }  else {
    return rb_exc_new3(rb_eRuntimeError, rb_obj_as_string(obj));
  }
}

/* Initialize Ruby runtime support */
SWIGRUNTIME void
SWIG_Ruby_InitRuntime(void)
{
  if (_mSWIG == Qnil) {
    _mSWIG = rb_define_module("SWIG");
    swig_call_id  = rb_intern("call");
    swig_arity_id = rb_intern("arity");
  }
}

/* Define Ruby class for C type */
SWIGRUNTIME void
SWIG_Ruby_define_class(swig_type_info *type)
{
  char *klass_name = (char *) malloc(4 + strlen(type->name) + 1);
  sprintf(klass_name, "TYPE%s", type->name);
  if (NIL_P(_cSWIG_Pointer)) {
    _cSWIG_Pointer = rb_define_class_under(_mSWIG, "Pointer", rb_cObject);
    rb_undef_method(CLASS_OF(_cSWIG_Pointer), "new");
  }
  rb_define_class_under(_mSWIG, klass_name, _cSWIG_Pointer);
  free((void *) klass_name);
}

/* Create a new pointer object */
SWIGRUNTIME VALUE
SWIG_Ruby_NewPointerObj(void *ptr, swig_type_info *type, int flags)
{
  int own =  flags & SWIG_POINTER_OWN; 
  int track;
  char *klass_name;
  swig_class *sklass;
  VALUE klass;
  VALUE obj;
  
  if (!ptr)
    return Qnil;
  
  if (type->clientdata) {
    sklass = (swig_class *) type->clientdata;
		
    /* Are we tracking this class and have we already returned this Ruby object? */
    track = sklass->trackObjects;
    if (track) {
      obj = SWIG_RubyInstanceFor(ptr);
      
      /* Check the object's type and make sure it has the correct type.
        It might not in cases where methods do things like 
        downcast methods. */
      if (obj != Qnil) {
        VALUE value = rb_iv_get(obj, "@__swigtype__");
        const char* type_name = RSTRING_PTR(value);
				
        if (strcmp(type->name, type_name) == 0) {
          return obj;
        }
      }
    }

    /* Create a new Ruby object */
    obj = Data_Wrap_Struct(sklass->klass, VOIDFUNC(sklass->mark), 
			   ( own ? VOIDFUNC(sklass->destroy) : 
			     (track ? VOIDFUNC(SWIG_RubyRemoveTracking) : 0 )
			     ), ptr);

    /* If tracking is on for this class then track this object. */
    if (track) {
      SWIG_RubyAddTracking(ptr, obj);
    }
  } else {
    klass_name = (char *) malloc(4 + strlen(type->name) + 1);
    sprintf(klass_name, "TYPE%s", type->name);
    klass = rb_const_get(_mSWIG, rb_intern(klass_name));
    free((void *) klass_name);
    obj = Data_Wrap_Struct(klass, 0, 0, ptr);
  }
  rb_iv_set(obj, "@__swigtype__", rb_str_new2(type->name));
  
  return obj;
}

/* Create a new class instance (always owned) */
SWIGRUNTIME VALUE
SWIG_Ruby_NewClassInstance(VALUE klass, swig_type_info *type)
{
  VALUE obj;
  swig_class *sklass = (swig_class *) type->clientdata;
  obj = Data_Wrap_Struct(klass, VOIDFUNC(sklass->mark), VOIDFUNC(sklass->destroy), 0);
  rb_iv_set(obj, "@__swigtype__", rb_str_new2(type->name));
  return obj;
}

/* Get type mangle from class name */
SWIGRUNTIMEINLINE char *
SWIG_Ruby_MangleStr(VALUE obj)
{
  VALUE stype = rb_iv_get(obj, "@__swigtype__");
  return StringValuePtr(stype);
}

/* Acquire a pointer value */
typedef struct {
  void (*datafree)(void *);
  int own;
} swig_ruby_owntype;

SWIGRUNTIME swig_ruby_owntype
SWIG_Ruby_AcquirePtr(VALUE obj, swig_ruby_owntype own) {
  swig_ruby_owntype oldown = {0, 0};
  if (obj) {
    oldown.datafree = RDATA(obj)->dfree;
    RDATA(obj)->dfree = own.datafree;
  }
  return oldown;
}

/* Convert a pointer value */
SWIGRUNTIME int
SWIG_Ruby_ConvertPtrAndOwn(VALUE obj, void **ptr, swig_type_info *ty, int flags, swig_ruby_owntype *own)
{
  char *c;
  swig_cast_info *tc;
  void *vptr = 0;

  /* Grab the pointer */
  if (NIL_P(obj)) {
    if (ptr)
      *ptr = 0;
    return SWIG_OK;
  } else {
    if (TYPE(obj) != T_DATA) {
      return SWIG_ERROR;
    }
    Data_Get_Struct(obj, void, vptr);
  }
  
  if (own) {
    own->datafree = RDATA(obj)->dfree;
    own->own = 0;
  }
    
  /* Check to see if the input object is giving up ownership
     of the underlying C struct or C++ object.  If so then we
     need to reset the destructor since the Ruby object no 
     longer owns the underlying C++ object.*/ 
  if (flags & SWIG_POINTER_DISOWN) {
    /* Is tracking on for this class? */
    int track = 0;
    if (ty && ty->clientdata) {
      swig_class *sklass = (swig_class *) ty->clientdata;
      track = sklass->trackObjects;
    }
		
    if (track) {
      /* We are tracking objects for this class.  Thus we change the destructor
       * to SWIG_RubyRemoveTracking.  This allows us to
       * remove the mapping from the C++ to Ruby object
       * when the Ruby object is garbage collected.  If we don't
       * do this, then it is possible we will return a reference 
       * to a Ruby object that no longer exists thereby crashing Ruby. */
      RDATA(obj)->dfree = SWIG_RubyRemoveTracking;
    } else {    
      RDATA(obj)->dfree = 0;
    }
  }

  /* Do type-checking if type info was provided */
  if (ty) {
    if (ty->clientdata) {
      if (rb_obj_is_kind_of(obj, ((swig_class *) (ty->clientdata))->klass)) {
        if (vptr == 0) {
          /* The object has already been deleted */
          return SWIG_ObjectPreviouslyDeletedError;
        }
      }
    }
    if ((c = SWIG_MangleStr(obj)) == NULL) {
      return SWIG_ERROR;
    }
    tc = SWIG_TypeCheck(c, ty);
    if (!tc) {
      return SWIG_ERROR;
    } else {
      if (ptr) {
        if (tc->type == ty) {
          *ptr = vptr;
        } else {
          int newmemory = 0;
          *ptr = SWIG_TypeCast(tc, vptr, &newmemory);
          if (newmemory == SWIG_CAST_NEW_MEMORY) {
            assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
            if (own)
              own->own = own->own | SWIG_CAST_NEW_MEMORY;
          }
        }
      }
    }
  } else {
    if (ptr)
      *ptr = vptr;
  }
  
  return SWIG_OK;
}

/* Check convert */
SWIGRUNTIMEINLINE int
SWIG_Ruby_CheckConvert(VALUE obj, swig_type_info *ty)
{
  char *c = SWIG_MangleStr(obj);
  if (!c) return 0;
  return SWIG_TypeCheck(c,ty) != 0;
}

SWIGRUNTIME VALUE
SWIG_Ruby_NewPackedObj(void *ptr, int sz, swig_type_info *type) {
  char result[1024];
  char *r = result;
  if ((2*sz + 1 + strlen(type->name)) > 1000) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r, ptr, sz);
  strcpy(r, type->name);
  return rb_str_new2(result);
}

/* Convert a packed value value */
SWIGRUNTIME int
SWIG_Ruby_ConvertPacked(VALUE obj, void *ptr, int sz, swig_type_info *ty) {
  swig_cast_info *tc;
  const char  *c;

  if (TYPE(obj) != T_STRING) goto type_error;
  c = StringValuePtr(obj);
  /* Pointer values must start with leading underscore */
  if (*c != '_') goto type_error;
  c++;
  c = SWIG_UnpackData(c, ptr, sz);
  if (ty) {
    tc = SWIG_TypeCheck(c, ty);
    if (!tc) goto type_error;
  }
  return SWIG_OK;

 type_error:
  return SWIG_ERROR;
}

SWIGRUNTIME swig_module_info *
SWIG_Ruby_GetModule(void *SWIGUNUSEDPARM(clientdata))
{
  VALUE pointer;
  swig_module_info *ret = 0;
  VALUE verbose = rb_gv_get("VERBOSE");

 /* temporarily disable warnings, since the pointer check causes warnings with 'ruby -w' */
  rb_gv_set("VERBOSE", Qfalse);
  
  /* first check if pointer already created */
  pointer = rb_gv_get("$swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME);
  if (pointer != Qnil) {
    Data_Get_Struct(pointer, swig_module_info, ret);
  }

  /* reinstate warnings */
  rb_gv_set("VERBOSE", verbose);
  return ret;
}

SWIGRUNTIME void 
SWIG_Ruby_SetModule(swig_module_info *pointer)
{
  /* register a new class */
  VALUE cl = rb_define_class("swig_runtime_data", rb_cObject);
  /* create and store the structure pointer to a global variable */
  swig_runtime_data_type_pointer = Data_Wrap_Struct(cl, 0, 0, pointer);
  rb_define_readonly_variable("$swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, &swig_runtime_data_type_pointer);
}

/* This function can be used to check whether a proc or method or similarly
   callable function has been passed.  Usually used in a %typecheck, like:

   %typecheck(c_callback_t, precedence=SWIG_TYPECHECK_POINTER) {
        $result = SWIG_Ruby_isCallable( $input );
   }
 */
SWIGINTERN
int SWIG_Ruby_isCallable( VALUE proc )
{
  if ( rb_respond_to( proc, swig_call_id ) )
    return 1;
  return 0;
}

/* This function can be used to check the arity (number of arguments)
   a proc or method can take.  Usually used in a %typecheck.
   Valid arities will be that equal to minimal or those < 0
   which indicate a variable number of parameters at the end.
 */
SWIGINTERN
int SWIG_Ruby_arity( VALUE proc, int minimal )
{
  if ( rb_respond_to( proc, swig_arity_id ) )
    {
      VALUE num = rb_funcall( proc, swig_arity_id, 0 );
      int arity = NUM2INT(num);
      if ( arity < 0 && (arity+1) < -minimal ) return 1;
      if ( arity == minimal ) return 1;
      return 1;
    }
  return 0;
}


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg);; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_FILE swig_types[0]
#define SWIGTYPE_p_apr_allocator_t swig_types[1]
#define SWIGTYPE_p_apr_array_header_t swig_types[2]
#define SWIGTYPE_p_apr_dir_t swig_types[3]
#define SWIGTYPE_p_apr_file_t swig_types[4]
#define SWIGTYPE_p_apr_finfo_t swig_types[5]
#define SWIGTYPE_p_apr_getopt_option_t swig_types[6]
#define SWIGTYPE_p_apr_getopt_t swig_types[7]
#define SWIGTYPE_p_apr_hash_t swig_types[8]
#define SWIGTYPE_p_apr_int32_t swig_types[9]
#define SWIGTYPE_p_apr_int64_t swig_types[10]
#define SWIGTYPE_p_apr_off_t swig_types[11]
#define SWIGTYPE_p_apr_pool_t swig_types[12]
#define SWIGTYPE_p_apr_pool_wrapper_t swig_types[13]
#define SWIGTYPE_p_apr_proc_t swig_types[14]
#define SWIGTYPE_p_apr_size_t swig_types[15]
#define SWIGTYPE_p_apr_uint32_t swig_types[16]
#define SWIGTYPE_p_apr_uint64_t swig_types[17]
#define SWIGTYPE_p_char swig_types[18]
#define SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[19]
#define SWIGTYPE_p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t swig_types[20]
#define SWIGTYPE_p_f_p_p_char_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t swig_types[21]
#define SWIGTYPE_p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void swig_types[22]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[23]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[24]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[25]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[26]
#define SWIGTYPE_p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[27]
#define SWIGTYPE_p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[28]
#define SWIGTYPE_p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[29]
#define SWIGTYPE_p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[30]
#define SWIGTYPE_p_f_p_q_const__char_p_q_const__char_p_void__int swig_types[31]
#define SWIGTYPE_p_f_p_q_const__char_p_q_const__char_p_void_p_apr_pool_t__int swig_types[32]
#define SWIGTYPE_p_f_p_q_const__char_p_void__int swig_types[33]
#define SWIGTYPE_p_f_p_q_const__char_p_void_p_apr_pool_t__int swig_types[34]
#define SWIGTYPE_p_f_p_q_const__struct_svn_version_t_p_q_const__struct_svn_version_t__int swig_types[35]
#define SWIGTYPE_p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[36]
#define SWIGTYPE_p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t swig_types[37]
#define SWIGTYPE_p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[38]
#define SWIGTYPE_p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[39]
#define SWIGTYPE_p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[40]
#define SWIGTYPE_p_f_p_void__p_svn_error_t swig_types[41]
#define SWIGTYPE_p_f_p_void__void swig_types[42]
#define SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t swig_types[43]
#define SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t swig_types[44]
#define SWIGTYPE_p_f_p_void_apr_size_t__p_svn_error_t swig_types[45]
#define SWIGTYPE_p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[46]
#define SWIGTYPE_p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t swig_types[47]
#define SWIGTYPE_p_f_p_void_p_char_p_apr_size_t__p_svn_error_t swig_types[48]
#define SWIGTYPE_p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t swig_types[49]
#define SWIGTYPE_p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[50]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_size_t__p_svn_error_t swig_types[51]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t swig_types[52]
#define SWIGTYPE_p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t swig_types[53]
#define SWIGTYPE_p_f_p_void_p_svn_boolean_t__p_svn_error_t swig_types[54]
#define SWIGTYPE_p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t swig_types[55]
#define SWIGTYPE_p_f_p_void_p_void__void swig_types[56]
#define SWIGTYPE_p_f_p_void_p_void_p_void_p_int__p_svn_error_t swig_types[57]
#define SWIGTYPE_p_f_p_void_svn_diff_datasource_e__p_svn_error_t swig_types[58]
#define SWIGTYPE_p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t swig_types[59]
#define SWIGTYPE_p_f_void__p_svn_version_t swig_types[60]
#define SWIGTYPE_p_int swig_types[61]
#define SWIGTYPE_p_long swig_types[62]
#define SWIGTYPE_p_p_apr_array_header_t swig_types[63]
#define SWIGTYPE_p_p_apr_file_t swig_types[64]
#define SWIGTYPE_p_p_apr_hash_t swig_types[65]
#define SWIGTYPE_p_p_char swig_types[66]
#define SWIGTYPE_p_p_svn_auth_baton_t swig_types[67]
#define SWIGTYPE_p_p_svn_auth_cred_simple_t swig_types[68]
#define SWIGTYPE_p_p_svn_auth_cred_ssl_client_cert_pw_t swig_types[69]
#define SWIGTYPE_p_p_svn_auth_cred_ssl_client_cert_t swig_types[70]
#define SWIGTYPE_p_p_svn_auth_cred_ssl_server_trust_t swig_types[71]
#define SWIGTYPE_p_p_svn_auth_cred_username_t swig_types[72]
#define SWIGTYPE_p_p_svn_auth_iterstate_t swig_types[73]
#define SWIGTYPE_p_p_svn_auth_provider_object_t swig_types[74]
#define SWIGTYPE_p_p_svn_checksum_t swig_types[75]
#define SWIGTYPE_p_p_svn_config_t swig_types[76]
#define SWIGTYPE_p_p_svn_diff_t swig_types[77]
#define SWIGTYPE_p_p_svn_io_dirent2_t swig_types[78]
#define SWIGTYPE_p_p_svn_patch_file_t swig_types[79]
#define SWIGTYPE_p_p_svn_patch_t swig_types[80]
#define SWIGTYPE_p_p_svn_stream_mark_t swig_types[81]
#define SWIGTYPE_p_p_svn_stream_t swig_types[82]
#define SWIGTYPE_p_p_svn_string_t swig_types[83]
#define SWIGTYPE_p_p_svn_stringbuf_t swig_types[84]
#define SWIGTYPE_p_p_void swig_types[85]
#define SWIGTYPE_p_svn_auth_baton_t swig_types[86]
#define SWIGTYPE_p_svn_auth_cred_simple_t swig_types[87]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t swig_types[88]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t swig_types[89]
#define SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t swig_types[90]
#define SWIGTYPE_p_svn_auth_cred_username_t swig_types[91]
#define SWIGTYPE_p_svn_auth_iterstate_t swig_types[92]
#define SWIGTYPE_p_svn_auth_provider_object_t swig_types[93]
#define SWIGTYPE_p_svn_auth_provider_t swig_types[94]
#define SWIGTYPE_p_svn_auth_ssl_server_cert_info_t swig_types[95]
#define SWIGTYPE_p_svn_cache_config_t swig_types[96]
#define SWIGTYPE_p_svn_checksum_ctx_t swig_types[97]
#define SWIGTYPE_p_svn_checksum_kind_t swig_types[98]
#define SWIGTYPE_p_svn_checksum_t swig_types[99]
#define SWIGTYPE_p_svn_commit_info_t swig_types[100]
#define SWIGTYPE_p_svn_config_t swig_types[101]
#define SWIGTYPE_p_svn_depth_t swig_types[102]
#define SWIGTYPE_p_svn_diff_binary_patch_t swig_types[103]
#define SWIGTYPE_p_svn_diff_conflict_display_style_t swig_types[104]
#define SWIGTYPE_p_svn_diff_datasource_e swig_types[105]
#define SWIGTYPE_p_svn_diff_file_ignore_space_t swig_types[106]
#define SWIGTYPE_p_svn_diff_file_options_t swig_types[107]
#define SWIGTYPE_p_svn_diff_fns2_t swig_types[108]
#define SWIGTYPE_p_svn_diff_fns_t swig_types[109]
#define SWIGTYPE_p_svn_diff_hunk_t swig_types[110]
#define SWIGTYPE_p_svn_diff_operation_kind_e swig_types[111]
#define SWIGTYPE_p_svn_diff_output_fns_t swig_types[112]
#define SWIGTYPE_p_svn_diff_t swig_types[113]
#define SWIGTYPE_p_svn_dirent_t swig_types[114]
#define SWIGTYPE_p_svn_errno_t swig_types[115]
#define SWIGTYPE_p_svn_error_t swig_types[116]
#define SWIGTYPE_p_svn_io_dirent2_t swig_types[117]
#define SWIGTYPE_p_svn_io_dirent_t swig_types[118]
#define SWIGTYPE_p_svn_io_file_del_t swig_types[119]
#define SWIGTYPE_p_svn_location_segment_t swig_types[120]
#define SWIGTYPE_p_svn_lock_t swig_types[121]
#define SWIGTYPE_p_svn_log_changed_path2_t swig_types[122]
#define SWIGTYPE_p_svn_log_changed_path_t swig_types[123]
#define SWIGTYPE_p_svn_log_entry_t swig_types[124]
#define SWIGTYPE_p_svn_merge_range_t swig_types[125]
#define SWIGTYPE_p_svn_mergeinfo_inheritance_t swig_types[126]
#define SWIGTYPE_p_svn_node_kind_t swig_types[127]
#define SWIGTYPE_p_svn_opt_revision_range_t swig_types[128]
#define SWIGTYPE_p_svn_opt_revision_t swig_types[129]
#define SWIGTYPE_p_svn_opt_revision_value_t swig_types[130]
#define SWIGTYPE_p_svn_opt_subcommand_desc2_t swig_types[131]
#define SWIGTYPE_p_svn_opt_subcommand_desc2_t_desc_overrides swig_types[132]
#define SWIGTYPE_p_svn_opt_subcommand_desc3_t swig_types[133]
#define SWIGTYPE_p_svn_opt_subcommand_desc3_t_desc_overrides swig_types[134]
#define SWIGTYPE_p_svn_opt_subcommand_desc_t swig_types[135]
#define SWIGTYPE_p_svn_patch_file_t swig_types[136]
#define SWIGTYPE_p_svn_patch_t swig_types[137]
#define SWIGTYPE_p_svn_prop_inherited_item_t swig_types[138]
#define SWIGTYPE_p_svn_prop_kind swig_types[139]
#define SWIGTYPE_p_svn_prop_patch_t swig_types[140]
#define SWIGTYPE_p_svn_prop_t swig_types[141]
#define SWIGTYPE_p_svn_stream_mark_t swig_types[142]
#define SWIGTYPE_p_svn_stream_t swig_types[143]
#define SWIGTYPE_p_svn_string_t swig_types[144]
#define SWIGTYPE_p_svn_stringbuf_t swig_types[145]
#define SWIGTYPE_p_svn_tristate_t swig_types[146]
#define SWIGTYPE_p_svn_version_checklist_t swig_types[147]
#define SWIGTYPE_p_svn_version_ext_linked_lib_t swig_types[148]
#define SWIGTYPE_p_svn_version_ext_loaded_lib_t swig_types[149]
#define SWIGTYPE_p_svn_version_extended_t swig_types[150]
#define SWIGTYPE_p_svn_version_t swig_types[151]
#define SWIGTYPE_p_unsigned_char swig_types[152]
#define SWIGTYPE_p_unsigned_long swig_types[153]
#define SWIGTYPE_p_void swig_types[154]
static swig_type_info *swig_types[156];
static swig_module_info swig_module = {swig_types, 155, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#define SWIG_init    Init_core
#define SWIG_name    "Svn::Ext::Core"

static VALUE mCore;

#define SWIG_RUBY_THREAD_BEGIN_BLOCK
#define SWIG_RUBY_THREAD_END_BLOCK


#define SWIGVERSION 0x030012 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#ifdef __cplusplus
extern "C" {
#endif

/* Ruby 1.9 changed the file name of this header */
#ifdef HAVE_RUBY_IO_H
#include "ruby/io.h"
#else
#include "rubyio.h"
#endif

#ifdef __cplusplus
}
#endif


#ifdef __cplusplus
extern "C" {
#endif
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
struct timeval rb_time_timeval(VALUE);
#endif
#ifdef __cplusplus
}
#endif


#include "svn_time.h"
#include "svn_pools.h"


#include "swigutil_rb.h"


static VALUE _global_svn_swig_rb_pool = Qnil;
static apr_pool_t *_global_pool = NULL;
static VALUE vresult = Qnil;
static VALUE *_global_vresult_address = &vresult;


#include <apr.h>
#include <apr_general.h>

#include <apr_md5.h>
#include "svn_diff.h"
#include "svn_private_config.h"


#include <apu.h>
#include <apr_xlate.h>


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


  #define SWIG_From_long   LONG2NUM 


SWIGINTERNINLINE VALUE
SWIG_From_int  (int value)
{    
  return SWIG_From_long  (value);
}


#include "svn_error_codes.h"




#include "svn_time.h"


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(VALUE obj, char** cptr, size_t* psize, int *alloc)
{
  if (TYPE(obj) == T_STRING) {
    char *cstr = StringValuePtr(obj); 
    size_t size = RSTRING_LEN(obj) + 1;
    if (cptr)  {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = (char *)memcpy(malloc((size)*sizeof(char)), cstr, sizeof(char)*(size));
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      }
    }
    if (psize) *psize = size;
    return SWIG_OK;
  } else {
    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *)vptr;
	if (psize) *psize = vptr ? (strlen((char*)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }  
  return SWIG_TypeError;
}







#include "svn_types.h"


SWIGINTERN VALUE
SWIG_ruby_failed(void)
{
  return Qnil;
} 


/*@SWIG:/opt/svnrm/prefix/share/swig/3.0.12/ruby/rubyprimtypes.swg,19,%ruby_aux_method@*/
SWIGINTERN VALUE SWIG_AUX_NUM2LONG(VALUE *args)
{
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  long *res = (long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2LONG(obj) : rb_big2long(obj);
  return obj;
}
/*@SWIG@*/

SWIGINTERN int
SWIG_AsVal_long (VALUE obj, long* val)
{
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(RUBY_METHOD_FUNC(SWIG_AUX_NUM2LONG), (VALUE)a, RUBY_METHOD_FUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int (VALUE obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE VALUE 
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > LONG_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_NewPointerObj((char *)(carray), pchar_descriptor, 0) : Qnil;
    } else {
      return rb_str_new(carray, (long)(size));
    }
  } else {
    return Qnil;
  }
}


SWIGINTERNINLINE VALUE 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}

SWIGINTERN struct svn_commit_info_t *new_svn_commit_info_t(apr_pool_t *pool){
    return svn_create_commit_info(pool);
  }
SWIGINTERN svn_commit_info_t *svn_commit_info_t_dup(struct svn_commit_info_t *self,apr_pool_t *pool){
    return svn_commit_info_dup(self, pool);
  }

SWIGINTERN int
SWIG_AsCharArray(VALUE obj, char *val, size_t size)
{ 
  char* cptr = 0; size_t csize = 0; int alloc = SWIG_OLDOBJ;
  int res = SWIG_AsCharPtrAndSize(obj, &cptr, &csize, &alloc);
  if (SWIG_IsOK(res)) {
    /* special case of single char conversion when we don't need space for NUL */
    if (size == 1 && csize == 2 && cptr && !cptr[1]) --csize;
    if (csize <= size) {
      if (val) {
	if (csize) memcpy(val, cptr, csize*sizeof(char));
	if (csize < size) memset(val + csize, 0, (size - csize)*sizeof(char));
      }
      if (alloc == SWIG_NEWOBJ) {
	free((char*)cptr);
	res = SWIG_DelNewMask(res);
      }      
      return res;
    }
    if (alloc == SWIG_NEWOBJ) free((char*)cptr);
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_char (VALUE obj, char *val)
{    
  int res = SWIG_AsCharArray(obj, val, 1);
  if (!SWIG_IsOK(res)) {
    long v;
    res = SWIG_AddCast(SWIG_AsVal_long (obj, &v));
    if (SWIG_IsOK(res)) {
      if ((CHAR_MIN <= v) && (v <= CHAR_MAX)) {
	if (val) *val = (char)(v);
      } else {
	res = SWIG_OverflowError;
      }
    }
  }
  return res;
}


SWIGINTERNINLINE VALUE
SWIG_From_char  (char c) 
{ 
  return SWIG_FromCharPtrAndSize(&c,1);
}

SWIGINTERN struct svn_merge_range_t *new_svn_merge_range_t(svn_revnum_t start,svn_revnum_t end,svn_boolean_t inheritable,apr_pool_t *pool){
    svn_merge_range_t *self;
    self = apr_palloc(pool, sizeof(svn_merge_range_t));
    self->start = start;
    self->end = end;
    self->inheritable = inheritable;
    return self;
  }
SWIGINTERN svn_merge_range_t *svn_merge_range_t_dup(struct svn_merge_range_t *self,apr_pool_t *pool){
    return svn_merge_range_dup(self, pool);
  }

static svn_error_t * svn_log_invoke_entry_receiver(
  svn_log_entry_receiver_t _obj, void *baton, svn_log_entry_t *log_entry, apr_pool_t *pool) {
  return _obj(baton, log_entry, pool);
}

static svn_error_t * svn_log_invoke_message_receiver(
  svn_log_message_receiver_t _obj, void *baton, apr_hash_t *changed_paths, svn_revnum_t revision, const char *author, const char *date, const char *message, apr_pool_t *pool) {
  return _obj(baton, changed_paths, revision, author, date, message, pool);
}

static svn_error_t * svn_commit_invoke_callback2(
  svn_commit_callback2_t _obj, const svn_commit_info_t *commit_info, void *baton, apr_pool_t *pool) {
  return _obj(commit_info, baton, pool);
}

static svn_error_t * svn_commit_invoke_callback(
  svn_commit_callback_t _obj, svn_revnum_t new_revision, const char *date, const char *author, void *baton) {
  return _obj(new_revision, date, author, baton);
}

static svn_error_t * svn_cancel_invoke_func(
  svn_cancel_func_t _obj, void *cancel_baton) {
  return _obj(cancel_baton);
}

static svn_error_t * svn_location_invoke_segment_receiver(
  svn_location_segment_receiver_t _obj, svn_location_segment_t *segment, void *baton, apr_pool_t *pool) {
  return _obj(segment, baton, pool);
}



#include "svn_pools.h"




#include "svn_version.h"


static const svn_version_t * svn_version_checklist_invoke_version_query(
  svn_version_checklist_t * _obj) {
  return (_obj->version_query)();
}

static const svn_version_t * svn_version_invoke_func(
  svn_version_func_t _obj) {
  return _obj();
}



#include "svn_props.h"




#include "svn_opt.h"

typedef struct {
  int optch; const char *desc; 
} svn_opt_subcommand_desc3_t_desc_overrides;

typedef struct {
  int optch; const char *desc; 
} svn_opt_subcommand_desc2_t_desc_overrides;




#include "svn_cmdline.h"




#include "svn_auth.h"


/*@SWIG:/opt/svnrm/prefix/share/swig/3.0.12/ruby/rubyprimtypes.swg,19,%ruby_aux_method@*/
SWIGINTERN VALUE SWIG_AUX_NUM2ULONG(VALUE *args)
{
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  unsigned long *res = (unsigned long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2ULONG(obj) : rb_big2ulong(obj);
  return obj;
}
/*@SWIG@*/

SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (VALUE obj, unsigned long *val) 
{
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    unsigned long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(RUBY_METHOD_FUNC(SWIG_AUX_NUM2ULONG), (VALUE)a, RUBY_METHOD_FUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}


SWIGINTERNINLINE VALUE
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return ULONG2NUM(value); 
}


static svn_error_t * svn_auth_provider_invoke_first_credentials(
  svn_auth_provider_t * _obj, void **credentials, void **iter_baton, void *provider_baton, apr_hash_t *parameters, const char *realmstring, apr_pool_t *pool) {
  return (_obj->first_credentials)(credentials, iter_baton, provider_baton, parameters, realmstring, pool);
}

static svn_error_t * svn_auth_provider_invoke_next_credentials(
  svn_auth_provider_t * _obj, void **credentials, void *iter_baton, void *provider_baton, apr_hash_t *parameters, const char *realmstring, apr_pool_t *pool) {
  return (_obj->next_credentials)(credentials, iter_baton, provider_baton, parameters, realmstring, pool);
}

static svn_error_t * svn_auth_provider_invoke_save_credentials(
  svn_auth_provider_t * _obj, svn_boolean_t *saved, void *credentials, void *provider_baton, apr_hash_t *parameters, const char *realmstring, apr_pool_t *pool) {
  return (_obj->save_credentials)(saved, credentials, provider_baton, parameters, realmstring, pool);
}

static void svn_auth_invoke_simple_provider_func(
  svn_auth_simple_provider_func_t _obj, svn_auth_provider_object_t **provider, apr_pool_t *pool) {
  _obj(provider, pool);
}

static void svn_auth_invoke_ssl_client_cert_pw_provider_func(
  svn_auth_ssl_client_cert_pw_provider_func_t _obj, svn_auth_provider_object_t **provider, apr_pool_t *pool) {
  _obj(provider, pool);
}

static svn_error_t * svn_auth_invoke_simple_prompt_func(
  svn_auth_simple_prompt_func_t _obj, svn_auth_cred_simple_t **cred, void *baton, const char *realm, const char *username, svn_boolean_t may_save, apr_pool_t *pool) {
  return _obj(cred, baton, realm, username, may_save, pool);
}

static svn_error_t * svn_auth_invoke_username_prompt_func(
  svn_auth_username_prompt_func_t _obj, svn_auth_cred_username_t **cred, void *baton, const char *realm, svn_boolean_t may_save, apr_pool_t *pool) {
  return _obj(cred, baton, realm, may_save, pool);
}

static svn_error_t * svn_auth_invoke_ssl_server_trust_prompt_func(
  svn_auth_ssl_server_trust_prompt_func_t _obj, svn_auth_cred_ssl_server_trust_t **cred, void *baton, const char *realm, apr_uint32_t failures, const svn_auth_ssl_server_cert_info_t *cert_info, svn_boolean_t may_save, apr_pool_t *pool) {
  return _obj(cred, baton, realm, failures, cert_info, may_save, pool);
}

static svn_error_t * svn_auth_invoke_ssl_client_cert_prompt_func(
  svn_auth_ssl_client_cert_prompt_func_t _obj, svn_auth_cred_ssl_client_cert_t **cred, void *baton, const char *realm, svn_boolean_t may_save, apr_pool_t *pool) {
  return _obj(cred, baton, realm, may_save, pool);
}

static svn_error_t * svn_auth_invoke_ssl_client_cert_pw_prompt_func(
  svn_auth_ssl_client_cert_pw_prompt_func_t _obj, svn_auth_cred_ssl_client_cert_pw_t **cred, void *baton, const char *realm, svn_boolean_t may_save, apr_pool_t *pool) {
  return _obj(cred, baton, realm, may_save, pool);
}

static svn_error_t * svn_auth_invoke_plaintext_prompt_func(
  svn_auth_plaintext_prompt_func_t _obj, svn_boolean_t *may_save_plaintext, const char *realmstring, void *baton, apr_pool_t *pool) {
  return _obj(may_save_plaintext, realmstring, baton, pool);
}

static svn_error_t * svn_auth_invoke_plaintext_passphrase_prompt_func(
  svn_auth_plaintext_passphrase_prompt_func_t _obj, svn_boolean_t *may_save_plaintext, const char *realmstring, void *baton, apr_pool_t *pool) {
  return _obj(may_save_plaintext, realmstring, baton, pool);
}



#include "svn_config.h"


static svn_boolean_t svn_config_invoke_section_enumerator(
  svn_config_section_enumerator_t _obj, const char *name, void *baton) {
  return _obj(name, baton);
}

static svn_boolean_t svn_config_invoke_section_enumerator2(
  svn_config_section_enumerator2_t _obj, const char *name, void *baton, apr_pool_t *pool) {
  return _obj(name, baton, pool);
}

static svn_boolean_t svn_config_invoke_enumerator(
  svn_config_enumerator_t _obj, const char *name, const char *value, void *baton) {
  return _obj(name, value, baton);
}

static svn_boolean_t svn_config_invoke_enumerator2(
  svn_config_enumerator2_t _obj, const char *name, const char *value, void *baton, apr_pool_t *pool) {
  return _obj(name, value, baton, pool);
}

static svn_error_t * svn_config_invoke_auth_walk_func(
  svn_config_auth_walk_func_t _obj, svn_boolean_t *delete_cred, void *walk_baton, const char *cred_kind, const char *realmstring, apr_hash_t *hash, apr_pool_t *scratch_pool) {
  return _obj(delete_cred, walk_baton, cred_kind, realmstring, hash, scratch_pool);
}



#include "svn_utf.h"




#include "svn_nls.h"




#include "svn_path.h"




#include "svn_dirent_uri.h"




#include "svn_mergeinfo.h"




#include "svn_io.h"


static svn_error_t * svn_read_invoke_fn(
  svn_read_fn_t _obj, void *baton, char *buffer, apr_size_t *len) {
  return _obj(baton, buffer, len);
}

static svn_error_t * svn_stream_invoke_skip_fn(
  svn_stream_skip_fn_t _obj, void *baton, apr_size_t len) {
  return _obj(baton, len);
}

static svn_error_t * svn_write_invoke_fn(
  svn_write_fn_t _obj, void *baton, const char *data, apr_size_t *len) {
  return _obj(baton, data, len);
}

static svn_error_t * svn_close_invoke_fn(
  svn_close_fn_t _obj, void *baton) {
  return _obj(baton);
}

static svn_error_t * svn_stream_invoke_mark_fn(
  svn_stream_mark_fn_t _obj, void *baton, svn_stream_mark_t **mark, apr_pool_t *pool) {
  return _obj(baton, mark, pool);
}

static svn_error_t * svn_stream_invoke_seek_fn(
  svn_stream_seek_fn_t _obj, void *baton, const svn_stream_mark_t *mark) {
  return _obj(baton, mark);
}

static svn_error_t * svn_stream_invoke_data_available_fn(
  svn_stream_data_available_fn_t _obj, void *baton, svn_boolean_t *data_available) {
  return _obj(baton, data_available);
}

static svn_error_t * svn_stream_invoke_readline_fn(
  svn_stream_readline_fn_t _obj, void *baton, svn_stringbuf_t **stringbuf, const char *eol, svn_boolean_t *eof, apr_pool_t *pool) {
  return _obj(baton, stringbuf, eol, eof, pool);
}

static svn_error_t * svn_stream_invoke_lazyopen_func(
  svn_stream_lazyopen_func_t _obj, svn_stream_t **stream, void *baton, apr_pool_t *result_pool, apr_pool_t *scratch_pool) {
  return _obj(stream, baton, result_pool, scratch_pool);
}

static svn_error_t * svn_io_invoke_walk_func(
  svn_io_walk_func_t _obj, void *baton, const char *path, const apr_finfo_t *finfo, apr_pool_t *pool) {
  return _obj(baton, path, finfo, pool);
}



#include "svn_checksum.h"




#include "svn_cache_config.h"




/* Helper function to set the gnome-keyring unlock prompt function. This
 * C function accepts an auth baton, a function and a prompt baton, but
 * the below callback_typemap uses both the function and the prompt
 * baton, so the resulting binding has just two arguments: The auth
 * baton and the prompt function.
 * The prompt function should again have two arguments: The keyring name
 * (string) and a pool (except for the ruby version, which doesn't have
 * the pool argument). It should return the entered password (string).
 * This binding generated for this function generates a reference to the
 * prompt function that was passed into this. The caller should store
 * that reference somewhere, to prevent the function from being garbage
 * collected...
 */
static void svn_auth_set_gnome_keyring_unlock_prompt_func(svn_auth_baton_t *ab,
                                                          svn_auth_gnome_keyring_unlock_prompt_func_t prompt_func,
                                                          void *prompt_baton) {
    svn_auth_set_parameter(ab, SVN_AUTH_PARAM_GNOME_KEYRING_UNLOCK_PROMPT_FUNC,
                           prompt_func);
    svn_auth_set_parameter(ab, SVN_AUTH_PARAM_GNOME_KEYRING_UNLOCK_PROMPT_BATON,
                           prompt_baton);
}


#include "svn_md5.h"




static void apr_pool_wrapper_destroy(apr_pool_wrapper_t *self);
static void apr_pool_wrapper_destroy_children(apr_pool_wrapper_t *self);
static void apr_pool_wrapper_remove_from_parent(apr_pool_wrapper_t *self);

SWIGINTERN void apr_pool_wrapper_t_destroy(VALUE object){
    svn_swig_rb_destroy_internal_pool(object);
  }
SWIGINTERN void apr_pool_wrapper_t_set_default_max_free_size(apr_size_t size){
    apr_allocator_max_free_set(svn_swig_rb_allocator(), size);
  }
SWIGINTERN struct apr_pool_wrapper_t *new_apr_pool_wrapper_t(apr_pool_wrapper_t *parent){
    apr_pool_wrapper_t *self;
    apr_pool_t *parent_pool;

    self = ALLOC(apr_pool_wrapper_t);
    if (parent) {
      parent_pool = parent->pool;
      (*((apr_pool_wrapper_t * *)apr_array_push(parent->children))) = self;
    } else {
      parent_pool = svn_swig_rb_pool();
    }
    self->pool = svn_pool_create_ex(parent_pool, NULL);
    self->destroyed = 0;
    self->parent = parent;
    self->children = apr_array_make(self->pool, 0,
                                    sizeof(apr_pool_wrapper_t *));
    return self;
  }
SWIGINTERN void apr_pool_wrapper_t_set_max_free_size(struct apr_pool_wrapper_t *self,apr_size_t size){
    apr_allocator_t *allocator;

    allocator = apr_pool_allocator_get(self->pool);
    apr_allocator_max_free_set(allocator, size);
  }
SWIGINTERN void apr_pool_wrapper_t__destroy(struct apr_pool_wrapper_t *self){
    apr_pool_wrapper_destroy(self);
  }

static void
apr_pool_wrapper_destroy(apr_pool_wrapper_t *self)
{
  if (!self->destroyed) {
    self->destroyed = TRUE;
    apr_pool_wrapper_destroy_children(self);
    apr_pool_wrapper_remove_from_parent(self);
    apr_pool_destroy(self->pool);
  }
}

static void
apr_pool_wrapper_destroy_children(apr_pool_wrapper_t *self)
{
  apr_pool_wrapper_t **child;

  while ((child = apr_array_pop(self->children))) {
    if (*child) {
      apr_pool_wrapper_destroy(*child);
    }
  }
}

static void
apr_pool_wrapper_remove_from_parent(apr_pool_wrapper_t *self)
{
  if (self->parent) {
    apr_pool_wrapper_t *child;
    int i, len;

    len = self->parent->children->nelts;
    for (i = 0; i < len; i++) {
      child = APR_ARRAY_IDX(self->parent->children, i, apr_pool_wrapper_t *);
      if (child == self) {
        APR_ARRAY_IDX(self->parent->children, i, apr_pool_wrapper_t *) = NULL;
        self->parent = NULL;
        break;
      }
    }
  }
}

SWIGINTERN struct svn_stream_t *new_svn_stream_t(VALUE io){
    return svn_swig_rb_make_stream(io);
  }
SWIGINTERN struct svn_auth_baton_t *new_svn_auth_baton_t(apr_array_header_t *providers,apr_pool_t *pool){
    svn_auth_baton_t *self;
    svn_auth_open(&self, providers, pool);
    return self;
  }

#include "svn_diff.h"

SWIGINTERN struct svn_diff_file_options_t *new_svn_diff_file_options_t(apr_pool_t *pool){
    return svn_diff_file_options_create(pool);
  }

static svn_error_t * svn_diff_fns2_invoke_datasources_open(
  svn_diff_fns2_t * _obj, void *diff_baton, apr_off_t *prefix_lines, apr_off_t *suffix_lines, const svn_diff_datasource_e *datasources, apr_size_t datasources_len) {
  return (_obj->datasources_open)(diff_baton, prefix_lines, suffix_lines, datasources, datasources_len);
}

static svn_error_t * svn_diff_fns2_invoke_datasource_close(
  svn_diff_fns2_t * _obj, void *diff_baton, svn_diff_datasource_e datasource) {
  return (_obj->datasource_close)(diff_baton, datasource);
}

static svn_error_t * svn_diff_fns2_invoke_datasource_get_next_token(
  svn_diff_fns2_t * _obj, apr_uint32_t *hash, void **token, void *diff_baton, svn_diff_datasource_e datasource) {
  return (_obj->datasource_get_next_token)(hash, token, diff_baton, datasource);
}

static svn_error_t * svn_diff_fns2_invoke_token_compare(
  svn_diff_fns2_t * _obj, void *diff_baton, void *ltoken, void *rtoken, int *compare) {
  return (_obj->token_compare)(diff_baton, ltoken, rtoken, compare);
}

static void svn_diff_fns2_invoke_token_discard(
  svn_diff_fns2_t * _obj, void *diff_baton, void *token) {
  (_obj->token_discard)(diff_baton, token);
}

static void svn_diff_fns2_invoke_token_discard_all(
  svn_diff_fns2_t * _obj, void *diff_baton) {
  (_obj->token_discard_all)(diff_baton);
}

static svn_error_t * svn_diff_fns_invoke_datasource_open(
  svn_diff_fns_t * _obj, void *diff_baton, svn_diff_datasource_e datasource) {
  return (_obj->datasource_open)(diff_baton, datasource);
}

static svn_error_t * svn_diff_fns_invoke_datasource_close(
  svn_diff_fns_t * _obj, void *diff_baton, svn_diff_datasource_e datasource) {
  return (_obj->datasource_close)(diff_baton, datasource);
}

static svn_error_t * svn_diff_fns_invoke_datasource_get_next_token(
  svn_diff_fns_t * _obj, apr_uint32_t *hash, void **token, void *diff_baton, svn_diff_datasource_e datasource) {
  return (_obj->datasource_get_next_token)(hash, token, diff_baton, datasource);
}

static svn_error_t * svn_diff_fns_invoke_token_compare(
  svn_diff_fns_t * _obj, void *diff_baton, void *ltoken, void *rtoken, int *compare) {
  return (_obj->token_compare)(diff_baton, ltoken, rtoken, compare);
}

static void svn_diff_fns_invoke_token_discard(
  svn_diff_fns_t * _obj, void *diff_baton, void *token) {
  (_obj->token_discard)(diff_baton, token);
}

static void svn_diff_fns_invoke_token_discard_all(
  svn_diff_fns_t * _obj, void *diff_baton) {
  (_obj->token_discard_all)(diff_baton);
}

static svn_error_t * svn_diff_output_fns_invoke_output_common(
  svn_diff_output_fns_t * _obj, void *output_baton, apr_off_t original_start, apr_off_t original_length, apr_off_t modified_start, apr_off_t modified_length, apr_off_t latest_start, apr_off_t latest_length) {
  return (_obj->output_common)(output_baton, original_start, original_length, modified_start, modified_length, latest_start, latest_length);
}

static svn_error_t * svn_diff_output_fns_invoke_output_diff_modified(
  svn_diff_output_fns_t * _obj, void *output_baton, apr_off_t original_start, apr_off_t original_length, apr_off_t modified_start, apr_off_t modified_length, apr_off_t latest_start, apr_off_t latest_length) {
  return (_obj->output_diff_modified)(output_baton, original_start, original_length, modified_start, modified_length, latest_start, latest_length);
}

static svn_error_t * svn_diff_output_fns_invoke_output_diff_latest(
  svn_diff_output_fns_t * _obj, void *output_baton, apr_off_t original_start, apr_off_t original_length, apr_off_t modified_start, apr_off_t modified_length, apr_off_t latest_start, apr_off_t latest_length) {
  return (_obj->output_diff_latest)(output_baton, original_start, original_length, modified_start, modified_length, latest_start, latest_length);
}

static svn_error_t * svn_diff_output_fns_invoke_output_diff_common(
  svn_diff_output_fns_t * _obj, void *output_baton, apr_off_t original_start, apr_off_t original_length, apr_off_t modified_start, apr_off_t modified_length, apr_off_t latest_start, apr_off_t latest_length) {
  return (_obj->output_diff_common)(output_baton, original_start, original_length, modified_start, modified_length, latest_start, latest_length);
}

static svn_error_t * svn_diff_output_fns_invoke_output_conflict(
  svn_diff_output_fns_t * _obj, void *output_baton, apr_off_t original_start, apr_off_t original_length, apr_off_t modified_start, apr_off_t modified_length, apr_off_t latest_start, apr_off_t latest_length, svn_diff_t *resolved_diff) {
  return (_obj->output_conflict)(output_baton, original_start, original_length, modified_start, modified_length, latest_start, latest_length, resolved_diff);
}



static VALUE
svn_default_charset(void)
{
  return PTR2NUM(APR_DEFAULT_CHARSET);
}

static VALUE
svn_locale_charset(void)
{
  return PTR2NUM(APR_LOCALE_CHARSET);
}

/* prompt providers return baton for protecting GC */
static VALUE
svn_swig_rb_auth_get_simple_prompt_provider(
  svn_auth_provider_object_t **provider,
  svn_auth_simple_prompt_func_t prompt_func,
  void *prompt_baton,
  int retry_limit,
  apr_pool_t *pool)
{
  svn_auth_get_simple_prompt_provider(provider, prompt_func, prompt_baton,
                                      retry_limit, pool);
  return rb_ary_new3(1, (VALUE)prompt_baton);
}

static VALUE
svn_swig_rb_auth_get_ssl_client_cert_prompt_provider(
  svn_auth_provider_object_t **provider,
  svn_auth_ssl_client_cert_prompt_func_t prompt_func,
  void *prompt_baton,
  int retry_limit,
  apr_pool_t *pool)
{
  svn_auth_get_ssl_client_cert_prompt_provider(provider, prompt_func,
                                               prompt_baton, retry_limit,
                                               pool);
  return rb_ary_new3(1, (VALUE)prompt_baton);
}

static VALUE
svn_swig_rb_auth_get_ssl_client_cert_pw_prompt_provider(
  svn_auth_provider_object_t **provider,
  svn_auth_ssl_client_cert_pw_prompt_func_t prompt_func,
  void *prompt_baton,
  int retry_limit,
  apr_pool_t *pool)
{
  svn_auth_get_ssl_client_cert_pw_prompt_provider(provider, prompt_func,
                                                  prompt_baton, retry_limit,
                                                  pool);
  return rb_ary_new3(1, (VALUE)prompt_baton);
}

static VALUE
svn_swig_rb_auth_get_ssl_server_trust_prompt_provider(
  svn_auth_provider_object_t **provider,
  svn_auth_ssl_server_trust_prompt_func_t prompt_func,
  void *prompt_baton,
  apr_pool_t *pool)
{
  svn_auth_get_ssl_server_trust_prompt_provider(provider, prompt_func,
                                                prompt_baton, pool);
  return rb_ary_new3(1, (VALUE)prompt_baton);
}

static VALUE
svn_swig_rb_auth_get_username_prompt_provider(
  svn_auth_provider_object_t **provider,
  svn_auth_username_prompt_func_t prompt_func,
  void *prompt_baton,
  int retry_limit,
  apr_pool_t *pool)
{
  svn_auth_get_username_prompt_provider(provider, prompt_func, prompt_baton,
                                        retry_limit, pool);
  return rb_ary_new3(1, (VALUE)prompt_baton);
}


static svn_error_t *
svn_swig_mergeinfo_merge(apr_hash_t **mergeinfo_inout,
                         apr_hash_t *changes,
                         apr_pool_t *pool)
{
  return svn_mergeinfo_merge(*mergeinfo_inout, changes, pool);
}

static svn_error_t *
svn_swig_mergeinfo_sort(apr_hash_t **mergeinfo_inout, apr_pool_t *pool)
{
  return svn_mergeinfo_sort(*mergeinfo_inout, pool);
}

static svn_error_t *
svn_swig_rangelist_merge(svn_rangelist_t **rangelist_inout,
                         svn_rangelist_t *changes,
                         apr_pool_t *pool)
{
  return svn_rangelist_merge(rangelist_inout, changes, pool);
}

static svn_error_t *
svn_swig_rangelist_reverse(svn_rangelist_t **rangelist_inout,
                           apr_pool_t *pool)
{
  return svn_rangelist_reverse(*rangelist_inout, pool);
}

SWIGINTERN VALUE
_wrap_apr_initialize(int argc, VALUE *argv, VALUE self) {
  apr_status_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (apr_status_t)apr_initialize();
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_apr_terminate(int argc, VALUE *argv, VALUE self) {
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    apr_terminate();
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_apr_time_ansi_put(int argc, VALUE *argv, VALUE self) {
  apr_time_t *arg1 = (apr_time_t *) 0 ;
  time_t arg2 ;
  apr_time_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  apr_status_t result;
  VALUE vresult = Qnil;
  
  arg1 = &temp1;
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0]))
    arg2 = (time_t)-1;
    else
    arg2 = NUM2LONG(rb_funcall(argv[0], rb_intern("tv_sec"), 0));
  }
  {
    result = (apr_status_t)apr_time_ansi_put(arg1,arg2);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LL2NUM(*arg1));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_apr_pool_destroy(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_pool_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_pool_t *","apr_pool_destroy", 1, argv[0] )); 
  }
  arg1 = (apr_pool_t *)(argp1);
  {
    apr_pool_destroy(arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_apr_pool_clear(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_pool_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_pool_t *","apr_pool_clear", 1, argv[0] )); 
  }
  arg1 = (apr_pool_t *)(argp1);
  {
    apr_pool_clear(arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_apr_file_open_stdout(int argc, VALUE *argv, VALUE self) {
  apr_file_t **arg1 = (apr_file_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_file_t *temp1 ;
  apr_status_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (apr_status_t)apr_file_open_stdout(arg1,arg2);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_apr_file_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_apr_file_open_stderr(int argc, VALUE *argv, VALUE self) {
  apr_file_t **arg1 = (apr_file_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_file_t *temp1 ;
  apr_status_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (apr_status_t)apr_file_open_stderr(arg1,arg2);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_apr_file_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_time_to_cstring(int argc, VALUE *argv, VALUE self) {
  apr_time_t arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_time_to_cstring(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_time_from_cstring(int argc, VALUE *argv, VALUE self) {
  apr_time_t *arg1 = (apr_time_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_time_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_time_from_cstring", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_time_from_cstring(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LL2NUM(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_time_to_human_cstring(int argc, VALUE *argv, VALUE self) {
  apr_time_t arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_time_to_human_cstring(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_parse_date(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  apr_time_t *arg2 = (apr_time_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_time_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  apr_time_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_parse_date", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = (apr_time_t)NUM2LL(argv[1]);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_parse_date(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LL2NUM(*arg2));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_sleep_for_timestamps(int argc, VALUE *argv, VALUE self) {
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    svn_sleep_for_timestamps();
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_error_t;

SWIGINTERN VALUE
_wrap_svn_error_t_apr_err_set(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  apr_status_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","apr_err", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_status_t","apr_err", 2, argv[0] ));
  } 
  arg2 = (apr_status_t)(val2);
  if (arg1) (arg1)->apr_err = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_apr_err_get(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_status_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","apr_err", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  result = (apr_status_t) ((arg1)->apr_err);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_message_get(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","message", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  result = (char *) ((arg1)->message);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_child_set(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  struct svn_error_t *arg2 = (struct svn_error_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","child", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_error_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "struct svn_error_t *","child", 2, argv[0] )); 
  }
  arg2 = (struct svn_error_t *)(argp2);
  if (arg1) (arg1)->child = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_child_get(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","child", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  result = (struct svn_error_t *) ((arg1)->child);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_error_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_pool_set(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","pool", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  if (argc > 0) {
    
  }
  if (arg1) (arg1)->pool = arg2;
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_pool_get(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_pool_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","pool", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  result = (apr_pool_t *) ((arg1)->pool);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_pool_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_file_get(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","file", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  result = (char *) ((arg1)->file);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_line_set(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","line", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "long","line", 2, argv[0] ));
  } 
  arg2 = (long)(val2);
  if (arg1) (arg1)->line = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_error_t_line_get(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *arg1 = (struct svn_error_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_error_t *","line", 1, self )); 
  }
  arg1 = (struct svn_error_t *)(argp1);
  result = (long) ((arg1)->line);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_error_t_allocate(VALUE self)
#else
_wrap_svn_error_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_error_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_error_t(int argc, VALUE *argv, VALUE self) {
  struct svn_error_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_error_t *)calloc(1, sizeof(struct svn_error_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_error_t(void *self) {
    struct svn_error_t *arg1 = (struct svn_error_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_node_kind_to_word(int argc, VALUE *argv, VALUE self) {
  svn_node_kind_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_node_kind_to_word", 1, argv[0] ));
  } 
  arg1 = (svn_node_kind_t)(val1);
  {
    result = (char *)svn_node_kind_to_word(arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_node_kind_from_word(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_node_kind_from_word", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_node_kind_t)svn_node_kind_from_word((char const *)arg1);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_tristate__to_word(int argc, VALUE *argv, VALUE self) {
  svn_tristate_t arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "svn_tristate_t","svn_tristate__to_word", 1, argv[0] ));
  } 
  arg1 = (svn_tristate_t)(val1);
  {
    result = (char *)svn_tristate__to_word(arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_tristate__from_word(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_tristate_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_tristate__from_word", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_tristate_t)svn_tristate__from_word((char const *)arg1);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_revnum_parse(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = (svn_revnum_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char **arg3 = (char **) 0 ;
  svn_revnum_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *temp3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg1 = &temp1;
  arg3 = &temp3;
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_revnum_parse", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_error_t *)svn_revnum_parse(arg1,(char const *)arg2,(char const **)arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_long, new_flags));
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_depth_to_word(int argc, VALUE *argv, VALUE self) {
  svn_depth_t arg1 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_to_depth(argv[0]);
  }
  {
    result = (char *)svn_depth_to_word(arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_depth_from_word(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_depth_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_depth_from_word", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_depth_t)svn_depth_from_word((char const *)arg1);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


static swig_class SwigClassSvn_dirent_t;

SWIGINTERN VALUE
_wrap_svn_dirent_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}



/*
  Document-method: Svn::Ext::Core::svn_dirent_t.size

  call-seq:
    size -> svn_filesize_t

Size or Length of the svn_dirent_t.
*/

/*
  Document-method: Svn::Ext::Core::svn_dirent_t.size=

  call-seq:
    size=(x) -> svn_filesize_t

Size or Length of the svn_dirent_t.
*/
SWIGINTERN VALUE
_wrap_svn_dirent_t_size_set(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  svn_filesize_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","size", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  {
    arg2 = (svn_filesize_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->size = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_size_get(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_filesize_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","size", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  result =  ((arg1)->size);
  {
    vresult = LL2NUM((svn_filesize_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_has_props_set(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","has_props", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->has_props = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_has_props_get(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","has_props", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->has_props);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_created_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","created_rev", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","created_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->created_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_created_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","created_rev", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->created_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_time_set(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","time", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->time = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_time_get(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","time", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  result =  ((arg1)->time);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_t_last_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *arg1 = (struct svn_dirent_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_dirent_t *","last_author", 1, self )); 
  }
  arg1 = (struct svn_dirent_t *)(argp1);
  result = (char *) ((arg1)->last_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_dirent_t_allocate(VALUE self)
#else
_wrap_svn_dirent_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_dirent_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_dirent_t(int argc, VALUE *argv, VALUE self) {
  struct svn_dirent_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_dirent_t *)calloc(1, sizeof(struct svn_dirent_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_dirent_t(void *self) {
    struct svn_dirent_t *arg1 = (struct svn_dirent_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_dirent_dup(int argc, VALUE *argv, VALUE self) {
  svn_dirent_t *arg1 = (svn_dirent_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_dirent_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_dirent_t const *","svn_dirent_dup", 1, argv[0] )); 
  }
  arg1 = (svn_dirent_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_dirent_t *)svn_dirent_dup((struct svn_dirent_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_create(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_dirent_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_dirent_t *)svn_dirent_create(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_dirent_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_commit_info_t;

SWIGINTERN VALUE
_wrap_svn_commit_info_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","date", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","date", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->date) free((char *)arg1->date);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->date = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","date", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  result = (char *) ((arg1)->date);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","author", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->author) free((char *)arg1->author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","author", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  result = (char *) ((arg1)->author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_post_commit_err_set(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","post_commit_err", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","post_commit_err", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->post_commit_err) free((char *)arg1->post_commit_err);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->post_commit_err = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_post_commit_err_get(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","post_commit_err", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  result = (char *) ((arg1)->post_commit_err);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_repos_root_set(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","repos_root", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_root", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_root) free((char *)arg1->repos_root);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_root = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_info_t_repos_root_get(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","repos_root", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  result = (char *) ((arg1)->repos_root);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_commit_info_t_allocate(VALUE self)
#else
_wrap_svn_commit_info_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_commit_info_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_commit_info_t(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  struct svn_commit_info_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (struct svn_commit_info_t *)new_svn_commit_info_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_commit_info_t(struct svn_commit_info_t *self){
  }
SWIGINTERN void
free_svn_commit_info_t(void *self) {
    struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *)self;
    delete_svn_commit_info_t(arg1);
}


/*
  Document-method: Svn::Ext::Core::svn_commit_info_t.dup

  call-seq:
    dup(pool) -> svn_commit_info_t

Create a duplicate of the class and unfreeze it if needed.
*/
SWIGINTERN VALUE
_wrap_svn_commit_info_t_dup(int argc, VALUE *argv, VALUE self) {
  struct svn_commit_info_t *arg1 = (struct svn_commit_info_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_commit_info_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_commit_info_t *","dup", 1, self )); 
  }
  arg1 = (struct svn_commit_info_t *)(argp1);
  if (argc > 0) {
    
  }
  {
    result = (svn_commit_info_t *)svn_commit_info_t_dup(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_log_changed_path2_t;

SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_action_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","action", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  ecode2 = SWIG_AsVal_char(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "char","action", 2, argv[0] ));
  } 
  arg2 = (char)(val2);
  if (arg1) (arg1)->action = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_action_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","action", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  result = (char) ((arg1)->action);
  vresult = SWIG_From_char((char)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_copyfrom_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","copyfrom_path", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","copyfrom_path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_path) free((char *)arg1->copyfrom_path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_copyfrom_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","copyfrom_path", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  result = (char *) ((arg1)->copyfrom_path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_copyfrom_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","copyfrom_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->copyfrom_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_copyfrom_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_node_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","node_kind", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","node_kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->node_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_node_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","node_kind", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->node_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_text_modified_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  svn_tristate_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","text_modified", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_tristate_t","text_modified", 2, argv[0] ));
  } 
  arg2 = (svn_tristate_t)(val2);
  if (arg1) (arg1)->text_modified = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_text_modified_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_tristate_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","text_modified", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  result = (svn_tristate_t) ((arg1)->text_modified);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_props_modified_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  svn_tristate_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","props_modified", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_tristate_t","props_modified", 2, argv[0] ));
  } 
  arg2 = (svn_tristate_t)(val2);
  if (arg1) (arg1)->props_modified = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_t_props_modified_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_tristate_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path2_t *","props_modified", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path2_t *)(argp1);
  result = (svn_tristate_t) ((arg1)->props_modified);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_log_changed_path2_t_allocate(VALUE self)
#else
_wrap_svn_log_changed_path2_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_log_changed_path2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_log_changed_path2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_log_changed_path2_t *)calloc(1, sizeof(struct svn_log_changed_path2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_log_changed_path2_t(void *self) {
    struct svn_log_changed_path2_t *arg1 = (struct svn_log_changed_path2_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_log_changed_path2_create(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_log_changed_path2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_log_changed_path2_t *)svn_log_changed_path2_create(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path2_dup(int argc, VALUE *argv, VALUE self) {
  svn_log_changed_path2_t *arg1 = (svn_log_changed_path2_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_log_changed_path2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_log_changed_path2_t const *","svn_log_changed_path2_dup", 1, argv[0] )); 
  }
  arg1 = (svn_log_changed_path2_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_log_changed_path2_t *)svn_log_changed_path2_dup((struct svn_log_changed_path2_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_log_changed_path2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_log_changed_path_t;

SWIGINTERN VALUE
_wrap_svn_log_changed_path_t_action_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path_t *arg1 = (struct svn_log_changed_path_t *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path_t *","action", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path_t *)(argp1);
  ecode2 = SWIG_AsVal_char(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "char","action", 2, argv[0] ));
  } 
  arg2 = (char)(val2);
  if (arg1) (arg1)->action = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path_t_action_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path_t *arg1 = (struct svn_log_changed_path_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path_t *","action", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path_t *)(argp1);
  result = (char) ((arg1)->action);
  vresult = SWIG_From_char((char)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path_t_copyfrom_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path_t *arg1 = (struct svn_log_changed_path_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path_t *","copyfrom_path", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path_t *)(argp1);
  result = (char *) ((arg1)->copyfrom_path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path_t_copyfrom_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path_t *arg1 = (struct svn_log_changed_path_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","copyfrom_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->copyfrom_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_changed_path_t_copyfrom_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path_t *arg1 = (struct svn_log_changed_path_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_changed_path_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_changed_path_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_log_changed_path_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_log_changed_path_t_allocate(VALUE self)
#else
_wrap_svn_log_changed_path_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_log_changed_path_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_log_changed_path_t(int argc, VALUE *argv, VALUE self) {
  struct svn_log_changed_path_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_log_changed_path_t *)calloc(1, sizeof(struct svn_log_changed_path_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_log_changed_path_t(void *self) {
    struct svn_log_changed_path_t *arg1 = (struct svn_log_changed_path_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_log_changed_path_dup(int argc, VALUE *argv, VALUE self) {
  svn_log_changed_path_t *arg1 = (svn_log_changed_path_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_log_changed_path_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_log_changed_path_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_log_changed_path_t const *","svn_log_changed_path_dup", 1, argv[0] )); 
  }
  arg1 = (svn_log_changed_path_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_log_changed_path_t *)svn_log_changed_path_dup((struct svn_log_changed_path_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_log_changed_path_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_log_entry_t;

SWIGINTERN VALUE
_wrap_svn_log_entry_t_changed_paths_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","changed_paths", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","changed_paths", 2, argv[0] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  if (arg1) (arg1)->changed_paths = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_changed_paths_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","changed_paths", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  result = (apr_hash_t *) ((arg1)->changed_paths);
  {
    VALUE rb_changed_path_hash;
    rb_changed_path_hash =
    svn_swig_rb_apr_hash_to_hash_swig_type(result, "svn_log_changed_path_t *");
    vresult = SWIG_Ruby_AppendOutput(vresult, rb_changed_path_hash);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_revprops_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","revprops", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg2 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[0], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg2)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[0], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (arg1) (arg1)->revprops = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_revprops_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","revprops", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  result = (apr_hash_t *) ((arg1)->revprops);
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_has_children_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","has_children", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->has_children = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_has_children_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","has_children", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->has_children);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_changed_paths2_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","changed_paths2", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","changed_paths2", 2, argv[0] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  if (arg1) (arg1)->changed_paths2 = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_changed_paths2_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","changed_paths2", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  result = (apr_hash_t *) ((arg1)->changed_paths2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_non_inheritable_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","non_inheritable", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->non_inheritable = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_non_inheritable_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","non_inheritable", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->non_inheritable);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_subtractive_merge_set(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","subtractive_merge", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->subtractive_merge = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_t_subtractive_merge_get(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_log_entry_t *","subtractive_merge", 1, self )); 
  }
  arg1 = (struct svn_log_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->subtractive_merge);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_log_entry_t_allocate(VALUE self)
#else
_wrap_svn_log_entry_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_log_entry_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_log_entry_t(int argc, VALUE *argv, VALUE self) {
  struct svn_log_entry_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_log_entry_t *)calloc(1, sizeof(struct svn_log_entry_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_log_entry_t(void *self) {
    struct svn_log_entry_t *arg1 = (struct svn_log_entry_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_log_entry_create(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_log_entry_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_log_entry_t *)svn_log_entry_create(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_entry_dup(int argc, VALUE *argv, VALUE self) {
  svn_log_entry_t *arg1 = (svn_log_entry_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_log_entry_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_log_entry_t const *","svn_log_entry_dup", 1, argv[0] )); 
  }
  arg1 = (svn_log_entry_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_log_entry_t *)svn_log_entry_dup((struct svn_log_entry_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mime_type_validate(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_mime_type_validate", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_mime_type_validate((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mime_type_is_binary(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_mime_type_is_binary", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_mime_type_is_binary((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


static swig_class SwigClassSvn_lock_t;

SWIGINTERN VALUE
_wrap_svn_lock_t_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","path", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","path", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  result = (char *) ((arg1)->path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_token_set(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","token", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","token", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->token) free((char *)arg1->token);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->token = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_token_get(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","token", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  result = (char *) ((arg1)->token);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_owner_set(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","owner", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","owner", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->owner) free((char *)arg1->owner);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->owner = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_owner_get(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","owner", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  result = (char *) ((arg1)->owner);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_comment_set(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","comment", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","comment", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->comment) free((char *)arg1->comment);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->comment = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_comment_get(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","comment", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  result = (char *) ((arg1)->comment);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_is_dav_comment_set(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","is_dav_comment", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->is_dav_comment = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_is_dav_comment_get(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","is_dav_comment", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->is_dav_comment);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_creation_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","creation_date", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->creation_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_creation_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","creation_date", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  result =  ((arg1)->creation_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_expiration_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","expiration_date", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->expiration_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_t_expiration_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *arg1 = (struct svn_lock_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_lock_t *","expiration_date", 1, self )); 
  }
  arg1 = (struct svn_lock_t *)(argp1);
  result =  ((arg1)->expiration_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_lock_t_allocate(VALUE self)
#else
_wrap_svn_lock_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_lock_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_lock_t(int argc, VALUE *argv, VALUE self) {
  struct svn_lock_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_lock_t *)calloc(1, sizeof(struct svn_lock_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_lock_t(void *self) {
    struct svn_lock_t *arg1 = (struct svn_lock_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_lock_create(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_lock_t *)svn_lock_create(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_lock_dup(int argc, VALUE *argv, VALUE self) {
  svn_lock_t *arg1 = (svn_lock_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_lock_t const *","svn_lock_dup", 1, argv[0] )); 
  }
  arg1 = (svn_lock_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_lock_t *)svn_lock_dup((struct svn_lock_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uuid_generate(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (char *)svn_uuid_generate(arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_merge_range_t;

SWIGINTERN VALUE
_wrap_svn_merge_range_t_start_set(int argc, VALUE *argv, VALUE self) {
  struct svn_merge_range_t *arg1 = (struct svn_merge_range_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_merge_range_t *","start", 1, self )); 
  }
  arg1 = (struct svn_merge_range_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","start", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->start = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_merge_range_t_start_get(int argc, VALUE *argv, VALUE self) {
  struct svn_merge_range_t *arg1 = (struct svn_merge_range_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_merge_range_t *","start", 1, self )); 
  }
  arg1 = (struct svn_merge_range_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->start);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_merge_range_t_end_set(int argc, VALUE *argv, VALUE self) {
  struct svn_merge_range_t *arg1 = (struct svn_merge_range_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_merge_range_t *","end", 1, self )); 
  }
  arg1 = (struct svn_merge_range_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","end", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->end = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_merge_range_t_end_get(int argc, VALUE *argv, VALUE self) {
  struct svn_merge_range_t *arg1 = (struct svn_merge_range_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_merge_range_t *","end", 1, self )); 
  }
  arg1 = (struct svn_merge_range_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->end);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_merge_range_t_inheritable_set(int argc, VALUE *argv, VALUE self) {
  struct svn_merge_range_t *arg1 = (struct svn_merge_range_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_merge_range_t *","inheritable", 1, self )); 
  }
  arg1 = (struct svn_merge_range_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->inheritable = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_merge_range_t_inheritable_get(int argc, VALUE *argv, VALUE self) {
  struct svn_merge_range_t *arg1 = (struct svn_merge_range_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_merge_range_t *","inheritable", 1, self )); 
  }
  arg1 = (struct svn_merge_range_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->inheritable);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_merge_range_t_allocate(VALUE self)
#else
_wrap_svn_merge_range_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_merge_range_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_merge_range_t(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t arg1 ;
  svn_revnum_t arg2 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  long val1 ;
  int ecode1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  struct svn_merge_range_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_long(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "svn_revnum_t","svn_merge_range_t", 1, argv[0] ));
  } 
  arg1 = (svn_revnum_t)(val1);
  ecode2 = SWIG_AsVal_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","svn_merge_range_t", 2, argv[1] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  arg3 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (struct svn_merge_range_t *)new_svn_merge_range_t(arg1,arg2,arg3,arg4);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_merge_range_t(struct svn_merge_range_t *self){
  }
SWIGINTERN void
free_svn_merge_range_t(void *self) {
    struct svn_merge_range_t *arg1 = (struct svn_merge_range_t *)self;
    delete_svn_merge_range_t(arg1);
}


/*
  Document-method: Svn::Ext::Core::svn_merge_range_t.dup

  call-seq:
    dup(pool) -> svn_merge_range_t

Create a duplicate of the class and unfreeze it if needed.
*/
SWIGINTERN VALUE
_wrap_svn_merge_range_t_dup(int argc, VALUE *argv, VALUE self) {
  struct svn_merge_range_t *arg1 = (struct svn_merge_range_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_merge_range_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_merge_range_t *","dup", 1, self )); 
  }
  arg1 = (struct svn_merge_range_t *)(argp1);
  if (argc > 0) {
    
  }
  {
    result = (svn_merge_range_t *)svn_merge_range_t_dup(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_merge_range_dup(int argc, VALUE *argv, VALUE self) {
  svn_merge_range_t *arg1 = (svn_merge_range_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_merge_range_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_merge_range_t const *","svn_merge_range_dup", 1, argv[0] )); 
  }
  arg1 = (svn_merge_range_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_merge_range_t *)svn_merge_range_dup((struct svn_merge_range_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_merge_range_contains_rev(int argc, VALUE *argv, VALUE self) {
  svn_merge_range_t *arg1 = (svn_merge_range_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_merge_range_t const *","svn_merge_range_contains_rev", 1, argv[0] )); 
  }
  arg1 = (svn_merge_range_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","svn_merge_range_contains_rev", 2, argv[1] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  {
    result = (svn_boolean_t)svn_merge_range_contains_rev((struct svn_merge_range_t const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_location_segment_t;

SWIGINTERN VALUE
_wrap_svn_location_segment_t_range_start_set(int argc, VALUE *argv, VALUE self) {
  struct svn_location_segment_t *arg1 = (struct svn_location_segment_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_location_segment_t *","range_start", 1, self )); 
  }
  arg1 = (struct svn_location_segment_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","range_start", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->range_start = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_location_segment_t_range_start_get(int argc, VALUE *argv, VALUE self) {
  struct svn_location_segment_t *arg1 = (struct svn_location_segment_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_location_segment_t *","range_start", 1, self )); 
  }
  arg1 = (struct svn_location_segment_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->range_start);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_location_segment_t_range_end_set(int argc, VALUE *argv, VALUE self) {
  struct svn_location_segment_t *arg1 = (struct svn_location_segment_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_location_segment_t *","range_end", 1, self )); 
  }
  arg1 = (struct svn_location_segment_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","range_end", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->range_end = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_location_segment_t_range_end_get(int argc, VALUE *argv, VALUE self) {
  struct svn_location_segment_t *arg1 = (struct svn_location_segment_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_location_segment_t *","range_end", 1, self )); 
  }
  arg1 = (struct svn_location_segment_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->range_end);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_location_segment_t_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_location_segment_t *arg1 = (struct svn_location_segment_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_location_segment_t *","path", 1, self )); 
  }
  arg1 = (struct svn_location_segment_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_location_segment_t_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_location_segment_t *arg1 = (struct svn_location_segment_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_location_segment_t *","path", 1, self )); 
  }
  arg1 = (struct svn_location_segment_t *)(argp1);
  result = (char *) ((arg1)->path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_location_segment_t_allocate(VALUE self)
#else
_wrap_svn_location_segment_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_location_segment_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_location_segment_t(int argc, VALUE *argv, VALUE self) {
  struct svn_location_segment_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_location_segment_t *)calloc(1, sizeof(struct svn_location_segment_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_location_segment_t(void *self) {
    struct svn_location_segment_t *arg1 = (struct svn_location_segment_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_location_segment_dup(int argc, VALUE *argv, VALUE self) {
  svn_location_segment_t *arg1 = (svn_location_segment_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_location_segment_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_location_segment_t const *","svn_location_segment_dup", 1, argv[0] )); 
  }
  arg1 = (svn_location_segment_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_location_segment_t *)svn_location_segment_dup((struct svn_location_segment_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_invoke_entry_receiver(int argc, VALUE *argv, VALUE self) {
  svn_log_entry_receiver_t arg1 = (svn_log_entry_receiver_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_log_entry_t *arg3 = (svn_log_entry_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_log_entry_receiver_t","svn_log_invoke_entry_receiver", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_log_invoke_entry_receiver", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_log_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_log_entry_t *","svn_log_invoke_entry_receiver", 3, argv[2] )); 
  }
  arg3 = (svn_log_entry_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_log_invoke_entry_receiver(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_log_invoke_message_receiver(int argc, VALUE *argv, VALUE self) {
  svn_log_message_receiver_t arg1 = (svn_log_message_receiver_t) 0 ;
  void *arg2 = (void *) 0 ;
  apr_hash_t *arg3 = (apr_hash_t *) 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_log_message_receiver_t","svn_log_invoke_message_receiver", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_log_invoke_message_receiver", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_hash_t *","svn_log_invoke_message_receiver", 3, argv[2] )); 
  }
  arg3 = (apr_hash_t *)(argp3);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_log_invoke_message_receiver", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_log_invoke_message_receiver", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_log_invoke_message_receiver", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_log_invoke_message_receiver", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_log_invoke_message_receiver(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_invoke_callback2(int argc, VALUE *argv, VALUE self) {
  svn_commit_callback2_t arg1 = (svn_commit_callback2_t) 0 ;
  svn_commit_info_t *arg2 = (svn_commit_info_t *) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_commit_callback2_t","svn_commit_invoke_callback2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_commit_info_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_commit_info_t const *","svn_commit_invoke_callback2", 2, argv[1] )); 
  }
  arg2 = (svn_commit_info_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_commit_invoke_callback2", 3, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_commit_invoke_callback2(arg1,(struct svn_commit_info_t const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_commit_invoke_callback(int argc, VALUE *argv, VALUE self) {
  svn_commit_callback_t arg1 = (svn_commit_callback_t) 0 ;
  svn_revnum_t arg2 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  void *arg5 = (void *) 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 5) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_commit_callback_t","svn_commit_invoke_callback", 1, argv[0] )); 
    }
  }
  ecode2 = SWIG_AsVal_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","svn_commit_invoke_callback", 2, argv[1] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_commit_invoke_callback", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_commit_invoke_callback", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[4],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_commit_invoke_callback", 5, argv[4] )); 
  }
  {
    result = (svn_error_t *)svn_commit_invoke_callback(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cancel_invoke_func(int argc, VALUE *argv, VALUE self) {
  svn_cancel_func_t arg1 = (svn_cancel_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  int res2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_cancel_func_t","svn_cancel_invoke_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_cancel_invoke_func", 2, argv[1] )); 
  }
  {
    result = (svn_error_t *)svn_cancel_invoke_func(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_location_invoke_segment_receiver(int argc, VALUE *argv, VALUE self) {
  svn_location_segment_receiver_t arg1 = (svn_location_segment_receiver_t) 0 ;
  svn_location_segment_t *arg2 = (svn_location_segment_t *) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_location_segment_receiver_t","svn_location_invoke_segment_receiver", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_location_segment_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_location_segment_t *","svn_location_invoke_segment_receiver", 2, argv[1] )); 
  }
  arg2 = (svn_location_segment_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_location_invoke_segment_receiver", 3, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_location_invoke_segment_receiver(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_pool_create(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  apr_allocator_t *arg2 = (apr_allocator_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  apr_pool_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    arg2 = NULL;
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_pool_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_pool_t *","svn_pool_create_ex", 1, argv[0] )); 
  }
  arg1 = (apr_pool_t *)(argp1);
  if (argc > 1) {
    res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_apr_allocator_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_allocator_t *","svn_pool_create_ex", 2, argv[1] )); 
    }
    arg2 = (apr_allocator_t *)(argp2);
  }
  {
    result = (apr_pool_t *)svn_pool_create_ex(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_pool_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_pool_create_allocator(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t arg1 ;
  apr_allocator_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  arg1 = RTEST(argv[0]);
  {
    result = (apr_allocator_t *)svn_pool_create_allocator(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_allocator_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_version_t;

SWIGINTERN VALUE
_wrap_svn_version_t_major_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *arg1 = (struct svn_version_t *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_t *","major", 1, self )); 
  }
  arg1 = (struct svn_version_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","major", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->major = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_t_major_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *arg1 = (struct svn_version_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_t *","major", 1, self )); 
  }
  arg1 = (struct svn_version_t *)(argp1);
  result = (int) ((arg1)->major);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_t_minor_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *arg1 = (struct svn_version_t *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_t *","minor", 1, self )); 
  }
  arg1 = (struct svn_version_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","minor", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->minor = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_t_minor_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *arg1 = (struct svn_version_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_t *","minor", 1, self )); 
  }
  arg1 = (struct svn_version_t *)(argp1);
  result = (int) ((arg1)->minor);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_t_patch_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *arg1 = (struct svn_version_t *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_t *","patch", 1, self )); 
  }
  arg1 = (struct svn_version_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","patch", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->patch = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_t_patch_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *arg1 = (struct svn_version_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_t *","patch", 1, self )); 
  }
  arg1 = (struct svn_version_t *)(argp1);
  result = (int) ((arg1)->patch);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_t_tag_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *arg1 = (struct svn_version_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_t *","tag", 1, self )); 
  }
  arg1 = (struct svn_version_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","tag", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->tag) free((char *)arg1->tag);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->tag = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_t_tag_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *arg1 = (struct svn_version_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_t *","tag", 1, self )); 
  }
  arg1 = (struct svn_version_t *)(argp1);
  result = (char *) ((arg1)->tag);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_version_t_allocate(VALUE self)
#else
_wrap_svn_version_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_version_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_version_t(int argc, VALUE *argv, VALUE self) {
  struct svn_version_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_version_t *)calloc(1, sizeof(struct svn_version_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_version_t(void *self) {
    struct svn_version_t *arg1 = (struct svn_version_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_ver_compatible(int argc, VALUE *argv, VALUE self) {
  svn_version_t *arg1 = (svn_version_t *) 0 ;
  svn_version_t *arg2 = (svn_version_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_t const *","svn_ver_compatible", 1, argv[0] )); 
  }
  arg1 = (svn_version_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_version_t const *","svn_ver_compatible", 2, argv[1] )); 
  }
  arg2 = (svn_version_t *)(argp2);
  {
    result = (svn_boolean_t)svn_ver_compatible((struct svn_version_t const *)arg1,(struct svn_version_t const *)arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_ver_equal(int argc, VALUE *argv, VALUE self) {
  svn_version_t *arg1 = (svn_version_t *) 0 ;
  svn_version_t *arg2 = (svn_version_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_t const *","svn_ver_equal", 1, argv[0] )); 
  }
  arg1 = (svn_version_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_version_t const *","svn_ver_equal", 2, argv[1] )); 
  }
  arg2 = (svn_version_t *)(argp2);
  {
    result = (svn_boolean_t)svn_ver_equal((struct svn_version_t const *)arg1,(struct svn_version_t const *)arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_version_checklist_t;

SWIGINTERN VALUE
_wrap_svn_version_checklist_t_label_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_checklist_t *arg1 = (struct svn_version_checklist_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_checklist_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_checklist_t *","label", 1, self )); 
  }
  arg1 = (struct svn_version_checklist_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","label", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->label) free((char *)arg1->label);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->label = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_checklist_t_label_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_checklist_t *arg1 = (struct svn_version_checklist_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_checklist_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_checklist_t *","label", 1, self )); 
  }
  arg1 = (struct svn_version_checklist_t *)(argp1);
  result = (char *) ((arg1)->label);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_checklist_t_version_query_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_checklist_t *arg1 = (struct svn_version_checklist_t *) 0 ;
  svn_version_t *(*arg2)(void) = (svn_version_t *(*)(void)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_checklist_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_checklist_t *","version_query", 1, self )); 
  }
  arg1 = (struct svn_version_checklist_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_void__p_svn_version_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_version_t const *(*)(void)","version_query", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->version_query = (svn_version_t const *(*)(void))arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_checklist_t_version_query_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_checklist_t *arg1 = (struct svn_version_checklist_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_version_t *(*result)(void) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_checklist_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_checklist_t *","version_query", 1, self )); 
  }
  arg1 = (struct svn_version_checklist_t *)(argp1);
  result = (svn_version_t *(*)(void)) ((arg1)->version_query);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_void__p_svn_version_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_version_checklist_t_allocate(VALUE self)
#else
_wrap_svn_version_checklist_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_version_checklist_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_version_checklist_t(int argc, VALUE *argv, VALUE self) {
  struct svn_version_checklist_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_version_checklist_t *)calloc(1, sizeof(struct svn_version_checklist_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_version_checklist_t(void *self) {
    struct svn_version_checklist_t *arg1 = (struct svn_version_checklist_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_ver_check_list2(int argc, VALUE *argv, VALUE self) {
  svn_version_t *arg1 = (svn_version_t *) 0 ;
  svn_version_checklist_t *arg2 = (svn_version_checklist_t *) 0 ;
  svn_boolean_t (*arg3)(svn_version_t const *,svn_version_t const *) = (svn_boolean_t (*)(svn_version_t const *,svn_version_t const *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_t const *","svn_ver_check_list2", 1, argv[0] )); 
  }
  arg1 = (svn_version_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_version_checklist_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_version_checklist_t const *","svn_ver_check_list2", 2, argv[1] )); 
  }
  arg2 = (svn_version_checklist_t *)(argp2);
  {
    int res = SWIG_ConvertFunctionPtr(argv[2], (void**)(&arg3), SWIGTYPE_p_f_p_q_const__struct_svn_version_t_p_q_const__struct_svn_version_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_boolean_t (*)(svn_version_t const *,svn_version_t const *)","svn_ver_check_list2", 3, argv[2] )); 
    }
  }
  {
    result = (svn_error_t *)svn_ver_check_list2((struct svn_version_t const *)arg1,(struct svn_version_checklist_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_ver_check_list(int argc, VALUE *argv, VALUE self) {
  svn_version_t *arg1 = (svn_version_t *) 0 ;
  svn_version_checklist_t *arg2 = (svn_version_checklist_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_t const *","svn_ver_check_list", 1, argv[0] )); 
  }
  arg1 = (svn_version_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_version_checklist_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_version_checklist_t const *","svn_ver_check_list", 2, argv[1] )); 
  }
  arg2 = (svn_version_checklist_t *)(argp2);
  {
    result = (svn_error_t *)svn_ver_check_list((struct svn_version_t const *)arg1,(struct svn_version_checklist_t const *)arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_subr_version(int argc, VALUE *argv, VALUE self) {
  svn_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (svn_version_t *)svn_subr_version();
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_version_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_extended(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_version_extended_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  arg1 = RTEST(argv[0]);
  if (argc > 1) {
    
  }
  {
    result = (svn_version_extended_t *)svn_version_extended(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_build_date(int argc, VALUE *argv, VALUE self) {
  svn_version_extended_t *arg1 = (svn_version_extended_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_extended_t const *","svn_version_ext_build_date", 1, argv[0] )); 
  }
  arg1 = (svn_version_extended_t *)(argp1);
  {
    result = (char *)svn_version_ext_build_date((struct svn_version_extended_t const *)arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_build_time(int argc, VALUE *argv, VALUE self) {
  svn_version_extended_t *arg1 = (svn_version_extended_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_extended_t const *","svn_version_ext_build_time", 1, argv[0] )); 
  }
  arg1 = (svn_version_extended_t *)(argp1);
  {
    result = (char *)svn_version_ext_build_time((struct svn_version_extended_t const *)arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_build_host(int argc, VALUE *argv, VALUE self) {
  svn_version_extended_t *arg1 = (svn_version_extended_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_extended_t const *","svn_version_ext_build_host", 1, argv[0] )); 
  }
  arg1 = (svn_version_extended_t *)(argp1);
  {
    result = (char *)svn_version_ext_build_host((struct svn_version_extended_t const *)arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_copyright(int argc, VALUE *argv, VALUE self) {
  svn_version_extended_t *arg1 = (svn_version_extended_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_extended_t const *","svn_version_ext_copyright", 1, argv[0] )); 
  }
  arg1 = (svn_version_extended_t *)(argp1);
  {
    result = (char *)svn_version_ext_copyright((struct svn_version_extended_t const *)arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_runtime_host(int argc, VALUE *argv, VALUE self) {
  svn_version_extended_t *arg1 = (svn_version_extended_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_extended_t const *","svn_version_ext_runtime_host", 1, argv[0] )); 
  }
  arg1 = (svn_version_extended_t *)(argp1);
  {
    result = (char *)svn_version_ext_runtime_host((struct svn_version_extended_t const *)arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_runtime_osname(int argc, VALUE *argv, VALUE self) {
  svn_version_extended_t *arg1 = (svn_version_extended_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_extended_t const *","svn_version_ext_runtime_osname", 1, argv[0] )); 
  }
  arg1 = (svn_version_extended_t *)(argp1);
  {
    result = (char *)svn_version_ext_runtime_osname((struct svn_version_extended_t const *)arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_version_ext_linked_lib_t;

SWIGINTERN VALUE
_wrap_svn_version_ext_linked_lib_t_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_linked_lib_t *arg1 = (struct svn_version_ext_linked_lib_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_linked_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_linked_lib_t *","name", 1, self )); 
  }
  arg1 = (struct svn_version_ext_linked_lib_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_linked_lib_t_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_linked_lib_t *arg1 = (struct svn_version_ext_linked_lib_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_linked_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_linked_lib_t *","name", 1, self )); 
  }
  arg1 = (struct svn_version_ext_linked_lib_t *)(argp1);
  result = (char *) ((arg1)->name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_linked_lib_t_compiled_version_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_linked_lib_t *arg1 = (struct svn_version_ext_linked_lib_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_linked_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_linked_lib_t *","compiled_version", 1, self )); 
  }
  arg1 = (struct svn_version_ext_linked_lib_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","compiled_version", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->compiled_version) free((char *)arg1->compiled_version);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->compiled_version = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_linked_lib_t_compiled_version_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_linked_lib_t *arg1 = (struct svn_version_ext_linked_lib_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_linked_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_linked_lib_t *","compiled_version", 1, self )); 
  }
  arg1 = (struct svn_version_ext_linked_lib_t *)(argp1);
  result = (char *) ((arg1)->compiled_version);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_linked_lib_t_runtime_version_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_linked_lib_t *arg1 = (struct svn_version_ext_linked_lib_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_linked_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_linked_lib_t *","runtime_version", 1, self )); 
  }
  arg1 = (struct svn_version_ext_linked_lib_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","runtime_version", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->runtime_version) free((char *)arg1->runtime_version);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->runtime_version = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_linked_lib_t_runtime_version_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_linked_lib_t *arg1 = (struct svn_version_ext_linked_lib_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_linked_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_linked_lib_t *","runtime_version", 1, self )); 
  }
  arg1 = (struct svn_version_ext_linked_lib_t *)(argp1);
  result = (char *) ((arg1)->runtime_version);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_version_ext_linked_lib_t_allocate(VALUE self)
#else
_wrap_svn_version_ext_linked_lib_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_version_ext_linked_lib_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_version_ext_linked_lib_t(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_linked_lib_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_version_ext_linked_lib_t *)calloc(1, sizeof(struct svn_version_ext_linked_lib_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_version_ext_linked_lib_t(void *self) {
    struct svn_version_ext_linked_lib_t *arg1 = (struct svn_version_ext_linked_lib_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_version_ext_linked_libs(int argc, VALUE *argv, VALUE self) {
  svn_version_extended_t *arg1 = (svn_version_extended_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_extended_t const *","svn_version_ext_linked_libs", 1, argv[0] )); 
  }
  arg1 = (svn_version_extended_t *)(argp1);
  {
    result = (apr_array_header_t *)svn_version_ext_linked_libs((struct svn_version_extended_t const *)arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_version_ext_loaded_lib_t;

SWIGINTERN VALUE
_wrap_svn_version_ext_loaded_lib_t_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_loaded_lib_t *arg1 = (struct svn_version_ext_loaded_lib_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_loaded_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_loaded_lib_t *","name", 1, self )); 
  }
  arg1 = (struct svn_version_ext_loaded_lib_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_loaded_lib_t_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_loaded_lib_t *arg1 = (struct svn_version_ext_loaded_lib_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_loaded_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_loaded_lib_t *","name", 1, self )); 
  }
  arg1 = (struct svn_version_ext_loaded_lib_t *)(argp1);
  result = (char *) ((arg1)->name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_loaded_lib_t_version_set(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_loaded_lib_t *arg1 = (struct svn_version_ext_loaded_lib_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_loaded_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_loaded_lib_t *","version", 1, self )); 
  }
  arg1 = (struct svn_version_ext_loaded_lib_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","version", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->version) free((char *)arg1->version);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->version = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_ext_loaded_lib_t_version_get(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_loaded_lib_t *arg1 = (struct svn_version_ext_loaded_lib_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_version_ext_loaded_lib_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_version_ext_loaded_lib_t *","version", 1, self )); 
  }
  arg1 = (struct svn_version_ext_loaded_lib_t *)(argp1);
  result = (char *) ((arg1)->version);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_version_ext_loaded_lib_t_allocate(VALUE self)
#else
_wrap_svn_version_ext_loaded_lib_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_version_ext_loaded_lib_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_version_ext_loaded_lib_t(int argc, VALUE *argv, VALUE self) {
  struct svn_version_ext_loaded_lib_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_version_ext_loaded_lib_t *)calloc(1, sizeof(struct svn_version_ext_loaded_lib_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_version_ext_loaded_lib_t(void *self) {
    struct svn_version_ext_loaded_lib_t *arg1 = (struct svn_version_ext_loaded_lib_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_version_ext_loaded_libs(int argc, VALUE *argv, VALUE self) {
  svn_version_extended_t *arg1 = (svn_version_extended_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_extended_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_extended_t const *","svn_version_ext_loaded_libs", 1, argv[0] )); 
  }
  arg1 = (svn_version_extended_t *)(argp1);
  {
    result = (apr_array_header_t *)svn_version_ext_loaded_libs((struct svn_version_extended_t const *)arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_checklist_invoke_version_query(int argc, VALUE *argv, VALUE self) {
  svn_version_checklist_t *arg1 = (svn_version_checklist_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_version_checklist_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_version_checklist_t *","svn_version_checklist_invoke_version_query", 1, argv[0] )); 
  }
  arg1 = (svn_version_checklist_t *)(argp1);
  {
    result = (svn_version_t *)svn_version_checklist_invoke_version_query(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_version_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_version_invoke_func(int argc, VALUE *argv, VALUE self) {
  svn_version_func_t arg1 = (svn_version_func_t) 0 ;
  svn_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_void__p_svn_version_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_version_func_t","svn_version_invoke_func", 1, argv[0] )); 
    }
  }
  {
    result = (svn_version_t *)svn_version_invoke_func((struct svn_version_t const *(*)(void))arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_version_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_dup(int argc, VALUE *argv, VALUE self) {
  svn_prop_t *arg1 = (svn_prop_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_prop_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_prop_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_prop_t const *","svn_prop_dup", 1, argv[0] )); 
  }
  arg1 = (svn_prop_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_prop_t *)svn_prop_dup((struct svn_prop_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_prop_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_array_dup(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_prop_array_dup", 1, argv[0] )); 
  }
  arg1 = (apr_array_header_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (apr_array_header_t *)svn_prop_array_dup((apr_array_header_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_prop_inherited_item_t;

SWIGINTERN VALUE
_wrap_svn_prop_inherited_item_t_path_or_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_inherited_item_t *arg1 = (struct svn_prop_inherited_item_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_inherited_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_inherited_item_t *","path_or_url", 1, self )); 
  }
  arg1 = (struct svn_prop_inherited_item_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path_or_url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path_or_url) free((char *)arg1->path_or_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path_or_url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_inherited_item_t_path_or_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_inherited_item_t *arg1 = (struct svn_prop_inherited_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_inherited_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_inherited_item_t *","path_or_url", 1, self )); 
  }
  arg1 = (struct svn_prop_inherited_item_t *)(argp1);
  result = (char *) ((arg1)->path_or_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_inherited_item_t_prop_hash_set(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_inherited_item_t *arg1 = (struct svn_prop_inherited_item_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_inherited_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_inherited_item_t *","prop_hash", 1, self )); 
  }
  arg1 = (struct svn_prop_inherited_item_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","prop_hash", 2, argv[0] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  if (arg1) (arg1)->prop_hash = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_inherited_item_t_prop_hash_get(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_inherited_item_t *arg1 = (struct svn_prop_inherited_item_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_inherited_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_inherited_item_t *","prop_hash", 1, self )); 
  }
  arg1 = (struct svn_prop_inherited_item_t *)(argp1);
  result = (apr_hash_t *) ((arg1)->prop_hash);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_prop_inherited_item_t_allocate(VALUE self)
#else
_wrap_svn_prop_inherited_item_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_prop_inherited_item_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_prop_inherited_item_t(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_inherited_item_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_prop_inherited_item_t *)calloc(1, sizeof(struct svn_prop_inherited_item_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_prop_inherited_item_t(void *self) {
    struct svn_prop_inherited_item_t *arg1 = (struct svn_prop_inherited_item_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_prop_hash_to_array(int argc, VALUE *argv, VALUE self) {
  apr_hash_t *arg1 = (apr_hash_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg1 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[0], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg1)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[0], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (apr_array_header_t *)svn_prop_hash_to_array((apr_hash_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_array_to_hash(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_prop_array_to_hash", 1, argv[0] )); 
  }
  arg1 = (apr_array_header_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_apr_pool_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_pool_t *","svn_prop_array_to_hash", 2, argv[1] )); 
  }
  arg2 = (apr_pool_t *)(argp2);
  {
    result = (apr_hash_t *)svn_prop_array_to_hash((apr_array_header_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_hash_dup(int argc, VALUE *argv, VALUE self) {
  apr_hash_t *arg1 = (apr_hash_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg1 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[0], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg1)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[0], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (apr_hash_t *)svn_prop_hash_dup((apr_hash_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_get_value(int argc, VALUE *argv, VALUE self) {
  apr_hash_t *arg1 = (apr_hash_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_hash_t const *","svn_prop_get_value", 1, argv[0] )); 
  }
  arg1 = (apr_hash_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_prop_get_value", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (char *)svn_prop_get_value((apr_hash_t const *)arg1,(char const *)arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_property_kind2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_prop_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_property_kind2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_prop_kind_t)svn_property_kind2((char const *)arg1);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_property_kind(int argc, VALUE *argv, VALUE self) {
  int *arg1 = (int *) 0 ;
  char *arg2 = (char *) 0 ;
  int temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_prop_kind_t result;
  VALUE vresult = Qnil;
  
  arg1 = &temp1;
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_property_kind", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_prop_kind_t)svn_property_kind(arg1,(char const *)arg2);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_is_svn_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_prop_is_svn_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_prop_is_svn_prop((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_has_svn_prop(int argc, VALUE *argv, VALUE self) {
  apr_hash_t *arg1 = (apr_hash_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg1 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[0], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg1)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[0], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_boolean_t)svn_prop_has_svn_prop((apr_hash_t const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_is_boolean(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_prop_is_boolean", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_prop_is_boolean((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_is_known_svn_rev_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_prop_is_known_svn_rev_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_prop_is_known_svn_rev_prop((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_is_known_svn_node_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_prop_is_known_svn_node_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_prop_is_known_svn_node_prop((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_is_known_svn_file_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_prop_is_known_svn_file_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_prop_is_known_svn_file_prop((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_is_known_svn_dir_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_prop_is_known_svn_dir_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_prop_is_known_svn_dir_prop((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_needs_translation(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_prop_needs_translation", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_prop_needs_translation((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_categorize_props(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  apr_array_header_t **arg2 = (apr_array_header_t **) 0 ;
  apr_array_header_t **arg3 = (apr_array_header_t **) 0 ;
  apr_array_header_t **arg4 = (apr_array_header_t **) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp2 ;
  apr_array_header_t *temp3 ;
  apr_array_header_t *temp4 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_to_apr_array_row_prop(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_categorize_props((apr_array_header_t const *)arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_prop_apr_array_to_hash_prop(*arg2));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_prop_apr_array_to_hash_prop(*arg3));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_prop_apr_array_to_hash_prop(*arg4));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_diffs(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  apr_hash_t *arg3 = (apr_hash_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg2 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[0], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg2)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[0], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg3 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[1], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg3)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[1], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_prop_diffs(arg1,(apr_hash_t const *)arg2,(apr_hash_t const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_prop_apr_array_to_hash_prop(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_name_is_valid(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_prop_name_is_valid", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_prop_name_is_valid((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


static swig_class SwigClassSvn_opt_subcommand_desc3_t;

SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","name", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","name", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  result = (char *) ((arg1)->name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_cmd_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  svn_opt_subcommand_t *arg2 = (svn_opt_subcommand_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","cmd_func", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_opt_subcommand_t *","cmd_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->cmd_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_cmd_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_subcommand_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","cmd_func", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  result = (svn_opt_subcommand_t *) ((arg1)->cmd_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_aliases_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  char **arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","aliases", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *[3]","aliases", 2, argv[0] )); 
  } 
  arg2 = (char **)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)3; ++ii) *(char const * *)&arg1->aliases[ii] = *((char const * *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""aliases""' of type '""char const *[3]""'");
    }
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_aliases_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char **result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","aliases", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  result = (char **)(char **) ((arg1)->aliases);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_char, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_help_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  char **arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","help", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *[100]","help", 2, argv[0] )); 
  } 
  arg2 = (char **)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)100; ++ii) *(char const * *)&arg1->help[ii] = *((char const * *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""help""' of type '""char const *[100]""'");
    }
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_help_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char **result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","help", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  result = (char **)(char **) ((arg1)->help);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_char, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_valid_options_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  int *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","valid_options", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "int [50]","valid_options", 2, argv[0] )); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)50; ++ii) *(int *)&arg1->valid_options[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""valid_options""' of type '""int [50]""'");
    }
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_valid_options_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","valid_options", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  result = (int *)(int *) ((arg1)->valid_options);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_desc_overrides_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_subcommand_desc3_t_desc_overrides *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc3_t *","desc_overrides", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc3_t *)(argp1);
  result = (svn_opt_subcommand_desc3_t_desc_overrides *)(svn_opt_subcommand_desc3_t_desc_overrides *) ((arg1)->desc_overrides);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_subcommand_desc3_t_desc_overrides, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_opt_subcommand_desc3_t_allocate(VALUE self)
#else
_wrap_svn_opt_subcommand_desc3_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_opt_subcommand_desc3_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_opt_subcommand_desc3_t(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc3_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_opt_subcommand_desc3_t *)calloc(1, sizeof(struct svn_opt_subcommand_desc3_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_opt_subcommand_desc3_t(void *self) {
    struct svn_opt_subcommand_desc3_t *arg1 = (struct svn_opt_subcommand_desc3_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_opt_subcommand_desc3_t_desc_overrides;

SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_desc_overrides_optch_set(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc3_t_desc_overrides *arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t_desc_overrides, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t_desc_overrides *","optch", 1, self )); 
  }
  arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","optch", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->optch = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_desc_overrides_optch_get(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc3_t_desc_overrides *arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t_desc_overrides, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t_desc_overrides *","optch", 1, self )); 
  }
  arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *)(argp1);
  result = (int) ((arg1)->optch);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_desc_overrides_desc_set(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc3_t_desc_overrides *arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t_desc_overrides, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t_desc_overrides *","desc", 1, self )); 
  }
  arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","desc", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->desc) free((char *)arg1->desc);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->desc = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc3_t_desc_overrides_desc_get(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc3_t_desc_overrides *arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t_desc_overrides, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t_desc_overrides *","desc", 1, self )); 
  }
  arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *)(argp1);
  result = (char *) ((arg1)->desc);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_opt_subcommand_desc3_t_desc_overrides_allocate(VALUE self)
#else
_wrap_svn_opt_subcommand_desc3_t_desc_overrides_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_opt_subcommand_desc3_t_desc_overrides);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_opt_subcommand_desc3_t_desc_overrides(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc3_t_desc_overrides *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (svn_opt_subcommand_desc3_t_desc_overrides *)calloc(1, sizeof(svn_opt_subcommand_desc3_t_desc_overrides));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_opt_subcommand_desc3_t_desc_overrides(void *self) {
    svn_opt_subcommand_desc3_t_desc_overrides *arg1 = (svn_opt_subcommand_desc3_t_desc_overrides *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_opt_subcommand_desc2_t;

SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","name", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","name", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  result = (char *) ((arg1)->name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_cmd_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  svn_opt_subcommand_t *arg2 = (svn_opt_subcommand_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","cmd_func", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_opt_subcommand_t *","cmd_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->cmd_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_cmd_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_subcommand_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","cmd_func", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  result = (svn_opt_subcommand_t *) ((arg1)->cmd_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_aliases_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  char **arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","aliases", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *[3]","aliases", 2, argv[0] )); 
  } 
  arg2 = (char **)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)3; ++ii) *(char const * *)&arg1->aliases[ii] = *((char const * *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""aliases""' of type '""char const *[3]""'");
    }
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_aliases_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char **result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","aliases", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  result = (char **)(char **) ((arg1)->aliases);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_char, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_help_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","help", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","help", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->help) free((char *)arg1->help);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->help = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_help_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","help", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  result = (char *) ((arg1)->help);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_valid_options_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  int *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","valid_options", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "int [50]","valid_options", 2, argv[0] )); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)50; ++ii) *(int *)&arg1->valid_options[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""valid_options""' of type '""int [50]""'");
    }
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_valid_options_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","valid_options", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  result = (int *)(int *) ((arg1)->valid_options);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_desc_overrides_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_subcommand_desc2_t_desc_overrides *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc2_t *","desc_overrides", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc2_t *)(argp1);
  result = (svn_opt_subcommand_desc2_t_desc_overrides *)(svn_opt_subcommand_desc2_t_desc_overrides *) ((arg1)->desc_overrides);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_subcommand_desc2_t_desc_overrides, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_opt_subcommand_desc2_t_allocate(VALUE self)
#else
_wrap_svn_opt_subcommand_desc2_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_opt_subcommand_desc2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_opt_subcommand_desc2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_opt_subcommand_desc2_t *)calloc(1, sizeof(struct svn_opt_subcommand_desc2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_opt_subcommand_desc2_t(void *self) {
    struct svn_opt_subcommand_desc2_t *arg1 = (struct svn_opt_subcommand_desc2_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_opt_subcommand_desc2_t_desc_overrides;

SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_desc_overrides_optch_set(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc2_t_desc_overrides *arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t_desc_overrides, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t_desc_overrides *","optch", 1, self )); 
  }
  arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","optch", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->optch = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_desc_overrides_optch_get(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc2_t_desc_overrides *arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t_desc_overrides, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t_desc_overrides *","optch", 1, self )); 
  }
  arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *)(argp1);
  result = (int) ((arg1)->optch);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_desc_overrides_desc_set(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc2_t_desc_overrides *arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t_desc_overrides, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t_desc_overrides *","desc", 1, self )); 
  }
  arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","desc", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->desc) free((char *)arg1->desc);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->desc = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc2_t_desc_overrides_desc_get(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc2_t_desc_overrides *arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t_desc_overrides, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t_desc_overrides *","desc", 1, self )); 
  }
  arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *)(argp1);
  result = (char *) ((arg1)->desc);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_opt_subcommand_desc2_t_desc_overrides_allocate(VALUE self)
#else
_wrap_svn_opt_subcommand_desc2_t_desc_overrides_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_opt_subcommand_desc2_t_desc_overrides);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_opt_subcommand_desc2_t_desc_overrides(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc2_t_desc_overrides *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (svn_opt_subcommand_desc2_t_desc_overrides *)calloc(1, sizeof(svn_opt_subcommand_desc2_t_desc_overrides));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_opt_subcommand_desc2_t_desc_overrides(void *self) {
    svn_opt_subcommand_desc2_t_desc_overrides *arg1 = (svn_opt_subcommand_desc2_t_desc_overrides *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_opt_subcommand_desc_t;

SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","name", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","name", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  result = (char *) ((arg1)->name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_cmd_func_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  svn_opt_subcommand_t *arg2 = (svn_opt_subcommand_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","cmd_func", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_opt_subcommand_t *","cmd_func", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->cmd_func = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_cmd_func_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_subcommand_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","cmd_func", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  result = (svn_opt_subcommand_t *) ((arg1)->cmd_func);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_aliases_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  char **arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","aliases", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *[3]","aliases", 2, argv[0] )); 
  } 
  arg2 = (char **)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)3; ++ii) *(char const * *)&arg1->aliases[ii] = *((char const * *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""aliases""' of type '""char const *[3]""'");
    }
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_aliases_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char **result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","aliases", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  result = (char **)(char **) ((arg1)->aliases);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_char, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_help_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","help", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","help", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->help) free((char *)arg1->help);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->help = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_help_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","help", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  result = (char *) ((arg1)->help);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_valid_options_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  int *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","valid_options", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "int [50]","valid_options", 2, argv[0] )); 
  } 
  arg2 = (int *)(argp2);
  {
    if (arg2) {
      size_t ii = 0;
      for (; ii < (size_t)50; ++ii) *(int *)&arg1->valid_options[ii] = *((int *)arg2 + ii);
    } else {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in variable '""valid_options""' of type '""int [50]""'");
    }
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_desc_t_valid_options_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_subcommand_desc_t *","valid_options", 1, self )); 
  }
  arg1 = (struct svn_opt_subcommand_desc_t *)(argp1);
  result = (int *)(int *) ((arg1)->valid_options);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_opt_subcommand_desc_t_allocate(VALUE self)
#else
_wrap_svn_opt_subcommand_desc_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_opt_subcommand_desc_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_opt_subcommand_desc_t(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_subcommand_desc_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_opt_subcommand_desc_t *)calloc(1, sizeof(struct svn_opt_subcommand_desc_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_opt_subcommand_desc_t(void *self) {
    struct svn_opt_subcommand_desc_t *arg1 = (struct svn_opt_subcommand_desc_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_opt_get_canonical_subcommand3(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc3_t *arg1 = (svn_opt_subcommand_desc3_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_subcommand_desc3_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t const *","svn_opt_get_canonical_subcommand3", 1, argv[0] )); 
  }
  arg1 = (svn_opt_subcommand_desc3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_get_canonical_subcommand3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_opt_subcommand_desc3_t *)svn_opt_get_canonical_subcommand3((struct svn_opt_subcommand_desc3_t const *)arg1,(char const *)arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_get_canonical_subcommand2(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc2_t *arg1 = (svn_opt_subcommand_desc2_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_subcommand_desc2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_get_canonical_subcommand2", 1, argv[0] )); 
  }
  arg1 = (svn_opt_subcommand_desc2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_get_canonical_subcommand2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_opt_subcommand_desc2_t *)svn_opt_get_canonical_subcommand2((struct svn_opt_subcommand_desc2_t const *)arg1,(char const *)arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_get_canonical_subcommand(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc_t *arg1 = (svn_opt_subcommand_desc_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_opt_subcommand_desc_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc_t const *","svn_opt_get_canonical_subcommand", 1, argv[0] )); 
  }
  arg1 = (svn_opt_subcommand_desc_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_get_canonical_subcommand", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_opt_subcommand_desc_t *)svn_opt_get_canonical_subcommand((struct svn_opt_subcommand_desc_t const *)arg1,(char const *)arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_get_option_from_code3(int argc, VALUE *argv, VALUE self) {
  int arg1 ;
  apr_getopt_option_t *arg2 = (apr_getopt_option_t *) 0 ;
  svn_opt_subcommand_desc3_t *arg3 = (svn_opt_subcommand_desc3_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  apr_getopt_option_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "int","svn_opt_get_option_from_code3", 1, argv[0] ));
  } 
  arg1 = (int)(val1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_get_option_from_code3", 2, argv[1] )); 
  }
  arg2 = (apr_getopt_option_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t const *","svn_opt_get_option_from_code3", 3, argv[2] )); 
  }
  arg3 = (svn_opt_subcommand_desc3_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    result = (apr_getopt_option_t *)svn_opt_get_option_from_code3(arg1,(apr_getopt_option_t const *)arg2,(struct svn_opt_subcommand_desc3_t const *)arg3,arg4);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_get_option_from_code2(int argc, VALUE *argv, VALUE self) {
  int arg1 ;
  apr_getopt_option_t *arg2 = (apr_getopt_option_t *) 0 ;
  svn_opt_subcommand_desc2_t *arg3 = (svn_opt_subcommand_desc2_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  apr_getopt_option_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "int","svn_opt_get_option_from_code2", 1, argv[0] ));
  } 
  arg1 = (int)(val1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_get_option_from_code2", 2, argv[1] )); 
  }
  arg2 = (apr_getopt_option_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_get_option_from_code2", 3, argv[2] )); 
  }
  arg3 = (svn_opt_subcommand_desc2_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    result = (apr_getopt_option_t *)svn_opt_get_option_from_code2(arg1,(apr_getopt_option_t const *)arg2,(struct svn_opt_subcommand_desc2_t const *)arg3,arg4);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_get_option_from_code(int argc, VALUE *argv, VALUE self) {
  int arg1 ;
  apr_getopt_option_t *arg2 = (apr_getopt_option_t *) 0 ;
  int val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  apr_getopt_option_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "int","svn_opt_get_option_from_code", 1, argv[0] ));
  } 
  arg1 = (int)(val1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_get_option_from_code", 2, argv[1] )); 
  }
  arg2 = (apr_getopt_option_t *)(argp2);
  {
    result = (apr_getopt_option_t *)svn_opt_get_option_from_code(arg1,(apr_getopt_option_t const *)arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_takes_option4(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc3_t *arg1 = (svn_opt_subcommand_desc3_t *) 0 ;
  int arg2 ;
  int *arg3 = (int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int temp3 ;
  int res3 = SWIG_TMPOBJ ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  arg3 = &temp3;
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t const *","svn_opt_subcommand_takes_option4", 1, argv[0] )); 
  }
  arg1 = (svn_opt_subcommand_desc3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","svn_opt_subcommand_takes_option4", 2, argv[1] ));
  } 
  arg2 = (int)(val2);
  {
    result = (svn_boolean_t)svn_opt_subcommand_takes_option4((struct svn_opt_subcommand_desc3_t const *)arg1,arg2,(int const *)arg3);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (SWIG_IsTmpObj(res3)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_takes_option3(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc2_t *arg1 = (svn_opt_subcommand_desc2_t *) 0 ;
  int arg2 ;
  int *arg3 = (int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int temp3 ;
  int res3 = SWIG_TMPOBJ ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  arg3 = &temp3;
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_subcommand_takes_option3", 1, argv[0] )); 
  }
  arg1 = (svn_opt_subcommand_desc2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","svn_opt_subcommand_takes_option3", 2, argv[1] ));
  } 
  arg2 = (int)(val2);
  {
    result = (svn_boolean_t)svn_opt_subcommand_takes_option3((struct svn_opt_subcommand_desc2_t const *)arg1,arg2,(int const *)arg3);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (SWIG_IsTmpObj(res3)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg3)));
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_int, new_flags));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_takes_option2(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc2_t *arg1 = (svn_opt_subcommand_desc2_t *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_subcommand_takes_option2", 1, argv[0] )); 
  }
  arg1 = (svn_opt_subcommand_desc2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","svn_opt_subcommand_takes_option2", 2, argv[1] ));
  } 
  arg2 = (int)(val2);
  {
    result = (svn_boolean_t)svn_opt_subcommand_takes_option2((struct svn_opt_subcommand_desc2_t const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_takes_option(int argc, VALUE *argv, VALUE self) {
  svn_opt_subcommand_desc_t *arg1 = (svn_opt_subcommand_desc_t *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc_t const *","svn_opt_subcommand_takes_option", 1, argv[0] )); 
  }
  arg1 = (svn_opt_subcommand_desc_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","svn_opt_subcommand_takes_option", 2, argv[1] ));
  } 
  arg2 = (int)(val2);
  {
    result = (svn_boolean_t)svn_opt_subcommand_takes_option((struct svn_opt_subcommand_desc_t const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_print_generic_help3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_subcommand_desc3_t *arg2 = (svn_opt_subcommand_desc3_t *) 0 ;
  apr_getopt_option_t *arg3 = (apr_getopt_option_t *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  FILE *arg6 = (FILE *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_opt_print_generic_help3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t const *","svn_opt_print_generic_help3", 2, argv[1] )); 
  }
  arg2 = (svn_opt_subcommand_desc3_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_print_generic_help3", 3, argv[2] )); 
  }
  arg3 = (apr_getopt_option_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_opt_print_generic_help3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 4) {
    
  }
  if (argc > 5) {
    res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_FILE, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "FILE *","svn_opt_print_generic_help3", 6, argv[5] )); 
    }
    arg6 = (FILE *)(argp6);
  }
  {
    svn_opt_print_generic_help3((char const *)arg1,(struct svn_opt_subcommand_desc3_t const *)arg2,(apr_getopt_option_t const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_print_generic_help2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_subcommand_desc2_t *arg2 = (svn_opt_subcommand_desc2_t *) 0 ;
  apr_getopt_option_t *arg3 = (apr_getopt_option_t *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  FILE *arg6 = (FILE *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_opt_print_generic_help2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_print_generic_help2", 2, argv[1] )); 
  }
  arg2 = (svn_opt_subcommand_desc2_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_print_generic_help2", 3, argv[2] )); 
  }
  arg3 = (apr_getopt_option_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_opt_print_generic_help2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 4) {
    
  }
  if (argc > 5) {
    res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_FILE, 0 |  0 );
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "FILE *","svn_opt_print_generic_help2", 6, argv[5] )); 
    }
    arg6 = (FILE *)(argp6);
  }
  {
    svn_opt_print_generic_help2((char const *)arg1,(struct svn_opt_subcommand_desc2_t const *)arg2,(apr_getopt_option_t const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_format_option(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  apr_getopt_option_t *arg2 = (apr_getopt_option_t *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_format_option", 2, argv[0] )); 
  }
  arg2 = (apr_getopt_option_t *)(argp2);
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  {
    svn_opt_format_option((char const **)arg1,(apr_getopt_option_t const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_help4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_subcommand_desc3_t *arg2 = (svn_opt_subcommand_desc3_t *) 0 ;
  apr_getopt_option_t *arg3 = (apr_getopt_option_t *) 0 ;
  int *arg4 = (int *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int temp4 ;
  int res4 = SWIG_TMPOBJ ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg4 = &temp4;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_opt_subcommand_help4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t const *","svn_opt_subcommand_help4", 2, argv[1] )); 
  }
  arg2 = (svn_opt_subcommand_desc3_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_subcommand_help4", 3, argv[2] )); 
  }
  arg3 = (apr_getopt_option_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    svn_opt_subcommand_help4((char const *)arg1,(struct svn_opt_subcommand_desc3_t const *)arg2,(apr_getopt_option_t const *)arg3,(int const *)arg4,arg5);
    
    
    
  }
  if (SWIG_IsTmpObj(res4)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_help3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_subcommand_desc2_t *arg2 = (svn_opt_subcommand_desc2_t *) 0 ;
  apr_getopt_option_t *arg3 = (apr_getopt_option_t *) 0 ;
  int *arg4 = (int *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int temp4 ;
  int res4 = SWIG_TMPOBJ ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg4 = &temp4;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_opt_subcommand_help3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_subcommand_help3", 2, argv[1] )); 
  }
  arg2 = (svn_opt_subcommand_desc2_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_subcommand_help3", 3, argv[2] )); 
  }
  arg3 = (apr_getopt_option_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    svn_opt_subcommand_help3((char const *)arg1,(struct svn_opt_subcommand_desc2_t const *)arg2,(apr_getopt_option_t const *)arg3,(int const *)arg4,arg5);
    
    
    
  }
  if (SWIG_IsTmpObj(res4)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_int, new_flags));
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_help2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_subcommand_desc2_t *arg2 = (svn_opt_subcommand_desc2_t *) 0 ;
  apr_getopt_option_t *arg3 = (apr_getopt_option_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_opt_subcommand_help2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_subcommand_help2", 2, argv[1] )); 
  }
  arg2 = (svn_opt_subcommand_desc2_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_subcommand_help2", 3, argv[2] )); 
  }
  arg3 = (apr_getopt_option_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    svn_opt_subcommand_help2((char const *)arg1,(struct svn_opt_subcommand_desc2_t const *)arg2,(apr_getopt_option_t const *)arg3,arg4);
    
    
    
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_subcommand_help(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_opt_subcommand_desc_t *arg2 = (svn_opt_subcommand_desc_t *) 0 ;
  apr_getopt_option_t *arg3 = (apr_getopt_option_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_opt_subcommand_help", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc_t const *","svn_opt_subcommand_help", 2, argv[1] )); 
  }
  arg2 = (svn_opt_subcommand_desc_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_subcommand_help", 3, argv[2] )); 
  }
  arg3 = (apr_getopt_option_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    svn_opt_subcommand_help((char const *)arg1,(struct svn_opt_subcommand_desc_t const *)arg2,(apr_getopt_option_t const *)arg3,arg4);
    
    
    
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_opt_revision_value_t;

SWIGINTERN VALUE
_wrap_svn_opt_revision_value_t_number_set(int argc, VALUE *argv, VALUE self) {
  union svn_opt_revision_value_t *arg1 = (union svn_opt_revision_value_t *) 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "union svn_opt_revision_value_t *","number", 1, self )); 
  }
  arg1 = (union svn_opt_revision_value_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","number", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->number = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_value_t_number_get(int argc, VALUE *argv, VALUE self) {
  union svn_opt_revision_value_t *arg1 = (union svn_opt_revision_value_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "union svn_opt_revision_value_t *","number", 1, self )); 
  }
  arg1 = (union svn_opt_revision_value_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->number);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_value_t_date_set(int argc, VALUE *argv, VALUE self) {
  union svn_opt_revision_value_t *arg1 = (union svn_opt_revision_value_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "union svn_opt_revision_value_t *","date", 1, self )); 
  }
  arg1 = (union svn_opt_revision_value_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_value_t_date_get(int argc, VALUE *argv, VALUE self) {
  union svn_opt_revision_value_t *arg1 = (union svn_opt_revision_value_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_value_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "union svn_opt_revision_value_t *","date", 1, self )); 
  }
  arg1 = (union svn_opt_revision_value_t *)(argp1);
  result =  ((arg1)->date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_opt_revision_value_t_allocate(VALUE self)
#else
_wrap_svn_opt_revision_value_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_opt_revision_value_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_opt_revision_value_t(int argc, VALUE *argv, VALUE self) {
  union svn_opt_revision_value_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (union svn_opt_revision_value_t *)calloc(1, sizeof(union svn_opt_revision_value_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_opt_revision_value_t(void *self) {
    union svn_opt_revision_value_t *arg1 = (union svn_opt_revision_value_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_opt_revision_t;

SWIGINTERN VALUE
_wrap_svn_opt_revision_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_t *arg1 = (struct svn_opt_revision_t *) 0 ;
  enum svn_opt_revision_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_revision_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_opt_revision_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_opt_revision_kind","kind", 2, argv[0] ));
  } 
  arg2 = (enum svn_opt_revision_kind)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_t *arg1 = (struct svn_opt_revision_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_opt_revision_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_revision_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_opt_revision_t *)(argp1);
  result = (enum svn_opt_revision_kind) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_t_value_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_t *arg1 = (struct svn_opt_revision_t *) 0 ;
  svn_opt_revision_value_t *arg2 = (svn_opt_revision_value_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_revision_t *","value", 1, self )); 
  }
  arg1 = (struct svn_opt_revision_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_opt_revision_value_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_opt_revision_value_t *","value", 2, argv[0] )); 
  }
  arg2 = (svn_opt_revision_value_t *)(argp2);
  if (arg1) (arg1)->value = *arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_t_value_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_t *arg1 = (struct svn_opt_revision_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_value_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_revision_t *","value", 1, self )); 
  }
  arg1 = (struct svn_opt_revision_t *)(argp1);
  result = (svn_opt_revision_value_t *)& ((arg1)->value);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_value_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_opt_revision_t_allocate(VALUE self)
#else
_wrap_svn_opt_revision_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_opt_revision_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_opt_revision_t(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_opt_revision_t *)calloc(1, sizeof(struct svn_opt_revision_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_opt_revision_t(void *self) {
    struct svn_opt_revision_t *arg1 = (struct svn_opt_revision_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_opt_revision_range_t;

SWIGINTERN VALUE
_wrap_svn_opt_revision_range_t_start_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_range_t *arg1 = (struct svn_opt_revision_range_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t rev2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_revision_range_t *","start", 1, self )); 
  }
  arg1 = (struct svn_opt_revision_range_t *)(argp1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[0]);
  }
  if (arg1) (arg1)->start = *arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_range_t_start_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_range_t *arg1 = (struct svn_opt_revision_range_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_revision_range_t *","start", 1, self )); 
  }
  arg1 = (struct svn_opt_revision_range_t *)(argp1);
  result = (svn_opt_revision_t *)& ((arg1)->start);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_range_t_end_set(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_range_t *arg1 = (struct svn_opt_revision_range_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t rev2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_revision_range_t *","end", 1, self )); 
  }
  arg1 = (struct svn_opt_revision_range_t *)(argp1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[0]);
  }
  if (arg1) (arg1)->end = *arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_revision_range_t_end_get(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_range_t *arg1 = (struct svn_opt_revision_range_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_opt_revision_range_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_opt_revision_range_t *","end", 1, self )); 
  }
  arg1 = (struct svn_opt_revision_range_t *)(argp1);
  result = (svn_opt_revision_t *)& ((arg1)->end);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_opt_revision_range_t_allocate(VALUE self)
#else
_wrap_svn_opt_revision_range_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_opt_revision_range_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_opt_revision_range_t(int argc, VALUE *argv, VALUE self) {
  struct svn_opt_revision_range_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_opt_revision_range_t *)calloc(1, sizeof(struct svn_opt_revision_range_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_opt_revision_range_t(void *self) {
    struct svn_opt_revision_range_t *arg1 = (struct svn_opt_revision_range_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_opt_parse_revision(int argc, VALUE *argv, VALUE self) {
  svn_opt_revision_t *arg1 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_opt_revision_t rev1 ;
  svn_opt_revision_t rev2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = &rev1;
    svn_swig_rb_set_revision(&rev1, argv[0]);
  }
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_opt_parse_revision", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 3) {
    
  }
  {
    result = (int)svn_opt_parse_revision(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_parse_revision_to_range(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_array_header_t *","svn_opt_parse_revision_to_range", 1, argv[0] )); 
  }
  arg1 = (apr_array_header_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_parse_revision_to_range", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (int)svn_opt_parse_revision_to_range(arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_resolve_revisions(int argc, VALUE *argv, VALUE self) {
  svn_opt_revision_t *arg1 = (svn_opt_revision_t *) 0 ;
  svn_opt_revision_t *arg2 = (svn_opt_revision_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_opt_revision_t rev1 ;
  svn_opt_revision_t rev2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg1 = &rev1;
    svn_swig_rb_set_revision(&rev1, argv[0]);
  }
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[1]);
  }
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_opt_resolve_revisions(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_parse_revprop(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_parse_revprop", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_opt_parse_revprop(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_push_implicit_dot_target(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_array_header_t *","svn_opt_push_implicit_dot_target", 1, argv[0] )); 
  }
  arg1 = (apr_array_header_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    svn_opt_push_implicit_dot_target(arg1,arg2);
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_parse_path(int argc, VALUE *argv, VALUE self) {
  svn_opt_revision_t *arg1 = (svn_opt_revision_t *) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_opt_revision_t rev1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = &rev1;
    svn_swig_rb_set_revision(&rev1, argv[0]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_opt_parse_path", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_opt_parse_path(arg1,(char const **)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_print_help5(int argc, VALUE *argv, VALUE self) {
  apr_getopt_t *arg1 = (apr_getopt_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_opt_subcommand_desc3_t *arg8 = (svn_opt_subcommand_desc3_t *) 0 ;
  apr_getopt_option_t *arg9 = (apr_getopt_option_t *) 0 ;
  int *arg10 = (int *) 0 ;
  char *arg11 = (char *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int temp10 ;
  int res10 = SWIG_TMPOBJ ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg10 = &temp10;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_getopt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_getopt_t *","svn_opt_print_help5", 1, argv[0] )); 
  }
  arg1 = (apr_getopt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help5", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help5", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_ConvertPtr(argv[7], &argp8,SWIGTYPE_p_svn_opt_subcommand_desc3_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc3_t const *","svn_opt_print_help5", 8, argv[7] )); 
  }
  arg8 = (svn_opt_subcommand_desc3_t *)(argp8);
  res9 = SWIG_ConvertPtr(argv[8], &argp9,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_print_help5", 9, argv[8] )); 
  }
  arg9 = (apr_getopt_option_t *)(argp9);
  res11 = SWIG_AsCharPtrAndSize(argv[9], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help5", 11, argv[9] ));
  }
  arg11 = (char *)(buf11);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_opt_print_help5(arg1,(char const *)arg2,arg3,arg4,arg5,(char const *)arg6,(char const *)arg7,(struct svn_opt_subcommand_desc3_t const *)arg8,(apr_getopt_option_t const *)arg9,(int const *)arg10,(char const *)arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res10)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg10)));
  } else {
    int new_flags = SWIG_IsNewObj(res10) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg10), SWIGTYPE_p_int, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_print_help4(int argc, VALUE *argv, VALUE self) {
  apr_getopt_t *arg1 = (apr_getopt_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_opt_subcommand_desc2_t *arg8 = (svn_opt_subcommand_desc2_t *) 0 ;
  apr_getopt_option_t *arg9 = (apr_getopt_option_t *) 0 ;
  int *arg10 = (int *) 0 ;
  char *arg11 = (char *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int temp10 ;
  int res10 = SWIG_TMPOBJ ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg10 = &temp10;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_getopt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_getopt_t *","svn_opt_print_help4", 1, argv[0] )); 
  }
  arg1 = (apr_getopt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help4", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help4", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_ConvertPtr(argv[7], &argp8,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_print_help4", 8, argv[7] )); 
  }
  arg8 = (svn_opt_subcommand_desc2_t *)(argp8);
  res9 = SWIG_ConvertPtr(argv[8], &argp9,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_print_help4", 9, argv[8] )); 
  }
  arg9 = (apr_getopt_option_t *)(argp9);
  res11 = SWIG_AsCharPtrAndSize(argv[9], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help4", 11, argv[9] ));
  }
  arg11 = (char *)(buf11);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_opt_print_help4(arg1,(char const *)arg2,arg3,arg4,arg5,(char const *)arg6,(char const *)arg7,(struct svn_opt_subcommand_desc2_t const *)arg8,(apr_getopt_option_t const *)arg9,(int const *)arg10,(char const *)arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res10)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg10)));
  } else {
    int new_flags = SWIG_IsNewObj(res10) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg10), SWIGTYPE_p_int, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_print_help3(int argc, VALUE *argv, VALUE self) {
  apr_getopt_t *arg1 = (apr_getopt_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_opt_subcommand_desc2_t *arg7 = (svn_opt_subcommand_desc2_t *) 0 ;
  apr_getopt_option_t *arg8 = (apr_getopt_option_t *) 0 ;
  int *arg9 = (int *) 0 ;
  char *arg10 = (char *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  int temp9 ;
  int res9 = SWIG_TMPOBJ ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg9 = &temp9;
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_getopt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_getopt_t *","svn_opt_print_help3", 1, argv[0] )); 
  }
  arg1 = (apr_getopt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help3", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help3", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_print_help3", 7, argv[6] )); 
  }
  arg7 = (svn_opt_subcommand_desc2_t *)(argp7);
  res8 = SWIG_ConvertPtr(argv[7], &argp8,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_print_help3", 8, argv[7] )); 
  }
  arg8 = (apr_getopt_option_t *)(argp8);
  res10 = SWIG_AsCharPtrAndSize(argv[8], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help3", 10, argv[8] ));
  }
  arg10 = (char *)(buf10);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_opt_print_help3(arg1,(char const *)arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(struct svn_opt_subcommand_desc2_t const *)arg7,(apr_getopt_option_t const *)arg8,(int const *)arg9,(char const *)arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res9)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg9)));
  } else {
    int new_flags = SWIG_IsNewObj(res9) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg9), SWIGTYPE_p_int, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_print_help2(int argc, VALUE *argv, VALUE self) {
  apr_getopt_t *arg1 = (apr_getopt_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_opt_subcommand_desc2_t *arg7 = (svn_opt_subcommand_desc2_t *) 0 ;
  apr_getopt_option_t *arg8 = (apr_getopt_option_t *) 0 ;
  char *arg9 = (char *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  int res9 ;
  char *buf9 = 0 ;
  int alloc9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_getopt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_getopt_t *","svn_opt_print_help2", 1, argv[0] )); 
  }
  arg1 = (apr_getopt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help2", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help2", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_svn_opt_subcommand_desc2_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc2_t const *","svn_opt_print_help2", 7, argv[6] )); 
  }
  arg7 = (svn_opt_subcommand_desc2_t *)(argp7);
  res8 = SWIG_ConvertPtr(argv[7], &argp8,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_print_help2", 8, argv[7] )); 
  }
  arg8 = (apr_getopt_option_t *)(argp8);
  res9 = SWIG_AsCharPtrAndSize(argv[8], &buf9, NULL, &alloc9);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help2", 9, argv[8] ));
  }
  arg9 = (char *)(buf9);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_opt_print_help2(arg1,(char const *)arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(struct svn_opt_subcommand_desc2_t const *)arg7,(apr_getopt_option_t const *)arg8,(char const *)arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_opt_print_help(int argc, VALUE *argv, VALUE self) {
  apr_getopt_t *arg1 = (apr_getopt_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  svn_opt_subcommand_desc_t *arg7 = (svn_opt_subcommand_desc_t *) 0 ;
  apr_getopt_option_t *arg8 = (apr_getopt_option_t *) 0 ;
  char *arg9 = (char *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  int res9 ;
  char *buf9 = 0 ;
  int alloc9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_getopt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_getopt_t *","svn_opt_print_help", 1, argv[0] )); 
  }
  arg1 = (apr_getopt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_svn_opt_subcommand_desc_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "svn_opt_subcommand_desc_t const *","svn_opt_print_help", 7, argv[6] )); 
  }
  arg7 = (svn_opt_subcommand_desc_t *)(argp7);
  res8 = SWIG_ConvertPtr(argv[7], &argp8,SWIGTYPE_p_apr_getopt_option_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "apr_getopt_option_t const *","svn_opt_print_help", 8, argv[7] )); 
  }
  arg8 = (apr_getopt_option_t *)(argp8);
  res9 = SWIG_AsCharPtrAndSize(argv[8], &buf9, NULL, &alloc9);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "char const *","svn_opt_print_help", 9, argv[8] ));
  }
  arg9 = (char *)(buf9);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_opt_print_help(arg1,(char const *)arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(struct svn_opt_subcommand_desc_t const *)arg7,(apr_getopt_option_t const *)arg8,(char const *)arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cmdline_init(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  FILE *arg2 = (FILE *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_cmdline_init", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_FILE, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "FILE *","svn_cmdline_init", 2, argv[1] )); 
  }
  arg2 = (FILE *)(argp2);
  {
    result = (int)svn_cmdline_init((char const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cmdline_create_auth_baton2(int argc, VALUE *argv, VALUE self) {
  svn_auth_baton_t **arg1 = (svn_auth_baton_t **) 0 ;
  svn_boolean_t arg2 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_config_t *arg12 = (svn_config_t *) 0 ;
  svn_cancel_func_t arg13 = (svn_cancel_func_t) 0 ;
  void *arg14 = (void *) 0 ;
  apr_pool_t *arg15 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_baton_t *temp1 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg15);
    _global_pool = arg15;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 12) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  arg2 = RTEST(argv[0]);
  {
    if (NIL_P(argv[1])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[1]);
    }
  }
  {
    if (NIL_P(argv[2])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[2]);
    }
  }
  {
    if (NIL_P(argv[3])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[3]);
    }
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  arg10 = RTEST(argv[8]);
  arg11 = RTEST(argv[9]);
  res12 = SWIG_ConvertPtr(argv[10], &argp12,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_config_t *","svn_cmdline_create_auth_baton2", 12, argv[10] )); 
  }
  arg12 = (svn_config_t *)(argp12);
  {
    arg13 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg14 = (void *)svn_swig_rb_make_baton(argv[11], _global_svn_swig_rb_pool);
  }
  if (argc > 12) {
    
  }
  {
    result = (svn_error_t *)svn_cmdline_create_auth_baton2(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_baton_t, 0));
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg14);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cmdline_create_auth_baton(int argc, VALUE *argv, VALUE self) {
  svn_auth_baton_t **arg1 = (svn_auth_baton_t **) 0 ;
  svn_boolean_t arg2 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_config_t *arg8 = (svn_config_t *) 0 ;
  svn_cancel_func_t arg9 = (svn_cancel_func_t) 0 ;
  void *arg10 = (void *) 0 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_baton_t *temp1 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  arg2 = RTEST(argv[0]);
  {
    if (NIL_P(argv[1])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[1]);
    }
  }
  {
    if (NIL_P(argv[2])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[2]);
    }
  }
  {
    if (NIL_P(argv[3])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[3]);
    }
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  res8 = SWIG_ConvertPtr(argv[6], &argp8,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_config_t *","svn_cmdline_create_auth_baton", 8, argv[6] )); 
  }
  arg8 = (svn_config_t *)(argp8);
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_cmdline_create_auth_baton(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_baton_t, 0));
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_auth_provider_t;

SWIGINTERN VALUE
_wrap_svn_auth_provider_t_cred_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_t *","cred_kind", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","cred_kind", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->cred_kind) free((char *)arg1->cred_kind);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->cred_kind = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_t_cred_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_t *","cred_kind", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_t *)(argp1);
  result = (char *) ((arg1)->cred_kind);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_t_first_credentials_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *) 0 ;
  svn_error_t *(*arg2)(void **,void **,void *,apr_hash_t *,char const *,apr_pool_t *) = (svn_error_t *(*)(void **,void **,void *,apr_hash_t *,char const *,apr_pool_t *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_t *","first_credentials", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void **,void **,void *,apr_hash_t *,char const *,apr_pool_t *)","first_credentials", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->first_credentials = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_t_first_credentials_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void **,void **,void *,apr_hash_t *,char const *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_t *","first_credentials", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_t *)(argp1);
  result = (svn_error_t *(*)(void **,void **,void *,apr_hash_t *,char const *,apr_pool_t *)) ((arg1)->first_credentials);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_t_next_credentials_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *) 0 ;
  svn_error_t *(*arg2)(void **,void *,void *,apr_hash_t *,char const *,apr_pool_t *) = (svn_error_t *(*)(void **,void *,void *,apr_hash_t *,char const *,apr_pool_t *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_t *","next_credentials", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void **,void *,void *,apr_hash_t *,char const *,apr_pool_t *)","next_credentials", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->next_credentials = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_t_next_credentials_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void **,void *,void *,apr_hash_t *,char const *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_t *","next_credentials", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_t *)(argp1);
  result = (svn_error_t *(*)(void **,void *,void *,apr_hash_t *,char const *,apr_pool_t *)) ((arg1)->next_credentials);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_t_save_credentials_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *) 0 ;
  svn_error_t *(*arg2)(svn_boolean_t *,void *,void *,apr_hash_t *,char const *,apr_pool_t *) = (svn_error_t *(*)(svn_boolean_t *,void *,void *,apr_hash_t *,char const *,apr_pool_t *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_t *","save_credentials", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_boolean_t *,void *,void *,apr_hash_t *,char const *,apr_pool_t *)","save_credentials", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->save_credentials = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_t_save_credentials_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_boolean_t *,void *,void *,apr_hash_t *,char const *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_t *","save_credentials", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_t *)(argp1);
  result = (svn_error_t *(*)(svn_boolean_t *,void *,void *,apr_hash_t *,char const *,apr_pool_t *)) ((arg1)->save_credentials);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_provider_t_allocate(VALUE self)
#else
_wrap_svn_auth_provider_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_provider_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_provider_t(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_auth_provider_t *)calloc(1, sizeof(struct svn_auth_provider_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_auth_provider_t(void *self) {
    struct svn_auth_provider_t *arg1 = (struct svn_auth_provider_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_auth_provider_object_t;

SWIGINTERN VALUE
_wrap_svn_auth_provider_object_t_vtable_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_object_t *arg1 = (struct svn_auth_provider_object_t *) 0 ;
  svn_auth_provider_t *arg2 = (svn_auth_provider_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_object_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_object_t *","vtable", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_object_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_auth_provider_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_auth_provider_t const *","vtable", 2, argv[0] )); 
  }
  arg2 = (svn_auth_provider_t *)(argp2);
  if (arg1) (arg1)->vtable = (svn_auth_provider_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_object_t_vtable_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_object_t *arg1 = (struct svn_auth_provider_object_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_auth_provider_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_object_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_object_t *","vtable", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_object_t *)(argp1);
  result = (svn_auth_provider_t *) ((arg1)->vtable);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_object_t_provider_baton_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_object_t *arg1 = (struct svn_auth_provider_object_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_object_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_object_t *","provider_baton", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_object_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, SWIG_POINTER_DISOWN);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","provider_baton", 2, argv[0] )); 
  }
  if (arg1) (arg1)->provider_baton = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_object_t_provider_baton_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_object_t *arg1 = (struct svn_auth_provider_object_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_provider_object_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_provider_object_t *","provider_baton", 1, self )); 
  }
  arg1 = (struct svn_auth_provider_object_t *)(argp1);
  result = (void *) ((arg1)->provider_baton);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_provider_object_t_allocate(VALUE self)
#else
_wrap_svn_auth_provider_object_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_provider_object_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_provider_object_t(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_provider_object_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_auth_provider_object_t *)calloc(1, sizeof(struct svn_auth_provider_object_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_auth_provider_object_t(void *self) {
    struct svn_auth_provider_object_t *arg1 = (struct svn_auth_provider_object_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_auth_cred_simple_t;

SWIGINTERN VALUE
_wrap_svn_auth_cred_simple_t_username_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_simple_t *arg1 = (struct svn_auth_cred_simple_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_simple_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_simple_t *","username", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_simple_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->username) free((char *)arg1->username);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->username = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_simple_t_username_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_simple_t *arg1 = (struct svn_auth_cred_simple_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_simple_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_simple_t *","username", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_simple_t *)(argp1);
  result = (char *) ((arg1)->username);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_simple_t_password_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_simple_t *arg1 = (struct svn_auth_cred_simple_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_simple_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_simple_t *","password", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_simple_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->password) free((char *)arg1->password);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->password = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_simple_t_password_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_simple_t *arg1 = (struct svn_auth_cred_simple_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_simple_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_simple_t *","password", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_simple_t *)(argp1);
  result = (char *) ((arg1)->password);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_simple_t_may_save_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_simple_t *arg1 = (struct svn_auth_cred_simple_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_simple_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_simple_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_simple_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->may_save = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_simple_t_may_save_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_simple_t *arg1 = (struct svn_auth_cred_simple_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_simple_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_simple_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_simple_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->may_save);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_cred_simple_t_allocate(VALUE self)
#else
_wrap_svn_auth_cred_simple_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_cred_simple_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_cred_simple_t(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_simple_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_auth_cred_simple_t *)calloc(1, sizeof(struct svn_auth_cred_simple_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_auth_cred_simple_t(void *self) {
    struct svn_auth_cred_simple_t *arg1 = (struct svn_auth_cred_simple_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_auth_cred_username_t;

SWIGINTERN VALUE
_wrap_svn_auth_cred_username_t_username_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_username_t *arg1 = (struct svn_auth_cred_username_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_username_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_username_t *","username", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_username_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->username) free((char *)arg1->username);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->username = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_username_t_username_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_username_t *arg1 = (struct svn_auth_cred_username_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_username_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_username_t *","username", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_username_t *)(argp1);
  result = (char *) ((arg1)->username);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_username_t_may_save_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_username_t *arg1 = (struct svn_auth_cred_username_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_username_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_username_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_username_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->may_save = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_username_t_may_save_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_username_t *arg1 = (struct svn_auth_cred_username_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_username_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_username_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_username_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->may_save);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_cred_username_t_allocate(VALUE self)
#else
_wrap_svn_auth_cred_username_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_cred_username_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_cred_username_t(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_username_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_auth_cred_username_t *)calloc(1, sizeof(struct svn_auth_cred_username_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_auth_cred_username_t(void *self) {
    struct svn_auth_cred_username_t *arg1 = (struct svn_auth_cred_username_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_auth_cred_ssl_client_cert_t;

SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_client_cert_t_cert_file_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_t *arg1 = (struct svn_auth_cred_ssl_client_cert_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_client_cert_t *","cert_file", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_client_cert_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","cert_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->cert_file) free((char *)arg1->cert_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->cert_file = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_client_cert_t_cert_file_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_t *arg1 = (struct svn_auth_cred_ssl_client_cert_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_client_cert_t *","cert_file", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_client_cert_t *)(argp1);
  result = (char *) ((arg1)->cert_file);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_client_cert_t_may_save_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_t *arg1 = (struct svn_auth_cred_ssl_client_cert_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_client_cert_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_client_cert_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->may_save = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_client_cert_t_may_save_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_t *arg1 = (struct svn_auth_cred_ssl_client_cert_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_client_cert_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_client_cert_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->may_save);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_cred_ssl_client_cert_t_allocate(VALUE self)
#else
_wrap_svn_auth_cred_ssl_client_cert_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_cred_ssl_client_cert_t(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_auth_cred_ssl_client_cert_t *)calloc(1, sizeof(struct svn_auth_cred_ssl_client_cert_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_auth_cred_ssl_client_cert_t(void *self) {
    struct svn_auth_cred_ssl_client_cert_t *arg1 = (struct svn_auth_cred_ssl_client_cert_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_auth_cred_ssl_client_cert_pw_t;

SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_client_cert_pw_t_password_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_pw_t *arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_client_cert_pw_t *","password", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->password) free((char *)arg1->password);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->password = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_client_cert_pw_t_password_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_pw_t *arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_client_cert_pw_t *","password", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *)(argp1);
  result = (char *) ((arg1)->password);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_client_cert_pw_t_may_save_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_pw_t *arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_client_cert_pw_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->may_save = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_client_cert_pw_t_may_save_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_pw_t *arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_client_cert_pw_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->may_save);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_cred_ssl_client_cert_pw_t_allocate(VALUE self)
#else
_wrap_svn_auth_cred_ssl_client_cert_pw_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_cred_ssl_client_cert_pw_t(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_client_cert_pw_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_auth_cred_ssl_client_cert_pw_t *)calloc(1, sizeof(struct svn_auth_cred_ssl_client_cert_pw_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_auth_cred_ssl_client_cert_pw_t(void *self) {
    struct svn_auth_cred_ssl_client_cert_pw_t *arg1 = (struct svn_auth_cred_ssl_client_cert_pw_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_auth_ssl_server_cert_info_t;

SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_hostname_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","hostname", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","hostname", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->hostname) free((char *)arg1->hostname);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->hostname = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_hostname_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","hostname", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  result = (char *) ((arg1)->hostname);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_fingerprint_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","fingerprint", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","fingerprint", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->fingerprint) free((char *)arg1->fingerprint);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->fingerprint = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_fingerprint_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","fingerprint", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  result = (char *) ((arg1)->fingerprint);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_valid_from_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","valid_from", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","valid_from", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->valid_from) free((char *)arg1->valid_from);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->valid_from = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_valid_from_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","valid_from", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  result = (char *) ((arg1)->valid_from);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_valid_until_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","valid_until", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","valid_until", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->valid_until) free((char *)arg1->valid_until);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->valid_until = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_valid_until_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","valid_until", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  result = (char *) ((arg1)->valid_until);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_issuer_dname_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","issuer_dname", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","issuer_dname", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->issuer_dname) free((char *)arg1->issuer_dname);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->issuer_dname = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_issuer_dname_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","issuer_dname", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  result = (char *) ((arg1)->issuer_dname);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_ascii_cert_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","ascii_cert", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","ascii_cert", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->ascii_cert) free((char *)arg1->ascii_cert);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->ascii_cert = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_t_ascii_cert_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_ssl_server_cert_info_t *","ascii_cert", 1, self )); 
  }
  arg1 = (struct svn_auth_ssl_server_cert_info_t *)(argp1);
  result = (char *) ((arg1)->ascii_cert);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_ssl_server_cert_info_t_allocate(VALUE self)
#else
_wrap_svn_auth_ssl_server_cert_info_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_ssl_server_cert_info_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_ssl_server_cert_info_t(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_ssl_server_cert_info_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_auth_ssl_server_cert_info_t *)calloc(1, sizeof(struct svn_auth_ssl_server_cert_info_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_auth_ssl_server_cert_info_t(void *self) {
    struct svn_auth_ssl_server_cert_info_t *arg1 = (struct svn_auth_ssl_server_cert_info_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_auth_ssl_server_cert_info_dup(int argc, VALUE *argv, VALUE self) {
  svn_auth_ssl_server_cert_info_t *arg1 = (svn_auth_ssl_server_cert_info_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_auth_ssl_server_cert_info_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_ssl_server_cert_info_t const *","svn_auth_ssl_server_cert_info_dup", 1, argv[0] )); 
  }
  arg1 = (svn_auth_ssl_server_cert_info_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_auth_ssl_server_cert_info_t *)svn_auth_ssl_server_cert_info_dup((struct svn_auth_ssl_server_cert_info_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_auth_cred_ssl_server_trust_t;

SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_server_trust_t_may_save_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_server_trust_t *arg1 = (struct svn_auth_cred_ssl_server_trust_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_server_trust_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_server_trust_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->may_save = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_server_trust_t_may_save_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_server_trust_t *arg1 = (struct svn_auth_cred_ssl_server_trust_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_server_trust_t *","may_save", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_server_trust_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->may_save);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_server_trust_t_accepted_failures_set(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_server_trust_t *arg1 = (struct svn_auth_cred_ssl_server_trust_t *) 0 ;
  apr_uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_server_trust_t *","accepted_failures", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_server_trust_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_uint32_t","accepted_failures", 2, argv[0] ));
  } 
  arg2 = (apr_uint32_t)(val2);
  if (arg1) (arg1)->accepted_failures = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_cred_ssl_server_trust_t_accepted_failures_get(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_server_trust_t *arg1 = (struct svn_auth_cred_ssl_server_trust_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_uint32_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_auth_cred_ssl_server_trust_t *","accepted_failures", 1, self )); 
  }
  arg1 = (struct svn_auth_cred_ssl_server_trust_t *)(argp1);
  result =  ((arg1)->accepted_failures);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_cred_ssl_server_trust_t_allocate(VALUE self)
#else
_wrap_svn_auth_cred_ssl_server_trust_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_cred_ssl_server_trust_t(int argc, VALUE *argv, VALUE self) {
  struct svn_auth_cred_ssl_server_trust_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_auth_cred_ssl_server_trust_t *)calloc(1, sizeof(struct svn_auth_cred_ssl_server_trust_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_auth_cred_ssl_server_trust_t(void *self) {
    struct svn_auth_cred_ssl_server_trust_t *arg1 = (struct svn_auth_cred_ssl_server_trust_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_auth_set_parameter(int argc, VALUE *argv, VALUE self) {
  svn_auth_baton_t *arg1 = (svn_auth_baton_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_baton_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_baton_t *","svn_auth_set_parameter", 1, argv[0] )); 
  }
  arg1 = (svn_auth_baton_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_auth_set_parameter", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    if (NIL_P(argv[2])) {
      arg3 = (void *)NULL;
    } else {
      VALUE _rb_pool;
      apr_pool_t *_global_pool;
      char *value = StringValuePtr(argv[2]);
      
      svn_swig_rb_get_pool(1, argv, Qnil, &_rb_pool, &_global_pool);
      arg3 = (void *)apr_pstrdup(_global_pool, value);
    }
  }
  {
    svn_auth_set_parameter(arg1,(char const *)arg2,(void const *)arg3);
    
    
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_parameter(int argc, VALUE *argv, VALUE self) {
  svn_auth_baton_t *arg1 = (svn_auth_baton_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_baton_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_baton_t *","svn_auth_get_parameter", 1, argv[0] )); 
  }
  arg1 = (svn_auth_baton_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_auth_get_parameter", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (void *)svn_auth_get_parameter(arg1,(char const *)arg2);
    
    
    
  }
  {
    char *value = result;
    if (value) {
      vresult = rb_str_new2(value);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_first_credentials(int argc, VALUE *argv, VALUE self) {
  void **arg1 = (void **) 0 ;
  svn_auth_iterstate_t **arg2 = (svn_auth_iterstate_t **) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_auth_baton_t *arg5 = (svn_auth_baton_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *temp1 ;
  svn_auth_iterstate_t *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_auth_first_credentials", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_auth_first_credentials", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[2], &argp5,SWIGTYPE_p_svn_auth_baton_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_auth_baton_t *","svn_auth_first_credentials", 5, argv[2] )); 
  }
  arg5 = (svn_auth_baton_t *)(argp5);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_auth_first_credentials(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_void, 0));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_auth_iterstate_t, 0));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_next_credentials(int argc, VALUE *argv, VALUE self) {
  void **arg1 = (void **) 0 ;
  svn_auth_iterstate_t *arg2 = (svn_auth_iterstate_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_auth_iterstate_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_auth_iterstate_t *","svn_auth_next_credentials", 2, argv[0] )); 
  }
  arg2 = (svn_auth_iterstate_t *)(argp2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_auth_next_credentials(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_void, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_save_credentials(int argc, VALUE *argv, VALUE self) {
  svn_auth_iterstate_t *arg1 = (svn_auth_iterstate_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_iterstate_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_iterstate_t *","svn_auth_save_credentials", 1, argv[0] )); 
  }
  arg1 = (svn_auth_iterstate_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_auth_save_credentials(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_forget_credentials(int argc, VALUE *argv, VALUE self) {
  svn_auth_baton_t *arg1 = (svn_auth_baton_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_baton_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_baton_t *","svn_auth_forget_credentials", 1, argv[0] )); 
  }
  arg1 = (svn_auth_baton_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_auth_forget_credentials", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_auth_forget_credentials", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_auth_forget_credentials(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_simple_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_simple_prompt_func_t arg2 = (svn_auth_simple_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_simple_prompt_func_t) svn_swig_rb_auth_simple_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_auth_get_simple_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    svn_auth_get_simple_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_username_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_username_prompt_func_t arg2 = (svn_auth_username_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_username_prompt_func_t) svn_swig_rb_auth_username_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_auth_get_username_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    svn_auth_get_username_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_simple_provider2(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_plaintext_prompt_func_t arg2 = (svn_auth_plaintext_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int res3 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_plaintext_prompt_func_t","svn_auth_get_simple_provider2", 2, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_get_simple_provider2", 3, argv[1] )); 
  }
  if (argc > 2) {
    
  }
  {
    svn_auth_get_simple_provider2(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_simple_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_auth_get_simple_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_platform_specific_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_auth_get_platform_specific_provider", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_auth_get_platform_specific_provider", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_auth_get_platform_specific_provider(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_platform_specific_client_providers(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = (apr_array_header_t **) 0 ;
  svn_config_t *arg2 = (svn_config_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_config_t *","svn_auth_get_platform_specific_client_providers", 2, argv[0] )); 
  }
  arg2 = (svn_config_t *)(argp2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_auth_get_platform_specific_client_providers(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_auth_provider_object(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_username_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_auth_get_username_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_ssl_server_trust_file_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_auth_get_ssl_server_trust_file_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_ssl_client_cert_file_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_auth_get_ssl_client_cert_file_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_ssl_client_cert_pw_file_provider2(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_plaintext_passphrase_prompt_func_t arg2 = (svn_auth_plaintext_passphrase_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int res3 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_plaintext_passphrase_prompt_func_t","svn_auth_get_ssl_client_cert_pw_file_provider2", 2, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_get_ssl_client_cert_pw_file_provider2", 3, argv[1] )); 
  }
  if (argc > 2) {
    
  }
  {
    svn_auth_get_ssl_client_cert_pw_file_provider2(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_ssl_client_cert_pw_file_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_auth_get_ssl_client_cert_pw_file_provider(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_ssl_server_trust_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_server_trust_prompt_func_t arg2 = (svn_auth_ssl_server_trust_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_server_trust_prompt_func_t) svn_swig_rb_auth_ssl_server_trust_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  if (argc > 1) {
    
  }
  {
    svn_auth_get_ssl_server_trust_prompt_provider(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_ssl_client_cert_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_client_cert_prompt_func_t arg2 = (svn_auth_ssl_client_cert_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_client_cert_prompt_func_t) svn_swig_rb_auth_ssl_client_cert_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_auth_get_ssl_client_cert_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    svn_auth_get_ssl_client_cert_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_get_ssl_client_cert_pw_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_client_cert_pw_prompt_func_t arg2 = (svn_auth_ssl_client_cert_pw_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_client_cert_pw_prompt_func_t) svn_swig_rb_auth_ssl_client_cert_pw_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_auth_get_ssl_client_cert_pw_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    svn_auth_get_ssl_client_cert_pw_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_invoke_first_credentials(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_t *arg1 = (svn_auth_provider_t *) 0 ;
  void **arg2 = (void **) 0 ;
  void **arg3 = (void **) 0 ;
  void *arg4 = (void *) 0 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *temp2 ;
  void *temp3 ;
  int res4 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_provider_t *","svn_auth_provider_invoke_first_credentials", 1, argv[0] )); 
  }
  arg1 = (svn_auth_provider_t *)(argp1);
  res4 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_auth_provider_invoke_first_credentials", 4, argv[1] )); 
  }
  res5 = SWIG_ConvertPtr(argv[2], &argp5,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_hash_t *","svn_auth_provider_invoke_first_credentials", 5, argv[2] )); 
  }
  arg5 = (apr_hash_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[3], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_auth_provider_invoke_first_credentials", 6, argv[3] ));
  }
  arg6 = (char *)(buf6);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_auth_provider_invoke_first_credentials(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg3, SWIGTYPE_p_void, 0));
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_invoke_next_credentials(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_t *arg1 = (svn_auth_provider_t *) 0 ;
  void **arg2 = (void **) 0 ;
  void *arg3 = (void *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *temp2 ;
  int res3 ;
  int res4 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_provider_t *","svn_auth_provider_invoke_next_credentials", 1, argv[0] )); 
  }
  arg1 = (svn_auth_provider_t *)(argp1);
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_provider_invoke_next_credentials", 3, argv[1] )); 
  }
  res4 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_auth_provider_invoke_next_credentials", 4, argv[2] )); 
  }
  res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_hash_t *","svn_auth_provider_invoke_next_credentials", 5, argv[3] )); 
  }
  arg5 = (apr_hash_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_auth_provider_invoke_next_credentials", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_auth_provider_invoke_next_credentials(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0));
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_provider_invoke_save_credentials(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_t *arg1 = (svn_auth_provider_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  void *arg3 = (void *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_hash_t *arg5 = (apr_hash_t *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t temp2 ;
  int res3 ;
  int res4 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_provider_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_provider_t *","svn_auth_provider_invoke_save_credentials", 1, argv[0] )); 
  }
  arg1 = (svn_auth_provider_t *)(argp1);
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_provider_invoke_save_credentials", 3, argv[1] )); 
  }
  res4 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_auth_provider_invoke_save_credentials", 4, argv[2] )); 
  }
  res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_hash_t *","svn_auth_provider_invoke_save_credentials", 5, argv[3] )); 
  }
  arg5 = (apr_hash_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_auth_provider_invoke_save_credentials", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_auth_provider_invoke_save_credentials(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_simple_provider_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_simple_provider_func_t arg1 = (svn_auth_simple_provider_func_t) 0 ;
  svn_auth_provider_object_t **arg2 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp2 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_simple_provider_func_t","svn_auth_invoke_simple_provider_func", 1, argv[0] )); 
    }
  }
  if (argc > 1) {
    
  }
  {
    svn_auth_invoke_simple_provider_func(arg1,arg2,arg3);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_ssl_client_cert_pw_provider_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_ssl_client_cert_pw_provider_func_t arg1 = (svn_auth_ssl_client_cert_pw_provider_func_t) 0 ;
  svn_auth_provider_object_t **arg2 = (svn_auth_provider_object_t **) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp2 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_ssl_client_cert_pw_provider_func_t","svn_auth_invoke_ssl_client_cert_pw_provider_func", 1, argv[0] )); 
    }
  }
  if (argc > 1) {
    
  }
  {
    svn_auth_invoke_ssl_client_cert_pw_provider_func(arg1,arg2,arg3);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_simple_prompt_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_simple_prompt_func_t arg1 = (svn_auth_simple_prompt_func_t) 0 ;
  svn_auth_cred_simple_t **arg2 = (svn_auth_cred_simple_t **) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_cred_simple_t *temp2 ;
  int res3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_simple_prompt_func_t","svn_auth_invoke_simple_prompt_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_invoke_simple_prompt_func", 3, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_auth_invoke_simple_prompt_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  {
    if (NIL_P(argv[3])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[3]);
    }
  }
  arg6 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_auth_invoke_simple_prompt_func(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_auth_cred_simple_t, 0));
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_username_prompt_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_username_prompt_func_t arg1 = (svn_auth_username_prompt_func_t) 0 ;
  svn_auth_cred_username_t **arg2 = (svn_auth_cred_username_t **) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_cred_username_t *temp2 ;
  int res3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_username_prompt_func_t","svn_auth_invoke_username_prompt_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_invoke_username_prompt_func", 3, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_auth_invoke_username_prompt_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_auth_invoke_username_prompt_func(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_auth_cred_username_t, 0));
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_ssl_server_trust_prompt_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_ssl_server_trust_prompt_func_t arg1 = (svn_auth_ssl_server_trust_prompt_func_t) 0 ;
  svn_auth_cred_ssl_server_trust_t **arg2 = (svn_auth_cred_ssl_server_trust_t **) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_uint32_t arg5 ;
  svn_auth_ssl_server_cert_info_t *arg6 = (svn_auth_ssl_server_cert_info_t *) 0 ;
  svn_boolean_t arg7 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_cred_ssl_server_trust_t *temp2 ;
  int res3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_ssl_server_trust_prompt_func_t","svn_auth_invoke_ssl_server_trust_prompt_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_invoke_ssl_server_trust_prompt_func", 3, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_auth_invoke_ssl_server_trust_prompt_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_unsigned_SS_long(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "apr_uint32_t","svn_auth_invoke_ssl_server_trust_prompt_func", 5, argv[3] ));
  } 
  arg5 = (apr_uint32_t)(val5);
  res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_auth_ssl_server_cert_info_t const *","svn_auth_invoke_ssl_server_trust_prompt_func", 6, argv[4] )); 
  }
  arg6 = (svn_auth_ssl_server_cert_info_t *)(argp6);
  arg7 = RTEST(argv[5]);
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_auth_invoke_ssl_server_trust_prompt_func(arg1,arg2,arg3,(char const *)arg4,arg5,(struct svn_auth_ssl_server_cert_info_t const *)arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t, 0));
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_ssl_client_cert_prompt_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_ssl_client_cert_prompt_func_t arg1 = (svn_auth_ssl_client_cert_prompt_func_t) 0 ;
  svn_auth_cred_ssl_client_cert_t **arg2 = (svn_auth_cred_ssl_client_cert_t **) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_cred_ssl_client_cert_t *temp2 ;
  int res3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_ssl_client_cert_prompt_func_t","svn_auth_invoke_ssl_client_cert_prompt_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_invoke_ssl_client_cert_prompt_func", 3, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_auth_invoke_ssl_client_cert_prompt_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_auth_invoke_ssl_client_cert_prompt_func(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t, 0));
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_ssl_client_cert_pw_prompt_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_ssl_client_cert_pw_prompt_func_t arg1 = (svn_auth_ssl_client_cert_pw_prompt_func_t) 0 ;
  svn_auth_cred_ssl_client_cert_pw_t **arg2 = (svn_auth_cred_ssl_client_cert_pw_t **) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_cred_ssl_client_cert_pw_t *temp2 ;
  int res3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_ssl_client_cert_pw_prompt_func_t","svn_auth_invoke_ssl_client_cert_pw_prompt_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_auth_invoke_ssl_client_cert_pw_prompt_func", 3, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_auth_invoke_ssl_client_cert_pw_prompt_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_auth_invoke_ssl_client_cert_pw_prompt_func(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t, 0));
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_plaintext_prompt_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_plaintext_prompt_func_t arg1 = (svn_auth_plaintext_prompt_func_t) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  char *arg3 = (char *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_plaintext_prompt_func_t","svn_auth_invoke_plaintext_prompt_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_auth_invoke_plaintext_prompt_func", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_auth_invoke_plaintext_prompt_func", 4, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_auth_invoke_plaintext_prompt_func(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_invoke_plaintext_passphrase_prompt_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_plaintext_passphrase_prompt_func_t arg1 = (svn_auth_plaintext_passphrase_prompt_func_t) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  char *arg3 = (char *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_auth_plaintext_passphrase_prompt_func_t","svn_auth_invoke_plaintext_passphrase_prompt_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_auth_invoke_plaintext_passphrase_prompt_func", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_auth_invoke_plaintext_passphrase_prompt_func", 4, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_auth_invoke_plaintext_passphrase_prompt_func(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_config(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_config_get_config(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg1,
        "svn_config_t *"))
    
    ;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_create2(int argc, VALUE *argv, VALUE self) {
  svn_config_t **arg1 = (svn_config_t **) 0 ;
  svn_boolean_t arg2 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_config_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  arg2 = RTEST(argv[0]);
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_config_create2(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_config_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_create(int argc, VALUE *argv, VALUE self) {
  svn_config_t **arg1 = (svn_config_t **) 0 ;
  svn_boolean_t arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_config_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  arg2 = RTEST(argv[0]);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_config_create(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_config_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_read3(int argc, VALUE *argv, VALUE self) {
  svn_config_t **arg1 = (svn_config_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_config_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_read3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[1]);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_config_read3(arg1,(char const *)arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_config_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_read2(int argc, VALUE *argv, VALUE self) {
  svn_config_t **arg1 = (svn_config_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_config_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_read2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[1]);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_config_read2(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_config_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_read(int argc, VALUE *argv, VALUE self) {
  svn_config_t **arg1 = (svn_config_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_config_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_read", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_config_read(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_config_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_parse(int argc, VALUE *argv, VALUE self) {
  svn_config_t **arg1 = (svn_config_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_config_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[0]);
  }
  arg3 = RTEST(argv[1]);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_config_parse(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_config_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_merge(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_merge", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_merge", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  {
    result = (svn_error_t *)svn_config_merge(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE vresult = Qnil;
  
  arg2 = &temp2;
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_get", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_config_get", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  {
    if (NIL_P(argv[3])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[3]);
    }
  }
  {
    svn_config_get(arg1,(char const **)arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5);
    
    
    
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_set(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_set", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_set", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_set", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_config_set", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    svn_config_set(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4);
    
    
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_bool(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg2 = &temp2;
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_get_bool", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get_bool", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_config_get_bool", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[3]);
  {
    result = (svn_error_t *)svn_config_get_bool(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_set_bool(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_set_bool", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_set_bool", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_set_bool", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  {
    svn_config_set_bool(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_int64(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  apr_int64_t *arg2 = (apr_int64_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_int64_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_int64_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg2 = &temp2;
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_get_int64", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get_int64", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_config_get_int64", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = (apr_int64_t)NUM2LL(argv[3]);
  }
  {
    result = (svn_error_t *)svn_config_get_int64(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LL2NUM(*arg2));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_set_int64(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_int64_t arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_set_int64", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_set_int64", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_set_int64", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = (apr_int64_t)NUM2LL(argv[3]);
  }
  {
    svn_config_set_int64(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_yes_no_ask(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg2 = &temp2;
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_get_yes_no_ask", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get_yes_no_ask", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_config_get_yes_no_ask", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  {
    if (NIL_P(argv[3])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[3]);
    }
  }
  {
    result = (svn_error_t *)svn_config_get_yes_no_ask(arg1,(char const **)arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_tristate(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  svn_tristate_t *arg2 = (svn_tristate_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_tristate_t arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 6) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_get_tristate", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_tristate_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_tristate_t *","svn_config_get_tristate", 2, argv[1] )); 
  }
  arg2 = (svn_tristate_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get_tristate", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_config_get_tristate", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_config_get_tristate", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  ecode6 = SWIG_AsVal_int(argv[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "svn_tristate_t","svn_config_get_tristate", 6, argv[5] ));
  } 
  arg6 = (svn_tristate_t)(val6);
  {
    result = (svn_error_t *)svn_config_get_tristate(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_enumerate_sections(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  svn_config_section_enumerator_t arg2 = (svn_config_section_enumerator_t) 0 ;
  void *arg3 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_enumerate_sections", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[1], (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_config_section_enumerator_t","svn_config_enumerate_sections", 2, argv[1] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_config_enumerate_sections", 3, argv[2] )); 
  }
  {
    result = (int)svn_config_enumerate_sections(arg1,arg2,arg3);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_enumerate_sections2(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  svn_config_section_enumerator2_t arg2 = (svn_config_section_enumerator2_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_enumerate_sections2", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  {
    arg2 = (svn_config_section_enumerator2_t) svn_swig_rb_config_section_enumerator;
    arg3 = (void *)svn_swig_rb_make_baton(argv[1], _global_svn_swig_rb_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (int)svn_config_enumerate_sections2(arg1,arg2,arg3,arg4);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_enumerate(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_config_enumerator_t arg3 = (svn_config_enumerator_t) 0 ;
  void *arg4 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res4 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_enumerate", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_enumerate", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    int res = SWIG_ConvertFunctionPtr(argv[2], (void**)(&arg3), SWIGTYPE_p_f_p_q_const__char_p_q_const__char_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_config_enumerator_t","svn_config_enumerate", 3, argv[2] )); 
    }
  }
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_config_enumerate", 4, argv[3] )); 
  }
  {
    result = (int)svn_config_enumerate(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_enumerate2(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_config_enumerator2_t arg3 = (svn_config_enumerator2_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_enumerate2", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_enumerate2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_config_enumerator2_t) svn_swig_rb_config_enumerator;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  if (argc > 3) {
    
  }
  {
    result = (int)svn_config_enumerate2(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_has_section(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_has_section", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_has_section", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_boolean_t)svn_config_has_section(arg1,(char const *)arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_find_group(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_find_group", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_find_group", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_find_group", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 3) {
    
  }
  {
    result = (char *)svn_config_find_group(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_server_setting(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_get_server_setting", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_get_server_setting", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get_server_setting", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  {
    result = (char *)svn_config_get_server_setting(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_server_setting_int(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_int64_t arg4 ;
  apr_int64_t *arg5 = (apr_int64_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  apr_int64_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg5 = &temp5;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_get_server_setting_int", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_get_server_setting_int", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get_server_setting_int", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = (apr_int64_t)NUM2LL(argv[3]);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_config_get_server_setting_int(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LL2NUM(*arg5));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_server_setting_bool(int argc, VALUE *argv, VALUE self) {
  svn_config_t *arg1 = (svn_config_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg2 = &temp2;
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_config_t *","svn_config_get_server_setting_bool", 1, argv[0] )); 
  }
  arg1 = (svn_config_t *)(argp1);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get_server_setting_bool", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_config_get_server_setting_bool", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[3]);
  {
    result = (svn_error_t *)svn_config_get_server_setting_bool(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_ensure(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg1 = NULL;
    } else {
      arg1 = StringValuePtr(argv[0]);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_config_ensure((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_read_auth_data(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_read_auth_data", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_read_auth_data", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[2])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[2]);
    }
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_config_read_auth_data(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, !*arg1 ? Qnil : svn_swig_rb_apr_hash_to_hash_svn_string(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_write_auth_data(int argc, VALUE *argv, VALUE self) {
  apr_hash_t *arg1 = (apr_hash_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg1 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[0], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg1)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[0], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_write_auth_data", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_write_auth_data", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_config_write_auth_data(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_walk_auth_data(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_config_auth_walk_func_t arg2 = (svn_config_auth_walk_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg1 = NULL;
    } else {
      arg1 = StringValuePtr(argv[0]);
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[1], (void**)(&arg2), SWIGTYPE_p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_config_auth_walk_func_t","svn_config_walk_auth_data", 2, argv[1] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_config_walk_auth_data", 3, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_config_walk_auth_data((char const *)arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_get_user_config_path(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_get_user_config_path", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_config_get_user_config_path((char const **)arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_dup(int argc, VALUE *argv, VALUE self) {
  svn_config_t **arg1 = (svn_config_t **) 0 ;
  svn_config_t *arg2 = (svn_config_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_config_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_config_t const *","svn_config_dup", 2, argv[0] )); 
  }
  arg2 = (svn_config_t *)(argp2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_config_dup(arg1,(struct svn_config_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_config_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_copy_config(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","svn_config_copy_config", 2, argv[0] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_config_copy_config(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg1,
        "svn_config_t *"))
    
    ;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_invoke_section_enumerator(int argc, VALUE *argv, VALUE self) {
  svn_config_section_enumerator_t arg1 = (svn_config_section_enumerator_t) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_q_const__char_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_config_section_enumerator_t","svn_config_invoke_section_enumerator", 1, argv[0] )); 
    }
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_invoke_section_enumerator", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_config_invoke_section_enumerator", 3, argv[2] )); 
  }
  {
    result = (svn_boolean_t)svn_config_invoke_section_enumerator(arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_invoke_section_enumerator2(int argc, VALUE *argv, VALUE self) {
  svn_config_section_enumerator2_t arg1 = (svn_config_section_enumerator2_t) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_q_const__char_p_void_p_apr_pool_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_config_section_enumerator2_t","svn_config_invoke_section_enumerator2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_invoke_section_enumerator2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_config_invoke_section_enumerator2", 3, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_boolean_t)svn_config_invoke_section_enumerator2(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_invoke_enumerator(int argc, VALUE *argv, VALUE self) {
  svn_config_enumerator_t arg1 = (svn_config_enumerator_t) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  void *arg4 = (void *) 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_q_const__char_p_q_const__char_p_void__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_config_enumerator_t","svn_config_invoke_enumerator", 1, argv[0] )); 
    }
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_invoke_enumerator", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_invoke_enumerator", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_config_invoke_enumerator", 4, argv[3] )); 
  }
  {
    result = (svn_boolean_t)svn_config_invoke_enumerator(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_invoke_enumerator2(int argc, VALUE *argv, VALUE self) {
  svn_config_enumerator2_t arg1 = (svn_config_enumerator2_t) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_q_const__char_p_q_const__char_p_void_p_apr_pool_t__int);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_config_enumerator2_t","svn_config_invoke_enumerator2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_config_invoke_enumerator2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_config_invoke_enumerator2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_config_invoke_enumerator2", 4, argv[3] )); 
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_boolean_t)svn_config_invoke_enumerator2(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_config_invoke_auth_walk_func(int argc, VALUE *argv, VALUE self) {
  svn_config_auth_walk_func_t arg1 = (svn_config_auth_walk_func_t) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  void *arg3 = (void *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_hash_t *arg6 = (apr_hash_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp2 ;
  int res3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_config_auth_walk_func_t","svn_config_invoke_auth_walk_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_config_invoke_auth_walk_func", 3, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_config_invoke_auth_walk_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_config_invoke_auth_walk_func", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[4], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (NIL_P(arg6)) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[4], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_config_invoke_auth_walk_func(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_initialize2(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  arg1 = RTEST(argv[0]);
  if (argc > 1) {
    
  }
  {
    svn_utf_initialize2(arg1,arg2);
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_initialize(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    svn_utf_initialize(arg1);
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_stringbuf_to_utf8(int argc, VALUE *argv, VALUE self) {
  svn_stringbuf_t **arg1 = (svn_stringbuf_t **) 0 ;
  svn_stringbuf_t *arg2 = (svn_stringbuf_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stringbuf_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = svn_stringbuf_ncreate(StringValuePtr(argv[0]),
        RSTRING_LEN(argv[0]),
        _global_pool);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_utf_stringbuf_to_utf8(arg1,(struct svn_stringbuf_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_string_to_utf8(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  svn_string_t value2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[0]);
      value2.len = RSTRING_LEN(argv[0]);
      arg2 = &value2;
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_utf_string_to_utf8((struct svn_string_t const **)arg1,(struct svn_string_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_to_utf8(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_to_utf8", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_utf_cstring_to_utf8((char const **)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_to_utf8_ex2(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_to_utf8_ex2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_to_utf8_ex2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_utf_cstring_to_utf8_ex2((char const **)arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_to_utf8_ex(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_to_utf8_ex", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_to_utf8_ex", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_to_utf8_ex", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_utf_cstring_to_utf8_ex((char const **)arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_stringbuf_from_utf8(int argc, VALUE *argv, VALUE self) {
  svn_stringbuf_t **arg1 = (svn_stringbuf_t **) 0 ;
  svn_stringbuf_t *arg2 = (svn_stringbuf_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stringbuf_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = svn_stringbuf_ncreate(StringValuePtr(argv[0]),
        RSTRING_LEN(argv[0]),
        _global_pool);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_utf_stringbuf_from_utf8(arg1,(struct svn_stringbuf_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_string_from_utf8(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  svn_string_t value2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[0]);
      value2.len = RSTRING_LEN(argv[0]);
      arg2 = &value2;
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_utf_string_from_utf8((struct svn_string_t const **)arg1,(struct svn_string_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_from_utf8(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_from_utf8", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_utf_cstring_from_utf8((char const **)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_from_utf8_ex2(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_from_utf8_ex2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_from_utf8_ex2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_utf_cstring_from_utf8_ex2((char const **)arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_from_utf8_ex(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_from_utf8_ex", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_from_utf8_ex", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_from_utf8_ex", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_utf_cstring_from_utf8_ex((char const **)arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_from_utf8_fuzzy(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_from_utf8_fuzzy", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_utf_cstring_from_utf8_fuzzy((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_from_utf8_stringbuf(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  svn_stringbuf_t *arg2 = (svn_stringbuf_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = svn_stringbuf_ncreate(StringValuePtr(argv[0]),
        RSTRING_LEN(argv[0]),
        _global_pool);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_utf_cstring_from_utf8_stringbuf((char const **)arg1,(struct svn_stringbuf_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_from_utf8_string(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  svn_string_t value2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[0]);
      value2.len = RSTRING_LEN(argv[0]);
      arg2 = &value2;
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_utf_cstring_from_utf8_string((char const **)arg1,(struct svn_string_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_utf_cstring_utf8_width(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_utf_cstring_utf8_width", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (int)svn_utf_cstring_utf8_width((char const *)arg1);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_nls_init(int argc, VALUE *argv, VALUE self) {
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (svn_error_t *)svn_nls_init();
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_internal_style(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_internal_style", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_path_internal_style((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_local_style(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_local_style", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_path_local_style((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_splitext(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_path_splitext", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 1) {
    
  }
  {
    svn_path_splitext((char const **)arg1,(char const **)arg2,(char const *)arg3,arg4);
    
    
    
  }
  vresult = rb_ary_new();
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_is_empty(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_is_empty", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (int)svn_path_is_empty((char const *)arg1);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_is_root(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_size_t arg2 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_is_root", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_size_t","svn_dirent_is_root", 2, argv[1] ));
  } 
  arg2 = (apr_size_t)(val2);
  {
    result = (svn_boolean_t)svn_dirent_is_root((char const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_canonicalize(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_canonicalize", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_path_canonicalize((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_is_canonical(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_is_canonical", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_boolean_t)svn_path_is_canonical((char const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_compare_paths(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_compare_paths", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_path_compare_paths", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (int)svn_path_compare_paths((char const *)arg1,(char const *)arg2);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_get_longest_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_get_longest_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_path_get_longest_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_path_get_longest_ancestor((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_is_dotpath_present(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_is_dotpath_present", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_path_is_dotpath_present((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_is_url(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_is_url", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_path_is_url((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_is_uri_safe(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_is_uri_safe", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_path_is_uri_safe((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_url_add_component2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_url_add_component2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_path_url_add_component2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_path_url_add_component2((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_is_repos_relative_url(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_is_repos_relative_url", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_path_is_repos_relative_url((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_resolve_repos_relative_url(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_path_resolve_repos_relative_url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_path_resolve_repos_relative_url", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_path_resolve_repos_relative_url((char const **)arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_path_illegal_path_escape(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_path_illegal_path_escape", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_path_illegal_path_escape((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_internal_style(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_internal_style", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_dirent_internal_style((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_local_style(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_local_style", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_dirent_local_style((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath__internal_style(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath__internal_style", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_relpath__internal_style((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_join(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_join", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_dirent_join", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_dirent_join((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_join(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath_join", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_relpath_join", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_relpath_join((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_basename(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_basename", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_dirent_basename((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_dirname(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_dirname", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_dirent_dirname((char const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_split(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_dirent_split", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 1) {
    
  }
  {
    svn_dirent_split((char const **)arg1,(char const **)arg2,(char const *)arg3,arg4);
    
    
    
  }
  vresult = rb_ary_new();
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_split(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_relpath_split", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 1) {
    
  }
  {
    svn_relpath_split((char const **)arg1,(char const **)arg2,(char const *)arg3,arg4);
    
    
    
  }
  vresult = rb_ary_new();
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_basename(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath_basename", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_relpath_basename((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_dirname(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath_dirname", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_relpath_dirname((char const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_prefix(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath_prefix", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","svn_relpath_prefix", 2, argv[1] ));
  } 
  arg2 = (int)(val2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_relpath_prefix((char const *)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_split(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_uri_split", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 1) {
    
  }
  {
    svn_uri_split((char const **)arg1,(char const **)arg2,(char const *)arg3,arg4);
    
    
    
  }
  vresult = rb_ary_new();
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_basename(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_uri_basename", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_uri_basename((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_dirname(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_uri_dirname", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_uri_dirname((char const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_is_absolute(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_is_absolute", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_dirent_is_absolute((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_is_root(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_size_t arg2 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_uri_is_root", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_size_t","svn_uri_is_root", 2, argv[1] ));
  } 
  arg2 = (apr_size_t)(val2);
  {
    result = (svn_boolean_t)svn_uri_is_root((char const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_canonicalize(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_canonicalize", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_dirent_canonicalize((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_canonicalize(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath_canonicalize", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_relpath_canonicalize((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_canonicalize(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_uri_canonicalize", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_uri_canonicalize((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_is_canonical(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_is_canonical", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_boolean_t)svn_dirent_is_canonical((char const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_is_canonical(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath_is_canonical", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_relpath_is_canonical((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_is_canonical(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_uri_is_canonical", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_boolean_t)svn_uri_is_canonical((char const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_get_longest_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_get_longest_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_dirent_get_longest_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_dirent_get_longest_ancestor((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_get_longest_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath_get_longest_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_relpath_get_longest_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_relpath_get_longest_ancestor((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_get_longest_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_uri_get_longest_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_uri_get_longest_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_uri_get_longest_ancestor((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_get_absolute(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_dirent_get_absolute", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_dirent_get_absolute((char const **)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_is_child(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_is_child", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_dirent_is_child", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_dirent_is_child((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_is_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_is_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_dirent_is_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_boolean_t)svn_dirent_is_ancestor((char const *)arg1,(char const *)arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri__is_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_uri__is_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_uri__is_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_boolean_t)svn_uri__is_ancestor((char const *)arg1,(char const *)arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_dirent_skip_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_dirent_skip_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_dirent_skip_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (char *)svn_dirent_skip_ancestor((char const *)arg1,(char const *)arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_relpath_skip_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_relpath_skip_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_relpath_skip_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (char *)svn_relpath_skip_ancestor((char const *)arg1,(char const *)arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_skip_ancestor(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_uri_skip_ancestor", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_uri_skip_ancestor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_uri_skip_ancestor((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_get_dirent_from_file_url(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_uri_get_dirent_from_file_url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_uri_get_dirent_from_file_url((char const **)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_uri_get_file_url_from_dirent(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_uri_get_file_url_from_dirent", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_uri_get_file_url_from_dirent((char const **)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_parse(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_mergeinfo_parse", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_parse(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_diff2(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t *arg2 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t arg3 = (svn_mergeinfo_t) 0 ;
  svn_mergeinfo_t arg4 = (svn_mergeinfo_t) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  svn_mergeinfo_t temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg3 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  {
    arg4 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[1], _global_pool);
  }
  arg5 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_diff2(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg2));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_diff(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t *arg2 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t arg3 = (svn_mergeinfo_t) 0 ;
  svn_mergeinfo_t arg4 = (svn_mergeinfo_t) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  svn_mergeinfo_t temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg3 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  {
    arg4 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[1], _global_pool);
  }
  arg5 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_diff(arg1,arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg2));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_merge2(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t arg1 = (svn_mergeinfo_t) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  {
    arg2 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_merge2(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_catalog_merge(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_catalog_t arg1 = (svn_mergeinfo_catalog_t) 0 ;
  svn_mergeinfo_catalog_t arg2 = (svn_mergeinfo_catalog_t) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_mergeinfo_catalog_t","svn_mergeinfo_catalog_merge", 1, argv[0] )); 
  }
  arg1 = (svn_mergeinfo_catalog_t)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_mergeinfo_catalog_t","svn_mergeinfo_catalog_merge", 2, argv[1] )); 
  }
  arg2 = (svn_mergeinfo_catalog_t)(argp2);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_catalog_merge(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_remove(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  svn_mergeinfo_t arg3 = (svn_mergeinfo_t) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  {
    arg3 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_remove(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_remove2(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  svn_mergeinfo_t arg3 = (svn_mergeinfo_t) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  {
    arg3 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[1], _global_pool);
  }
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_remove2(arg1,arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_rangelist_diff(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t **arg1 = (svn_rangelist_t **) 0 ;
  svn_rangelist_t **arg2 = (svn_rangelist_t **) 0 ;
  svn_rangelist_t *arg3 = (svn_rangelist_t *) 0 ;
  svn_rangelist_t *arg4 = (svn_rangelist_t *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_rangelist_t *temp1 ;
  svn_rangelist_t *temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg3 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  {
    arg4 = svn_swig_rb_array_to_apr_array_merge_range(argv[1], _global_pool);
  }
  arg5 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_rangelist_diff(arg1,arg2,(apr_array_header_t const *)arg3,(apr_array_header_t const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_merge_range(*arg1));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_merge_range(*arg2));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_rangelist_merge2(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t *arg1 = (svn_rangelist_t *) 0 ;
  svn_rangelist_t *arg2 = (svn_rangelist_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_rangelist_t *","svn_rangelist_merge2", 1, argv[0] )); 
  }
  arg1 = (svn_rangelist_t *)(argp1);
  {
    arg2 = svn_swig_rb_array_to_apr_array_merge_range(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_rangelist_merge2(arg1,(apr_array_header_t const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_rangelist_remove(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t **arg1 = (svn_rangelist_t **) 0 ;
  svn_rangelist_t *arg2 = (svn_rangelist_t *) 0 ;
  svn_rangelist_t *arg3 = (svn_rangelist_t *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_rangelist_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  {
    arg3 = svn_swig_rb_array_to_apr_array_merge_range(argv[1], _global_pool);
  }
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_rangelist_remove(arg1,(apr_array_header_t const *)arg2,(apr_array_header_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_intersect2(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  svn_mergeinfo_t arg3 = (svn_mergeinfo_t) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_mergeinfo_t","svn_mergeinfo_intersect2", 2, argv[0] )); 
  }
  arg2 = (svn_mergeinfo_t)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_mergeinfo_t","svn_mergeinfo_intersect2", 3, argv[1] )); 
  }
  arg3 = (svn_mergeinfo_t)(argp3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_intersect2(arg1,arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_intersect(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  svn_mergeinfo_t arg3 = (svn_mergeinfo_t) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_mergeinfo_t","svn_mergeinfo_intersect", 2, argv[0] )); 
  }
  arg2 = (svn_mergeinfo_t)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_mergeinfo_t","svn_mergeinfo_intersect", 3, argv[1] )); 
  }
  arg3 = (svn_mergeinfo_t)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_intersect(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_rangelist_intersect(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t **arg1 = (svn_rangelist_t **) 0 ;
  svn_rangelist_t *arg2 = (svn_rangelist_t *) 0 ;
  svn_rangelist_t *arg3 = (svn_rangelist_t *) 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_rangelist_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  {
    arg3 = svn_swig_rb_array_to_apr_array_merge_range(argv[1], _global_pool);
  }
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_rangelist_intersect(arg1,(apr_array_header_t const *)arg2,(apr_array_header_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_rangelist_to_string(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  svn_rangelist_t *arg2 = (svn_rangelist_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_rangelist_to_string(arg1,(apr_array_header_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_rangelist_inheritable2(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t **arg1 = (svn_rangelist_t **) 0 ;
  svn_rangelist_t *arg2 = (svn_rangelist_t *) 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_rangelist_t *temp1 ;
  long val3 ;
  int ecode3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  ecode3 = SWIG_AsVal_long(argv[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_revnum_t","svn_rangelist_inheritable2", 3, argv[1] ));
  } 
  arg3 = (svn_revnum_t)(val3);
  ecode4 = SWIG_AsVal_long(argv[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_rangelist_inheritable2", 4, argv[2] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_rangelist_inheritable2(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_rangelist_inheritable(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t **arg1 = (svn_rangelist_t **) 0 ;
  svn_rangelist_t *arg2 = (svn_rangelist_t *) 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_rangelist_t *temp1 ;
  long val3 ;
  int ecode3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  ecode3 = SWIG_AsVal_long(argv[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_revnum_t","svn_rangelist_inheritable", 3, argv[1] ));
  } 
  arg3 = (svn_revnum_t)(val3);
  ecode4 = SWIG_AsVal_long(argv[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_rangelist_inheritable", 4, argv[2] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_rangelist_inheritable(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_inheritable2(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  svn_revnum_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_mergeinfo_inheritable2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_long(argv[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_mergeinfo_inheritable2", 4, argv[2] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  ecode5 = SWIG_AsVal_long(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_mergeinfo_inheritable2", 5, argv[3] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  arg6 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_inheritable2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_inheritable(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t *arg1 = (svn_mergeinfo_t *) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  char *arg3 = (char *) 0 ;
  svn_revnum_t arg4 ;
  svn_revnum_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t temp1 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_mergeinfo_inheritable", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_long(argv[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_mergeinfo_inheritable", 4, argv[2] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  ecode5 = SWIG_AsVal_long(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_mergeinfo_inheritable", 5, argv[3] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_inheritable(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_to_string(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_mergeinfo_to_string(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_catalog_dup(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_catalog_t arg1 = (svn_mergeinfo_catalog_t) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_mergeinfo_catalog_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_mergeinfo_catalog_t","svn_mergeinfo_catalog_dup", 1, argv[0] )); 
  }
  arg1 = (svn_mergeinfo_catalog_t)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_mergeinfo_catalog_t)svn_mergeinfo_catalog_dup(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_mergeinfo_dup(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_t arg1 = (svn_mergeinfo_t) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_mergeinfo_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_mergeinfo_t)svn_mergeinfo_dup(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_rangelist_dup(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t *arg1 = (svn_rangelist_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_rangelist_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_rangelist_t *)svn_rangelist_dup((apr_array_header_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_inheritance_to_word(int argc, VALUE *argv, VALUE self) {
  svn_mergeinfo_inheritance_t arg1 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_to_mergeinfo_inheritance(argv[0]);
  }
  {
    result = (char *)svn_inheritance_to_word(arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_inheritance_from_word(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_mergeinfo_inheritance_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_inheritance_from_word", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_mergeinfo_inheritance_t)svn_inheritance_from_word((char const *)arg1);
    
    
    
  }
  vresult = SWIG_From_int((int)(result));
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


static swig_class SwigClassSvn_io_dirent2_t;

SWIGINTERN VALUE
_wrap_svn_io_dirent2_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent2_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_io_dirent2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent2_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent2_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_io_dirent2_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent2_t_special_set(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent2_t *","special", 1, self )); 
  }
  arg1 = (struct svn_io_dirent2_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->special = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent2_t_special_get(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent2_t *","special", 1, self )); 
  }
  arg1 = (struct svn_io_dirent2_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->special);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent2_t_filesize_set(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *) 0 ;
  svn_filesize_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent2_t *","filesize", 1, self )); 
  }
  arg1 = (struct svn_io_dirent2_t *)(argp1);
  {
    arg2 = (svn_filesize_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->filesize = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent2_t_filesize_get(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_filesize_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent2_t *","filesize", 1, self )); 
  }
  arg1 = (struct svn_io_dirent2_t *)(argp1);
  result =  ((arg1)->filesize);
  {
    vresult = LL2NUM((svn_filesize_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent2_t_mtime_set(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *) 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent2_t *","mtime", 1, self )); 
  }
  arg1 = (struct svn_io_dirent2_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->mtime = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent2_t_mtime_get(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent2_t *","mtime", 1, self )); 
  }
  arg1 = (struct svn_io_dirent2_t *)(argp1);
  result =  ((arg1)->mtime);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_io_dirent2_t_allocate(VALUE self)
#else
_wrap_svn_io_dirent2_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_io_dirent2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_io_dirent2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_io_dirent2_t *)calloc(1, sizeof(struct svn_io_dirent2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_io_dirent2_t(void *self) {
    struct svn_io_dirent2_t *arg1 = (struct svn_io_dirent2_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_io_dirent2_create(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_io_dirent2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_io_dirent2_t *)svn_io_dirent2_create(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent2_dup(int argc, VALUE *argv, VALUE self) {
  svn_io_dirent2_t *arg1 = (svn_io_dirent2_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_io_dirent2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_io_dirent2_t const *","svn_io_dirent2_dup", 1, argv[0] )); 
  }
  arg1 = (svn_io_dirent2_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_io_dirent2_t *)svn_io_dirent2_dup((struct svn_io_dirent2_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_io_dirent2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_io_dirent_t;

SWIGINTERN VALUE
_wrap_svn_io_dirent_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent_t *arg1 = (struct svn_io_dirent_t *) 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_io_dirent_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent_t *arg1 = (struct svn_io_dirent_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_io_dirent_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent_t_special_set(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent_t *arg1 = (struct svn_io_dirent_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent_t *","special", 1, self )); 
  }
  arg1 = (struct svn_io_dirent_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->special = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dirent_t_special_get(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent_t *arg1 = (struct svn_io_dirent_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_io_dirent_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_io_dirent_t *","special", 1, self )); 
  }
  arg1 = (struct svn_io_dirent_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->special);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_io_dirent_t_allocate(VALUE self)
#else
_wrap_svn_io_dirent_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_io_dirent_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_io_dirent_t(int argc, VALUE *argv, VALUE self) {
  struct svn_io_dirent_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_io_dirent_t *)calloc(1, sizeof(struct svn_io_dirent_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_io_dirent_t(void *self) {
    struct svn_io_dirent_t *arg1 = (struct svn_io_dirent_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_io_open_uniquely_named(int argc, VALUE *argv, VALUE self) {
  apr_file_t **arg1 = (apr_file_t **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_io_file_del_t arg6 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_file_t *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_open_uniquely_named", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_open_uniquely_named", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[2], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_io_open_uniquely_named", 5, argv[2] ));
  }
  arg5 = (char *)(buf5);
  ecode6 = SWIG_AsVal_int(argv[3], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "svn_io_file_del_t","svn_io_open_uniquely_named", 6, argv[3] ));
  } 
  arg6 = (svn_io_file_del_t)(val6);
  if (argc > 4) {
    
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_io_open_uniquely_named(arg1,(char const **)arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_apr_file_t, 0));
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_open_unique_file3(int argc, VALUE *argv, VALUE self) {
  apr_file_t **arg1 = (apr_file_t **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  svn_io_file_del_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_file_t *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_open_unique_file3", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_io_file_del_t","svn_io_open_unique_file3", 4, argv[1] ));
  } 
  arg4 = (svn_io_file_del_t)(val4);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_open_unique_file3(arg1,(char const **)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_apr_file_t, 0));
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_open_unique_file2(int argc, VALUE *argv, VALUE self) {
  apr_file_t **arg1 = (apr_file_t **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_io_file_del_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_file_t *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_open_unique_file2", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_open_unique_file2", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_int(argv[2], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_io_file_del_t","svn_io_open_unique_file2", 5, argv[2] ));
  } 
  arg5 = (svn_io_file_del_t)(val5);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_open_unique_file2(arg1,(char const **)arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_apr_file_t, 0));
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_open_unique_file(int argc, VALUE *argv, VALUE self) {
  apr_file_t **arg1 = (apr_file_t **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_file_t *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_open_unique_file", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_open_unique_file", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_open_unique_file(arg1,(char const **)arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_apr_file_t, 0));
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_copy_perms(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_copy_perms", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_copy_perms", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_io_copy_perms((char const *)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_sleep_for_timestamps(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_sleep_for_timestamps", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    svn_io_sleep_for_timestamps((char const *)arg1,arg2);
    
    
    
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_filesizes_three_different_p(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_boolean_t temp2 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res4 = SWIG_AsCharPtrAndSize(argv[0], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_filesizes_three_different_p", 4, argv[0] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[1], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_io_filesizes_three_different_p", 5, argv[1] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[2], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_io_filesizes_three_different_p", 6, argv[2] ));
  }
  arg6 = (char *)(buf6);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_filesizes_three_different_p(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_checksum2(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t **arg1 = (svn_checksum_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_checksum_kind_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_checksum_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_file_checksum2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(argv[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_checksum_kind_t","svn_io_file_checksum2", 3, argv[1] ));
  } 
  arg3 = (svn_checksum_kind_t)(val3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_checksum2(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_io_file_checksum2 arg 1 (svn_checksum_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_io_file_checksum2 is not implemented yet");
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_checksum(int argc, VALUE *argv, VALUE self) {
  unsigned char *arg1 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  unsigned char temp1[APR_MD5_DIGESTSIZE] ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_file_checksum", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_checksum(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    char *digest_string = (char *)svn_md5_digest_to_cstring(arg1, _global_pool);
    vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(digest_string ? digest_string : ""));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_files_contents_same_p(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_files_contents_same_p", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_files_contents_same_p", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_io_files_contents_same_p(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_files_contents_three_same_p(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_boolean_t temp2 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res4 = SWIG_AsCharPtrAndSize(argv[0], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_files_contents_three_same_p", 4, argv[0] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[1], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_io_files_contents_three_same_p", 5, argv[1] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[2], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_io_files_contents_three_same_p", 6, argv[2] ));
  }
  arg6 = (char *)(buf6);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_files_contents_three_same_p(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_create_bytes(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_size_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_file_create_bytes", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void const *","svn_io_file_create_bytes", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_unsigned_SS_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "apr_size_t","svn_io_file_create_bytes", 3, argv[2] ));
  } 
  arg3 = (apr_size_t)(val3);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_create_bytes((char const *)arg1,(void const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_create_empty(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_file_create_empty", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_create_empty((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_lock_open_file(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  svn_boolean_t arg2 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  arg2 = RTEST(argv[1]);
  arg3 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_lock_open_file(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_unlock_open_file(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_io_unlock_open_file(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_set_read2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_read_fn_t arg2 = (svn_read_fn_t) 0 ;
  svn_read_fn_t arg3 = (svn_read_fn_t) 0 ;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[1], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_char_p_apr_size_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_read_fn_t","svn_stream_set_read2", 2, argv[1] )); 
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[2], (void**)(&arg3), SWIGTYPE_p_f_p_void_p_char_p_apr_size_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_read_fn_t","svn_stream_set_read2", 3, argv[2] )); 
    }
  }
  {
    svn_stream_set_read2(arg1,arg2,arg3);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_set_skip(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_skip_fn_t arg2 = (svn_stream_skip_fn_t) 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[1], (void**)(&arg2), SWIGTYPE_p_f_p_void_apr_size_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_skip_fn_t","svn_stream_set_skip", 2, argv[1] )); 
    }
  }
  {
    svn_stream_set_skip(arg1,arg2);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_set_mark(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_mark_fn_t arg2 = (svn_stream_mark_fn_t) 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[1], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_mark_fn_t","svn_stream_set_mark", 2, argv[1] )); 
    }
  }
  {
    svn_stream_set_mark(arg1,arg2);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_set_seek(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_seek_fn_t arg2 = (svn_stream_seek_fn_t) 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[1], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_seek_fn_t","svn_stream_set_seek", 2, argv[1] )); 
    }
  }
  {
    svn_stream_set_seek(arg1,arg2);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_set_data_available(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_data_available_fn_t arg2 = (svn_stream_data_available_fn_t) 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[1], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_svn_boolean_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_data_available_fn_t","svn_stream_set_data_available", 2, argv[1] )); 
    }
  }
  {
    svn_stream_set_data_available(arg1,arg2);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_set_readline(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_readline_fn_t arg2 = (svn_stream_readline_fn_t) 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[1], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_readline_fn_t","svn_stream_set_readline", 2, argv[1] )); 
    }
  }
  {
    svn_stream_set_readline(arg1,arg2);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_empty(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_empty(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_disown(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_disown(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_open_readonly(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_stream_open_readonly", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_stream_open_readonly(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_open_writable(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_stream_open_writable", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_stream_open_writable(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_open_unique(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  char **arg2 = (char **) 0 ;
  char *arg3 = (char *) 0 ;
  svn_io_file_del_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_stream_open_unique", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_io_file_del_t","svn_stream_open_unique", 4, argv[1] ));
  } 
  arg4 = (svn_io_file_del_t)(val4);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_stream_open_unique(arg1,(char const **)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0));
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_from_aprfile2(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  svn_boolean_t arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  arg2 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_from_aprfile2(arg1,arg2,arg3);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_from_aprfile(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_from_aprfile(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_for_stdin2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  svn_boolean_t arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  arg2 = RTEST(argv[0]);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_stream_for_stdin2(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_for_stdin(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_error_t *)svn_stream_for_stdin(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_for_stderr(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_error_t *)svn_stream_for_stderr(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_for_stdout(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = (svn_stream_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_error_t *)svn_stream_for_stdout(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stringbuf_from_stream(int argc, VALUE *argv, VALUE self) {
  svn_stringbuf_t **arg1 = (svn_stringbuf_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  apr_size_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stringbuf_t *temp1 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[0]);
  }
  ecode3 = SWIG_AsVal_unsigned_SS_long(argv[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "apr_size_t","svn_stringbuf_from_stream", 3, argv[1] ));
  } 
  arg3 = (apr_size_t)(val3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_stringbuf_from_stream(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_from_stringbuf(int argc, VALUE *argv, VALUE self) {
  svn_stringbuf_t *arg1 = (svn_stringbuf_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg1 = NULL;
    } else {
      arg1 = svn_stringbuf_ncreate(StringValuePtr(argv[0]),
        RSTRING_LEN(argv[0]),
        _global_pool);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_from_stringbuf(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_from_string(int argc, VALUE *argv, VALUE self) {
  svn_string_t *arg1 = (svn_string_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t value1 ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg1 = NULL;
    } else {
      value1.data = StringValuePtr(argv[0]);
      value1.len = RSTRING_LEN(argv[0]);
      arg1 = &value1;
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_from_string((struct svn_string_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_buffered(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_buffered(arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_compressed(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_compressed(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_checksummed2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_checksum_t **arg2 = (svn_checksum_t **) 0 ;
  svn_checksum_t **arg3 = (svn_checksum_t **) 0 ;
  svn_checksum_kind_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_checksum_t *temp2 ;
  svn_checksum_t *temp3 ;
  int val4 ;
  int ecode4 = 0 ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_checksum_kind_t","svn_stream_checksummed2", 4, argv[1] ));
  } 
  arg4 = (svn_checksum_kind_t)(val4);
  arg5 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_checksummed2(arg1,arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    /* FIXME: Missing argout typemap: svn_stream_checksummed2 arg 2 (svn_checksum_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_stream_checksummed2 is not implemented yet");
    
  }
  {
    /* FIXME: Missing argout typemap: svn_stream_checksummed2 arg 3 (svn_checksum_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_stream_checksummed2 is not implemented yet");
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_contents_checksum(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t **arg1 = (svn_checksum_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_checksum_kind_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_checksum_t *temp1 ;
  int val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[0]);
  }
  ecode3 = SWIG_AsVal_int(argv[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_checksum_kind_t","svn_stream_contents_checksum", 3, argv[1] ));
  } 
  arg3 = (svn_checksum_kind_t)(val3);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_stream_contents_checksum(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_stream_contents_checksum arg 1 (svn_checksum_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_stream_contents_checksum is not implemented yet");
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_read_full(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_size_t *arg3 = (apr_size_t *) 0 ;
  apr_size_t temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    temp2 = NUM2LONG(argv[1]);
    arg2 = malloc(temp2);
    arg3 = (apr_size_t *)&temp2;
  }
  {
    result = (svn_error_t *)svn_stream_read_full(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 == 0 ? Qnil : rb_str_new(arg2, *arg3));
    free(arg2);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_supports_partial_read(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    result = (svn_boolean_t)svn_stream_supports_partial_read(arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_read2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_size_t *arg3 = (apr_size_t *) 0 ;
  apr_size_t temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    temp2 = NUM2LONG(argv[1]);
    arg2 = malloc(temp2);
    arg3 = (apr_size_t *)&temp2;
  }
  {
    result = (svn_error_t *)svn_stream_read2(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 == 0 ? Qnil : rb_str_new(arg2, *arg3));
    free(arg2);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_read(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_size_t *arg3 = (apr_size_t *) 0 ;
  apr_size_t temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    temp2 = NUM2LONG(argv[1]);
    arg2 = malloc(temp2);
    arg3 = (apr_size_t *)&temp2;
  }
  {
    result = (svn_error_t *)svn_stream_read(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 == 0 ? Qnil : rb_str_new(arg2, *arg3));
    free(arg2);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_skip(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  apr_size_t arg2 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_size_t","svn_stream_skip", 2, argv[1] ));
  } 
  arg2 = (apr_size_t)(val2);
  {
    result = (svn_error_t *)svn_stream_skip(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_write(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  char *arg2 = (char *) 0 ;
  apr_size_t *arg3 = (apr_size_t *) 0 ;
  apr_size_t temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    arg2 = StringValuePtr(argv[1]);
    temp2 = RSTRING_LEN(argv[1]);
    arg3 = (apr_size_t *)&temp2;
  }
  {
    result = (svn_error_t *)svn_stream_write(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM(*arg3));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_close(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    result = (svn_error_t *)svn_stream_close(arg1);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_reset(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    result = (svn_error_t *)svn_stream_reset(arg1);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_supports_mark(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    result = (svn_boolean_t)svn_stream_supports_mark(arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_supports_reset(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    result = (svn_boolean_t)svn_stream_supports_reset(arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_mark(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_mark_t **arg2 = (svn_stream_mark_t **) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_mark_t *temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_stream_mark(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_stream_mark arg 2 (svn_stream_mark_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_stream_mark is not implemented yet");
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_seek(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_mark_t *arg2 = (svn_stream_mark_t *) 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_stream_mark_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_stream_mark_t const *","svn_stream_seek", 2, argv[1] )); 
  }
  arg2 = (svn_stream_mark_t *)(argp2);
  {
    result = (svn_error_t *)svn_stream_seek(arg1,(struct svn_stream_mark_t const *)arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_data_available(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_boolean_t *arg2 = (svn_boolean_t *) 0 ;
  svn_boolean_t temp2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg2 = &temp2;
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    result = (svn_error_t *)svn_stream_data_available(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_tee(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[1]);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_tee(arg1,arg2,arg3);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_puts(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  char *arg2 = (char *) 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_stream_puts", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    result = (svn_error_t *)svn_stream_puts(arg1,(char const *)arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_readline(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stringbuf_t **arg2 = (svn_stringbuf_t **) 0 ;
  char *arg3 = (char *) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stringbuf_t *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_boolean_t temp4 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg4 = &temp4;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_stream_readline", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_stream_readline(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg2)->data, (*arg2)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_copy3(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[1]);
  }
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_stream_copy3(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_copy2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[1]);
  }
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_stream_copy2(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_copy(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[1]);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_stream_copy(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_contents_same2(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    arg3 = svn_swig_rb_make_stream(argv[1]);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_stream_contents_same2(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_contents_same(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = (svn_boolean_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    arg3 = svn_swig_rb_make_stream(argv[1]);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_stream_contents_same(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_string_from_stream2(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  apr_size_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[0]);
  }
  ecode3 = SWIG_AsVal_unsigned_SS_long(argv[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "apr_size_t","svn_string_from_stream2", 3, argv[1] ));
  } 
  arg3 = (apr_size_t)(val3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_string_from_stream2(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_string_from_stream(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = (svn_string_t **) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[0]);
  }
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_string_from_stream(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_lazyopen_create(int argc, VALUE *argv, VALUE self) {
  svn_stream_lazyopen_func_t arg1 = (svn_stream_lazyopen_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_lazyopen_func_t","svn_stream_lazyopen_create", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_stream_lazyopen_create", 2, argv[1] )); 
  }
  arg3 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_stream_t *)svn_stream_lazyopen_create(arg1,arg2,arg3,arg4);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stringbuf_from_file2(int argc, VALUE *argv, VALUE self) {
  svn_stringbuf_t **arg1 = (svn_stringbuf_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stringbuf_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_stringbuf_from_file2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_stringbuf_from_file2(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stringbuf_from_file(int argc, VALUE *argv, VALUE self) {
  svn_stringbuf_t **arg1 = (svn_stringbuf_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stringbuf_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_stringbuf_from_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_stringbuf_from_file(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stringbuf_from_aprfile(int argc, VALUE *argv, VALUE self) {
  svn_stringbuf_t **arg1 = (svn_stringbuf_t **) 0 ;
  apr_file_t *arg2 = (apr_file_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stringbuf_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_stringbuf_from_aprfile(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_remove_file2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_boolean_t arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_remove_file2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  arg2 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_io_remove_file2((char const *)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_remove_dir2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  svn_boolean_t arg2 ;
  svn_cancel_func_t arg3 = (svn_cancel_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_remove_dir2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  arg2 = RTEST(argv[1]);
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_remove_dir2((char const *)arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_get_dirents3(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_get_dirents3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_get_dirents3(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg1,
        "svn_dirent_t *"))
    
    ;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_stat_dirent2(int argc, VALUE *argv, VALUE self) {
  svn_io_dirent2_t **arg1 = (svn_io_dirent2_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_io_dirent2_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_stat_dirent2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[1]);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_io_stat_dirent2((struct svn_io_dirent2_t const **)arg1,(char const *)arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_io_stat_dirent2 arg 1 (svn_io_dirent2_t const **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_io_stat_dirent2 is not implemented yet");
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_stat_dirent(int argc, VALUE *argv, VALUE self) {
  svn_io_dirent2_t **arg1 = (svn_io_dirent2_t **) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_io_dirent2_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_stat_dirent", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_stat_dirent((struct svn_io_dirent2_t const **)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_io_stat_dirent arg 1 (svn_io_dirent2_t const **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_io_stat_dirent is not implemented yet");
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dir_walk2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  apr_int32_t arg2 ;
  svn_io_walk_func_t arg3 = (svn_io_walk_func_t) 0 ;
  void *arg4 = (void *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res4 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_dir_walk2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_int32_t","svn_io_dir_walk2", 2, argv[1] ));
  } 
  arg2 = (apr_int32_t)(val2);
  {
    int res = SWIG_ConvertFunctionPtr(argv[2], (void**)(&arg3), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_io_walk_func_t","svn_io_dir_walk2", 3, argv[2] )); 
    }
  }
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_io_dir_walk2", 4, argv[3] )); 
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_io_dir_walk2((char const *)arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_start_cmd3(int argc, VALUE *argv, VALUE self) {
  apr_proc_t *arg1 = (apr_proc_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char **arg4 = (char **) 0 ;
  char **arg5 = (char **) 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  apr_file_t *arg8 = (apr_file_t *) 0 ;
  svn_boolean_t arg9 ;
  apr_file_t *arg10 = (apr_file_t *) 0 ;
  svn_boolean_t arg11 ;
  apr_file_t *arg12 = (apr_file_t *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 12) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_proc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_proc_t *","svn_io_start_cmd3", 1, argv[0] )); 
  }
  arg1 = (apr_proc_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_start_cmd3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_start_cmd3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *const *","svn_io_start_cmd3", 4, argv[3] )); 
  }
  arg4 = (char **)(argp4);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *const *","svn_io_start_cmd3", 5, argv[4] )); 
  }
  arg5 = (char **)(argp5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = svn_swig_rb_make_file(argv[7], _global_pool);
  }
  arg9 = RTEST(argv[8]);
  {
    arg10 = svn_swig_rb_make_file(argv[9], _global_pool);
  }
  arg11 = RTEST(argv[10]);
  {
    arg12 = svn_swig_rb_make_file(argv[11], _global_pool);
  }
  if (argc > 12) {
    
  }
  {
    result = (svn_error_t *)svn_io_start_cmd3(arg1,(char const *)arg2,(char const *)arg3,(char const *const *)arg4,(char const *const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_start_cmd2(int argc, VALUE *argv, VALUE self) {
  apr_proc_t *arg1 = (apr_proc_t *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char **arg4 = (char **) 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_file_t *arg7 = (apr_file_t *) 0 ;
  svn_boolean_t arg8 ;
  apr_file_t *arg9 = (apr_file_t *) 0 ;
  svn_boolean_t arg10 ;
  apr_file_t *arg11 = (apr_file_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_proc_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_proc_t *","svn_io_start_cmd2", 1, argv[0] )); 
  }
  arg1 = (apr_proc_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_start_cmd2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_start_cmd2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *const *","svn_io_start_cmd2", 4, argv[3] )); 
  }
  arg4 = (char **)(argp4);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    arg7 = svn_swig_rb_make_file(argv[6], _global_pool);
  }
  arg8 = RTEST(argv[7]);
  {
    arg9 = svn_swig_rb_make_file(argv[8], _global_pool);
  }
  arg10 = RTEST(argv[9]);
  {
    arg11 = svn_swig_rb_make_file(argv[10], _global_pool);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_io_start_cmd2(arg1,(char const *)arg2,(char const *)arg3,(char const *const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_run_diff2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char **arg2 = (char **) 0 ;
  int arg3 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  int *arg8 = (int *) 0 ;
  apr_file_t *arg9 = (apr_file_t *) 0 ;
  apr_file_t *arg10 = (apr_file_t *) 0 ;
  char *arg11 = (char *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int temp8 ;
  int res8 = SWIG_TMPOBJ ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg8 = &temp8;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_p_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *const *","svn_io_run_diff2", 2, argv[1] )); 
  }
  arg2 = (char **)(argp2);
  ecode3 = SWIG_AsVal_int(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "int","svn_io_run_diff2", 3, argv[2] ));
  } 
  arg3 = (int)(val3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff2", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff2", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff2", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  {
    arg9 = svn_swig_rb_make_file(argv[7], _global_pool);
  }
  {
    arg10 = svn_swig_rb_make_file(argv[8], _global_pool);
  }
  res11 = SWIG_AsCharPtrAndSize(argv[9], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff2", 11, argv[9] ));
  }
  arg11 = (char *)(buf11);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_io_run_diff2((char const *)arg1,(char const *const *)arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,arg10,(char const *)arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res8)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg8)));
  } else {
    int new_flags = SWIG_IsNewObj(res8) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg8), SWIGTYPE_p_int, new_flags));
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_run_diff3_3(int argc, VALUE *argv, VALUE self) {
  int *arg1 = (int *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  apr_file_t *arg9 = (apr_file_t *) 0 ;
  char *arg10 = (char *) 0 ;
  apr_array_header_t *arg11 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff3_3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff3_3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff3_3", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff3_3", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff3_3", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff3_3", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff3_3", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  {
    arg9 = svn_swig_rb_make_file(argv[7], _global_pool);
  }
  res10 = SWIG_AsCharPtrAndSize(argv[8], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_io_run_diff3_3", 10, argv[8] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_ConvertPtr(argv[9], &argp11,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_io_run_diff3_3", 11, argv[9] )); 
  }
  arg11 = (apr_array_header_t *)(argp11);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_io_run_diff3_3(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,(char const *)arg10,(apr_array_header_t const *)arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg1)));
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_parse_mimetypes_file(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_parse_mimetypes_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_io_parse_mimetypes_file(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_string(*arg1));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_detect_mimetype2(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_hash_t *arg3 = (apr_hash_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_detect_mimetype2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_hash_to_apr_hash_string(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_io_detect_mimetype2((char const **)arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_detect_mimetype(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_detect_mimetype", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_io_detect_mimetype((char const **)arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_is_binary_data(int argc, VALUE *argv, VALUE self) {
  void *arg1 = (void *) 0 ;
  apr_size_t arg2 ;
  int res1 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "void const *","svn_io_is_binary_data", 1, argv[0] )); 
  }
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_size_t","svn_io_is_binary_data", 2, argv[1] ));
  } 
  arg2 = (apr_size_t)(val2);
  {
    result = (svn_boolean_t)svn_io_is_binary_data((void const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_putc(int argc, VALUE *argv, VALUE self) {
  char arg1 ;
  apr_file_t *arg2 = (apr_file_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char val1 ;
  int ecode1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_char(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "char","svn_io_file_putc", 1, argv[0] ));
  } 
  arg1 = (char)(val1);
  {
    arg2 = svn_swig_rb_make_file(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_putc(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_size_get(int argc, VALUE *argv, VALUE self) {
  svn_filesize_t *arg1 = (svn_filesize_t *) 0 ;
  apr_file_t *arg2 = (apr_file_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_filesize_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_size_get(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LL2NUM(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_get_offset(int argc, VALUE *argv, VALUE self) {
  apr_off_t *arg1 = (apr_off_t *) 0 ;
  apr_file_t *arg2 = (apr_file_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_off_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_off_t *","svn_io_file_get_offset", 1, argv[0] )); 
  }
  arg1 = (apr_off_t *)(argp1);
  {
    arg2 = svn_swig_rb_make_file(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_get_offset(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_read_full2(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_size_t arg3 ;
  apr_size_t *arg4 = (apr_size_t *) 0 ;
  svn_boolean_t *arg5 = (svn_boolean_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  apr_size_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  svn_boolean_t temp5 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg4 = &temp4;
  arg5 = &temp5;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_io_file_read_full2", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_unsigned_SS_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "apr_size_t","svn_io_file_read_full2", 3, argv[2] ));
  } 
  arg3 = (apr_size_t)(val3);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_read_full2(arg1,arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res4)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_unsigned_SS_long((*arg4)));
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_apr_size_t, new_flags));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg5 ? Qtrue : Qfalse);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_aligned_seek(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  apr_off_t arg2 ;
  apr_off_t *arg3 = (apr_off_t *) 0 ;
  apr_off_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  {
    arg2 = (apr_off_t)NUM2LL(argv[1]);
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_off_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_off_t *","svn_io_file_aligned_seek", 3, argv[2] )); 
  }
  arg3 = (apr_off_t *)(argp3);
  {
    arg4 = (apr_off_t)NUM2LL(argv[3]);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_aligned_seek(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_flush(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_flush(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_write_atomic2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_size_t arg3 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_write_atomic2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void const *","svn_io_write_atomic2", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_unsigned_SS_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "apr_size_t","svn_io_write_atomic2", 3, argv[2] ));
  } 
  arg3 = (apr_size_t)(val3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_write_atomic2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_io_write_atomic2((char const *)arg1,(void const *)arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_write_atomic(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_size_t arg3 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_write_atomic", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void const *","svn_io_write_atomic", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_unsigned_SS_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "apr_size_t","svn_io_write_atomic", 3, argv[2] ));
  } 
  arg3 = (apr_size_t)(val3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_io_write_atomic", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_io_write_atomic((char const *)arg1,(void const *)arg2,arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_write_unique(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  char *arg2 = (char *) 0 ;
  void *arg3 = (void *) 0 ;
  apr_size_t arg4 ;
  svn_io_file_del_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  unsigned long val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_write_unique", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void const *","svn_io_write_unique", 3, argv[1] )); 
  }
  ecode4 = SWIG_AsVal_unsigned_SS_long(argv[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "apr_size_t","svn_io_write_unique", 4, argv[2] ));
  } 
  arg4 = (apr_size_t)(val4);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_io_file_del_t","svn_io_write_unique", 5, argv[3] ));
  } 
  arg5 = (svn_io_file_del_t)(val5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_io_write_unique((char const **)arg1,(char const *)arg2,(void const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_trunc(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  apr_off_t arg2 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  {
    arg2 = (apr_off_t)NUM2LL(argv[1]);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_trunc(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_rename2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_io_file_rename2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_io_file_rename2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_rename2((char const *)arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_dir_close(int argc, VALUE *argv, VALUE self) {
  apr_dir_t *arg1 = (apr_dir_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_dir_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_dir_t *","svn_io_dir_close", 1, argv[0] )); 
  }
  arg1 = (apr_dir_t *)(argp1);
  {
    result = (svn_error_t *)svn_io_dir_close(arg1);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_name_get(int argc, VALUE *argv, VALUE self) {
  char **arg1 = (char **) 0 ;
  apr_file_t *arg2 = (apr_file_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_name_get((char const **)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_file_readline(int argc, VALUE *argv, VALUE self) {
  apr_file_t *arg1 = (apr_file_t *) 0 ;
  svn_stringbuf_t **arg2 = (svn_stringbuf_t **) 0 ;
  char **arg3 = (char **) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  apr_size_t arg5 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stringbuf_t *temp2 ;
  char *temp3 ;
  svn_boolean_t temp4 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_file(argv[0], _global_pool);
  }
  ecode5 = SWIG_AsVal_unsigned_SS_long(argv[1], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "apr_size_t","svn_io_file_readline", 5, argv[1] ));
  } 
  arg5 = (apr_size_t)(val5);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_io_file_readline(arg1,arg2,(char const **)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg2)->data, (*arg2)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_read_invoke_fn(int argc, VALUE *argv, VALUE self) {
  svn_read_fn_t arg1 = (svn_read_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_size_t *arg4 = (apr_size_t *) 0 ;
  int res2 ;
  apr_size_t temp3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_char_p_apr_size_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_read_fn_t","svn_read_invoke_fn", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_read_invoke_fn", 2, argv[1] )); 
  }
  {
    temp3 = NUM2LONG(argv[2]);
    arg3 = malloc(temp3);
    arg4 = (apr_size_t *)&temp3;
  }
  {
    result = (svn_error_t *)svn_read_invoke_fn(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 == 0 ? Qnil : rb_str_new(arg3, *arg4));
    free(arg3);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_invoke_skip_fn(int argc, VALUE *argv, VALUE self) {
  svn_stream_skip_fn_t arg1 = (svn_stream_skip_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  apr_size_t arg3 ;
  int res2 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_apr_size_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_skip_fn_t","svn_stream_invoke_skip_fn", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_stream_invoke_skip_fn", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_unsigned_SS_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "apr_size_t","svn_stream_invoke_skip_fn", 3, argv[2] ));
  } 
  arg3 = (apr_size_t)(val3);
  {
    result = (svn_error_t *)svn_stream_invoke_skip_fn(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_write_invoke_fn(int argc, VALUE *argv, VALUE self) {
  svn_write_fn_t arg1 = (svn_write_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_size_t *arg4 = (apr_size_t *) 0 ;
  int res2 ;
  apr_size_t temp3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_size_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_write_fn_t","svn_write_invoke_fn", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_write_invoke_fn", 2, argv[1] )); 
  }
  {
    arg3 = StringValuePtr(argv[2]);
    temp3 = RSTRING_LEN(argv[2]);
    arg4 = (apr_size_t *)&temp3;
  }
  {
    result = (svn_error_t *)svn_write_invoke_fn(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM(*arg4));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_close_invoke_fn(int argc, VALUE *argv, VALUE self) {
  svn_close_fn_t arg1 = (svn_close_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  int res2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_close_fn_t","svn_close_invoke_fn", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_close_invoke_fn", 2, argv[1] )); 
  }
  {
    result = (svn_error_t *)svn_close_invoke_fn(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_invoke_mark_fn(int argc, VALUE *argv, VALUE self) {
  svn_stream_mark_fn_t arg1 = (svn_stream_mark_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_stream_mark_t **arg3 = (svn_stream_mark_t **) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  svn_stream_mark_t *temp3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_mark_fn_t","svn_stream_invoke_mark_fn", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_stream_invoke_mark_fn", 2, argv[1] )); 
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_stream_invoke_mark_fn(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_stream_invoke_mark_fn arg 3 (svn_stream_mark_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_stream_invoke_mark_fn is not implemented yet");
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_invoke_seek_fn(int argc, VALUE *argv, VALUE self) {
  svn_stream_seek_fn_t arg1 = (svn_stream_seek_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_stream_mark_t *arg3 = (svn_stream_mark_t *) 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_seek_fn_t","svn_stream_invoke_seek_fn", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_stream_invoke_seek_fn", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_stream_mark_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_stream_mark_t const *","svn_stream_invoke_seek_fn", 3, argv[2] )); 
  }
  arg3 = (svn_stream_mark_t *)(argp3);
  {
    result = (svn_error_t *)svn_stream_invoke_seek_fn(arg1,arg2,(struct svn_stream_mark_t const *)arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_invoke_data_available_fn(int argc, VALUE *argv, VALUE self) {
  svn_stream_data_available_fn_t arg1 = (svn_stream_data_available_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_boolean_t *arg3 = (svn_boolean_t *) 0 ;
  int res2 ;
  svn_boolean_t temp3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg3 = &temp3;
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_svn_boolean_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_data_available_fn_t","svn_stream_invoke_data_available_fn", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_stream_invoke_data_available_fn", 2, argv[1] )); 
  }
  {
    result = (svn_error_t *)svn_stream_invoke_data_available_fn(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_invoke_readline_fn(int argc, VALUE *argv, VALUE self) {
  svn_stream_readline_fn_t arg1 = (svn_stream_readline_fn_t) 0 ;
  void *arg2 = (void *) 0 ;
  svn_stringbuf_t **arg3 = (svn_stringbuf_t **) 0 ;
  char *arg4 = (char *) 0 ;
  svn_boolean_t *arg5 = (svn_boolean_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  svn_stringbuf_t *temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_boolean_t temp5 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  arg5 = &temp5;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_readline_fn_t","svn_stream_invoke_readline_fn", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_stream_invoke_readline_fn", 2, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_stream_invoke_readline_fn", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_stream_invoke_readline_fn(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg3)->data, (*arg3)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg5 ? Qtrue : Qfalse);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_stream_invoke_lazyopen_func(int argc, VALUE *argv, VALUE self) {
  svn_stream_lazyopen_func_t arg1 = (svn_stream_lazyopen_func_t) 0 ;
  svn_stream_t **arg2 = (svn_stream_t **) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp2 ;
  int res3 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_stream_lazyopen_func_t","svn_stream_invoke_lazyopen_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_stream_invoke_lazyopen_func", 3, argv[1] )); 
  }
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_stream_invoke_lazyopen_func(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_stream_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_io_invoke_walk_func(int argc, VALUE *argv, VALUE self) {
  svn_io_walk_func_t arg1 = (svn_io_walk_func_t) 0 ;
  void *arg2 = (void *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_finfo_t *arg4 = (apr_finfo_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_io_walk_func_t","svn_io_invoke_walk_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_io_invoke_walk_func", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_io_invoke_walk_func", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_apr_finfo_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "apr_finfo_t const *","svn_io_invoke_walk_func", 4, argv[3] )); 
  }
  arg4 = (apr_finfo_t *)(argp4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_io_invoke_walk_func(arg1,arg2,(char const *)arg3,(apr_finfo_t const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_checksum_t;

SWIGINTERN VALUE
_wrap_svn_checksum_t_digest_set(int argc, VALUE *argv, VALUE self) {
  struct svn_checksum_t *arg1 = (struct svn_checksum_t *) 0 ;
  unsigned char *arg2 = (unsigned char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_checksum_t *","digest", 1, self )); 
  }
  arg1 = (struct svn_checksum_t *)(argp1);
  
  if (arg1) (arg1)->digest = (unsigned char const *)arg2;
  
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_t_digest_get(int argc, VALUE *argv, VALUE self) {
  struct svn_checksum_t *arg1 = (struct svn_checksum_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_checksum_t *","digest", 1, self )); 
  }
  arg1 = (struct svn_checksum_t *)(argp1);
  result = (unsigned char *) ((arg1)->digest);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_checksum_t *arg1 = (struct svn_checksum_t *) 0 ;
  svn_checksum_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_checksum_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_checksum_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_checksum_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_checksum_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_checksum_t *arg1 = (struct svn_checksum_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_checksum_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_checksum_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_checksum_t *)(argp1);
  result = (svn_checksum_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_checksum_t_allocate(VALUE self)
#else
_wrap_svn_checksum_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_checksum_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_checksum_t(int argc, VALUE *argv, VALUE self) {
  struct svn_checksum_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_checksum_t *)calloc(1, sizeof(struct svn_checksum_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_checksum_t(void *self) {
    struct svn_checksum_t *arg1 = (struct svn_checksum_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_checksum_create(int argc, VALUE *argv, VALUE self) {
  svn_checksum_kind_t arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int val1 ;
  int ecode1 = 0 ;
  svn_checksum_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "svn_checksum_kind_t","svn_checksum_create", 1, argv[0] ));
  } 
  arg1 = (svn_checksum_kind_t)(val1);
  if (argc > 1) {
    
  }
  {
    result = (svn_checksum_t *)svn_checksum_create(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_clear(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t *","svn_checksum_clear", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  {
    result = (svn_error_t *)svn_checksum_clear(arg1);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_match(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  svn_checksum_t *arg2 = (svn_checksum_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_match", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_match", 2, argv[1] )); 
  }
  arg2 = (svn_checksum_t *)(argp2);
  {
    result = (svn_boolean_t)svn_checksum_match((struct svn_checksum_t const *)arg1,(struct svn_checksum_t const *)arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_dup(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_checksum_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_dup", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_checksum_t *)svn_checksum_dup((struct svn_checksum_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_to_cstring_display(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_to_cstring_display", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_checksum_to_cstring_display((struct svn_checksum_t const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_to_cstring(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_to_cstring", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_checksum_to_cstring((struct svn_checksum_t const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_serialize(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_serialize", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (char *)svn_checksum_serialize((struct svn_checksum_t const *)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_deserialize(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t **arg1 = (svn_checksum_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_checksum_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_checksum_deserialize", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_checksum_deserialize((struct svn_checksum_t const **)arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_checksum_deserialize arg 1 (svn_checksum_t const **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_checksum_deserialize is not implemented yet");
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_parse_hex(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t **arg1 = (svn_checksum_t **) 0 ;
  svn_checksum_kind_t arg2 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_checksum_t *temp1 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_checksum_kind_t","svn_checksum_parse_hex", 2, argv[0] ));
  } 
  arg2 = (svn_checksum_kind_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_checksum_parse_hex", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_checksum_parse_hex(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_checksum_parse_hex arg 1 (svn_checksum_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_checksum_parse_hex is not implemented yet");
    
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t **arg1 = (svn_checksum_t **) 0 ;
  svn_checksum_kind_t arg2 ;
  void *arg3 = (void *) 0 ;
  apr_size_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_checksum_t *temp1 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  unsigned long val4 ;
  int ecode4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_checksum_kind_t","svn_checksum", 2, argv[0] ));
  } 
  arg2 = (svn_checksum_kind_t)(val2);
  res3 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void const *","svn_checksum", 3, argv[1] )); 
  }
  ecode4 = SWIG_AsVal_unsigned_SS_long(argv[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "apr_size_t","svn_checksum", 4, argv[2] ));
  } 
  arg4 = (apr_size_t)(val4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_checksum(arg1,arg2,(void const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_checksum arg 1 (svn_checksum_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_checksum is not implemented yet");
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_empty_checksum(int argc, VALUE *argv, VALUE self) {
  svn_checksum_kind_t arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int val1 ;
  int ecode1 = 0 ;
  svn_checksum_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "svn_checksum_kind_t","svn_checksum_empty_checksum", 1, argv[0] ));
  } 
  arg1 = (svn_checksum_kind_t)(val1);
  if (argc > 1) {
    
  }
  {
    result = (svn_checksum_t *)svn_checksum_empty_checksum(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_ctx_create(int argc, VALUE *argv, VALUE self) {
  svn_checksum_kind_t arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int val1 ;
  int ecode1 = 0 ;
  svn_checksum_ctx_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "svn_checksum_kind_t","svn_checksum_ctx_create", 1, argv[0] ));
  } 
  arg1 = (svn_checksum_kind_t)(val1);
  if (argc > 1) {
    
  }
  {
    result = (svn_checksum_ctx_t *)svn_checksum_ctx_create(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_checksum_ctx_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_ctx_reset(int argc, VALUE *argv, VALUE self) {
  svn_checksum_ctx_t *arg1 = (svn_checksum_ctx_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_ctx_t *","svn_checksum_ctx_reset", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_ctx_t *)(argp1);
  {
    result = (svn_error_t *)svn_checksum_ctx_reset(arg1);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_update(int argc, VALUE *argv, VALUE self) {
  svn_checksum_ctx_t *arg1 = (svn_checksum_ctx_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_ctx_t *","svn_checksum_update", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_ctx_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void const *","svn_checksum_update", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_unsigned_SS_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "apr_size_t","svn_checksum_update", 3, argv[2] ));
  } 
  arg3 = (apr_size_t)(val3);
  {
    result = (svn_error_t *)svn_checksum_update(arg1,(void const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_final(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t **arg1 = (svn_checksum_t **) 0 ;
  svn_checksum_ctx_t *arg2 = (svn_checksum_ctx_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_checksum_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_checksum_ctx_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_checksum_ctx_t const *","svn_checksum_final", 2, argv[0] )); 
  }
  arg2 = (svn_checksum_ctx_t *)(argp2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_checksum_final(arg1,(struct svn_checksum_ctx_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    /* FIXME: Missing argout typemap: svn_checksum_final arg 1 (svn_checksum_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_checksum_final is not implemented yet");
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_size(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_size_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_size", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  {
    result = svn_checksum_size((struct svn_checksum_t const *)arg1);
    
    
    
  }
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_is_empty_checksum(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t *","svn_checksum_is_empty_checksum", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  {
    result = (svn_boolean_t)svn_checksum_is_empty_checksum(arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_checksum_mismatch_err(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t *arg1 = (svn_checksum_t *) 0 ;
  svn_checksum_t *arg2 = (svn_checksum_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  char *arg4 = (char *) 0 ;
  void *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if (argc < 2) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_mismatch_err", 1, argv[0] )); 
  }
  arg1 = (svn_checksum_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_checksum_mismatch_err", 2, argv[1] )); 
  }
  arg2 = (svn_checksum_t *)(argp2);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_checksum_mismatch_err", 4, argv[3] ));
    }
    arg4 = (char *)(buf4);
  }
  {
    result = (svn_error_t *)svn_checksum_mismatch_err((struct svn_checksum_t const *)arg1,(struct svn_checksum_t const *)arg2,arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


static swig_class SwigClassSvn_cache_config_t;

SWIGINTERN VALUE
_wrap_svn_cache_config_t_cache_size_set(int argc, VALUE *argv, VALUE self) {
  struct svn_cache_config_t *arg1 = (struct svn_cache_config_t *) 0 ;
  apr_uint64_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_cache_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_cache_config_t *","cache_size", 1, self )); 
  }
  arg1 = (struct svn_cache_config_t *)(argp1);
  {
    arg2 = (apr_uint64_t)NUM2ULL(argv[0]);
  }
  if (arg1) (arg1)->cache_size = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cache_config_t_cache_size_get(int argc, VALUE *argv, VALUE self) {
  struct svn_cache_config_t *arg1 = (struct svn_cache_config_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_uint64_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_cache_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_cache_config_t *","cache_size", 1, self )); 
  }
  arg1 = (struct svn_cache_config_t *)(argp1);
  result =  ((arg1)->cache_size);
  {
    vresult = ULL2NUM((apr_uint64_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cache_config_t_file_handle_count_set(int argc, VALUE *argv, VALUE self) {
  struct svn_cache_config_t *arg1 = (struct svn_cache_config_t *) 0 ;
  apr_size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_cache_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_cache_config_t *","file_handle_count", 1, self )); 
  }
  arg1 = (struct svn_cache_config_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_size_t","file_handle_count", 2, argv[0] ));
  } 
  arg2 = (apr_size_t)(val2);
  if (arg1) (arg1)->file_handle_count = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cache_config_t_file_handle_count_get(int argc, VALUE *argv, VALUE self) {
  struct svn_cache_config_t *arg1 = (struct svn_cache_config_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_size_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_cache_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_cache_config_t *","file_handle_count", 1, self )); 
  }
  arg1 = (struct svn_cache_config_t *)(argp1);
  result =  ((arg1)->file_handle_count);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cache_config_t_single_threaded_set(int argc, VALUE *argv, VALUE self) {
  struct svn_cache_config_t *arg1 = (struct svn_cache_config_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_cache_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_cache_config_t *","single_threaded", 1, self )); 
  }
  arg1 = (struct svn_cache_config_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->single_threaded = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cache_config_t_single_threaded_get(int argc, VALUE *argv, VALUE self) {
  struct svn_cache_config_t *arg1 = (struct svn_cache_config_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_cache_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_cache_config_t *","single_threaded", 1, self )); 
  }
  arg1 = (struct svn_cache_config_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->single_threaded);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_cache_config_t_allocate(VALUE self)
#else
_wrap_svn_cache_config_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_cache_config_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_cache_config_t(int argc, VALUE *argv, VALUE self) {
  struct svn_cache_config_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_cache_config_t *)calloc(1, sizeof(struct svn_cache_config_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_cache_config_t(void *self) {
    struct svn_cache_config_t *arg1 = (struct svn_cache_config_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_cache_config_get(int argc, VALUE *argv, VALUE self) {
  svn_cache_config_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (svn_cache_config_t *)svn_cache_config_get();
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_cache_config_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_cache_config_set(int argc, VALUE *argv, VALUE self) {
  svn_cache_config_t *arg1 = (svn_cache_config_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_cache_config_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_cache_config_t const *","svn_cache_config_set", 1, argv[0] )); 
  }
  arg1 = (svn_cache_config_t *)(argp1);
  {
    svn_cache_config_set((struct svn_cache_config_t const *)arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_auth_set_gnome_keyring_unlock_prompt_func(int argc, VALUE *argv, VALUE self) {
  svn_auth_baton_t *arg1 = (svn_auth_baton_t *) 0 ;
  svn_auth_gnome_keyring_unlock_prompt_func_t arg2 = (svn_auth_gnome_keyring_unlock_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_auth_baton_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_auth_baton_t *","svn_auth_set_gnome_keyring_unlock_prompt_func", 1, argv[0] )); 
  }
  arg1 = (svn_auth_baton_t *)(argp1);
  {
    arg2 = (svn_auth_gnome_keyring_unlock_prompt_func_t) svn_swig_rb_auth_gnome_keyring_unlock_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[1], _global_svn_swig_rb_pool);
  }
  {
    svn_auth_set_gnome_keyring_unlock_prompt_func(arg1,arg2,arg3);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_md5_empty_string_digest(int argc, VALUE *argv, VALUE self) {
  unsigned char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (unsigned char *)svn_md5_empty_string_digest();
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_unsigned_char, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_md5_digest_to_cstring_display(int argc, VALUE *argv, VALUE self) {
  unsigned char *arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg1 = NULL;
    } else if (RSTRING_LEN(argv[0]) != APR_MD5_DIGESTSIZE) {
      rb_raise(rb_eArgError, "digest size (%d) must be %d",
        RSTRING_LEN(argv[0]), APR_MD5_DIGESTSIZE);
    } else {
      arg1 = (unsigned char *)StringValuePtr(argv[0]);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_md5_digest_to_cstring_display((unsigned char const (*))arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_md5_digest_to_cstring(int argc, VALUE *argv, VALUE self) {
  unsigned char *arg1 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg1 = NULL;
    } else if (RSTRING_LEN(argv[0]) != APR_MD5_DIGESTSIZE) {
      rb_raise(rb_eArgError, "digest size (%d) must be %d",
        RSTRING_LEN(argv[0]), APR_MD5_DIGESTSIZE);
    } else {
      arg1 = (unsigned char *)StringValuePtr(argv[0]);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (char *)svn_md5_digest_to_cstring((unsigned char const (*))arg1,arg2);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_md5_digests_match(int argc, VALUE *argv, VALUE self) {
  unsigned char *arg1 ;
  unsigned char *arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "unsigned char const []","svn_md5_digests_match", 1, argv[0] )); 
  } 
  arg1 = (unsigned char *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "unsigned char const []","svn_md5_digests_match", 2, argv[1] )); 
  } 
  arg2 = (unsigned char *)(argp2);
  {
    result = (svn_boolean_t)svn_md5_digests_match((unsigned char const (*))arg1,(unsigned char const (*))arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassApr_pool_wrapper_t;

SWIGINTERN VALUE
_wrap_apr_pool_wrapper_t_destroy(int argc, VALUE *argv, VALUE self) {
  VALUE arg1 = (VALUE) 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  arg1 = argv[0];
  {
    apr_pool_wrapper_t_destroy(arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_apr_pool_wrapper_t_set_default_max_free_size(int argc, VALUE *argv, VALUE self) {
  apr_size_t arg1 ;
  unsigned long val1 ;
  int ecode1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_unsigned_SS_long(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "apr_size_t","apr_pool_wrapper_t_set_default_max_free_size", 1, argv[0] ));
  } 
  arg1 = (apr_size_t)(val1);
  {
    apr_pool_wrapper_t_set_default_max_free_size(arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_apr_pool_wrapper_t_allocate(VALUE self)
#else
_wrap_apr_pool_wrapper_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_apr_pool_wrapper_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_apr_pool_wrapper_t(int argc, VALUE *argv, VALUE self) {
  apr_pool_wrapper_t *arg1 = (apr_pool_wrapper_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct apr_pool_wrapper_t *result = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_apr_pool_wrapper_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "apr_pool_wrapper_t *","apr_pool_wrapper_t", 1, argv[0] )); 
  }
  arg1 = (apr_pool_wrapper_t *)(argp1);
  {
    result = (struct apr_pool_wrapper_t *)new_apr_pool_wrapper_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void delete_apr_pool_wrapper_t(struct apr_pool_wrapper_t *self){
    apr_pool_wrapper_destroy(self);
    xfree(self);
  }
SWIGINTERN void
free_apr_pool_wrapper_t(void *self) {
    struct apr_pool_wrapper_t *arg1 = (struct apr_pool_wrapper_t *)self;
    delete_apr_pool_wrapper_t(arg1);
}

SWIGINTERN VALUE
_wrap_apr_pool_wrapper_t_set_max_free_size(int argc, VALUE *argv, VALUE self) {
  struct apr_pool_wrapper_t *arg1 = (struct apr_pool_wrapper_t *) 0 ;
  apr_size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_apr_pool_wrapper_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct apr_pool_wrapper_t *","set_max_free_size", 1, self )); 
  }
  arg1 = (struct apr_pool_wrapper_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "apr_size_t","set_max_free_size", 2, argv[0] ));
  } 
  arg2 = (apr_size_t)(val2);
  {
    apr_pool_wrapper_t_set_max_free_size(arg1,arg2);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_apr_pool_wrapper_t__destroy(int argc, VALUE *argv, VALUE self) {
  struct apr_pool_wrapper_t *arg1 = (struct apr_pool_wrapper_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_apr_pool_wrapper_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct apr_pool_wrapper_t *","_destroy", 1, self )); 
  }
  arg1 = (struct apr_pool_wrapper_t *)(argp1);
  {
    apr_pool_wrapper_t__destroy(arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_stream_t;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_stream_t_allocate(VALUE self)
#else
_wrap_svn_stream_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_stream_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_stream_t(int argc, VALUE *argv, VALUE self) {
  VALUE arg1 = (VALUE) 0 ;
  struct svn_stream_t *result = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  arg1 = argv[0];
  {
    result = (struct svn_stream_t *)new_svn_stream_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void delete_svn_stream_t(struct svn_stream_t *self){
  }
SWIGINTERN void
free_svn_stream_t(void *self) {
    struct svn_stream_t *arg1 = (struct svn_stream_t *)self;
    delete_svn_stream_t(arg1);
}

static swig_class SwigClassSvn_auth_baton_t;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_auth_baton_t_allocate(VALUE self)
#else
_wrap_svn_auth_baton_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_auth_baton_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_auth_baton_t(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t *arg1 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  struct svn_auth_baton_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_array_to_auth_provider_object_apr_array(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (struct svn_auth_baton_t *)new_svn_auth_baton_t(arg1,arg2);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_auth_baton_t(struct svn_auth_baton_t *self){
  }
SWIGINTERN void
free_svn_auth_baton_t(void *self) {
    struct svn_auth_baton_t *arg1 = (struct svn_auth_baton_t *)self;
    delete_svn_auth_baton_t(arg1);
}

SWIGINTERN VALUE
_wrap_svn_diff_version(int argc, VALUE *argv, VALUE self) {
  svn_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (svn_version_t *)svn_diff_version();
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_version_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_diff_fns2_t;

SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_datasources_open_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_off_t *,apr_off_t *,svn_diff_datasource_e const *,apr_size_t) = (svn_error_t *(*)(void *,apr_off_t *,apr_off_t *,svn_diff_datasource_e const *,apr_size_t)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","datasources_open", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,apr_off_t *,apr_off_t *,svn_diff_datasource_e const *,apr_size_t)","datasources_open", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->datasources_open = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_datasources_open_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,apr_off_t *,apr_off_t *,svn_diff_datasource_e const *,apr_size_t) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","datasources_open", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  result = (svn_error_t *(*)(void *,apr_off_t *,apr_off_t *,svn_diff_datasource_e const *,apr_size_t)) ((arg1)->datasources_open);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_datasource_close_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  svn_error_t *(*arg2)(void *,svn_diff_datasource_e) = (svn_error_t *(*)(void *,svn_diff_datasource_e)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","datasource_close", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_svn_diff_datasource_e__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,svn_diff_datasource_e)","datasource_close", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->datasource_close = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_datasource_close_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,svn_diff_datasource_e) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","datasource_close", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  result = (svn_error_t *(*)(void *,svn_diff_datasource_e)) ((arg1)->datasource_close);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_svn_diff_datasource_e__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_datasource_get_next_token_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  svn_error_t *(*arg2)(apr_uint32_t *,void **,void *,svn_diff_datasource_e) = (svn_error_t *(*)(apr_uint32_t *,void **,void *,svn_diff_datasource_e)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","datasource_get_next_token", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(apr_uint32_t *,void **,void *,svn_diff_datasource_e)","datasource_get_next_token", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->datasource_get_next_token = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_datasource_get_next_token_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(apr_uint32_t *,void **,void *,svn_diff_datasource_e) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","datasource_get_next_token", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  result = (svn_error_t *(*)(apr_uint32_t *,void **,void *,svn_diff_datasource_e)) ((arg1)->datasource_get_next_token);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_token_compare_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  svn_error_t *(*arg2)(void *,void *,void *,int *) = (svn_error_t *(*)(void *,void *,void *,int *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","token_compare", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_void_p_void_p_int__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,void *,void *,int *)","token_compare", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->token_compare = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_token_compare_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,void *,void *,int *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","token_compare", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  result = (svn_error_t *(*)(void *,void *,void *,int *)) ((arg1)->token_compare);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_void_p_void_p_int__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_token_discard_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  void (*arg2)(void *,void *) = (void (*)(void *,void *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","token_discard", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_void__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "void (*)(void *,void *)","token_discard", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->token_discard = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_token_discard_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void (*result)(void *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","token_discard", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  result = (void (*)(void *,void *)) ((arg1)->token_discard);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_void__void);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_token_discard_all_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  void (*arg2)(void *) = (void (*)(void *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","token_discard_all", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "void (*)(void *)","token_discard_all", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->token_discard_all = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_t_token_discard_all_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void (*result)(void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns2_t *","token_discard_all", 1, self )); 
  }
  arg1 = (struct svn_diff_fns2_t *)(argp1);
  result = (void (*)(void *)) ((arg1)->token_discard_all);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void__void);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_diff_fns2_t_allocate(VALUE self)
#else
_wrap_svn_diff_fns2_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_diff_fns2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_diff_fns2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_diff_fns2_t *)calloc(1, sizeof(struct svn_diff_fns2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_diff_fns2_t(void *self) {
    struct svn_diff_fns2_t *arg1 = (struct svn_diff_fns2_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_diff_fns_t;

SWIGINTERN VALUE
_wrap_svn_diff_fns_t_datasource_open_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  svn_error_t *(*arg2)(void *,svn_diff_datasource_e) = (svn_error_t *(*)(void *,svn_diff_datasource_e)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","datasource_open", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_svn_diff_datasource_e__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,svn_diff_datasource_e)","datasource_open", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->datasource_open = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_datasource_open_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,svn_diff_datasource_e) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","datasource_open", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  result = (svn_error_t *(*)(void *,svn_diff_datasource_e)) ((arg1)->datasource_open);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_svn_diff_datasource_e__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_datasource_close_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  svn_error_t *(*arg2)(void *,svn_diff_datasource_e) = (svn_error_t *(*)(void *,svn_diff_datasource_e)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","datasource_close", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_svn_diff_datasource_e__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,svn_diff_datasource_e)","datasource_close", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->datasource_close = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_datasource_close_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,svn_diff_datasource_e) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","datasource_close", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  result = (svn_error_t *(*)(void *,svn_diff_datasource_e)) ((arg1)->datasource_close);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_svn_diff_datasource_e__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_datasource_get_next_token_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  svn_error_t *(*arg2)(apr_uint32_t *,void **,void *,svn_diff_datasource_e) = (svn_error_t *(*)(apr_uint32_t *,void **,void *,svn_diff_datasource_e)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","datasource_get_next_token", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(apr_uint32_t *,void **,void *,svn_diff_datasource_e)","datasource_get_next_token", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->datasource_get_next_token = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_datasource_get_next_token_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(apr_uint32_t *,void **,void *,svn_diff_datasource_e) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","datasource_get_next_token", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  result = (svn_error_t *(*)(apr_uint32_t *,void **,void *,svn_diff_datasource_e)) ((arg1)->datasource_get_next_token);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_token_compare_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  svn_error_t *(*arg2)(void *,void *,void *,int *) = (svn_error_t *(*)(void *,void *,void *,int *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","token_compare", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_void_p_void_p_int__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,void *,void *,int *)","token_compare", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->token_compare = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_token_compare_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,void *,void *,int *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","token_compare", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  result = (svn_error_t *(*)(void *,void *,void *,int *)) ((arg1)->token_compare);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_void_p_void_p_int__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_token_discard_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  void (*arg2)(void *,void *) = (void (*)(void *,void *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","token_discard", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_p_void__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "void (*)(void *,void *)","token_discard", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->token_discard = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_token_discard_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void (*result)(void *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","token_discard", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  result = (void (*)(void *,void *)) ((arg1)->token_discard);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_void__void);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_token_discard_all_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  void (*arg2)(void *) = (void (*)(void *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","token_discard_all", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "void (*)(void *)","token_discard_all", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->token_discard_all = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_t_token_discard_all_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void (*result)(void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_fns_t *","token_discard_all", 1, self )); 
  }
  arg1 = (struct svn_diff_fns_t *)(argp1);
  result = (void (*)(void *)) ((arg1)->token_discard_all);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void__void);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_diff_fns_t_allocate(VALUE self)
#else
_wrap_svn_diff_fns_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_diff_fns_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_diff_fns_t(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_fns_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_diff_fns_t *)calloc(1, sizeof(struct svn_diff_fns_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_diff_fns_t(void *self) {
    struct svn_diff_fns_t *arg1 = (struct svn_diff_fns_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_diff_diff_2(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_fns2_t *arg3 = (svn_diff_fns2_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_diff_2", 2, argv[0] )); 
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_diff_fns2_t const *","svn_diff_diff_2", 3, argv[1] )); 
  }
  arg3 = (svn_diff_fns2_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_diff_2(arg1,arg2,(struct svn_diff_fns2_t const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_diff(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_fns_t *arg3 = (svn_diff_fns_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_diff", 2, argv[0] )); 
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_diff_fns_t const *","svn_diff_diff", 3, argv[1] )); 
  }
  arg3 = (svn_diff_fns_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_diff(arg1,arg2,(struct svn_diff_fns_t const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_diff3_2(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_fns2_t *arg3 = (svn_diff_fns2_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_diff3_2", 2, argv[0] )); 
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_diff_fns2_t const *","svn_diff_diff3_2", 3, argv[1] )); 
  }
  arg3 = (svn_diff_fns2_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_diff3_2(arg1,arg2,(struct svn_diff_fns2_t const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_diff3(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_fns_t *arg3 = (svn_diff_fns_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_diff3", 2, argv[0] )); 
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_diff_fns_t const *","svn_diff_diff3", 3, argv[1] )); 
  }
  arg3 = (svn_diff_fns_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_diff3(arg1,arg2,(struct svn_diff_fns_t const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_diff4_2(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_fns2_t *arg3 = (svn_diff_fns2_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_diff4_2", 2, argv[0] )); 
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_diff_fns2_t const *","svn_diff_diff4_2", 3, argv[1] )); 
  }
  arg3 = (svn_diff_fns2_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_diff4_2(arg1,arg2,(struct svn_diff_fns2_t const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_diff4(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_fns_t *arg3 = (svn_diff_fns_t *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_diff4", 2, argv[0] )); 
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_diff_fns_t const *","svn_diff_diff4", 3, argv[1] )); 
  }
  arg3 = (svn_diff_fns_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_diff4(arg1,arg2,(struct svn_diff_fns_t const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_contains_conflicts(int argc, VALUE *argv, VALUE self) {
  svn_diff_t *arg1 = (svn_diff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_contains_conflicts", 1, argv[0] )); 
  }
  arg1 = (svn_diff_t *)(argp1);
  {
    result = (svn_boolean_t)svn_diff_contains_conflicts(arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_contains_diffs(int argc, VALUE *argv, VALUE self) {
  svn_diff_t *arg1 = (svn_diff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_contains_diffs", 1, argv[0] )); 
  }
  arg1 = (svn_diff_t *)(argp1);
  {
    result = (svn_boolean_t)svn_diff_contains_diffs(arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_diff_output_fns_t;

SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_common_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t) = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_common", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)","output_common", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->output_common = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_common_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_common", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  result = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)) ((arg1)->output_common);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_diff_modified_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t) = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_diff_modified", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)","output_diff_modified", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->output_diff_modified = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_diff_modified_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_diff_modified", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  result = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)) ((arg1)->output_diff_modified);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_diff_latest_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t) = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_diff_latest", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)","output_diff_latest", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->output_diff_latest = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_diff_latest_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_diff_latest", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  result = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)) ((arg1)->output_diff_latest);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_diff_common_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t) = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_diff_common", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)","output_diff_common", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->output_diff_common = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_diff_common_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_diff_common", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  result = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)) ((arg1)->output_diff_common);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_conflict_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  svn_error_t *(*arg2)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,svn_diff_t *) = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,svn_diff_t *)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_conflict", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,svn_diff_t *)","output_conflict", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->output_conflict = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_t_output_conflict_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,svn_diff_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_output_fns_t *","output_conflict", 1, self )); 
  }
  arg1 = (struct svn_diff_output_fns_t *)(argp1);
  result = (svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,svn_diff_t *)) ((arg1)->output_conflict);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_diff_output_fns_t_allocate(VALUE self)
#else
_wrap_svn_diff_output_fns_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_diff_output_fns_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_diff_output_fns_t(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_output_fns_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_diff_output_fns_t *)calloc(1, sizeof(struct svn_diff_output_fns_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_diff_output_fns_t(void *self) {
    struct svn_diff_output_fns_t *arg1 = (struct svn_diff_output_fns_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_diff_output2(int argc, VALUE *argv, VALUE self) {
  svn_diff_t *arg1 = (svn_diff_t *) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_output_fns_t *arg3 = (svn_diff_output_fns_t *) 0 ;
  svn_cancel_func_t arg4 = (svn_cancel_func_t) 0 ;
  void *arg5 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_output2", 1, argv[0] )); 
  }
  arg1 = (svn_diff_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_output2", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_diff_output_fns_t const *","svn_diff_output2", 3, argv[2] )); 
  }
  arg3 = (svn_diff_output_fns_t *)(argp3);
  {
    arg4 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  {
    result = (svn_error_t *)svn_diff_output2(arg1,arg2,(struct svn_diff_output_fns_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output(int argc, VALUE *argv, VALUE self) {
  svn_diff_t *arg1 = (svn_diff_t *) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_output_fns_t *arg3 = (svn_diff_output_fns_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_output", 1, argv[0] )); 
  }
  arg1 = (svn_diff_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_output", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_diff_output_fns_t const *","svn_diff_output", 3, argv[2] )); 
  }
  arg3 = (svn_diff_output_fns_t *)(argp3);
  {
    result = (svn_error_t *)svn_diff_output(arg1,arg2,(struct svn_diff_output_fns_t const *)arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_diff_file_options_t;

SWIGINTERN VALUE
_wrap_svn_diff_file_options_t_ignore_space_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *) 0 ;
  svn_diff_file_ignore_space_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_file_options_t *","ignore_space", 1, self )); 
  }
  arg1 = (struct svn_diff_file_options_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_diff_file_ignore_space_t","ignore_space", 2, argv[0] ));
  } 
  arg2 = (svn_diff_file_ignore_space_t)(val2);
  if (arg1) (arg1)->ignore_space = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_options_t_ignore_space_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_diff_file_ignore_space_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_file_options_t *","ignore_space", 1, self )); 
  }
  arg1 = (struct svn_diff_file_options_t *)(argp1);
  result = (svn_diff_file_ignore_space_t) ((arg1)->ignore_space);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_options_t_ignore_eol_style_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_file_options_t *","ignore_eol_style", 1, self )); 
  }
  arg1 = (struct svn_diff_file_options_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->ignore_eol_style = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_options_t_ignore_eol_style_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_file_options_t *","ignore_eol_style", 1, self )); 
  }
  arg1 = (struct svn_diff_file_options_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->ignore_eol_style);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_options_t_show_c_function_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_file_options_t *","show_c_function", 1, self )); 
  }
  arg1 = (struct svn_diff_file_options_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->show_c_function = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_options_t_show_c_function_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_file_options_t *","show_c_function", 1, self )); 
  }
  arg1 = (struct svn_diff_file_options_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->show_c_function);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_options_t_context_size_set(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_file_options_t *","context_size", 1, self )); 
  }
  arg1 = (struct svn_diff_file_options_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","context_size", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->context_size = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_options_t_context_size_get(int argc, VALUE *argv, VALUE self) {
  struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_diff_file_options_t *","context_size", 1, self )); 
  }
  arg1 = (struct svn_diff_file_options_t *)(argp1);
  result = (int) ((arg1)->context_size);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_diff_file_options_t_allocate(VALUE self)
#else
_wrap_svn_diff_file_options_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_diff_file_options_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_diff_file_options_t(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  struct svn_diff_file_options_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (struct svn_diff_file_options_t *)new_svn_diff_file_options_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_diff_file_options_t(struct svn_diff_file_options_t *self){
  }
SWIGINTERN void
free_svn_diff_file_options_t(void *self) {
    struct svn_diff_file_options_t *arg1 = (struct svn_diff_file_options_t *)self;
    delete_svn_diff_file_options_t(arg1);
}

SWIGINTERN VALUE
_wrap_svn_diff_file_options_parse(int argc, VALUE *argv, VALUE self) {
  svn_diff_file_options_t *arg1 = (svn_diff_file_options_t *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_file_options_t *","svn_diff_file_options_parse", 1, argv[0] )); 
  }
  arg1 = (svn_diff_file_options_t *)(argp1);
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_options_parse(arg1,(apr_array_header_t const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_diff_2(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  svn_diff_file_options_t *arg4 = (svn_diff_file_options_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff_2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff_2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_diff_file_diff_2", 4, argv[2] )); 
  }
  arg4 = (svn_diff_file_options_t *)(argp4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_diff_2(arg1,(char const *)arg2,(char const *)arg3,(struct svn_diff_file_options_t const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_diff(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_diff(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_diff3_2(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  svn_diff_file_options_t *arg5 = (svn_diff_file_options_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff3_2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff3_2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff3_2", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_diff_file_diff3_2", 5, argv[3] )); 
  }
  arg5 = (svn_diff_file_options_t *)(argp5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_diff3_2(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,(struct svn_diff_file_options_t const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_diff3(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff3", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_diff3(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_diff4_2(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_diff_file_options_t *arg6 = (svn_diff_file_options_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff4_2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff4_2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff4_2", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff4_2", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_diff_file_diff4_2", 6, argv[4] )); 
  }
  arg6 = (svn_diff_file_options_t *)(argp6);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_diff4_2(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(struct svn_diff_file_options_t const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_diff4(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff4", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff4", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff4", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_diff4", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_diff4(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_output_unified4(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  svn_boolean_t arg9 ;
  int arg10 ;
  svn_cancel_func_t arg11 = (svn_cancel_func_t) 0 ;
  void *arg12 = (void *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  int val10 ;
  int ecode10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_file_output_unified4", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified4", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified4", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified4", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified4", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  {
    arg7 = NULL;
    
    if (NIL_P(argv[6])) {
      
    } else if (TYPE(argv[6]) == T_FIXNUM) {
      arg7 = (char *)NUM2INT(argv[6]);
      if (!(arg7 == APR_LOCALE_CHARSET || arg7 == APR_DEFAULT_CHARSET)) {
        arg7 = NULL;
      }
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
    
    if (!arg7) {
      arg7 = (char *)APR_LOCALE_CHARSET;
    }
  }
  res8 = SWIG_AsCharPtrAndSize(argv[7], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified4", 8, argv[7] ));
  }
  arg8 = (char *)(buf8);
  arg9 = RTEST(argv[8]);
  ecode10 = SWIG_AsVal_int(argv[9], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), Ruby_Format_TypeError( "", "int","svn_diff_file_output_unified4", 10, argv[9] ));
  } 
  arg10 = (int)(val10);
  {
    arg11 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg12 = (void *)svn_swig_rb_make_baton(argv[10], _global_svn_swig_rb_pool);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_output_unified4(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg12);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_output_unified3(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  svn_boolean_t arg9 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_file_output_unified3", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified3", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified3", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  {
    arg7 = NULL;
    
    if (NIL_P(argv[6])) {
      
    } else if (TYPE(argv[6]) == T_FIXNUM) {
      arg7 = (char *)NUM2INT(argv[6]);
      if (!(arg7 == APR_LOCALE_CHARSET || arg7 == APR_DEFAULT_CHARSET)) {
        arg7 = NULL;
      }
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
    
    if (!arg7) {
      arg7 = (char *)APR_LOCALE_CHARSET;
    }
  }
  res8 = SWIG_AsCharPtrAndSize(argv[7], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified3", 8, argv[7] ));
  }
  arg8 = (char *)(buf8);
  arg9 = RTEST(argv[8]);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_output_unified3(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_output_unified2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_file_output_unified2", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified2", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified2", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  {
    arg7 = NULL;
    
    if (NIL_P(argv[6])) {
      
    } else if (TYPE(argv[6]) == T_FIXNUM) {
      arg7 = (char *)NUM2INT(argv[6]);
      if (!(arg7 == APR_LOCALE_CHARSET || arg7 == APR_DEFAULT_CHARSET)) {
        arg7 = NULL;
      }
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
    
    if (!arg7) {
      arg7 = (char *)APR_LOCALE_CHARSET;
    }
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_output_unified2(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_output_unified(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_file_output_unified", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_unified", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_output_unified(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_output_merge3(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  svn_diff_conflict_display_style_t arg10 ;
  svn_cancel_func_t arg11 = (svn_cancel_func_t) 0 ;
  void *arg12 = (void *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int val10 ;
  int ecode10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_file_output_merge3", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge3", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  {
    if (NIL_P(argv[7])) {
      arg8 = NULL;
    } else {
      arg8 = StringValuePtr(argv[7]);
    }
  }
  {
    if (NIL_P(argv[8])) {
      arg9 = NULL;
    } else {
      arg9 = StringValuePtr(argv[8]);
    }
  }
  ecode10 = SWIG_AsVal_int(argv[9], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), Ruby_Format_TypeError( "", "svn_diff_conflict_display_style_t","svn_diff_file_output_merge3", 10, argv[9] ));
  } 
  arg10 = (svn_diff_conflict_display_style_t)(val10);
  {
    arg11 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg12 = (void *)svn_swig_rb_make_baton(argv[10], _global_svn_swig_rb_pool);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_output_merge3(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg12);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_output_merge2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  svn_diff_conflict_display_style_t arg10 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int val10 ;
  int ecode10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_file_output_merge2", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge2", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  {
    if (NIL_P(argv[7])) {
      arg8 = NULL;
    } else {
      arg8 = StringValuePtr(argv[7]);
    }
  }
  {
    if (NIL_P(argv[8])) {
      arg9 = NULL;
    } else {
      arg9 = StringValuePtr(argv[8]);
    }
  }
  ecode10 = SWIG_AsVal_int(argv[9], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), Ruby_Format_TypeError( "", "svn_diff_conflict_display_style_t","svn_diff_file_output_merge2", 10, argv[9] ));
  } 
  arg10 = (svn_diff_conflict_display_style_t)(val10);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_output_merge2(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_file_output_merge(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_file_output_merge", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_file_output_merge", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  {
    if (NIL_P(argv[7])) {
      arg8 = NULL;
    } else {
      arg8 = StringValuePtr(argv[7]);
    }
  }
  {
    if (NIL_P(argv[8])) {
      arg9 = NULL;
    } else {
      arg9 = StringValuePtr(argv[8]);
    }
  }
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_diff_file_output_merge(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_binary(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_stream_t *arg2 = (svn_stream_t *) 0 ;
  svn_stream_t *arg3 = (svn_stream_t *) 0 ;
  svn_cancel_func_t arg4 = (svn_cancel_func_t) 0 ;
  void *arg5 = (void *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  {
    arg2 = svn_swig_rb_make_stream(argv[1]);
  }
  {
    arg3 = svn_swig_rb_make_stream(argv[2]);
  }
  {
    arg4 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_diff_output_binary(arg1,arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_diff(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  svn_diff_file_options_t *arg4 = (svn_diff_file_options_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  svn_string_t value2 ;
  svn_string_t value3 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[0]);
      value2.len = RSTRING_LEN(argv[0]);
      arg2 = &value2;
    }
  }
  {
    if (NIL_P(argv[1])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[1]);
      value3.len = RSTRING_LEN(argv[1]);
      arg3 = &value3;
    }
  }
  res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_diff_mem_string_diff", 4, argv[2] )); 
  }
  arg4 = (svn_diff_file_options_t *)(argp4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_diff(arg1,(struct svn_string_t const *)arg2,(struct svn_string_t const *)arg3,(struct svn_diff_file_options_t const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_diff3(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  svn_diff_file_options_t *arg5 = (svn_diff_file_options_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  svn_string_t value2 ;
  svn_string_t value3 ;
  svn_string_t value4 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[0]);
      value2.len = RSTRING_LEN(argv[0]);
      arg2 = &value2;
    }
  }
  {
    if (NIL_P(argv[1])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[1]);
      value3.len = RSTRING_LEN(argv[1]);
      arg3 = &value3;
    }
  }
  {
    if (NIL_P(argv[2])) {
      arg4 = NULL;
    } else {
      value4.data = StringValuePtr(argv[2]);
      value4.len = RSTRING_LEN(argv[2]);
      arg4 = &value4;
    }
  }
  res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_diff_mem_string_diff3", 5, argv[3] )); 
  }
  arg5 = (svn_diff_file_options_t *)(argp5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_diff3(arg1,(struct svn_string_t const *)arg2,(struct svn_string_t const *)arg3,(struct svn_string_t const *)arg4,(struct svn_diff_file_options_t const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_diff4(int argc, VALUE *argv, VALUE self) {
  svn_diff_t **arg1 = (svn_diff_t **) 0 ;
  svn_string_t *arg2 = (svn_string_t *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  svn_string_t *arg5 = (svn_string_t *) 0 ;
  svn_diff_file_options_t *arg6 = (svn_diff_file_options_t *) 0 ;
  apr_pool_t *arg7 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_diff_t *temp1 ;
  svn_string_t value2 ;
  svn_string_t value3 ;
  svn_string_t value4 ;
  svn_string_t value5 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[0]);
      value2.len = RSTRING_LEN(argv[0]);
      arg2 = &value2;
    }
  }
  {
    if (NIL_P(argv[1])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[1]);
      value3.len = RSTRING_LEN(argv[1]);
      arg3 = &value3;
    }
  }
  {
    if (NIL_P(argv[2])) {
      arg4 = NULL;
    } else {
      value4.data = StringValuePtr(argv[2]);
      value4.len = RSTRING_LEN(argv[2]);
      arg4 = &value4;
    }
  }
  {
    if (NIL_P(argv[3])) {
      arg5 = NULL;
    } else {
      value5.data = StringValuePtr(argv[3]);
      value5.len = RSTRING_LEN(argv[3]);
      arg5 = &value5;
    }
  }
  res6 = SWIG_ConvertPtr(argv[4], &argp6,SWIGTYPE_p_svn_diff_file_options_t, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_diff_file_options_t const *","svn_diff_mem_string_diff4", 6, argv[4] )); 
  }
  arg6 = (svn_diff_file_options_t *)(argp6);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_diff4(arg1,(struct svn_string_t const *)arg2,(struct svn_string_t const *)arg3,(struct svn_string_t const *)arg4,(struct svn_string_t const *)arg5,(struct svn_diff_file_options_t const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_diff_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_output_unified3(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  svn_boolean_t arg3 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_string_t *arg8 = (svn_string_t *) 0 ;
  svn_string_t *arg9 = (svn_string_t *) 0 ;
  int arg10 ;
  svn_cancel_func_t arg11 = (svn_cancel_func_t) 0 ;
  void *arg12 = (void *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_string_t value8 ;
  svn_string_t value9 ;
  int val10 ;
  int ecode10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_mem_string_output_unified3", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  arg3 = RTEST(argv[2]);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_mem_string_output_unified3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_mem_string_output_unified3", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_diff_mem_string_output_unified3", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  {
    arg7 = NULL;
    
    if (NIL_P(argv[6])) {
      
    } else if (TYPE(argv[6]) == T_FIXNUM) {
      arg7 = (char *)NUM2INT(argv[6]);
      if (!(arg7 == APR_LOCALE_CHARSET || arg7 == APR_DEFAULT_CHARSET)) {
        arg7 = NULL;
      }
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
    
    if (!arg7) {
      arg7 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    if (NIL_P(argv[7])) {
      arg8 = NULL;
    } else {
      value8.data = StringValuePtr(argv[7]);
      value8.len = RSTRING_LEN(argv[7]);
      arg8 = &value8;
    }
  }
  {
    if (NIL_P(argv[8])) {
      arg9 = NULL;
    } else {
      value9.data = StringValuePtr(argv[8]);
      value9.len = RSTRING_LEN(argv[8]);
      arg9 = &value9;
    }
  }
  ecode10 = SWIG_AsVal_int(argv[9], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), Ruby_Format_TypeError( "", "int","svn_diff_mem_string_output_unified3", 10, argv[9] ));
  } 
  arg10 = (int)(val10);
  {
    arg11 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg12 = (void *)svn_swig_rb_make_baton(argv[10], _global_svn_swig_rb_pool);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_output_unified3(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(struct svn_string_t const *)arg8,(struct svn_string_t const *)arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg12);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_output_unified2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  svn_boolean_t arg3 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  svn_string_t *arg8 = (svn_string_t *) 0 ;
  svn_string_t *arg9 = (svn_string_t *) 0 ;
  apr_pool_t *arg10 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  svn_string_t value8 ;
  svn_string_t value9 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_mem_string_output_unified2", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  arg3 = RTEST(argv[2]);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_mem_string_output_unified2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_diff_mem_string_output_unified2", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_diff_mem_string_output_unified2", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  {
    arg7 = NULL;
    
    if (NIL_P(argv[6])) {
      
    } else if (TYPE(argv[6]) == T_FIXNUM) {
      arg7 = (char *)NUM2INT(argv[6]);
      if (!(arg7 == APR_LOCALE_CHARSET || arg7 == APR_DEFAULT_CHARSET)) {
        arg7 = NULL;
      }
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
    
    if (!arg7) {
      arg7 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    if (NIL_P(argv[7])) {
      arg8 = NULL;
    } else {
      value8.data = StringValuePtr(argv[7]);
      value8.len = RSTRING_LEN(argv[7]);
      arg8 = &value8;
    }
  }
  {
    if (NIL_P(argv[8])) {
      arg9 = NULL;
    } else {
      value9.data = StringValuePtr(argv[8]);
      value9.len = RSTRING_LEN(argv[8]);
      arg9 = &value9;
    }
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_output_unified2(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(struct svn_string_t const *)arg8,(struct svn_string_t const *)arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_output_unified(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  svn_string_t *arg6 = (svn_string_t *) 0 ;
  svn_string_t *arg7 = (svn_string_t *) 0 ;
  apr_pool_t *arg8 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_string_t value6 ;
  svn_string_t value7 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_mem_string_output_unified", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_diff_mem_string_output_unified", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_diff_mem_string_output_unified", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = NULL;
    
    if (NIL_P(argv[4])) {
      
    } else if (TYPE(argv[4]) == T_FIXNUM) {
      arg5 = (char *)NUM2INT(argv[4]);
      if (!(arg5 == APR_LOCALE_CHARSET || arg5 == APR_DEFAULT_CHARSET)) {
        arg5 = NULL;
      }
    } else {
      arg5 = StringValuePtr(argv[4]);
    }
    
    if (!arg5) {
      arg5 = (char *)APR_LOCALE_CHARSET;
    }
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      value6.data = StringValuePtr(argv[5]);
      value6.len = RSTRING_LEN(argv[5]);
      arg6 = &value6;
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      value7.data = StringValuePtr(argv[6]);
      value7.len = RSTRING_LEN(argv[6]);
      arg7 = &value7;
    }
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_output_unified(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(struct svn_string_t const *)arg6,(struct svn_string_t const *)arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_output_merge3(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  svn_string_t *arg5 = (svn_string_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  svn_diff_conflict_display_style_t arg10 ;
  svn_cancel_func_t arg11 = (svn_cancel_func_t) 0 ;
  void *arg12 = (void *) 0 ;
  apr_pool_t *arg13 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_string_t value3 ;
  svn_string_t value4 ;
  svn_string_t value5 ;
  int val10 ;
  int ecode10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_mem_string_output_merge3", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[2]);
      value3.len = RSTRING_LEN(argv[2]);
      arg3 = &value3;
    }
  }
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      value4.data = StringValuePtr(argv[3]);
      value4.len = RSTRING_LEN(argv[3]);
      arg4 = &value4;
    }
  }
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      value5.data = StringValuePtr(argv[4]);
      value5.len = RSTRING_LEN(argv[4]);
      arg5 = &value5;
    }
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  {
    if (NIL_P(argv[7])) {
      arg8 = NULL;
    } else {
      arg8 = StringValuePtr(argv[7]);
    }
  }
  {
    if (NIL_P(argv[8])) {
      arg9 = NULL;
    } else {
      arg9 = StringValuePtr(argv[8]);
    }
  }
  ecode10 = SWIG_AsVal_int(argv[9], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), Ruby_Format_TypeError( "", "svn_diff_conflict_display_style_t","svn_diff_mem_string_output_merge3", 10, argv[9] ));
  } 
  arg10 = (svn_diff_conflict_display_style_t)(val10);
  {
    arg11 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg12 = (void *)svn_swig_rb_make_baton(argv[10], _global_svn_swig_rb_pool);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_output_merge3(arg1,arg2,(struct svn_string_t const *)arg3,(struct svn_string_t const *)arg4,(struct svn_string_t const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg12);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_output_merge2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  svn_string_t *arg5 = (svn_string_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  svn_diff_conflict_display_style_t arg10 ;
  apr_pool_t *arg11 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_string_t value3 ;
  svn_string_t value4 ;
  svn_string_t value5 ;
  int val10 ;
  int ecode10 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_mem_string_output_merge2", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[2]);
      value3.len = RSTRING_LEN(argv[2]);
      arg3 = &value3;
    }
  }
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      value4.data = StringValuePtr(argv[3]);
      value4.len = RSTRING_LEN(argv[3]);
      arg4 = &value4;
    }
  }
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      value5.data = StringValuePtr(argv[4]);
      value5.len = RSTRING_LEN(argv[4]);
      arg5 = &value5;
    }
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  {
    if (NIL_P(argv[7])) {
      arg8 = NULL;
    } else {
      arg8 = StringValuePtr(argv[7]);
    }
  }
  {
    if (NIL_P(argv[8])) {
      arg9 = NULL;
    } else {
      arg9 = StringValuePtr(argv[8]);
    }
  }
  ecode10 = SWIG_AsVal_int(argv[9], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), Ruby_Format_TypeError( "", "svn_diff_conflict_display_style_t","svn_diff_mem_string_output_merge2", 10, argv[9] ));
  } 
  arg10 = (svn_diff_conflict_display_style_t)(val10);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_output_merge2(arg1,arg2,(struct svn_string_t const *)arg3,(struct svn_string_t const *)arg4,(struct svn_string_t const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_mem_string_output_merge(int argc, VALUE *argv, VALUE self) {
  svn_stream_t *arg1 = (svn_stream_t *) 0 ;
  svn_diff_t *arg2 = (svn_diff_t *) 0 ;
  svn_string_t *arg3 = (svn_string_t *) 0 ;
  svn_string_t *arg4 = (svn_string_t *) 0 ;
  svn_string_t *arg5 = (svn_string_t *) 0 ;
  char *arg6 = (char *) 0 ;
  char *arg7 = (char *) 0 ;
  char *arg8 = (char *) 0 ;
  char *arg9 = (char *) 0 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  apr_pool_t *arg12 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_string_t value3 ;
  svn_string_t value4 ;
  svn_string_t value5 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = svn_swig_rb_make_stream(argv[0]);
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_mem_string_output_merge", 2, argv[1] )); 
  }
  arg2 = (svn_diff_t *)(argp2);
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[2]);
      value3.len = RSTRING_LEN(argv[2]);
      arg3 = &value3;
    }
  }
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      value4.data = StringValuePtr(argv[3]);
      value4.len = RSTRING_LEN(argv[3]);
      arg4 = &value4;
    }
  }
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      value5.data = StringValuePtr(argv[4]);
      value5.len = RSTRING_LEN(argv[4]);
      arg5 = &value5;
    }
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  {
    if (NIL_P(argv[7])) {
      arg8 = NULL;
    } else {
      arg8 = StringValuePtr(argv[7]);
    }
  }
  {
    if (NIL_P(argv[8])) {
      arg9 = NULL;
    } else {
      arg9 = StringValuePtr(argv[8]);
    }
  }
  arg10 = RTEST(argv[9]);
  arg11 = RTEST(argv[10]);
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_diff_mem_string_output_merge(arg1,arg2,(struct svn_string_t const *)arg3,(struct svn_string_t const *)arg4,(struct svn_string_t const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_readline_diff_text(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  svn_stringbuf_t **arg2 = (svn_stringbuf_t **) 0 ;
  char **arg3 = (char **) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_stringbuf_t *temp2 ;
  char *temp3 ;
  svn_boolean_t temp4 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t *","svn_diff_hunk_readline_diff_text", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_hunk_readline_diff_text(arg1,arg2,(char const **)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg2)->data, (*arg2)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_readline_original_text(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  svn_stringbuf_t **arg2 = (svn_stringbuf_t **) 0 ;
  char **arg3 = (char **) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_stringbuf_t *temp2 ;
  char *temp3 ;
  svn_boolean_t temp4 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t *","svn_diff_hunk_readline_original_text", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_hunk_readline_original_text(arg1,arg2,(char const **)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg2)->data, (*arg2)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_readline_modified_text(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  svn_stringbuf_t **arg2 = (svn_stringbuf_t **) 0 ;
  char **arg3 = (char **) 0 ;
  svn_boolean_t *arg4 = (svn_boolean_t *) 0 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_stringbuf_t *temp2 ;
  char *temp3 ;
  svn_boolean_t temp4 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t *","svn_diff_hunk_readline_modified_text", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_diff_hunk_readline_modified_text(arg1,arg2,(char const **)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg2)->data, (*arg2)->len));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3));
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil);
    }
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_reset_diff_text(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t *","svn_diff_hunk_reset_diff_text", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    svn_diff_hunk_reset_diff_text(arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_reset_original_text(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t *","svn_diff_hunk_reset_original_text", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    svn_diff_hunk_reset_original_text(arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_reset_modified_text(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t *","svn_diff_hunk_reset_modified_text", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    svn_diff_hunk_reset_modified_text(arg1);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_get_original_start(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t const *","svn_diff_hunk_get_original_start", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    result = (svn_linenum_t)svn_diff_hunk_get_original_start((struct svn_diff_hunk_t const *)arg1);
    
    
    
  }
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_get_original_length(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t const *","svn_diff_hunk_get_original_length", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    result = (svn_linenum_t)svn_diff_hunk_get_original_length((struct svn_diff_hunk_t const *)arg1);
    
    
    
  }
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_get_modified_start(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t const *","svn_diff_hunk_get_modified_start", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    result = (svn_linenum_t)svn_diff_hunk_get_modified_start((struct svn_diff_hunk_t const *)arg1);
    
    
    
  }
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_get_modified_length(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t const *","svn_diff_hunk_get_modified_length", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    result = (svn_linenum_t)svn_diff_hunk_get_modified_length((struct svn_diff_hunk_t const *)arg1);
    
    
    
  }
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_get_leading_context(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t const *","svn_diff_hunk_get_leading_context", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    result = (svn_linenum_t)svn_diff_hunk_get_leading_context((struct svn_diff_hunk_t const *)arg1);
    
    
    
  }
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_hunk_get_trailing_context(int argc, VALUE *argv, VALUE self) {
  svn_diff_hunk_t *arg1 = (svn_diff_hunk_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_hunk_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_hunk_t const *","svn_diff_hunk_get_trailing_context", 1, argv[0] )); 
  }
  arg1 = (svn_diff_hunk_t *)(argp1);
  {
    result = (svn_linenum_t)svn_diff_hunk_get_trailing_context((struct svn_diff_hunk_t const *)arg1);
    
    
    
  }
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_prop_patch_t;

SWIGINTERN VALUE
_wrap_svn_prop_patch_t_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_patch_t *arg1 = (struct svn_prop_patch_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_patch_t *","name", 1, self )); 
  }
  arg1 = (struct svn_prop_patch_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_patch_t_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_patch_t *arg1 = (struct svn_prop_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_patch_t *","name", 1, self )); 
  }
  arg1 = (struct svn_prop_patch_t *)(argp1);
  result = (char *) ((arg1)->name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_patch_t_operation_set(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_patch_t *arg1 = (struct svn_prop_patch_t *) 0 ;
  svn_diff_operation_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_patch_t *","operation", 1, self )); 
  }
  arg1 = (struct svn_prop_patch_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_diff_operation_kind_t","operation", 2, argv[0] ));
  } 
  arg2 = (svn_diff_operation_kind_t)(val2);
  if (arg1) (arg1)->operation = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_patch_t_operation_get(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_patch_t *arg1 = (struct svn_prop_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_diff_operation_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_patch_t *","operation", 1, self )); 
  }
  arg1 = (struct svn_prop_patch_t *)(argp1);
  result = (svn_diff_operation_kind_t) ((arg1)->operation);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_patch_t_hunks_set(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_patch_t *arg1 = (struct svn_prop_patch_t *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_patch_t *","hunks", 1, self )); 
  }
  arg1 = (struct svn_prop_patch_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_array_header_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_array_header_t *","hunks", 2, argv[0] )); 
  }
  arg2 = (apr_array_header_t *)(argp2);
  if (arg1) (arg1)->hunks = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_prop_patch_t_hunks_get(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_patch_t *arg1 = (struct svn_prop_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_prop_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_prop_patch_t *","hunks", 1, self )); 
  }
  arg1 = (struct svn_prop_patch_t *)(argp1);
  result = (apr_array_header_t *) ((arg1)->hunks);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_prop_patch_t_allocate(VALUE self)
#else
_wrap_svn_prop_patch_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_prop_patch_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_prop_patch_t(int argc, VALUE *argv, VALUE self) {
  struct svn_prop_patch_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_prop_patch_t *)calloc(1, sizeof(struct svn_prop_patch_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_prop_patch_t(void *self) {
    struct svn_prop_patch_t *arg1 = (struct svn_prop_patch_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_diff_get_binary_diff_original_stream(int argc, VALUE *argv, VALUE self) {
  svn_diff_binary_patch_t *arg1 = (svn_diff_binary_patch_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_binary_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_binary_patch_t const *","svn_diff_get_binary_diff_original_stream", 1, argv[0] )); 
  }
  arg1 = (svn_diff_binary_patch_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_stream_t *)svn_diff_get_binary_diff_original_stream((struct svn_diff_binary_patch_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_get_binary_diff_result_stream(int argc, VALUE *argv, VALUE self) {
  svn_diff_binary_patch_t *arg1 = (svn_diff_binary_patch_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_stream_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_binary_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_binary_patch_t const *","svn_diff_get_binary_diff_result_stream", 1, argv[0] )); 
  }
  arg1 = (svn_diff_binary_patch_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_stream_t *)svn_diff_get_binary_diff_result_stream((struct svn_diff_binary_patch_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_stream_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_patch_t;

SWIGINTERN VALUE
_wrap_svn_patch_t_old_filename_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","old_filename", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","old_filename", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->old_filename) free((char *)arg1->old_filename);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->old_filename = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_old_filename_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","old_filename", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (char *) ((arg1)->old_filename);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_new_filename_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","new_filename", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","new_filename", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->new_filename) free((char *)arg1->new_filename);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->new_filename = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_new_filename_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","new_filename", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (char *) ((arg1)->new_filename);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_hunks_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  apr_array_header_t *arg2 = (apr_array_header_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","hunks", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_array_header_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_array_header_t *","hunks", 2, argv[0] )); 
  }
  arg2 = (apr_array_header_t *)(argp2);
  if (arg1) (arg1)->hunks = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_hunks_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","hunks", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (apr_array_header_t *) ((arg1)->hunks);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_prop_patches_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","prop_patches", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","prop_patches", 2, argv[0] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  if (arg1) (arg1)->prop_patches = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_prop_patches_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","prop_patches", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (apr_hash_t *) ((arg1)->prop_patches);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_operation_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_diff_operation_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","operation", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_diff_operation_kind_t","operation", 2, argv[0] ));
  } 
  arg2 = (svn_diff_operation_kind_t)(val2);
  if (arg1) (arg1)->operation = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_operation_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_diff_operation_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","operation", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_diff_operation_kind_t) ((arg1)->operation);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_reverse_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","reverse", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->reverse = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_reverse_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","reverse", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->reverse);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_mergeinfo_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","mergeinfo", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  {
    arg2 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  if (arg1) (arg1)->mergeinfo = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_mergeinfo_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_mergeinfo_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","mergeinfo", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_mergeinfo_t) ((arg1)->mergeinfo);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_reverse_mergeinfo_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_mergeinfo_t arg2 = (svn_mergeinfo_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","reverse_mergeinfo", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_mergeinfo_t","reverse_mergeinfo", 2, argv[0] )); 
  }
  arg2 = (svn_mergeinfo_t)(argp2);
  if (arg1) (arg1)->reverse_mergeinfo = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_reverse_mergeinfo_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_mergeinfo_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","reverse_mergeinfo", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_mergeinfo_t) ((arg1)->reverse_mergeinfo);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_binary_patch_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_diff_binary_patch_t *arg2 = (svn_diff_binary_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","binary_patch", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_diff_binary_patch_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_diff_binary_patch_t *","binary_patch", 2, argv[0] )); 
  }
  arg2 = (svn_diff_binary_patch_t *)(argp2);
  if (arg1) (arg1)->binary_patch = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_binary_patch_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_diff_binary_patch_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","binary_patch", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_diff_binary_patch_t *) ((arg1)->binary_patch);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_diff_binary_patch_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_old_executable_bit_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_tristate_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","old_executable_bit", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_tristate_t","old_executable_bit", 2, argv[0] ));
  } 
  arg2 = (svn_tristate_t)(val2);
  if (arg1) (arg1)->old_executable_bit = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_old_executable_bit_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_tristate_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","old_executable_bit", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_tristate_t) ((arg1)->old_executable_bit);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_new_executable_bit_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_tristate_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","new_executable_bit", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_tristate_t","new_executable_bit", 2, argv[0] ));
  } 
  arg2 = (svn_tristate_t)(val2);
  if (arg1) (arg1)->new_executable_bit = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_new_executable_bit_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_tristate_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","new_executable_bit", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_tristate_t) ((arg1)->new_executable_bit);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_old_symlink_bit_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_tristate_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","old_symlink_bit", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_tristate_t","old_symlink_bit", 2, argv[0] ));
  } 
  arg2 = (svn_tristate_t)(val2);
  if (arg1) (arg1)->old_symlink_bit = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_old_symlink_bit_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_tristate_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","old_symlink_bit", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_tristate_t) ((arg1)->old_symlink_bit);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_new_symlink_bit_set(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  svn_tristate_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","new_symlink_bit", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_tristate_t","new_symlink_bit", 2, argv[0] ));
  } 
  arg2 = (svn_tristate_t)(val2);
  if (arg1) (arg1)->new_symlink_bit = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_patch_t_new_symlink_bit_get(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *arg1 = (struct svn_patch_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_tristate_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_patch_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_patch_t *","new_symlink_bit", 1, self )); 
  }
  arg1 = (struct svn_patch_t *)(argp1);
  result = (svn_tristate_t) ((arg1)->new_symlink_bit);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_patch_t_allocate(VALUE self)
#else
_wrap_svn_patch_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_patch_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_patch_t(int argc, VALUE *argv, VALUE self) {
  struct svn_patch_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_patch_t *)calloc(1, sizeof(struct svn_patch_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_patch_t(void *self) {
    struct svn_patch_t *arg1 = (struct svn_patch_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_diff_open_patch_file(int argc, VALUE *argv, VALUE self) {
  svn_patch_file_t **arg1 = (svn_patch_file_t **) 0 ;
  char *arg2 = (char *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_patch_file_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_diff_open_patch_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_diff_open_patch_file(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_patch_file_t, 0));
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_parse_next_patch(int argc, VALUE *argv, VALUE self) {
  svn_patch_t **arg1 = (svn_patch_t **) 0 ;
  svn_patch_file_t *arg2 = (svn_patch_file_t *) 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  apr_pool_t *arg6 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_patch_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_patch_file_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_patch_file_t *","svn_diff_parse_next_patch", 2, argv[0] )); 
  }
  arg2 = (svn_patch_file_t *)(argp2);
  arg3 = RTEST(argv[1]);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_diff_parse_next_patch(arg1,arg2,arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_patch_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_close_patch_file(int argc, VALUE *argv, VALUE self) {
  svn_patch_file_t *arg1 = (svn_patch_file_t *) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_patch_file_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_patch_file_t *","svn_diff_close_patch_file", 1, argv[0] )); 
  }
  arg1 = (svn_patch_file_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_diff_close_patch_file(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_invoke_datasources_open(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns2_t *arg1 = (svn_diff_fns2_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_off_t *arg3 = (apr_off_t *) 0 ;
  apr_off_t *arg4 = (apr_off_t *) 0 ;
  svn_diff_datasource_e *arg5 = (svn_diff_datasource_e *) 0 ;
  apr_size_t arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  unsigned long val6 ;
  int ecode6 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 6) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns2_t *","svn_diff_fns2_invoke_datasources_open", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_datasources_open", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_apr_off_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "apr_off_t *","svn_diff_fns2_invoke_datasources_open", 3, argv[2] )); 
  }
  arg3 = (apr_off_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_apr_off_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "apr_off_t *","svn_diff_fns2_invoke_datasources_open", 4, argv[3] )); 
  }
  arg4 = (apr_off_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_diff_datasource_e, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_diff_datasource_e const *","svn_diff_fns2_invoke_datasources_open", 5, argv[4] )); 
  }
  arg5 = (svn_diff_datasource_e *)(argp5);
  ecode6 = SWIG_AsVal_unsigned_SS_long(argv[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "apr_size_t","svn_diff_fns2_invoke_datasources_open", 6, argv[5] ));
  } 
  arg6 = (apr_size_t)(val6);
  {
    result = (svn_error_t *)svn_diff_fns2_invoke_datasources_open(arg1,arg2,arg3,arg4,(enum svn_diff_datasource_e const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_invoke_datasource_close(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns2_t *arg1 = (svn_diff_fns2_t *) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_datasource_e arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  int val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns2_t *","svn_diff_fns2_invoke_datasource_close", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_datasource_close", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_int(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_diff_datasource_e","svn_diff_fns2_invoke_datasource_close", 3, argv[2] ));
  } 
  arg3 = (svn_diff_datasource_e)(val3);
  {
    result = (svn_error_t *)svn_diff_fns2_invoke_datasource_close(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_invoke_datasource_get_next_token(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns2_t *arg1 = (svn_diff_fns2_t *) 0 ;
  apr_uint32_t *arg2 = (apr_uint32_t *) 0 ;
  void **arg3 = (void **) 0 ;
  void *arg4 = (void *) 0 ;
  svn_diff_datasource_e arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_uint32_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  void *temp3 ;
  int res4 ;
  int val5 ;
  int ecode5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns2_t *","svn_diff_fns2_invoke_datasource_get_next_token", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns2_t *)(argp1);
  res4 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_datasource_get_next_token", 4, argv[1] )); 
  }
  ecode5 = SWIG_AsVal_int(argv[2], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_diff_datasource_e","svn_diff_fns2_invoke_datasource_get_next_token", 5, argv[2] ));
  } 
  arg5 = (svn_diff_datasource_e)(val5);
  {
    result = (svn_error_t *)svn_diff_fns2_invoke_datasource_get_next_token(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res2)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_unsigned_SS_long((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_apr_uint32_t, new_flags));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg3, SWIGTYPE_p_void, 0));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_invoke_token_compare(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns2_t *arg1 = (svn_diff_fns2_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *arg3 = (void *) 0 ;
  void *arg4 = (void *) 0 ;
  int *arg5 = (int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  int res3 ;
  int res4 ;
  int temp5 ;
  int res5 = SWIG_TMPOBJ ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg5 = &temp5;
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns2_t *","svn_diff_fns2_invoke_token_compare", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_token_compare", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_token_compare", 3, argv[2] )); 
  }
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_token_compare", 4, argv[3] )); 
  }
  {
    result = (svn_error_t *)svn_diff_fns2_invoke_token_compare(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res5)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_invoke_token_discard(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns2_t *arg1 = (svn_diff_fns2_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *arg3 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  int res3 ;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns2_t *","svn_diff_fns2_invoke_token_discard", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_token_discard", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_token_discard", 3, argv[2] )); 
  }
  {
    svn_diff_fns2_invoke_token_discard(arg1,arg2,arg3);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns2_invoke_token_discard_all(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns2_t *arg1 = (svn_diff_fns2_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns2_t *","svn_diff_fns2_invoke_token_discard_all", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns2_invoke_token_discard_all", 2, argv[1] )); 
  }
  {
    svn_diff_fns2_invoke_token_discard_all(arg1,arg2);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_invoke_datasource_open(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns_t *arg1 = (svn_diff_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_datasource_e arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  int val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns_t *","svn_diff_fns_invoke_datasource_open", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_datasource_open", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_int(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_diff_datasource_e","svn_diff_fns_invoke_datasource_open", 3, argv[2] ));
  } 
  arg3 = (svn_diff_datasource_e)(val3);
  {
    result = (svn_error_t *)svn_diff_fns_invoke_datasource_open(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_invoke_datasource_close(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns_t *arg1 = (svn_diff_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  svn_diff_datasource_e arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  int val3 ;
  int ecode3 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns_t *","svn_diff_fns_invoke_datasource_close", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_datasource_close", 2, argv[1] )); 
  }
  ecode3 = SWIG_AsVal_int(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_diff_datasource_e","svn_diff_fns_invoke_datasource_close", 3, argv[2] ));
  } 
  arg3 = (svn_diff_datasource_e)(val3);
  {
    result = (svn_error_t *)svn_diff_fns_invoke_datasource_close(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_invoke_datasource_get_next_token(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns_t *arg1 = (svn_diff_fns_t *) 0 ;
  apr_uint32_t *arg2 = (apr_uint32_t *) 0 ;
  void **arg3 = (void **) 0 ;
  void *arg4 = (void *) 0 ;
  svn_diff_datasource_e arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_uint32_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  void *temp3 ;
  int res4 ;
  int val5 ;
  int ecode5 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns_t *","svn_diff_fns_invoke_datasource_get_next_token", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns_t *)(argp1);
  res4 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_datasource_get_next_token", 4, argv[1] )); 
  }
  ecode5 = SWIG_AsVal_int(argv[2], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_diff_datasource_e","svn_diff_fns_invoke_datasource_get_next_token", 5, argv[2] ));
  } 
  arg5 = (svn_diff_datasource_e)(val5);
  {
    result = (svn_error_t *)svn_diff_fns_invoke_datasource_get_next_token(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res2)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_unsigned_SS_long((*arg2)));
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_apr_uint32_t, new_flags));
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg3, SWIGTYPE_p_void, 0));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_invoke_token_compare(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns_t *arg1 = (svn_diff_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *arg3 = (void *) 0 ;
  void *arg4 = (void *) 0 ;
  int *arg5 = (int *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  int res3 ;
  int res4 ;
  int temp5 ;
  int res5 = SWIG_TMPOBJ ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg5 = &temp5;
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns_t *","svn_diff_fns_invoke_token_compare", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_token_compare", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_token_compare", 3, argv[2] )); 
  }
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_token_compare", 4, argv[3] )); 
  }
  {
    result = (svn_error_t *)svn_diff_fns_invoke_token_compare(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  if (SWIG_IsTmpObj(res5)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg5)));
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_int, new_flags));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_invoke_token_discard(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns_t *arg1 = (svn_diff_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *arg3 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  int res3 ;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns_t *","svn_diff_fns_invoke_token_discard", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_token_discard", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg3), 0, 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_token_discard", 3, argv[2] )); 
  }
  {
    svn_diff_fns_invoke_token_discard(arg1,arg2,arg3);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_fns_invoke_token_discard_all(int argc, VALUE *argv, VALUE self) {
  svn_diff_fns_t *arg1 = (svn_diff_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_fns_t *","svn_diff_fns_invoke_token_discard_all", 1, argv[0] )); 
  }
  arg1 = (svn_diff_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_fns_invoke_token_discard_all", 2, argv[1] )); 
  }
  {
    svn_diff_fns_invoke_token_discard_all(arg1,arg2);
    
    
    
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_invoke_output_common(int argc, VALUE *argv, VALUE self) {
  svn_diff_output_fns_t *arg1 = (svn_diff_output_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_off_t arg3 ;
  apr_off_t arg4 ;
  apr_off_t arg5 ;
  apr_off_t arg6 ;
  apr_off_t arg7 ;
  apr_off_t arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 8) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_output_fns_t *","svn_diff_output_fns_invoke_output_common", 1, argv[0] )); 
  }
  arg1 = (svn_diff_output_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_output_fns_invoke_output_common", 2, argv[1] )); 
  }
  {
    arg3 = (apr_off_t)NUM2LL(argv[2]);
  }
  {
    arg4 = (apr_off_t)NUM2LL(argv[3]);
  }
  {
    arg5 = (apr_off_t)NUM2LL(argv[4]);
  }
  {
    arg6 = (apr_off_t)NUM2LL(argv[5]);
  }
  {
    arg7 = (apr_off_t)NUM2LL(argv[6]);
  }
  {
    arg8 = (apr_off_t)NUM2LL(argv[7]);
  }
  {
    result = (svn_error_t *)svn_diff_output_fns_invoke_output_common(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_invoke_output_diff_modified(int argc, VALUE *argv, VALUE self) {
  svn_diff_output_fns_t *arg1 = (svn_diff_output_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_off_t arg3 ;
  apr_off_t arg4 ;
  apr_off_t arg5 ;
  apr_off_t arg6 ;
  apr_off_t arg7 ;
  apr_off_t arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 8) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_output_fns_t *","svn_diff_output_fns_invoke_output_diff_modified", 1, argv[0] )); 
  }
  arg1 = (svn_diff_output_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_output_fns_invoke_output_diff_modified", 2, argv[1] )); 
  }
  {
    arg3 = (apr_off_t)NUM2LL(argv[2]);
  }
  {
    arg4 = (apr_off_t)NUM2LL(argv[3]);
  }
  {
    arg5 = (apr_off_t)NUM2LL(argv[4]);
  }
  {
    arg6 = (apr_off_t)NUM2LL(argv[5]);
  }
  {
    arg7 = (apr_off_t)NUM2LL(argv[6]);
  }
  {
    arg8 = (apr_off_t)NUM2LL(argv[7]);
  }
  {
    result = (svn_error_t *)svn_diff_output_fns_invoke_output_diff_modified(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_invoke_output_diff_latest(int argc, VALUE *argv, VALUE self) {
  svn_diff_output_fns_t *arg1 = (svn_diff_output_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_off_t arg3 ;
  apr_off_t arg4 ;
  apr_off_t arg5 ;
  apr_off_t arg6 ;
  apr_off_t arg7 ;
  apr_off_t arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 8) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_output_fns_t *","svn_diff_output_fns_invoke_output_diff_latest", 1, argv[0] )); 
  }
  arg1 = (svn_diff_output_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_output_fns_invoke_output_diff_latest", 2, argv[1] )); 
  }
  {
    arg3 = (apr_off_t)NUM2LL(argv[2]);
  }
  {
    arg4 = (apr_off_t)NUM2LL(argv[3]);
  }
  {
    arg5 = (apr_off_t)NUM2LL(argv[4]);
  }
  {
    arg6 = (apr_off_t)NUM2LL(argv[5]);
  }
  {
    arg7 = (apr_off_t)NUM2LL(argv[6]);
  }
  {
    arg8 = (apr_off_t)NUM2LL(argv[7]);
  }
  {
    result = (svn_error_t *)svn_diff_output_fns_invoke_output_diff_latest(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_invoke_output_diff_common(int argc, VALUE *argv, VALUE self) {
  svn_diff_output_fns_t *arg1 = (svn_diff_output_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_off_t arg3 ;
  apr_off_t arg4 ;
  apr_off_t arg5 ;
  apr_off_t arg6 ;
  apr_off_t arg7 ;
  apr_off_t arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 8) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_output_fns_t *","svn_diff_output_fns_invoke_output_diff_common", 1, argv[0] )); 
  }
  arg1 = (svn_diff_output_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_output_fns_invoke_output_diff_common", 2, argv[1] )); 
  }
  {
    arg3 = (apr_off_t)NUM2LL(argv[2]);
  }
  {
    arg4 = (apr_off_t)NUM2LL(argv[3]);
  }
  {
    arg5 = (apr_off_t)NUM2LL(argv[4]);
  }
  {
    arg6 = (apr_off_t)NUM2LL(argv[5]);
  }
  {
    arg7 = (apr_off_t)NUM2LL(argv[6]);
  }
  {
    arg8 = (apr_off_t)NUM2LL(argv[7]);
  }
  {
    result = (svn_error_t *)svn_diff_output_fns_invoke_output_diff_common(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_diff_output_fns_invoke_output_conflict(int argc, VALUE *argv, VALUE self) {
  svn_diff_output_fns_t *arg1 = (svn_diff_output_fns_t *) 0 ;
  void *arg2 = (void *) 0 ;
  apr_off_t arg3 ;
  apr_off_t arg4 ;
  apr_off_t arg5 ;
  apr_off_t arg6 ;
  apr_off_t arg7 ;
  apr_off_t arg8 ;
  svn_diff_t *arg9 = (svn_diff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 9) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_diff_output_fns_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_diff_output_fns_t *","svn_diff_output_fns_invoke_output_conflict", 1, argv[0] )); 
  }
  arg1 = (svn_diff_output_fns_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_diff_output_fns_invoke_output_conflict", 2, argv[1] )); 
  }
  {
    arg3 = (apr_off_t)NUM2LL(argv[2]);
  }
  {
    arg4 = (apr_off_t)NUM2LL(argv[3]);
  }
  {
    arg5 = (apr_off_t)NUM2LL(argv[4]);
  }
  {
    arg6 = (apr_off_t)NUM2LL(argv[5]);
  }
  {
    arg7 = (apr_off_t)NUM2LL(argv[6]);
  }
  {
    arg8 = (apr_off_t)NUM2LL(argv[7]);
  }
  res9 = SWIG_ConvertPtr(argv[8], &argp9,SWIGTYPE_p_svn_diff_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_diff_t *","svn_diff_output_fns_invoke_output_conflict", 9, argv[8] )); 
  }
  arg9 = (svn_diff_t *)(argp9);
  {
    result = (svn_error_t *)svn_diff_output_fns_invoke_output_conflict(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_default_charset(int argc, VALUE *argv, VALUE self) {
  VALUE result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (VALUE)svn_default_charset();
    
    
    
  }
  vresult = result;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_locale_charset(int argc, VALUE *argv, VALUE self) {
  VALUE result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (VALUE)svn_locale_charset();
    
    
    
  }
  vresult = result;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_rb_auth_get_simple_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_simple_prompt_func_t arg2 = (svn_auth_simple_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_simple_prompt_func_t) svn_swig_rb_auth_simple_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_swig_rb_auth_get_simple_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    result = (VALUE)svn_swig_rb_auth_get_simple_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  vresult = result;
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_rb_auth_get_ssl_client_cert_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_client_cert_prompt_func_t arg2 = (svn_auth_ssl_client_cert_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_client_cert_prompt_func_t) svn_swig_rb_auth_ssl_client_cert_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_swig_rb_auth_get_ssl_client_cert_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    result = (VALUE)svn_swig_rb_auth_get_ssl_client_cert_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  vresult = result;
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_rb_auth_get_ssl_client_cert_pw_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_client_cert_pw_prompt_func_t arg2 = (svn_auth_ssl_client_cert_pw_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_client_cert_pw_prompt_func_t) svn_swig_rb_auth_ssl_client_cert_pw_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_swig_rb_auth_get_ssl_client_cert_pw_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    result = (VALUE)svn_swig_rb_auth_get_ssl_client_cert_pw_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  vresult = result;
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_rb_auth_get_ssl_server_trust_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_ssl_server_trust_prompt_func_t arg2 = (svn_auth_ssl_server_trust_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  apr_pool_t *arg4 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_ssl_server_trust_prompt_func_t) svn_swig_rb_auth_ssl_server_trust_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (VALUE)svn_swig_rb_auth_get_ssl_server_trust_prompt_provider(arg1,arg2,arg3,arg4);
    
    
    
  }
  vresult = result;
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_rb_auth_get_username_prompt_provider(int argc, VALUE *argv, VALUE self) {
  svn_auth_provider_object_t **arg1 = (svn_auth_provider_object_t **) 0 ;
  svn_auth_username_prompt_func_t arg2 = (svn_auth_username_prompt_func_t) 0 ;
  void *arg3 = (void *) 0 ;
  int arg4 ;
  apr_pool_t *arg5 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_auth_provider_object_t *temp1 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg2 = (svn_auth_username_prompt_func_t) svn_swig_rb_auth_username_prompt_func;
    arg3 = (void *)svn_swig_rb_make_baton(argv[0], _global_svn_swig_rb_pool);
  }
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","svn_swig_rb_auth_get_username_prompt_provider", 4, argv[1] ));
  } 
  arg4 = (int)(val4);
  if (argc > 2) {
    
  }
  {
    result = (VALUE)svn_swig_rb_auth_get_username_prompt_provider(arg1,arg2,arg3,arg4,arg5);
    
    
    
  }
  vresult = result;
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_auth_provider_object_t, 0));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_mergeinfo_merge(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  apr_hash_t *arg2 = (apr_hash_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = &temp1;
    *arg1 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  {
    arg2 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_swig_mergeinfo_merge(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_mergeinfo_sort(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = (apr_hash_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = &temp1;
    *arg1 = svn_swig_rb_hash_to_apr_hash_merge_range(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_swig_mergeinfo_sort(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_rangelist_merge(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t **arg1 = (svn_rangelist_t **) 0 ;
  svn_rangelist_t *arg2 = (svn_rangelist_t *) 0 ;
  apr_pool_t *arg3 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_rangelist_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    arg1 = &temp1;
    *arg1 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  {
    arg2 = svn_swig_rb_array_to_apr_array_merge_range(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_swig_rangelist_merge(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_swig_rangelist_reverse(int argc, VALUE *argv, VALUE self) {
  svn_rangelist_t **arg1 = (svn_rangelist_t **) 0 ;
  apr_pool_t *arg2 = (apr_pool_t *) 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_rangelist_t *temp1 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    arg1 = &temp1;
    *arg1 = svn_swig_rb_array_to_apr_array_merge_range(argv[0], _global_pool);
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_swig_rangelist_reverse(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = Qnil;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_merge_range(*arg1));
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}



/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_FILE = {"_p_FILE", "FILE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_allocator_t = {"_p_apr_allocator_t", "apr_allocator_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_array_header_t = {"_p_apr_array_header_t", "apr_array_header_t *|svn_rangelist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_dir_t = {"_p_apr_dir_t", "apr_dir_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_file_t = {"_p_apr_file_t", "apr_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_finfo_t = {"_p_apr_finfo_t", "apr_finfo_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_getopt_option_t = {"_p_apr_getopt_option_t", "apr_getopt_option_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_getopt_t = {"_p_apr_getopt_t", "apr_getopt_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_hash_t = {"_p_apr_hash_t", "svn_mergeinfo_catalog_t|svn_mergeinfo_t|apr_hash_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int32_t = {"_p_apr_int32_t", "apr_int32_t *|time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int64_t = {"_p_apr_int64_t", "apr_int64_t *|svn_filesize_t *|apr_time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_off_t = {"_p_apr_off_t", "apr_off_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_pool_t = {"_p_apr_pool_t", "apr_pool_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_pool_wrapper_t = {"_p_apr_pool_wrapper_t", "apr_pool_wrapper_t *|struct apr_pool_wrapper_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_proc_t = {"_p_apr_proc_t", "apr_proc_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_size_t = {"_p_apr_size_t", "apr_size_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_uint32_t = {"_p_apr_uint32_t", "apr_uint32_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_uint64_t = {"_p_apr_uint64_t", "apr_uint64_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_opt_subcommand_t *|struct svn_error_t *(*)(apr_getopt_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t = {"_p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t", "struct svn_error_t *(*)(apr_uint32_t *,void **,void *,svn_diff_datasource_e)|svn_error_t *(*)(apr_uint32_t *,void **,void *,svn_diff_datasource_e)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_char_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_char_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t", "svn_auth_gnome_keyring_unlock_prompt_func_t|struct svn_error_t *(*)(char **,char const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void = {"_p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void", "svn_auth_simple_provider_func_t|svn_auth_ssl_client_cert_pw_provider_func_t|void (*)(struct svn_auth_provider_object_t **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_auth_simple_prompt_func_t|struct svn_error_t *(*)(svn_auth_cred_simple_t **,void *,char const *,char const *,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_auth_ssl_client_cert_pw_prompt_func_t|struct svn_error_t *(*)(svn_auth_cred_ssl_client_cert_pw_t **,void *,char const *,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_auth_ssl_client_cert_prompt_func_t|struct svn_error_t *(*)(svn_auth_cred_ssl_client_cert_t **,void *,char const *,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_auth_cred_ssl_server_trust_t **,void *,char const *,apr_uint32_t,svn_auth_ssl_server_cert_info_t const *,svn_boolean_t,apr_pool_t *)|svn_auth_ssl_server_trust_prompt_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_auth_cred_username_t **,void *,char const *,svn_boolean_t,apr_pool_t *)|svn_auth_username_prompt_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_stream_t **,void *,apr_pool_t *,apr_pool_t *)|svn_stream_lazyopen_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void **,void **,void *,apr_hash_t *,char const *,apr_pool_t *)|svn_error_t *(*)(void **,void **,void *,apr_hash_t *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void **,void *,void *,apr_hash_t *,char const *,apr_pool_t *)|struct svn_error_t *(*)(void **,void *,void *,apr_hash_t *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_q_const__char_p_void__int = {"_p_f_p_q_const__char_p_q_const__char_p_void__int", "svn_config_enumerator_t|int (*)(char const *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_q_const__char_p_void_p_apr_pool_t__int = {"_p_f_p_q_const__char_p_q_const__char_p_void_p_apr_pool_t__int", "svn_config_enumerator2_t|int (*)(char const *,char const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_void__int = {"_p_f_p_q_const__char_p_void__int", "svn_config_section_enumerator_t|int (*)(char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_void_p_apr_pool_t__int = {"_p_f_p_q_const__char_p_void_p_apr_pool_t__int", "int (*)(char const *,void *,apr_pool_t *)|svn_config_section_enumerator2_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__struct_svn_version_t_p_q_const__struct_svn_version_t__int = {"_p_f_p_q_const__struct_svn_version_t_p_q_const__struct_svn_version_t__int", "svn_boolean_t (*)(svn_version_t const *,svn_version_t const *)|int (*)(struct svn_version_t const *,struct svn_version_t const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_commit_callback2_t|struct svn_error_t *(*)(svn_commit_info_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_boolean_t *,char const *,void *,apr_pool_t *)|svn_auth_plaintext_prompt_func_t|svn_auth_plaintext_passphrase_prompt_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_boolean_t *,void *,char const *,char const *,apr_hash_t *,apr_pool_t *)|svn_config_auth_walk_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_boolean_t *,void *,void *,apr_hash_t *,char const *,apr_pool_t *)|struct svn_error_t *(*)(svn_boolean_t *,void *,void *,apr_hash_t *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(svn_location_segment_t *,void *,apr_pool_t *)|svn_location_segment_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void__p_svn_error_t = {"_p_f_p_void__p_svn_error_t", "svn_cancel_func_t|svn_close_fn_t|struct svn_error_t *(*)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void__void = {"_p_f_p_void__void", "void (*)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t = {"_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t", "struct svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)|svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t = {"_p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t", "struct svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,svn_diff_t *)|svn_error_t *(*)(void *,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,apr_off_t,svn_diff_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_apr_size_t__p_svn_error_t = {"_p_f_p_void_apr_size_t__p_svn_error_t", "struct svn_error_t *(*)(void *,apr_size_t)|svn_stream_skip_fn_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,apr_hash_t *,svn_revnum_t,char const *,char const *,char const *,apr_pool_t *)|svn_log_message_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t = {"_p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t", "svn_error_t *(*)(void *,apr_off_t *,apr_off_t *,svn_diff_datasource_e const *,apr_size_t)|struct svn_error_t *(*)(void *,apr_off_t *,apr_off_t *,svn_diff_datasource_e const *,apr_size_t)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_char_p_apr_size_t__p_svn_error_t = {"_p_f_p_void_p_char_p_apr_size_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char *,apr_size_t *)|svn_read_fn_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,svn_stream_mark_t **,apr_pool_t *)|svn_stream_mark_fn_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,svn_stringbuf_t **,char const *,svn_boolean_t *,apr_pool_t *)|svn_stream_readline_fn_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_apr_size_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_apr_size_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,apr_size_t *)|svn_write_fn_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,char const *,apr_finfo_t const *,apr_pool_t *)|svn_io_walk_func_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t = {"_p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t", "svn_stream_seek_fn_t|struct svn_error_t *(*)(void *,svn_stream_mark_t const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_boolean_t__p_svn_error_t = {"_p_f_p_void_p_svn_boolean_t__p_svn_error_t", "svn_stream_data_available_fn_t|struct svn_error_t *(*)(void *,svn_boolean_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t", "struct svn_error_t *(*)(void *,svn_log_entry_t *,apr_pool_t *)|svn_log_entry_receiver_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_void__void = {"_p_f_p_void_p_void__void", "void (*)(void *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_void_p_void_p_int__p_svn_error_t = {"_p_f_p_void_p_void_p_void_p_int__p_svn_error_t", "svn_error_t *(*)(void *,void *,void *,int *)|struct svn_error_t *(*)(void *,void *,void *,int *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_svn_diff_datasource_e__p_svn_error_t = {"_p_f_p_void_svn_diff_datasource_e__p_svn_error_t", "struct svn_error_t *(*)(void *,svn_diff_datasource_e)|svn_error_t *(*)(void *,svn_diff_datasource_e)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t = {"_p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t", "struct svn_error_t *(*)(svn_revnum_t,char const *,char const *,void *)|svn_commit_callback_t", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_void__p_svn_version_t = {"_p_f_void__p_svn_version_t", "svn_version_func_t|struct svn_version_t *(*)(void)|svn_version_t *(*)(void)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "int *|svn_boolean_t *|apr_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "long *|svn_revnum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_array_header_t = {"_p_p_apr_array_header_t", "apr_array_header_t **|svn_rangelist_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_file_t = {"_p_p_apr_file_t", "apr_file_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_hash_t = {"_p_p_apr_hash_t", "apr_hash_t **|svn_mergeinfo_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_baton_t = {"_p_p_svn_auth_baton_t", "struct svn_auth_baton_t **|svn_auth_baton_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_cred_simple_t = {"_p_p_svn_auth_cred_simple_t", "struct svn_auth_cred_simple_t **|svn_auth_cred_simple_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_cred_ssl_client_cert_pw_t = {"_p_p_svn_auth_cred_ssl_client_cert_pw_t", "struct svn_auth_cred_ssl_client_cert_pw_t **|svn_auth_cred_ssl_client_cert_pw_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_cred_ssl_client_cert_t = {"_p_p_svn_auth_cred_ssl_client_cert_t", "struct svn_auth_cred_ssl_client_cert_t **|svn_auth_cred_ssl_client_cert_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_cred_ssl_server_trust_t = {"_p_p_svn_auth_cred_ssl_server_trust_t", "struct svn_auth_cred_ssl_server_trust_t **|svn_auth_cred_ssl_server_trust_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_cred_username_t = {"_p_p_svn_auth_cred_username_t", "struct svn_auth_cred_username_t **|svn_auth_cred_username_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_iterstate_t = {"_p_p_svn_auth_iterstate_t", "struct svn_auth_iterstate_t **|svn_auth_iterstate_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_auth_provider_object_t = {"_p_p_svn_auth_provider_object_t", "struct svn_auth_provider_object_t **|svn_auth_provider_object_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_checksum_t = {"_p_p_svn_checksum_t", "struct svn_checksum_t **|svn_checksum_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_config_t = {"_p_p_svn_config_t", "struct svn_config_t **|svn_config_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_diff_t = {"_p_p_svn_diff_t", "struct svn_diff_t **|svn_diff_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_io_dirent2_t = {"_p_p_svn_io_dirent2_t", "struct svn_io_dirent2_t **|svn_io_dirent2_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_patch_file_t = {"_p_p_svn_patch_file_t", "struct svn_patch_file_t **|svn_patch_file_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_patch_t = {"_p_p_svn_patch_t", "struct svn_patch_t **|svn_patch_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_stream_mark_t = {"_p_p_svn_stream_mark_t", "struct svn_stream_mark_t **|svn_stream_mark_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_stream_t = {"_p_p_svn_stream_t", "struct svn_stream_t **|svn_stream_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_string_t = {"_p_p_svn_string_t", "struct svn_string_t **|svn_string_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_stringbuf_t = {"_p_p_svn_stringbuf_t", "struct svn_stringbuf_t **|svn_stringbuf_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_void = {"_p_p_void", "void **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_baton_t = {"_p_svn_auth_baton_t", "struct svn_auth_baton_t *|svn_auth_baton_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_simple_t = {"_p_svn_auth_cred_simple_t", "struct svn_auth_cred_simple_t *|svn_auth_cred_simple_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_pw_t = {"_p_svn_auth_cred_ssl_client_cert_pw_t", "struct svn_auth_cred_ssl_client_cert_pw_t *|svn_auth_cred_ssl_client_cert_pw_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_t = {"_p_svn_auth_cred_ssl_client_cert_t", "struct svn_auth_cred_ssl_client_cert_t *|svn_auth_cred_ssl_client_cert_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_server_trust_t = {"_p_svn_auth_cred_ssl_server_trust_t", "struct svn_auth_cred_ssl_server_trust_t *|svn_auth_cred_ssl_server_trust_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_username_t = {"_p_svn_auth_cred_username_t", "struct svn_auth_cred_username_t *|svn_auth_cred_username_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_iterstate_t = {"_p_svn_auth_iterstate_t", "struct svn_auth_iterstate_t *|svn_auth_iterstate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_object_t = {"_p_svn_auth_provider_object_t", "struct svn_auth_provider_object_t *|svn_auth_provider_object_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_t = {"_p_svn_auth_provider_t", "struct svn_auth_provider_t *|svn_auth_provider_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_ssl_server_cert_info_t = {"_p_svn_auth_ssl_server_cert_info_t", "struct svn_auth_ssl_server_cert_info_t *|svn_auth_ssl_server_cert_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_cache_config_t = {"_p_svn_cache_config_t", "struct svn_cache_config_t *|svn_cache_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_ctx_t = {"_p_svn_checksum_ctx_t", "struct svn_checksum_ctx_t *|svn_checksum_ctx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_kind_t = {"_p_svn_checksum_kind_t", "enum svn_checksum_kind_t *|svn_checksum_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_t = {"_p_svn_checksum_t", "struct svn_checksum_t *|svn_checksum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_commit_info_t = {"_p_svn_commit_info_t", "struct svn_commit_info_t *|svn_commit_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_config_t = {"_p_svn_config_t", "struct svn_config_t *|svn_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_depth_t = {"_p_svn_depth_t", "enum svn_depth_t *|svn_depth_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_binary_patch_t = {"_p_svn_diff_binary_patch_t", "struct svn_diff_binary_patch_t *|svn_diff_binary_patch_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_conflict_display_style_t = {"_p_svn_diff_conflict_display_style_t", "enum svn_diff_conflict_display_style_t *|svn_diff_conflict_display_style_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_datasource_e = {"_p_svn_diff_datasource_e", "enum svn_diff_datasource_e *|svn_diff_datasource_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_file_ignore_space_t = {"_p_svn_diff_file_ignore_space_t", "enum svn_diff_file_ignore_space_t *|svn_diff_file_ignore_space_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_file_options_t = {"_p_svn_diff_file_options_t", "struct svn_diff_file_options_t *|svn_diff_file_options_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_fns2_t = {"_p_svn_diff_fns2_t", "struct svn_diff_fns2_t *|svn_diff_fns2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_fns_t = {"_p_svn_diff_fns_t", "struct svn_diff_fns_t *|svn_diff_fns_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_hunk_t = {"_p_svn_diff_hunk_t", "struct svn_diff_hunk_t *|svn_diff_hunk_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_operation_kind_e = {"_p_svn_diff_operation_kind_e", "svn_diff_operation_kind_t *|enum svn_diff_operation_kind_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_output_fns_t = {"_p_svn_diff_output_fns_t", "struct svn_diff_output_fns_t *|svn_diff_output_fns_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_t = {"_p_svn_diff_t", "struct svn_diff_t *|svn_diff_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_dirent_t = {"_p_svn_dirent_t", "struct svn_dirent_t *|svn_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_errno_t = {"_p_svn_errno_t", "enum svn_errno_t *|svn_errno_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_error_t = {"_p_svn_error_t", "struct svn_error_t *|svn_error_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent2_t = {"_p_svn_io_dirent2_t", "struct svn_io_dirent2_t *|svn_io_dirent2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent_t = {"_p_svn_io_dirent_t", "struct svn_io_dirent_t *|svn_io_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_file_del_t = {"_p_svn_io_file_del_t", "enum svn_io_file_del_t *|svn_io_file_del_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_location_segment_t = {"_p_svn_location_segment_t", "struct svn_location_segment_t *|svn_location_segment_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_lock_t = {"_p_svn_lock_t", "struct svn_lock_t *|svn_lock_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path2_t = {"_p_svn_log_changed_path2_t", "struct svn_log_changed_path2_t *|svn_log_changed_path2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path_t = {"_p_svn_log_changed_path_t", "struct svn_log_changed_path_t *|svn_log_changed_path_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_entry_t = {"_p_svn_log_entry_t", "struct svn_log_entry_t *|svn_log_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_merge_range_t = {"_p_svn_merge_range_t", "struct svn_merge_range_t *|svn_merge_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_mergeinfo_inheritance_t = {"_p_svn_mergeinfo_inheritance_t", "enum svn_mergeinfo_inheritance_t *|svn_mergeinfo_inheritance_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_node_kind_t = {"_p_svn_node_kind_t", "enum svn_node_kind_t *|svn_node_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_range_t = {"_p_svn_opt_revision_range_t", "struct svn_opt_revision_range_t *|svn_opt_revision_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_t = {"_p_svn_opt_revision_t", "struct svn_opt_revision_t *|svn_opt_revision_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_value_t = {"_p_svn_opt_revision_value_t", "union svn_opt_revision_value_t *|svn_opt_revision_value_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc2_t = {"_p_svn_opt_subcommand_desc2_t", "struct svn_opt_subcommand_desc2_t *|svn_opt_subcommand_desc2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc2_t_desc_overrides = {"_p_svn_opt_subcommand_desc2_t_desc_overrides", "svn_opt_subcommand_desc2_t_desc_overrides *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc3_t = {"_p_svn_opt_subcommand_desc3_t", "struct svn_opt_subcommand_desc3_t *|svn_opt_subcommand_desc3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc3_t_desc_overrides = {"_p_svn_opt_subcommand_desc3_t_desc_overrides", "svn_opt_subcommand_desc3_t_desc_overrides *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc_t = {"_p_svn_opt_subcommand_desc_t", "struct svn_opt_subcommand_desc_t *|svn_opt_subcommand_desc_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_patch_file_t = {"_p_svn_patch_file_t", "struct svn_patch_file_t *|svn_patch_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_patch_t = {"_p_svn_patch_t", "struct svn_patch_t *|svn_patch_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_inherited_item_t = {"_p_svn_prop_inherited_item_t", "struct svn_prop_inherited_item_t *|svn_prop_inherited_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_kind = {"_p_svn_prop_kind", "svn_prop_kind_t *|enum svn_prop_kind *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_patch_t = {"_p_svn_prop_patch_t", "struct svn_prop_patch_t *|svn_prop_patch_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_t = {"_p_svn_prop_t", "struct svn_prop_t *|svn_prop_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_mark_t = {"_p_svn_stream_mark_t", "struct svn_stream_mark_t *|svn_stream_mark_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_t = {"_p_svn_stream_t", "struct svn_stream_t *|svn_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_string_t = {"_p_svn_string_t", "struct svn_string_t *|svn_string_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stringbuf_t = {"_p_svn_stringbuf_t", "struct svn_stringbuf_t *|svn_stringbuf_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_tristate_t = {"_p_svn_tristate_t", "enum svn_tristate_t *|svn_tristate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_checklist_t = {"_p_svn_version_checklist_t", "struct svn_version_checklist_t *|svn_version_checklist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_linked_lib_t = {"_p_svn_version_ext_linked_lib_t", "struct svn_version_ext_linked_lib_t *|svn_version_ext_linked_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_loaded_lib_t = {"_p_svn_version_ext_loaded_lib_t", "struct svn_version_ext_loaded_lib_t *|svn_version_ext_loaded_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_extended_t = {"_p_svn_version_extended_t", "struct svn_version_extended_t *|svn_version_extended_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_t = {"_p_svn_version_t", "struct svn_version_t *|svn_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "unsigned long *|svn_linenum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_FILE,
  &_swigt__p_apr_allocator_t,
  &_swigt__p_apr_array_header_t,
  &_swigt__p_apr_dir_t,
  &_swigt__p_apr_file_t,
  &_swigt__p_apr_finfo_t,
  &_swigt__p_apr_getopt_option_t,
  &_swigt__p_apr_getopt_t,
  &_swigt__p_apr_hash_t,
  &_swigt__p_apr_int32_t,
  &_swigt__p_apr_int64_t,
  &_swigt__p_apr_off_t,
  &_swigt__p_apr_pool_t,
  &_swigt__p_apr_pool_wrapper_t,
  &_swigt__p_apr_proc_t,
  &_swigt__p_apr_size_t,
  &_swigt__p_apr_uint32_t,
  &_swigt__p_apr_uint64_t,
  &_swigt__p_char,
  &_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t,
  &_swigt__p_f_p_p_char_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void,
  &_swigt__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_q_const__char_p_void__int,
  &_swigt__p_f_p_q_const__char_p_q_const__char_p_void_p_apr_pool_t__int,
  &_swigt__p_f_p_q_const__char_p_void__int,
  &_swigt__p_f_p_q_const__char_p_void_p_apr_pool_t__int,
  &_swigt__p_f_p_q_const__struct_svn_version_t_p_q_const__struct_svn_version_t__int,
  &_swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void__p_svn_error_t,
  &_swigt__p_f_p_void__void,
  &_swigt__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t,
  &_swigt__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t,
  &_swigt__p_f_p_void_apr_size_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_char_p_apr_size_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_apr_size_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_svn_boolean_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_void__void,
  &_swigt__p_f_p_void_p_void_p_void_p_int__p_svn_error_t,
  &_swigt__p_f_p_void_svn_diff_datasource_e__p_svn_error_t,
  &_swigt__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_void__p_svn_version_t,
  &_swigt__p_int,
  &_swigt__p_long,
  &_swigt__p_p_apr_array_header_t,
  &_swigt__p_p_apr_file_t,
  &_swigt__p_p_apr_hash_t,
  &_swigt__p_p_char,
  &_swigt__p_p_svn_auth_baton_t,
  &_swigt__p_p_svn_auth_cred_simple_t,
  &_swigt__p_p_svn_auth_cred_ssl_client_cert_pw_t,
  &_swigt__p_p_svn_auth_cred_ssl_client_cert_t,
  &_swigt__p_p_svn_auth_cred_ssl_server_trust_t,
  &_swigt__p_p_svn_auth_cred_username_t,
  &_swigt__p_p_svn_auth_iterstate_t,
  &_swigt__p_p_svn_auth_provider_object_t,
  &_swigt__p_p_svn_checksum_t,
  &_swigt__p_p_svn_config_t,
  &_swigt__p_p_svn_diff_t,
  &_swigt__p_p_svn_io_dirent2_t,
  &_swigt__p_p_svn_patch_file_t,
  &_swigt__p_p_svn_patch_t,
  &_swigt__p_p_svn_stream_mark_t,
  &_swigt__p_p_svn_stream_t,
  &_swigt__p_p_svn_string_t,
  &_swigt__p_p_svn_stringbuf_t,
  &_swigt__p_p_void,
  &_swigt__p_svn_auth_baton_t,
  &_swigt__p_svn_auth_cred_simple_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_pw_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_t,
  &_swigt__p_svn_auth_cred_ssl_server_trust_t,
  &_swigt__p_svn_auth_cred_username_t,
  &_swigt__p_svn_auth_iterstate_t,
  &_swigt__p_svn_auth_provider_object_t,
  &_swigt__p_svn_auth_provider_t,
  &_swigt__p_svn_auth_ssl_server_cert_info_t,
  &_swigt__p_svn_cache_config_t,
  &_swigt__p_svn_checksum_ctx_t,
  &_swigt__p_svn_checksum_kind_t,
  &_swigt__p_svn_checksum_t,
  &_swigt__p_svn_commit_info_t,
  &_swigt__p_svn_config_t,
  &_swigt__p_svn_depth_t,
  &_swigt__p_svn_diff_binary_patch_t,
  &_swigt__p_svn_diff_conflict_display_style_t,
  &_swigt__p_svn_diff_datasource_e,
  &_swigt__p_svn_diff_file_ignore_space_t,
  &_swigt__p_svn_diff_file_options_t,
  &_swigt__p_svn_diff_fns2_t,
  &_swigt__p_svn_diff_fns_t,
  &_swigt__p_svn_diff_hunk_t,
  &_swigt__p_svn_diff_operation_kind_e,
  &_swigt__p_svn_diff_output_fns_t,
  &_swigt__p_svn_diff_t,
  &_swigt__p_svn_dirent_t,
  &_swigt__p_svn_errno_t,
  &_swigt__p_svn_error_t,
  &_swigt__p_svn_io_dirent2_t,
  &_swigt__p_svn_io_dirent_t,
  &_swigt__p_svn_io_file_del_t,
  &_swigt__p_svn_location_segment_t,
  &_swigt__p_svn_lock_t,
  &_swigt__p_svn_log_changed_path2_t,
  &_swigt__p_svn_log_changed_path_t,
  &_swigt__p_svn_log_entry_t,
  &_swigt__p_svn_merge_range_t,
  &_swigt__p_svn_mergeinfo_inheritance_t,
  &_swigt__p_svn_node_kind_t,
  &_swigt__p_svn_opt_revision_range_t,
  &_swigt__p_svn_opt_revision_t,
  &_swigt__p_svn_opt_revision_value_t,
  &_swigt__p_svn_opt_subcommand_desc2_t,
  &_swigt__p_svn_opt_subcommand_desc2_t_desc_overrides,
  &_swigt__p_svn_opt_subcommand_desc3_t,
  &_swigt__p_svn_opt_subcommand_desc3_t_desc_overrides,
  &_swigt__p_svn_opt_subcommand_desc_t,
  &_swigt__p_svn_patch_file_t,
  &_swigt__p_svn_patch_t,
  &_swigt__p_svn_prop_inherited_item_t,
  &_swigt__p_svn_prop_kind,
  &_swigt__p_svn_prop_patch_t,
  &_swigt__p_svn_prop_t,
  &_swigt__p_svn_stream_mark_t,
  &_swigt__p_svn_stream_t,
  &_swigt__p_svn_string_t,
  &_swigt__p_svn_stringbuf_t,
  &_swigt__p_svn_tristate_t,
  &_swigt__p_svn_version_checklist_t,
  &_swigt__p_svn_version_ext_linked_lib_t,
  &_swigt__p_svn_version_ext_loaded_lib_t,
  &_swigt__p_svn_version_extended_t,
  &_swigt__p_svn_version_t,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_long,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_FILE[] = {  {&_swigt__p_FILE, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_allocator_t[] = {  {&_swigt__p_apr_allocator_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_array_header_t[] = {  {&_swigt__p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_dir_t[] = {  {&_swigt__p_apr_dir_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_file_t[] = {  {&_swigt__p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_finfo_t[] = {  {&_swigt__p_apr_finfo_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_getopt_option_t[] = {  {&_swigt__p_apr_getopt_option_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_getopt_t[] = {  {&_swigt__p_apr_getopt_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_hash_t[] = {  {&_swigt__p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int32_t[] = {  {&_swigt__p_apr_int32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int64_t[] = {  {&_swigt__p_apr_int64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_off_t[] = {  {&_swigt__p_apr_off_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_pool_t[] = {  {&_swigt__p_apr_pool_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_pool_wrapper_t[] = {  {&_swigt__p_apr_pool_wrapper_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_proc_t[] = {  {&_swigt__p_apr_proc_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_size_t[] = {  {&_swigt__p_apr_size_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_uint32_t[] = {  {&_swigt__p_apr_uint32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_uint64_t[] = {  {&_swigt__p_apr_uint64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t[] = {  {&_swigt__p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_char_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_char_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void[] = {  {&_swigt__p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_q_const__char_p_void__int[] = {  {&_swigt__p_f_p_q_const__char_p_q_const__char_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_q_const__char_p_void_p_apr_pool_t__int[] = {  {&_swigt__p_f_p_q_const__char_p_q_const__char_p_void_p_apr_pool_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_void__int[] = {  {&_swigt__p_f_p_q_const__char_p_void__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_void_p_apr_pool_t__int[] = {  {&_swigt__p_f_p_q_const__char_p_void_p_apr_pool_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__struct_svn_version_t_p_q_const__struct_svn_version_t__int[] = {  {&_swigt__p_f_p_q_const__struct_svn_version_t_p_q_const__struct_svn_version_t__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void__void[] = {  {&_swigt__p_f_p_void__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_apr_size_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_apr_size_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_char_p_apr_size_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_char_p_apr_size_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_apr_size_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_apr_size_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_boolean_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_boolean_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_void__void[] = {  {&_swigt__p_f_p_void_p_void__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_void_p_void_p_int__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_void_p_void_p_int__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_svn_diff_datasource_e__p_svn_error_t[] = {  {&_swigt__p_f_p_void_svn_diff_datasource_e__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_void__p_svn_version_t[] = {  {&_swigt__p_f_void__p_svn_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_array_header_t[] = {  {&_swigt__p_p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_file_t[] = {  {&_swigt__p_p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_hash_t[] = {  {&_swigt__p_p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_baton_t[] = {  {&_swigt__p_p_svn_auth_baton_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_cred_simple_t[] = {  {&_swigt__p_p_svn_auth_cred_simple_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_cred_ssl_client_cert_pw_t[] = {  {&_swigt__p_p_svn_auth_cred_ssl_client_cert_pw_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_cred_ssl_client_cert_t[] = {  {&_swigt__p_p_svn_auth_cred_ssl_client_cert_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_cred_ssl_server_trust_t[] = {  {&_swigt__p_p_svn_auth_cred_ssl_server_trust_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_cred_username_t[] = {  {&_swigt__p_p_svn_auth_cred_username_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_iterstate_t[] = {  {&_swigt__p_p_svn_auth_iterstate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_auth_provider_object_t[] = {  {&_swigt__p_p_svn_auth_provider_object_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_checksum_t[] = {  {&_swigt__p_p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_config_t[] = {  {&_swigt__p_p_svn_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_diff_t[] = {  {&_swigt__p_p_svn_diff_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_io_dirent2_t[] = {  {&_swigt__p_p_svn_io_dirent2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_patch_file_t[] = {  {&_swigt__p_p_svn_patch_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_patch_t[] = {  {&_swigt__p_p_svn_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_stream_mark_t[] = {  {&_swigt__p_p_svn_stream_mark_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_stream_t[] = {  {&_swigt__p_p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_string_t[] = {  {&_swigt__p_p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_stringbuf_t[] = {  {&_swigt__p_p_svn_stringbuf_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_void[] = {  {&_swigt__p_p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_baton_t[] = {  {&_swigt__p_svn_auth_baton_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_simple_t[] = {  {&_swigt__p_svn_auth_cred_simple_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_pw_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_pw_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_server_trust_t[] = {  {&_swigt__p_svn_auth_cred_ssl_server_trust_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_username_t[] = {  {&_swigt__p_svn_auth_cred_username_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_iterstate_t[] = {  {&_swigt__p_svn_auth_iterstate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_object_t[] = {  {&_swigt__p_svn_auth_provider_object_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_t[] = {  {&_swigt__p_svn_auth_provider_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_ssl_server_cert_info_t[] = {  {&_swigt__p_svn_auth_ssl_server_cert_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_cache_config_t[] = {  {&_swigt__p_svn_cache_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_ctx_t[] = {  {&_swigt__p_svn_checksum_ctx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_kind_t[] = {  {&_swigt__p_svn_checksum_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_t[] = {  {&_swigt__p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_commit_info_t[] = {  {&_swigt__p_svn_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_config_t[] = {  {&_swigt__p_svn_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_depth_t[] = {  {&_swigt__p_svn_depth_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_binary_patch_t[] = {  {&_swigt__p_svn_diff_binary_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_conflict_display_style_t[] = {  {&_swigt__p_svn_diff_conflict_display_style_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_datasource_e[] = {  {&_swigt__p_svn_diff_datasource_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_file_ignore_space_t[] = {  {&_swigt__p_svn_diff_file_ignore_space_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_file_options_t[] = {  {&_swigt__p_svn_diff_file_options_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_fns2_t[] = {  {&_swigt__p_svn_diff_fns2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_fns_t[] = {  {&_swigt__p_svn_diff_fns_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_hunk_t[] = {  {&_swigt__p_svn_diff_hunk_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_operation_kind_e[] = {  {&_swigt__p_svn_diff_operation_kind_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_output_fns_t[] = {  {&_swigt__p_svn_diff_output_fns_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_t[] = {  {&_swigt__p_svn_diff_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_dirent_t[] = {  {&_swigt__p_svn_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_errno_t[] = {  {&_swigt__p_svn_errno_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_error_t[] = {  {&_swigt__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent2_t[] = {  {&_swigt__p_svn_io_dirent2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent_t[] = {  {&_swigt__p_svn_io_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_file_del_t[] = {  {&_swigt__p_svn_io_file_del_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_location_segment_t[] = {  {&_swigt__p_svn_location_segment_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_lock_t[] = {  {&_swigt__p_svn_lock_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path2_t[] = {  {&_swigt__p_svn_log_changed_path2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path_t[] = {  {&_swigt__p_svn_log_changed_path_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_entry_t[] = {  {&_swigt__p_svn_log_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_merge_range_t[] = {  {&_swigt__p_svn_merge_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_mergeinfo_inheritance_t[] = {  {&_swigt__p_svn_mergeinfo_inheritance_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_node_kind_t[] = {  {&_swigt__p_svn_node_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_range_t[] = {  {&_swigt__p_svn_opt_revision_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_t[] = {  {&_swigt__p_svn_opt_revision_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_value_t[] = {  {&_swigt__p_svn_opt_revision_value_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc2_t[] = {  {&_swigt__p_svn_opt_subcommand_desc2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc2_t_desc_overrides[] = {  {&_swigt__p_svn_opt_subcommand_desc2_t_desc_overrides, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc3_t[] = {  {&_swigt__p_svn_opt_subcommand_desc3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc3_t_desc_overrides[] = {  {&_swigt__p_svn_opt_subcommand_desc3_t_desc_overrides, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc_t[] = {  {&_swigt__p_svn_opt_subcommand_desc_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_patch_file_t[] = {  {&_swigt__p_svn_patch_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_patch_t[] = {  {&_swigt__p_svn_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_inherited_item_t[] = {  {&_swigt__p_svn_prop_inherited_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_kind[] = {  {&_swigt__p_svn_prop_kind, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_patch_t[] = {  {&_swigt__p_svn_prop_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_t[] = {  {&_swigt__p_svn_prop_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_mark_t[] = {  {&_swigt__p_svn_stream_mark_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_t[] = {  {&_swigt__p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_string_t[] = {  {&_swigt__p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stringbuf_t[] = {  {&_swigt__p_svn_stringbuf_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_tristate_t[] = {  {&_swigt__p_svn_tristate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_checklist_t[] = {  {&_swigt__p_svn_version_checklist_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_linked_lib_t[] = {  {&_swigt__p_svn_version_ext_linked_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_loaded_lib_t[] = {  {&_swigt__p_svn_version_ext_loaded_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_extended_t[] = {  {&_swigt__p_svn_version_extended_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_t[] = {  {&_swigt__p_svn_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_FILE,
  _swigc__p_apr_allocator_t,
  _swigc__p_apr_array_header_t,
  _swigc__p_apr_dir_t,
  _swigc__p_apr_file_t,
  _swigc__p_apr_finfo_t,
  _swigc__p_apr_getopt_option_t,
  _swigc__p_apr_getopt_t,
  _swigc__p_apr_hash_t,
  _swigc__p_apr_int32_t,
  _swigc__p_apr_int64_t,
  _swigc__p_apr_off_t,
  _swigc__p_apr_pool_t,
  _swigc__p_apr_pool_wrapper_t,
  _swigc__p_apr_proc_t,
  _swigc__p_apr_size_t,
  _swigc__p_apr_uint32_t,
  _swigc__p_apr_uint64_t,
  _swigc__p_char,
  _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_apr_uint32_t_p_p_void_p_void_svn_diff_datasource_e__p_svn_error_t,
  _swigc__p_f_p_p_char_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_struct_svn_auth_provider_object_t_p_apr_pool_t__void,
  _swigc__p_f_p_p_svn_auth_cred_simple_t_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_ssl_client_cert_pw_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_ssl_client_cert_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_ssl_server_trust_t_p_void_p_q_const__char_apr_uint32_t_p_q_const__svn_auth_ssl_server_cert_info_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_auth_cred_username_t_p_void_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_stream_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_void_p_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_void_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_q_const__char_p_void__int,
  _swigc__p_f_p_q_const__char_p_q_const__char_p_void_p_apr_pool_t__int,
  _swigc__p_f_p_q_const__char_p_void__int,
  _swigc__p_f_p_q_const__char_p_void_p_apr_pool_t__int,
  _swigc__p_f_p_q_const__struct_svn_version_t_p_q_const__struct_svn_version_t__int,
  _swigc__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_boolean_t_p_void_p_q_const__char_p_q_const__char_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_boolean_t_p_void_p_void_p_apr_hash_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void__p_svn_error_t,
  _swigc__p_f_p_void__void,
  _swigc__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t__p_svn_error_t,
  _swigc__p_f_p_void_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_apr_off_t_p_svn_diff_t__p_svn_error_t,
  _swigc__p_f_p_void_apr_size_t__p_svn_error_t,
  _swigc__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_apr_off_t_p_apr_off_t_p_q_const__svn_diff_datasource_e_apr_size_t__p_svn_error_t,
  _swigc__p_f_p_void_p_char_p_apr_size_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_svn_stream_mark_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_svn_stringbuf_t_p_q_const__char_p_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_apr_size_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__apr_finfo_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__svn_stream_mark_t__p_svn_error_t,
  _swigc__p_f_p_void_p_svn_boolean_t__p_svn_error_t,
  _swigc__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_void__void,
  _swigc__p_f_p_void_p_void_p_void_p_int__p_svn_error_t,
  _swigc__p_f_p_void_svn_diff_datasource_e__p_svn_error_t,
  _swigc__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_void__p_svn_version_t,
  _swigc__p_int,
  _swigc__p_long,
  _swigc__p_p_apr_array_header_t,
  _swigc__p_p_apr_file_t,
  _swigc__p_p_apr_hash_t,
  _swigc__p_p_char,
  _swigc__p_p_svn_auth_baton_t,
  _swigc__p_p_svn_auth_cred_simple_t,
  _swigc__p_p_svn_auth_cred_ssl_client_cert_pw_t,
  _swigc__p_p_svn_auth_cred_ssl_client_cert_t,
  _swigc__p_p_svn_auth_cred_ssl_server_trust_t,
  _swigc__p_p_svn_auth_cred_username_t,
  _swigc__p_p_svn_auth_iterstate_t,
  _swigc__p_p_svn_auth_provider_object_t,
  _swigc__p_p_svn_checksum_t,
  _swigc__p_p_svn_config_t,
  _swigc__p_p_svn_diff_t,
  _swigc__p_p_svn_io_dirent2_t,
  _swigc__p_p_svn_patch_file_t,
  _swigc__p_p_svn_patch_t,
  _swigc__p_p_svn_stream_mark_t,
  _swigc__p_p_svn_stream_t,
  _swigc__p_p_svn_string_t,
  _swigc__p_p_svn_stringbuf_t,
  _swigc__p_p_void,
  _swigc__p_svn_auth_baton_t,
  _swigc__p_svn_auth_cred_simple_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_pw_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_t,
  _swigc__p_svn_auth_cred_ssl_server_trust_t,
  _swigc__p_svn_auth_cred_username_t,
  _swigc__p_svn_auth_iterstate_t,
  _swigc__p_svn_auth_provider_object_t,
  _swigc__p_svn_auth_provider_t,
  _swigc__p_svn_auth_ssl_server_cert_info_t,
  _swigc__p_svn_cache_config_t,
  _swigc__p_svn_checksum_ctx_t,
  _swigc__p_svn_checksum_kind_t,
  _swigc__p_svn_checksum_t,
  _swigc__p_svn_commit_info_t,
  _swigc__p_svn_config_t,
  _swigc__p_svn_depth_t,
  _swigc__p_svn_diff_binary_patch_t,
  _swigc__p_svn_diff_conflict_display_style_t,
  _swigc__p_svn_diff_datasource_e,
  _swigc__p_svn_diff_file_ignore_space_t,
  _swigc__p_svn_diff_file_options_t,
  _swigc__p_svn_diff_fns2_t,
  _swigc__p_svn_diff_fns_t,
  _swigc__p_svn_diff_hunk_t,
  _swigc__p_svn_diff_operation_kind_e,
  _swigc__p_svn_diff_output_fns_t,
  _swigc__p_svn_diff_t,
  _swigc__p_svn_dirent_t,
  _swigc__p_svn_errno_t,
  _swigc__p_svn_error_t,
  _swigc__p_svn_io_dirent2_t,
  _swigc__p_svn_io_dirent_t,
  _swigc__p_svn_io_file_del_t,
  _swigc__p_svn_location_segment_t,
  _swigc__p_svn_lock_t,
  _swigc__p_svn_log_changed_path2_t,
  _swigc__p_svn_log_changed_path_t,
  _swigc__p_svn_log_entry_t,
  _swigc__p_svn_merge_range_t,
  _swigc__p_svn_mergeinfo_inheritance_t,
  _swigc__p_svn_node_kind_t,
  _swigc__p_svn_opt_revision_range_t,
  _swigc__p_svn_opt_revision_t,
  _swigc__p_svn_opt_revision_value_t,
  _swigc__p_svn_opt_subcommand_desc2_t,
  _swigc__p_svn_opt_subcommand_desc2_t_desc_overrides,
  _swigc__p_svn_opt_subcommand_desc3_t,
  _swigc__p_svn_opt_subcommand_desc3_t_desc_overrides,
  _swigc__p_svn_opt_subcommand_desc_t,
  _swigc__p_svn_patch_file_t,
  _swigc__p_svn_patch_t,
  _swigc__p_svn_prop_inherited_item_t,
  _swigc__p_svn_prop_kind,
  _swigc__p_svn_prop_patch_t,
  _swigc__p_svn_prop_t,
  _swigc__p_svn_stream_mark_t,
  _swigc__p_svn_stream_t,
  _swigc__p_svn_string_t,
  _swigc__p_svn_stringbuf_t,
  _swigc__p_svn_tristate_t,
  _swigc__p_svn_version_checklist_t,
  _swigc__p_svn_version_ext_linked_lib_t,
  _swigc__p_svn_version_ext_loaded_lib_t,
  _swigc__p_svn_version_extended_t,
  _swigc__p_svn_version_t,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_long,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;

  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }

  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);

    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }

  /* When multiple interpreters are used, a module could have already been initialized in
     a different interpreter, but not yet have a pointer in this interpreter.
     In this case, we do not want to continue adding types... everything should be
     set up already */
  if (init == 0) return;

  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %d\n", swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;

#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
#endif

    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
	type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }

    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {

      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
	if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
	if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
	  printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
	  cast->type = ret;
	  ret = 0;
	} else {
	  /* Check for casting already in the list */
	  swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
	  if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
	  if (!ocast) ret = 0;
	}
      }

      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
	printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;

#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
  printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;

  if (init_run) return;
  init_run = 1;

  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
            SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{ /* c-mode */
#endif
}
#endif

/*

*/
#ifdef __cplusplus
extern "C"
#endif
SWIGEXPORT void Init_core(void) {
  size_t i;
  
  SWIG_InitRuntime();
  mCore = rb_define_module("Svn");
  mCore = rb_define_module_under(mCore, "Ext");
  mCore = rb_define_module_under(mCore, "Core");
  
  SWIG_InitializeModule(0);
  for (i = 0; i < swig_module.size; i++) {
    SWIG_define_class(swig_module.types[i]);
  }
  
  SWIG_RubyInitializeTrackings();
  rb_define_const(mCore, "SWIG_SVN_INVALID_REVNUM", SWIG_From_long((long)(-1)));
  rb_define_const(mCore, "SWIG_SVN_IGNORED_REVNUM", SWIG_From_long((long)(-1)));
  rb_define_module_function(mCore, "apr_initialize", _wrap_apr_initialize, -1);
  rb_define_module_function(mCore, "apr_terminate", _wrap_apr_terminate, -1);
  rb_define_module_function(mCore, "apr_time_ansi_put", _wrap_apr_time_ansi_put, -1);
  rb_define_module_function(mCore, "apr_pool_destroy", _wrap_apr_pool_destroy, -1);
  rb_define_module_function(mCore, "apr_pool_clear", _wrap_apr_pool_clear, -1);
  rb_define_module_function(mCore, "apr_file_open_stdout", _wrap_apr_file_open_stdout, -1);
  rb_define_module_function(mCore, "apr_file_open_stderr", _wrap_apr_file_open_stderr, -1);
  rb_define_const(mCore, "SVN_ERR_CATEGORY_SIZE", SWIG_From_int((int)(5000)));
  rb_define_const(mCore, "SVN_WARNING", SWIG_From_int((int)(SVN_WARNING)));
  rb_define_const(mCore, "SVN_ERR_BAD_CONTAINING_POOL", SWIG_From_int((int)(SVN_ERR_BAD_CONTAINING_POOL)));
  rb_define_const(mCore, "SVN_ERR_BAD_FILENAME", SWIG_From_int((int)(SVN_ERR_BAD_FILENAME)));
  rb_define_const(mCore, "SVN_ERR_BAD_URL", SWIG_From_int((int)(SVN_ERR_BAD_URL)));
  rb_define_const(mCore, "SVN_ERR_BAD_DATE", SWIG_From_int((int)(SVN_ERR_BAD_DATE)));
  rb_define_const(mCore, "SVN_ERR_BAD_MIME_TYPE", SWIG_From_int((int)(SVN_ERR_BAD_MIME_TYPE)));
  rb_define_const(mCore, "SVN_ERR_BAD_PROPERTY_VALUE", SWIG_From_int((int)(SVN_ERR_BAD_PROPERTY_VALUE)));
  rb_define_const(mCore, "SVN_ERR_BAD_VERSION_FILE_FORMAT", SWIG_From_int((int)(SVN_ERR_BAD_VERSION_FILE_FORMAT)));
  rb_define_const(mCore, "SVN_ERR_BAD_RELATIVE_PATH", SWIG_From_int((int)(SVN_ERR_BAD_RELATIVE_PATH)));
  rb_define_const(mCore, "SVN_ERR_BAD_UUID", SWIG_From_int((int)(SVN_ERR_BAD_UUID)));
  rb_define_const(mCore, "SVN_ERR_BAD_CONFIG_VALUE", SWIG_From_int((int)(SVN_ERR_BAD_CONFIG_VALUE)));
  rb_define_const(mCore, "SVN_ERR_BAD_SERVER_SPECIFICATION", SWIG_From_int((int)(SVN_ERR_BAD_SERVER_SPECIFICATION)));
  rb_define_const(mCore, "SVN_ERR_BAD_CHECKSUM_KIND", SWIG_From_int((int)(SVN_ERR_BAD_CHECKSUM_KIND)));
  rb_define_const(mCore, "SVN_ERR_BAD_CHECKSUM_PARSE", SWIG_From_int((int)(SVN_ERR_BAD_CHECKSUM_PARSE)));
  rb_define_const(mCore, "SVN_ERR_BAD_TOKEN", SWIG_From_int((int)(SVN_ERR_BAD_TOKEN)));
  rb_define_const(mCore, "SVN_ERR_BAD_CHANGELIST_NAME", SWIG_From_int((int)(SVN_ERR_BAD_CHANGELIST_NAME)));
  rb_define_const(mCore, "SVN_ERR_BAD_ATOMIC", SWIG_From_int((int)(SVN_ERR_BAD_ATOMIC)));
  rb_define_const(mCore, "SVN_ERR_BAD_COMPRESSION_METHOD", SWIG_From_int((int)(SVN_ERR_BAD_COMPRESSION_METHOD)));
  rb_define_const(mCore, "SVN_ERR_BAD_PROPERTY_VALUE_EOL", SWIG_From_int((int)(SVN_ERR_BAD_PROPERTY_VALUE_EOL)));
  rb_define_const(mCore, "SVN_ERR_XML_ATTRIB_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_XML_ATTRIB_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_XML_MISSING_ANCESTRY", SWIG_From_int((int)(SVN_ERR_XML_MISSING_ANCESTRY)));
  rb_define_const(mCore, "SVN_ERR_XML_UNKNOWN_ENCODING", SWIG_From_int((int)(SVN_ERR_XML_UNKNOWN_ENCODING)));
  rb_define_const(mCore, "SVN_ERR_XML_MALFORMED", SWIG_From_int((int)(SVN_ERR_XML_MALFORMED)));
  rb_define_const(mCore, "SVN_ERR_XML_UNESCAPABLE_DATA", SWIG_From_int((int)(SVN_ERR_XML_UNESCAPABLE_DATA)));
  rb_define_const(mCore, "SVN_ERR_XML_UNEXPECTED_ELEMENT", SWIG_From_int((int)(SVN_ERR_XML_UNEXPECTED_ELEMENT)));
  rb_define_const(mCore, "SVN_ERR_IO_INCONSISTENT_EOL", SWIG_From_int((int)(SVN_ERR_IO_INCONSISTENT_EOL)));
  rb_define_const(mCore, "SVN_ERR_IO_UNKNOWN_EOL", SWIG_From_int((int)(SVN_ERR_IO_UNKNOWN_EOL)));
  rb_define_const(mCore, "SVN_ERR_IO_CORRUPT_EOL", SWIG_From_int((int)(SVN_ERR_IO_CORRUPT_EOL)));
  rb_define_const(mCore, "SVN_ERR_IO_UNIQUE_NAMES_EXHAUSTED", SWIG_From_int((int)(SVN_ERR_IO_UNIQUE_NAMES_EXHAUSTED)));
  rb_define_const(mCore, "SVN_ERR_IO_PIPE_FRAME_ERROR", SWIG_From_int((int)(SVN_ERR_IO_PIPE_FRAME_ERROR)));
  rb_define_const(mCore, "SVN_ERR_IO_PIPE_READ_ERROR", SWIG_From_int((int)(SVN_ERR_IO_PIPE_READ_ERROR)));
  rb_define_const(mCore, "SVN_ERR_IO_WRITE_ERROR", SWIG_From_int((int)(SVN_ERR_IO_WRITE_ERROR)));
  rb_define_const(mCore, "SVN_ERR_IO_PIPE_WRITE_ERROR", SWIG_From_int((int)(SVN_ERR_IO_PIPE_WRITE_ERROR)));
  rb_define_const(mCore, "SVN_ERR_STREAM_UNEXPECTED_EOF", SWIG_From_int((int)(SVN_ERR_STREAM_UNEXPECTED_EOF)));
  rb_define_const(mCore, "SVN_ERR_STREAM_MALFORMED_DATA", SWIG_From_int((int)(SVN_ERR_STREAM_MALFORMED_DATA)));
  rb_define_const(mCore, "SVN_ERR_STREAM_UNRECOGNIZED_DATA", SWIG_From_int((int)(SVN_ERR_STREAM_UNRECOGNIZED_DATA)));
  rb_define_const(mCore, "SVN_ERR_STREAM_SEEK_NOT_SUPPORTED", SWIG_From_int((int)(SVN_ERR_STREAM_SEEK_NOT_SUPPORTED)));
  rb_define_const(mCore, "SVN_ERR_STREAM_NOT_SUPPORTED", SWIG_From_int((int)(SVN_ERR_STREAM_NOT_SUPPORTED)));
  rb_define_const(mCore, "SVN_ERR_NODE_UNKNOWN_KIND", SWIG_From_int((int)(SVN_ERR_NODE_UNKNOWN_KIND)));
  rb_define_const(mCore, "SVN_ERR_NODE_UNEXPECTED_KIND", SWIG_From_int((int)(SVN_ERR_NODE_UNEXPECTED_KIND)));
  rb_define_const(mCore, "SVN_ERR_ENTRY_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_ENTRY_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_ENTRY_EXISTS", SWIG_From_int((int)(SVN_ERR_ENTRY_EXISTS)));
  rb_define_const(mCore, "SVN_ERR_ENTRY_MISSING_REVISION", SWIG_From_int((int)(SVN_ERR_ENTRY_MISSING_REVISION)));
  rb_define_const(mCore, "SVN_ERR_ENTRY_MISSING_URL", SWIG_From_int((int)(SVN_ERR_ENTRY_MISSING_URL)));
  rb_define_const(mCore, "SVN_ERR_ENTRY_ATTRIBUTE_INVALID", SWIG_From_int((int)(SVN_ERR_ENTRY_ATTRIBUTE_INVALID)));
  rb_define_const(mCore, "SVN_ERR_ENTRY_FORBIDDEN", SWIG_From_int((int)(SVN_ERR_ENTRY_FORBIDDEN)));
  rb_define_const(mCore, "SVN_ERR_WC_OBSTRUCTED_UPDATE", SWIG_From_int((int)(SVN_ERR_WC_OBSTRUCTED_UPDATE)));
  rb_define_const(mCore, "SVN_ERR_WC_UNWIND_MISMATCH", SWIG_From_int((int)(SVN_ERR_WC_UNWIND_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_WC_UNWIND_EMPTY", SWIG_From_int((int)(SVN_ERR_WC_UNWIND_EMPTY)));
  rb_define_const(mCore, "SVN_ERR_WC_UNWIND_NOT_EMPTY", SWIG_From_int((int)(SVN_ERR_WC_UNWIND_NOT_EMPTY)));
  rb_define_const(mCore, "SVN_ERR_WC_LOCKED", SWIG_From_int((int)(SVN_ERR_WC_LOCKED)));
  rb_define_const(mCore, "SVN_ERR_WC_NOT_LOCKED", SWIG_From_int((int)(SVN_ERR_WC_NOT_LOCKED)));
  rb_define_const(mCore, "SVN_ERR_WC_INVALID_LOCK", SWIG_From_int((int)(SVN_ERR_WC_INVALID_LOCK)));
  rb_define_const(mCore, "SVN_ERR_WC_NOT_WORKING_COPY", SWIG_From_int((int)(SVN_ERR_WC_NOT_WORKING_COPY)));
  rb_define_const(mCore, "SVN_ERR_WC_NOT_DIRECTORY", SWIG_From_int((int)(SVN_ERR_WC_NOT_DIRECTORY)));
  rb_define_const(mCore, "SVN_ERR_WC_NOT_FILE", SWIG_From_int((int)(SVN_ERR_WC_NOT_FILE)));
  rb_define_const(mCore, "SVN_ERR_WC_BAD_ADM_LOG", SWIG_From_int((int)(SVN_ERR_WC_BAD_ADM_LOG)));
  rb_define_const(mCore, "SVN_ERR_WC_PATH_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_WC_PATH_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_WC_NOT_UP_TO_DATE", SWIG_From_int((int)(SVN_ERR_WC_NOT_UP_TO_DATE)));
  rb_define_const(mCore, "SVN_ERR_WC_LEFT_LOCAL_MOD", SWIG_From_int((int)(SVN_ERR_WC_LEFT_LOCAL_MOD)));
  rb_define_const(mCore, "SVN_ERR_WC_SCHEDULE_CONFLICT", SWIG_From_int((int)(SVN_ERR_WC_SCHEDULE_CONFLICT)));
  rb_define_const(mCore, "SVN_ERR_WC_PATH_FOUND", SWIG_From_int((int)(SVN_ERR_WC_PATH_FOUND)));
  rb_define_const(mCore, "SVN_ERR_WC_FOUND_CONFLICT", SWIG_From_int((int)(SVN_ERR_WC_FOUND_CONFLICT)));
  rb_define_const(mCore, "SVN_ERR_WC_CORRUPT", SWIG_From_int((int)(SVN_ERR_WC_CORRUPT)));
  rb_define_const(mCore, "SVN_ERR_WC_CORRUPT_TEXT_BASE", SWIG_From_int((int)(SVN_ERR_WC_CORRUPT_TEXT_BASE)));
  rb_define_const(mCore, "SVN_ERR_WC_NODE_KIND_CHANGE", SWIG_From_int((int)(SVN_ERR_WC_NODE_KIND_CHANGE)));
  rb_define_const(mCore, "SVN_ERR_WC_INVALID_OP_ON_CWD", SWIG_From_int((int)(SVN_ERR_WC_INVALID_OP_ON_CWD)));
  rb_define_const(mCore, "SVN_ERR_WC_BAD_ADM_LOG_START", SWIG_From_int((int)(SVN_ERR_WC_BAD_ADM_LOG_START)));
  rb_define_const(mCore, "SVN_ERR_WC_UNSUPPORTED_FORMAT", SWIG_From_int((int)(SVN_ERR_WC_UNSUPPORTED_FORMAT)));
  rb_define_const(mCore, "SVN_ERR_WC_BAD_PATH", SWIG_From_int((int)(SVN_ERR_WC_BAD_PATH)));
  rb_define_const(mCore, "SVN_ERR_WC_INVALID_SCHEDULE", SWIG_From_int((int)(SVN_ERR_WC_INVALID_SCHEDULE)));
  rb_define_const(mCore, "SVN_ERR_WC_INVALID_RELOCATION", SWIG_From_int((int)(SVN_ERR_WC_INVALID_RELOCATION)));
  rb_define_const(mCore, "SVN_ERR_WC_INVALID_SWITCH", SWIG_From_int((int)(SVN_ERR_WC_INVALID_SWITCH)));
  rb_define_const(mCore, "SVN_ERR_WC_MISMATCHED_CHANGELIST", SWIG_From_int((int)(SVN_ERR_WC_MISMATCHED_CHANGELIST)));
  rb_define_const(mCore, "SVN_ERR_WC_CONFLICT_RESOLVER_FAILURE", SWIG_From_int((int)(SVN_ERR_WC_CONFLICT_RESOLVER_FAILURE)));
  rb_define_const(mCore, "SVN_ERR_WC_COPYFROM_PATH_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_WC_COPYFROM_PATH_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_WC_CHANGELIST_MOVE", SWIG_From_int((int)(SVN_ERR_WC_CHANGELIST_MOVE)));
  rb_define_const(mCore, "SVN_ERR_WC_CANNOT_DELETE_FILE_EXTERNAL", SWIG_From_int((int)(SVN_ERR_WC_CANNOT_DELETE_FILE_EXTERNAL)));
  rb_define_const(mCore, "SVN_ERR_WC_CANNOT_MOVE_FILE_EXTERNAL", SWIG_From_int((int)(SVN_ERR_WC_CANNOT_MOVE_FILE_EXTERNAL)));
  rb_define_const(mCore, "SVN_ERR_WC_DB_ERROR", SWIG_From_int((int)(SVN_ERR_WC_DB_ERROR)));
  rb_define_const(mCore, "SVN_ERR_WC_MISSING", SWIG_From_int((int)(SVN_ERR_WC_MISSING)));
  rb_define_const(mCore, "SVN_ERR_WC_NOT_SYMLINK", SWIG_From_int((int)(SVN_ERR_WC_NOT_SYMLINK)));
  rb_define_const(mCore, "SVN_ERR_WC_PATH_UNEXPECTED_STATUS", SWIG_From_int((int)(SVN_ERR_WC_PATH_UNEXPECTED_STATUS)));
  rb_define_const(mCore, "SVN_ERR_WC_UPGRADE_REQUIRED", SWIG_From_int((int)(SVN_ERR_WC_UPGRADE_REQUIRED)));
  rb_define_const(mCore, "SVN_ERR_WC_CLEANUP_REQUIRED", SWIG_From_int((int)(SVN_ERR_WC_CLEANUP_REQUIRED)));
  rb_define_const(mCore, "SVN_ERR_WC_INVALID_OPERATION_DEPTH", SWIG_From_int((int)(SVN_ERR_WC_INVALID_OPERATION_DEPTH)));
  rb_define_const(mCore, "SVN_ERR_WC_PATH_ACCESS_DENIED", SWIG_From_int((int)(SVN_ERR_WC_PATH_ACCESS_DENIED)));
  rb_define_const(mCore, "SVN_ERR_WC_MIXED_REVISIONS", SWIG_From_int((int)(SVN_ERR_WC_MIXED_REVISIONS)));
  rb_define_const(mCore, "SVN_ERR_WC_DUPLICATE_EXTERNALS_TARGET", SWIG_From_int((int)(SVN_ERR_WC_DUPLICATE_EXTERNALS_TARGET)));
  rb_define_const(mCore, "SVN_ERR_FS_GENERAL", SWIG_From_int((int)(SVN_ERR_FS_GENERAL)));
  rb_define_const(mCore, "SVN_ERR_FS_CLEANUP", SWIG_From_int((int)(SVN_ERR_FS_CLEANUP)));
  rb_define_const(mCore, "SVN_ERR_FS_ALREADY_OPEN", SWIG_From_int((int)(SVN_ERR_FS_ALREADY_OPEN)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_OPEN", SWIG_From_int((int)(SVN_ERR_FS_NOT_OPEN)));
  rb_define_const(mCore, "SVN_ERR_FS_CORRUPT", SWIG_From_int((int)(SVN_ERR_FS_CORRUPT)));
  rb_define_const(mCore, "SVN_ERR_FS_PATH_SYNTAX", SWIG_From_int((int)(SVN_ERR_FS_PATH_SYNTAX)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_REVISION", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_REVISION)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_TRANSACTION", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_TRANSACTION)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_ENTRY", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_ENTRY)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_REPRESENTATION", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_REPRESENTATION)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_STRING", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_STRING)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_COPY", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_COPY)));
  rb_define_const(mCore, "SVN_ERR_FS_TRANSACTION_NOT_MUTABLE", SWIG_From_int((int)(SVN_ERR_FS_TRANSACTION_NOT_MUTABLE)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_FS_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_FS_ID_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_FS_ID_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_ID", SWIG_From_int((int)(SVN_ERR_FS_NOT_ID)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_DIRECTORY", SWIG_From_int((int)(SVN_ERR_FS_NOT_DIRECTORY)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_FILE", SWIG_From_int((int)(SVN_ERR_FS_NOT_FILE)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_SINGLE_PATH_COMPONENT", SWIG_From_int((int)(SVN_ERR_FS_NOT_SINGLE_PATH_COMPONENT)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_MUTABLE", SWIG_From_int((int)(SVN_ERR_FS_NOT_MUTABLE)));
  rb_define_const(mCore, "SVN_ERR_FS_ALREADY_EXISTS", SWIG_From_int((int)(SVN_ERR_FS_ALREADY_EXISTS)));
  rb_define_const(mCore, "SVN_ERR_FS_ROOT_DIR", SWIG_From_int((int)(SVN_ERR_FS_ROOT_DIR)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_TXN_ROOT", SWIG_From_int((int)(SVN_ERR_FS_NOT_TXN_ROOT)));
  rb_define_const(mCore, "SVN_ERR_FS_NOT_REVISION_ROOT", SWIG_From_int((int)(SVN_ERR_FS_NOT_REVISION_ROOT)));
  rb_define_const(mCore, "SVN_ERR_FS_CONFLICT", SWIG_From_int((int)(SVN_ERR_FS_CONFLICT)));
  rb_define_const(mCore, "SVN_ERR_FS_REP_CHANGED", SWIG_From_int((int)(SVN_ERR_FS_REP_CHANGED)));
  rb_define_const(mCore, "SVN_ERR_FS_REP_NOT_MUTABLE", SWIG_From_int((int)(SVN_ERR_FS_REP_NOT_MUTABLE)));
  rb_define_const(mCore, "SVN_ERR_FS_MALFORMED_SKEL", SWIG_From_int((int)(SVN_ERR_FS_MALFORMED_SKEL)));
  rb_define_const(mCore, "SVN_ERR_FS_TXN_OUT_OF_DATE", SWIG_From_int((int)(SVN_ERR_FS_TXN_OUT_OF_DATE)));
  rb_define_const(mCore, "SVN_ERR_FS_BERKELEY_DB", SWIG_From_int((int)(SVN_ERR_FS_BERKELEY_DB)));
  rb_define_const(mCore, "SVN_ERR_FS_BERKELEY_DB_DEADLOCK", SWIG_From_int((int)(SVN_ERR_FS_BERKELEY_DB_DEADLOCK)));
  rb_define_const(mCore, "SVN_ERR_FS_TRANSACTION_DEAD", SWIG_From_int((int)(SVN_ERR_FS_TRANSACTION_DEAD)));
  rb_define_const(mCore, "SVN_ERR_FS_TRANSACTION_NOT_DEAD", SWIG_From_int((int)(SVN_ERR_FS_TRANSACTION_NOT_DEAD)));
  rb_define_const(mCore, "SVN_ERR_FS_UNKNOWN_FS_TYPE", SWIG_From_int((int)(SVN_ERR_FS_UNKNOWN_FS_TYPE)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_USER", SWIG_From_int((int)(SVN_ERR_FS_NO_USER)));
  rb_define_const(mCore, "SVN_ERR_FS_PATH_ALREADY_LOCKED", SWIG_From_int((int)(SVN_ERR_FS_PATH_ALREADY_LOCKED)));
  rb_define_const(mCore, "SVN_ERR_FS_PATH_NOT_LOCKED", SWIG_From_int((int)(SVN_ERR_FS_PATH_NOT_LOCKED)));
  rb_define_const(mCore, "SVN_ERR_FS_BAD_LOCK_TOKEN", SWIG_From_int((int)(SVN_ERR_FS_BAD_LOCK_TOKEN)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_LOCK_TOKEN", SWIG_From_int((int)(SVN_ERR_FS_NO_LOCK_TOKEN)));
  rb_define_const(mCore, "SVN_ERR_FS_LOCK_OWNER_MISMATCH", SWIG_From_int((int)(SVN_ERR_FS_LOCK_OWNER_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_LOCK", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_LOCK)));
  rb_define_const(mCore, "SVN_ERR_FS_LOCK_EXPIRED", SWIG_From_int((int)(SVN_ERR_FS_LOCK_EXPIRED)));
  rb_define_const(mCore, "SVN_ERR_FS_OUT_OF_DATE", SWIG_From_int((int)(SVN_ERR_FS_OUT_OF_DATE)));
  rb_define_const(mCore, "SVN_ERR_FS_UNSUPPORTED_FORMAT", SWIG_From_int((int)(SVN_ERR_FS_UNSUPPORTED_FORMAT)));
  rb_define_const(mCore, "SVN_ERR_FS_REP_BEING_WRITTEN", SWIG_From_int((int)(SVN_ERR_FS_REP_BEING_WRITTEN)));
  rb_define_const(mCore, "SVN_ERR_FS_TXN_NAME_TOO_LONG", SWIG_From_int((int)(SVN_ERR_FS_TXN_NAME_TOO_LONG)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_NODE_ORIGIN", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_NODE_ORIGIN)));
  rb_define_const(mCore, "SVN_ERR_FS_UNSUPPORTED_UPGRADE", SWIG_From_int((int)(SVN_ERR_FS_UNSUPPORTED_UPGRADE)));
  rb_define_const(mCore, "SVN_ERR_FS_NO_SUCH_CHECKSUM_REP", SWIG_From_int((int)(SVN_ERR_FS_NO_SUCH_CHECKSUM_REP)));
  rb_define_const(mCore, "SVN_ERR_FS_PROP_BASEVALUE_MISMATCH", SWIG_From_int((int)(SVN_ERR_FS_PROP_BASEVALUE_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_FS_INCORRECT_EDITOR_COMPLETION", SWIG_From_int((int)(SVN_ERR_FS_INCORRECT_EDITOR_COMPLETION)));
  rb_define_const(mCore, "SVN_ERR_FS_PACKED_REVPROP_READ_FAILURE", SWIG_From_int((int)(SVN_ERR_FS_PACKED_REVPROP_READ_FAILURE)));
  rb_define_const(mCore, "SVN_ERR_FS_REVPROP_CACHE_INIT_FAILURE", SWIG_From_int((int)(SVN_ERR_FS_REVPROP_CACHE_INIT_FAILURE)));
  rb_define_const(mCore, "SVN_ERR_FS_MALFORMED_TXN_ID", SWIG_From_int((int)(SVN_ERR_FS_MALFORMED_TXN_ID)));
  rb_define_const(mCore, "SVN_ERR_FS_INDEX_CORRUPTION", SWIG_From_int((int)(SVN_ERR_FS_INDEX_CORRUPTION)));
  rb_define_const(mCore, "SVN_ERR_FS_INDEX_REVISION", SWIG_From_int((int)(SVN_ERR_FS_INDEX_REVISION)));
  rb_define_const(mCore, "SVN_ERR_FS_INDEX_OVERFLOW", SWIG_From_int((int)(SVN_ERR_FS_INDEX_OVERFLOW)));
  rb_define_const(mCore, "SVN_ERR_FS_CONTAINER_INDEX", SWIG_From_int((int)(SVN_ERR_FS_CONTAINER_INDEX)));
  rb_define_const(mCore, "SVN_ERR_FS_INDEX_INCONSISTENT", SWIG_From_int((int)(SVN_ERR_FS_INDEX_INCONSISTENT)));
  rb_define_const(mCore, "SVN_ERR_FS_LOCK_OPERATION_FAILED", SWIG_From_int((int)(SVN_ERR_FS_LOCK_OPERATION_FAILED)));
  rb_define_const(mCore, "SVN_ERR_FS_UNSUPPORTED_TYPE", SWIG_From_int((int)(SVN_ERR_FS_UNSUPPORTED_TYPE)));
  rb_define_const(mCore, "SVN_ERR_FS_CONTAINER_SIZE", SWIG_From_int((int)(SVN_ERR_FS_CONTAINER_SIZE)));
  rb_define_const(mCore, "SVN_ERR_FS_MALFORMED_NODEREV_ID", SWIG_From_int((int)(SVN_ERR_FS_MALFORMED_NODEREV_ID)));
  rb_define_const(mCore, "SVN_ERR_FS_INVALID_GENERATION", SWIG_From_int((int)(SVN_ERR_FS_INVALID_GENERATION)));
  rb_define_const(mCore, "SVN_ERR_FS_CORRUPT_REVPROP_MANIFEST", SWIG_From_int((int)(SVN_ERR_FS_CORRUPT_REVPROP_MANIFEST)));
  rb_define_const(mCore, "SVN_ERR_FS_CORRUPT_PROPLIST", SWIG_From_int((int)(SVN_ERR_FS_CORRUPT_PROPLIST)));
  rb_define_const(mCore, "SVN_ERR_FS_AMBIGUOUS_CHECKSUM_REP", SWIG_From_int((int)(SVN_ERR_FS_AMBIGUOUS_CHECKSUM_REP)));
  rb_define_const(mCore, "SVN_ERR_REPOS_LOCKED", SWIG_From_int((int)(SVN_ERR_REPOS_LOCKED)));
  rb_define_const(mCore, "SVN_ERR_REPOS_HOOK_FAILURE", SWIG_From_int((int)(SVN_ERR_REPOS_HOOK_FAILURE)));
  rb_define_const(mCore, "SVN_ERR_REPOS_BAD_ARGS", SWIG_From_int((int)(SVN_ERR_REPOS_BAD_ARGS)));
  rb_define_const(mCore, "SVN_ERR_REPOS_NO_DATA_FOR_REPORT", SWIG_From_int((int)(SVN_ERR_REPOS_NO_DATA_FOR_REPORT)));
  rb_define_const(mCore, "SVN_ERR_REPOS_BAD_REVISION_REPORT", SWIG_From_int((int)(SVN_ERR_REPOS_BAD_REVISION_REPORT)));
  rb_define_const(mCore, "SVN_ERR_REPOS_UNSUPPORTED_VERSION", SWIG_From_int((int)(SVN_ERR_REPOS_UNSUPPORTED_VERSION)));
  rb_define_const(mCore, "SVN_ERR_REPOS_DISABLED_FEATURE", SWIG_From_int((int)(SVN_ERR_REPOS_DISABLED_FEATURE)));
  rb_define_const(mCore, "SVN_ERR_REPOS_POST_COMMIT_HOOK_FAILED", SWIG_From_int((int)(SVN_ERR_REPOS_POST_COMMIT_HOOK_FAILED)));
  rb_define_const(mCore, "SVN_ERR_REPOS_POST_LOCK_HOOK_FAILED", SWIG_From_int((int)(SVN_ERR_REPOS_POST_LOCK_HOOK_FAILED)));
  rb_define_const(mCore, "SVN_ERR_REPOS_POST_UNLOCK_HOOK_FAILED", SWIG_From_int((int)(SVN_ERR_REPOS_POST_UNLOCK_HOOK_FAILED)));
  rb_define_const(mCore, "SVN_ERR_REPOS_UNSUPPORTED_UPGRADE", SWIG_From_int((int)(SVN_ERR_REPOS_UNSUPPORTED_UPGRADE)));
  rb_define_const(mCore, "SVN_ERR_RA_ILLEGAL_URL", SWIG_From_int((int)(SVN_ERR_RA_ILLEGAL_URL)));
  rb_define_const(mCore, "SVN_ERR_RA_NOT_AUTHORIZED", SWIG_From_int((int)(SVN_ERR_RA_NOT_AUTHORIZED)));
  rb_define_const(mCore, "SVN_ERR_RA_UNKNOWN_AUTH", SWIG_From_int((int)(SVN_ERR_RA_UNKNOWN_AUTH)));
  rb_define_const(mCore, "SVN_ERR_RA_NOT_IMPLEMENTED", SWIG_From_int((int)(SVN_ERR_RA_NOT_IMPLEMENTED)));
  rb_define_const(mCore, "SVN_ERR_RA_OUT_OF_DATE", SWIG_From_int((int)(SVN_ERR_RA_OUT_OF_DATE)));
  rb_define_const(mCore, "SVN_ERR_RA_NO_REPOS_UUID", SWIG_From_int((int)(SVN_ERR_RA_NO_REPOS_UUID)));
  rb_define_const(mCore, "SVN_ERR_RA_UNSUPPORTED_ABI_VERSION", SWIG_From_int((int)(SVN_ERR_RA_UNSUPPORTED_ABI_VERSION)));
  rb_define_const(mCore, "SVN_ERR_RA_NOT_LOCKED", SWIG_From_int((int)(SVN_ERR_RA_NOT_LOCKED)));
  rb_define_const(mCore, "SVN_ERR_RA_PARTIAL_REPLAY_NOT_SUPPORTED", SWIG_From_int((int)(SVN_ERR_RA_PARTIAL_REPLAY_NOT_SUPPORTED)));
  rb_define_const(mCore, "SVN_ERR_RA_UUID_MISMATCH", SWIG_From_int((int)(SVN_ERR_RA_UUID_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_RA_REPOS_ROOT_URL_MISMATCH", SWIG_From_int((int)(SVN_ERR_RA_REPOS_ROOT_URL_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_RA_SESSION_URL_MISMATCH", SWIG_From_int((int)(SVN_ERR_RA_SESSION_URL_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_RA_CANNOT_CREATE_TUNNEL", SWIG_From_int((int)(SVN_ERR_RA_CANNOT_CREATE_TUNNEL)));
  rb_define_const(mCore, "SVN_ERR_RA_CANNOT_CREATE_SESSION", SWIG_From_int((int)(SVN_ERR_RA_CANNOT_CREATE_SESSION)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_SOCK_INIT", SWIG_From_int((int)(SVN_ERR_RA_DAV_SOCK_INIT)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_CREATING_REQUEST", SWIG_From_int((int)(SVN_ERR_RA_DAV_CREATING_REQUEST)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_REQUEST_FAILED", SWIG_From_int((int)(SVN_ERR_RA_DAV_REQUEST_FAILED)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_OPTIONS_REQ_FAILED", SWIG_From_int((int)(SVN_ERR_RA_DAV_OPTIONS_REQ_FAILED)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_PROPS_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_RA_DAV_PROPS_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_ALREADY_EXISTS", SWIG_From_int((int)(SVN_ERR_RA_DAV_ALREADY_EXISTS)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_INVALID_CONFIG_VALUE", SWIG_From_int((int)(SVN_ERR_RA_DAV_INVALID_CONFIG_VALUE)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_PATH_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_RA_DAV_PATH_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_PROPPATCH_FAILED", SWIG_From_int((int)(SVN_ERR_RA_DAV_PROPPATCH_FAILED)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_MALFORMED_DATA", SWIG_From_int((int)(SVN_ERR_RA_DAV_MALFORMED_DATA)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_RESPONSE_HEADER_BADNESS", SWIG_From_int((int)(SVN_ERR_RA_DAV_RESPONSE_HEADER_BADNESS)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_RELOCATED", SWIG_From_int((int)(SVN_ERR_RA_DAV_RELOCATED)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_CONN_TIMEOUT", SWIG_From_int((int)(SVN_ERR_RA_DAV_CONN_TIMEOUT)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_FORBIDDEN", SWIG_From_int((int)(SVN_ERR_RA_DAV_FORBIDDEN)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_PRECONDITION_FAILED", SWIG_From_int((int)(SVN_ERR_RA_DAV_PRECONDITION_FAILED)));
  rb_define_const(mCore, "SVN_ERR_RA_DAV_METHOD_NOT_ALLOWED", SWIG_From_int((int)(SVN_ERR_RA_DAV_METHOD_NOT_ALLOWED)));
  rb_define_const(mCore, "SVN_ERR_RA_LOCAL_REPOS_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_RA_LOCAL_REPOS_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_RA_LOCAL_REPOS_OPEN_FAILED", SWIG_From_int((int)(SVN_ERR_RA_LOCAL_REPOS_OPEN_FAILED)));
  rb_define_const(mCore, "SVN_ERR_SVNDIFF_INVALID_HEADER", SWIG_From_int((int)(SVN_ERR_SVNDIFF_INVALID_HEADER)));
  rb_define_const(mCore, "SVN_ERR_SVNDIFF_CORRUPT_WINDOW", SWIG_From_int((int)(SVN_ERR_SVNDIFF_CORRUPT_WINDOW)));
  rb_define_const(mCore, "SVN_ERR_SVNDIFF_BACKWARD_VIEW", SWIG_From_int((int)(SVN_ERR_SVNDIFF_BACKWARD_VIEW)));
  rb_define_const(mCore, "SVN_ERR_SVNDIFF_INVALID_OPS", SWIG_From_int((int)(SVN_ERR_SVNDIFF_INVALID_OPS)));
  rb_define_const(mCore, "SVN_ERR_SVNDIFF_UNEXPECTED_END", SWIG_From_int((int)(SVN_ERR_SVNDIFF_UNEXPECTED_END)));
  rb_define_const(mCore, "SVN_ERR_SVNDIFF_INVALID_COMPRESSED_DATA", SWIG_From_int((int)(SVN_ERR_SVNDIFF_INVALID_COMPRESSED_DATA)));
  rb_define_const(mCore, "SVN_ERR_APMOD_MISSING_PATH_TO_FS", SWIG_From_int((int)(SVN_ERR_APMOD_MISSING_PATH_TO_FS)));
  rb_define_const(mCore, "SVN_ERR_APMOD_MALFORMED_URI", SWIG_From_int((int)(SVN_ERR_APMOD_MALFORMED_URI)));
  rb_define_const(mCore, "SVN_ERR_APMOD_ACTIVITY_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_APMOD_ACTIVITY_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_APMOD_BAD_BASELINE", SWIG_From_int((int)(SVN_ERR_APMOD_BAD_BASELINE)));
  rb_define_const(mCore, "SVN_ERR_APMOD_CONNECTION_ABORTED", SWIG_From_int((int)(SVN_ERR_APMOD_CONNECTION_ABORTED)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_VERSIONED_PATH_REQUIRED", SWIG_From_int((int)(SVN_ERR_CLIENT_VERSIONED_PATH_REQUIRED)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_RA_ACCESS_REQUIRED", SWIG_From_int((int)(SVN_ERR_CLIENT_RA_ACCESS_REQUIRED)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_BAD_REVISION", SWIG_From_int((int)(SVN_ERR_CLIENT_BAD_REVISION)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_DUPLICATE_COMMIT_URL", SWIG_From_int((int)(SVN_ERR_CLIENT_DUPLICATE_COMMIT_URL)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_IS_BINARY_FILE", SWIG_From_int((int)(SVN_ERR_CLIENT_IS_BINARY_FILE)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_INVALID_EXTERNALS_DESCRIPTION", SWIG_From_int((int)(SVN_ERR_CLIENT_INVALID_EXTERNALS_DESCRIPTION)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_MODIFIED", SWIG_From_int((int)(SVN_ERR_CLIENT_MODIFIED)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_IS_DIRECTORY", SWIG_From_int((int)(SVN_ERR_CLIENT_IS_DIRECTORY)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_REVISION_RANGE", SWIG_From_int((int)(SVN_ERR_CLIENT_REVISION_RANGE)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_INVALID_RELOCATION", SWIG_From_int((int)(SVN_ERR_CLIENT_INVALID_RELOCATION)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_REVISION_AUTHOR_CONTAINS_NEWLINE", SWIG_From_int((int)(SVN_ERR_CLIENT_REVISION_AUTHOR_CONTAINS_NEWLINE)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_PROPERTY_NAME", SWIG_From_int((int)(SVN_ERR_CLIENT_PROPERTY_NAME)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_UNRELATED_RESOURCES", SWIG_From_int((int)(SVN_ERR_CLIENT_UNRELATED_RESOURCES)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_MISSING_LOCK_TOKEN", SWIG_From_int((int)(SVN_ERR_CLIENT_MISSING_LOCK_TOKEN)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_MULTIPLE_SOURCES_DISALLOWED", SWIG_From_int((int)(SVN_ERR_CLIENT_MULTIPLE_SOURCES_DISALLOWED)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_NO_VERSIONED_PARENT", SWIG_From_int((int)(SVN_ERR_CLIENT_NO_VERSIONED_PARENT)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_NOT_READY_TO_MERGE", SWIG_From_int((int)(SVN_ERR_CLIENT_NOT_READY_TO_MERGE)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_FILE_EXTERNAL_OVERWRITE_VERSIONED", SWIG_From_int((int)(SVN_ERR_CLIENT_FILE_EXTERNAL_OVERWRITE_VERSIONED)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_PATCH_BAD_STRIP_COUNT", SWIG_From_int((int)(SVN_ERR_CLIENT_PATCH_BAD_STRIP_COUNT)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_CYCLE_DETECTED", SWIG_From_int((int)(SVN_ERR_CLIENT_CYCLE_DETECTED)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_MERGE_UPDATE_REQUIRED", SWIG_From_int((int)(SVN_ERR_CLIENT_MERGE_UPDATE_REQUIRED)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_INVALID_MERGEINFO_NO_MERGETRACKING", SWIG_From_int((int)(SVN_ERR_CLIENT_INVALID_MERGEINFO_NO_MERGETRACKING)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_NO_LOCK_TOKEN", SWIG_From_int((int)(SVN_ERR_CLIENT_NO_LOCK_TOKEN)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_FORBIDDEN_BY_SERVER", SWIG_From_int((int)(SVN_ERR_CLIENT_FORBIDDEN_BY_SERVER)));
  rb_define_const(mCore, "SVN_ERR_CLIENT_CONFLICT_OPTION_NOT_APPLICABLE", SWIG_From_int((int)(SVN_ERR_CLIENT_CONFLICT_OPTION_NOT_APPLICABLE)));
  rb_define_const(mCore, "SVN_ERR_BASE", SWIG_From_int((int)(SVN_ERR_BASE)));
  rb_define_const(mCore, "SVN_ERR_PLUGIN_LOAD_FAILURE", SWIG_From_int((int)(SVN_ERR_PLUGIN_LOAD_FAILURE)));
  rb_define_const(mCore, "SVN_ERR_MALFORMED_FILE", SWIG_From_int((int)(SVN_ERR_MALFORMED_FILE)));
  rb_define_const(mCore, "SVN_ERR_INCOMPLETE_DATA", SWIG_From_int((int)(SVN_ERR_INCOMPLETE_DATA)));
  rb_define_const(mCore, "SVN_ERR_INCORRECT_PARAMS", SWIG_From_int((int)(SVN_ERR_INCORRECT_PARAMS)));
  rb_define_const(mCore, "SVN_ERR_UNVERSIONED_RESOURCE", SWIG_From_int((int)(SVN_ERR_UNVERSIONED_RESOURCE)));
  rb_define_const(mCore, "SVN_ERR_TEST_FAILED", SWIG_From_int((int)(SVN_ERR_TEST_FAILED)));
  rb_define_const(mCore, "SVN_ERR_UNSUPPORTED_FEATURE", SWIG_From_int((int)(SVN_ERR_UNSUPPORTED_FEATURE)));
  rb_define_const(mCore, "SVN_ERR_BAD_PROP_KIND", SWIG_From_int((int)(SVN_ERR_BAD_PROP_KIND)));
  rb_define_const(mCore, "SVN_ERR_ILLEGAL_TARGET", SWIG_From_int((int)(SVN_ERR_ILLEGAL_TARGET)));
  rb_define_const(mCore, "SVN_ERR_DELTA_MD5_CHECKSUM_ABSENT", SWIG_From_int((int)(SVN_ERR_DELTA_MD5_CHECKSUM_ABSENT)));
  rb_define_const(mCore, "SVN_ERR_DIR_NOT_EMPTY", SWIG_From_int((int)(SVN_ERR_DIR_NOT_EMPTY)));
  rb_define_const(mCore, "SVN_ERR_EXTERNAL_PROGRAM", SWIG_From_int((int)(SVN_ERR_EXTERNAL_PROGRAM)));
  rb_define_const(mCore, "SVN_ERR_SWIG_PY_EXCEPTION_SET", SWIG_From_int((int)(SVN_ERR_SWIG_PY_EXCEPTION_SET)));
  rb_define_const(mCore, "SVN_ERR_CHECKSUM_MISMATCH", SWIG_From_int((int)(SVN_ERR_CHECKSUM_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_CANCELLED", SWIG_From_int((int)(SVN_ERR_CANCELLED)));
  rb_define_const(mCore, "SVN_ERR_INVALID_DIFF_OPTION", SWIG_From_int((int)(SVN_ERR_INVALID_DIFF_OPTION)));
  rb_define_const(mCore, "SVN_ERR_PROPERTY_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_PROPERTY_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_NO_AUTH_FILE_PATH", SWIG_From_int((int)(SVN_ERR_NO_AUTH_FILE_PATH)));
  rb_define_const(mCore, "SVN_ERR_VERSION_MISMATCH", SWIG_From_int((int)(SVN_ERR_VERSION_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_MERGEINFO_PARSE_ERROR", SWIG_From_int((int)(SVN_ERR_MERGEINFO_PARSE_ERROR)));
  rb_define_const(mCore, "SVN_ERR_CEASE_INVOCATION", SWIG_From_int((int)(SVN_ERR_CEASE_INVOCATION)));
  rb_define_const(mCore, "SVN_ERR_REVNUM_PARSE_FAILURE", SWIG_From_int((int)(SVN_ERR_REVNUM_PARSE_FAILURE)));
  rb_define_const(mCore, "SVN_ERR_ITER_BREAK", SWIG_From_int((int)(SVN_ERR_ITER_BREAK)));
  rb_define_const(mCore, "SVN_ERR_UNKNOWN_CHANGELIST", SWIG_From_int((int)(SVN_ERR_UNKNOWN_CHANGELIST)));
  rb_define_const(mCore, "SVN_ERR_RESERVED_FILENAME_SPECIFIED", SWIG_From_int((int)(SVN_ERR_RESERVED_FILENAME_SPECIFIED)));
  rb_define_const(mCore, "SVN_ERR_UNKNOWN_CAPABILITY", SWIG_From_int((int)(SVN_ERR_UNKNOWN_CAPABILITY)));
  rb_define_const(mCore, "SVN_ERR_TEST_SKIPPED", SWIG_From_int((int)(SVN_ERR_TEST_SKIPPED)));
  rb_define_const(mCore, "SVN_ERR_NO_APR_MEMCACHE", SWIG_From_int((int)(SVN_ERR_NO_APR_MEMCACHE)));
  rb_define_const(mCore, "SVN_ERR_ATOMIC_INIT_FAILURE", SWIG_From_int((int)(SVN_ERR_ATOMIC_INIT_FAILURE)));
  rb_define_const(mCore, "SVN_ERR_SQLITE_ERROR", SWIG_From_int((int)(SVN_ERR_SQLITE_ERROR)));
  rb_define_const(mCore, "SVN_ERR_SQLITE_READONLY", SWIG_From_int((int)(SVN_ERR_SQLITE_READONLY)));
  rb_define_const(mCore, "SVN_ERR_SQLITE_UNSUPPORTED_SCHEMA", SWIG_From_int((int)(SVN_ERR_SQLITE_UNSUPPORTED_SCHEMA)));
  rb_define_const(mCore, "SVN_ERR_SQLITE_BUSY", SWIG_From_int((int)(SVN_ERR_SQLITE_BUSY)));
  rb_define_const(mCore, "SVN_ERR_SQLITE_RESETTING_FOR_ROLLBACK", SWIG_From_int((int)(SVN_ERR_SQLITE_RESETTING_FOR_ROLLBACK)));
  rb_define_const(mCore, "SVN_ERR_SQLITE_CONSTRAINT", SWIG_From_int((int)(SVN_ERR_SQLITE_CONSTRAINT)));
  rb_define_const(mCore, "SVN_ERR_TOO_MANY_MEMCACHED_SERVERS", SWIG_From_int((int)(SVN_ERR_TOO_MANY_MEMCACHED_SERVERS)));
  rb_define_const(mCore, "SVN_ERR_MALFORMED_VERSION_STRING", SWIG_From_int((int)(SVN_ERR_MALFORMED_VERSION_STRING)));
  rb_define_const(mCore, "SVN_ERR_CORRUPTED_ATOMIC_STORAGE", SWIG_From_int((int)(SVN_ERR_CORRUPTED_ATOMIC_STORAGE)));
  rb_define_const(mCore, "SVN_ERR_UTF8PROC_ERROR", SWIG_From_int((int)(SVN_ERR_UTF8PROC_ERROR)));
  rb_define_const(mCore, "SVN_ERR_UTF8_GLOB", SWIG_From_int((int)(SVN_ERR_UTF8_GLOB)));
  rb_define_const(mCore, "SVN_ERR_CORRUPT_PACKED_DATA", SWIG_From_int((int)(SVN_ERR_CORRUPT_PACKED_DATA)));
  rb_define_const(mCore, "SVN_ERR_COMPOSED_ERROR", SWIG_From_int((int)(SVN_ERR_COMPOSED_ERROR)));
  rb_define_const(mCore, "SVN_ERR_INVALID_INPUT", SWIG_From_int((int)(SVN_ERR_INVALID_INPUT)));
  rb_define_const(mCore, "SVN_ERR_SQLITE_ROLLBACK_FAILED", SWIG_From_int((int)(SVN_ERR_SQLITE_ROLLBACK_FAILED)));
  rb_define_const(mCore, "SVN_ERR_LZ4_COMPRESSION_FAILED", SWIG_From_int((int)(SVN_ERR_LZ4_COMPRESSION_FAILED)));
  rb_define_const(mCore, "SVN_ERR_LZ4_DECOMPRESSION_FAILED", SWIG_From_int((int)(SVN_ERR_LZ4_DECOMPRESSION_FAILED)));
  rb_define_const(mCore, "SVN_ERR_CL_ARG_PARSING_ERROR", SWIG_From_int((int)(SVN_ERR_CL_ARG_PARSING_ERROR)));
  rb_define_const(mCore, "SVN_ERR_CL_INSUFFICIENT_ARGS", SWIG_From_int((int)(SVN_ERR_CL_INSUFFICIENT_ARGS)));
  rb_define_const(mCore, "SVN_ERR_CL_MUTUALLY_EXCLUSIVE_ARGS", SWIG_From_int((int)(SVN_ERR_CL_MUTUALLY_EXCLUSIVE_ARGS)));
  rb_define_const(mCore, "SVN_ERR_CL_ADM_DIR_RESERVED", SWIG_From_int((int)(SVN_ERR_CL_ADM_DIR_RESERVED)));
  rb_define_const(mCore, "SVN_ERR_CL_LOG_MESSAGE_IS_VERSIONED_FILE", SWIG_From_int((int)(SVN_ERR_CL_LOG_MESSAGE_IS_VERSIONED_FILE)));
  rb_define_const(mCore, "SVN_ERR_CL_LOG_MESSAGE_IS_PATHNAME", SWIG_From_int((int)(SVN_ERR_CL_LOG_MESSAGE_IS_PATHNAME)));
  rb_define_const(mCore, "SVN_ERR_CL_COMMIT_IN_ADDED_DIR", SWIG_From_int((int)(SVN_ERR_CL_COMMIT_IN_ADDED_DIR)));
  rb_define_const(mCore, "SVN_ERR_CL_NO_EXTERNAL_EDITOR", SWIG_From_int((int)(SVN_ERR_CL_NO_EXTERNAL_EDITOR)));
  rb_define_const(mCore, "SVN_ERR_CL_BAD_LOG_MESSAGE", SWIG_From_int((int)(SVN_ERR_CL_BAD_LOG_MESSAGE)));
  rb_define_const(mCore, "SVN_ERR_CL_UNNECESSARY_LOG_MESSAGE", SWIG_From_int((int)(SVN_ERR_CL_UNNECESSARY_LOG_MESSAGE)));
  rb_define_const(mCore, "SVN_ERR_CL_NO_EXTERNAL_MERGE_TOOL", SWIG_From_int((int)(SVN_ERR_CL_NO_EXTERNAL_MERGE_TOOL)));
  rb_define_const(mCore, "SVN_ERR_CL_ERROR_PROCESSING_EXTERNALS", SWIG_From_int((int)(SVN_ERR_CL_ERROR_PROCESSING_EXTERNALS)));
  rb_define_const(mCore, "SVN_ERR_CL_REPOS_VERIFY_FAILED", SWIG_From_int((int)(SVN_ERR_CL_REPOS_VERIFY_FAILED)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_CMD_ERR", SWIG_From_int((int)(SVN_ERR_RA_SVN_CMD_ERR)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_UNKNOWN_CMD", SWIG_From_int((int)(SVN_ERR_RA_SVN_UNKNOWN_CMD)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_CONNECTION_CLOSED", SWIG_From_int((int)(SVN_ERR_RA_SVN_CONNECTION_CLOSED)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_IO_ERROR", SWIG_From_int((int)(SVN_ERR_RA_SVN_IO_ERROR)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_MALFORMED_DATA", SWIG_From_int((int)(SVN_ERR_RA_SVN_MALFORMED_DATA)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_REPOS_NOT_FOUND", SWIG_From_int((int)(SVN_ERR_RA_SVN_REPOS_NOT_FOUND)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_BAD_VERSION", SWIG_From_int((int)(SVN_ERR_RA_SVN_BAD_VERSION)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_NO_MECHANISMS", SWIG_From_int((int)(SVN_ERR_RA_SVN_NO_MECHANISMS)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_EDIT_ABORTED", SWIG_From_int((int)(SVN_ERR_RA_SVN_EDIT_ABORTED)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_REQUEST_SIZE", SWIG_From_int((int)(SVN_ERR_RA_SVN_REQUEST_SIZE)));
  rb_define_const(mCore, "SVN_ERR_RA_SVN_RESPONSE_SIZE", SWIG_From_int((int)(SVN_ERR_RA_SVN_RESPONSE_SIZE)));
  rb_define_const(mCore, "SVN_ERR_AUTHN_CREDS_UNAVAILABLE", SWIG_From_int((int)(SVN_ERR_AUTHN_CREDS_UNAVAILABLE)));
  rb_define_const(mCore, "SVN_ERR_AUTHN_NO_PROVIDER", SWIG_From_int((int)(SVN_ERR_AUTHN_NO_PROVIDER)));
  rb_define_const(mCore, "SVN_ERR_AUTHN_PROVIDERS_EXHAUSTED", SWIG_From_int((int)(SVN_ERR_AUTHN_PROVIDERS_EXHAUSTED)));
  rb_define_const(mCore, "SVN_ERR_AUTHN_CREDS_NOT_SAVED", SWIG_From_int((int)(SVN_ERR_AUTHN_CREDS_NOT_SAVED)));
  rb_define_const(mCore, "SVN_ERR_AUTHN_FAILED", SWIG_From_int((int)(SVN_ERR_AUTHN_FAILED)));
  rb_define_const(mCore, "SVN_ERR_AUTHZ_ROOT_UNREADABLE", SWIG_From_int((int)(SVN_ERR_AUTHZ_ROOT_UNREADABLE)));
  rb_define_const(mCore, "SVN_ERR_AUTHZ_UNREADABLE", SWIG_From_int((int)(SVN_ERR_AUTHZ_UNREADABLE)));
  rb_define_const(mCore, "SVN_ERR_AUTHZ_PARTIALLY_READABLE", SWIG_From_int((int)(SVN_ERR_AUTHZ_PARTIALLY_READABLE)));
  rb_define_const(mCore, "SVN_ERR_AUTHZ_INVALID_CONFIG", SWIG_From_int((int)(SVN_ERR_AUTHZ_INVALID_CONFIG)));
  rb_define_const(mCore, "SVN_ERR_AUTHZ_UNWRITABLE", SWIG_From_int((int)(SVN_ERR_AUTHZ_UNWRITABLE)));
  rb_define_const(mCore, "SVN_ERR_DIFF_DATASOURCE_MODIFIED", SWIG_From_int((int)(SVN_ERR_DIFF_DATASOURCE_MODIFIED)));
  rb_define_const(mCore, "SVN_ERR_DIFF_UNEXPECTED_DATA", SWIG_From_int((int)(SVN_ERR_DIFF_UNEXPECTED_DATA)));
  rb_define_const(mCore, "SVN_ERR_RA_SERF_SSPI_INITIALISATION_FAILED", SWIG_From_int((int)(SVN_ERR_RA_SERF_SSPI_INITIALISATION_FAILED)));
  rb_define_const(mCore, "SVN_ERR_RA_SERF_SSL_CERT_UNTRUSTED", SWIG_From_int((int)(SVN_ERR_RA_SERF_SSL_CERT_UNTRUSTED)));
  rb_define_const(mCore, "SVN_ERR_RA_SERF_GSSAPI_INITIALISATION_FAILED", SWIG_From_int((int)(SVN_ERR_RA_SERF_GSSAPI_INITIALISATION_FAILED)));
  rb_define_const(mCore, "SVN_ERR_RA_SERF_WRAPPED_ERROR", SWIG_From_int((int)(SVN_ERR_RA_SERF_WRAPPED_ERROR)));
  rb_define_const(mCore, "SVN_ERR_RA_SERF_STREAM_BUCKET_READ_ERROR", SWIG_From_int((int)(SVN_ERR_RA_SERF_STREAM_BUCKET_READ_ERROR)));
  rb_define_const(mCore, "SVN_ERR_ASSERTION_FAIL", SWIG_From_int((int)(SVN_ERR_ASSERTION_FAIL)));
  rb_define_const(mCore, "SVN_ERR_ASSERTION_ONLY_TRACING_LINKS", SWIG_From_int((int)(SVN_ERR_ASSERTION_ONLY_TRACING_LINKS)));
  rb_define_const(mCore, "SVN_ERR_ASN1_OUT_OF_DATA", SWIG_From_int((int)(SVN_ERR_ASN1_OUT_OF_DATA)));
  rb_define_const(mCore, "SVN_ERR_ASN1_UNEXPECTED_TAG", SWIG_From_int((int)(SVN_ERR_ASN1_UNEXPECTED_TAG)));
  rb_define_const(mCore, "SVN_ERR_ASN1_INVALID_LENGTH", SWIG_From_int((int)(SVN_ERR_ASN1_INVALID_LENGTH)));
  rb_define_const(mCore, "SVN_ERR_ASN1_LENGTH_MISMATCH", SWIG_From_int((int)(SVN_ERR_ASN1_LENGTH_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_ASN1_INVALID_DATA", SWIG_From_int((int)(SVN_ERR_ASN1_INVALID_DATA)));
  rb_define_const(mCore, "SVN_ERR_X509_FEATURE_UNAVAILABLE", SWIG_From_int((int)(SVN_ERR_X509_FEATURE_UNAVAILABLE)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_PEM", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_PEM)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_FORMAT", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_FORMAT)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_VERSION", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_VERSION)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_SERIAL", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_SERIAL)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_ALG", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_ALG)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_NAME", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_NAME)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_DATE", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_DATE)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_PUBKEY", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_PUBKEY)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_SIGNATURE", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_SIGNATURE)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_INVALID_EXTENSIONS", SWIG_From_int((int)(SVN_ERR_X509_CERT_INVALID_EXTENSIONS)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_UNKNOWN_VERSION", SWIG_From_int((int)(SVN_ERR_X509_CERT_UNKNOWN_VERSION)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_UNKNOWN_PK_ALG", SWIG_From_int((int)(SVN_ERR_X509_CERT_UNKNOWN_PK_ALG)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_SIG_MISMATCH", SWIG_From_int((int)(SVN_ERR_X509_CERT_SIG_MISMATCH)));
  rb_define_const(mCore, "SVN_ERR_X509_CERT_VERIFY_FAILED", SWIG_From_int((int)(SVN_ERR_X509_CERT_VERIFY_FAILED)));
  rb_define_const(mCore, "SVN_ERR_LAST", SWIG_From_int((int)(SVN_ERR_LAST)));
  rb_define_module_function(mCore, "svn_time_to_cstring", _wrap_svn_time_to_cstring, -1);
  rb_define_module_function(mCore, "svn_time_from_cstring", _wrap_svn_time_from_cstring, -1);
  rb_define_module_function(mCore, "svn_time_to_human_cstring", _wrap_svn_time_to_human_cstring, -1);
  rb_define_module_function(mCore, "svn_parse_date", _wrap_svn_parse_date, -1);
  rb_define_module_function(mCore, "svn_sleep_for_timestamps", _wrap_svn_sleep_for_timestamps, -1);
  rb_define_const(mCore, "SVN_UNALIGNED_ACCESS_IS_OK", SWIG_From_int((int)(0)));
  rb_define_const(mCore, "TRUE", SWIG_From_int((int)(1)));
  rb_define_const(mCore, "FALSE", SWIG_From_int((int)(0)));
  
  SwigClassSvn_error_t.klass = rb_define_class_under(mCore, "Svn_error_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_error_t, (void *) &SwigClassSvn_error_t);
  rb_define_alloc_func(SwigClassSvn_error_t.klass, _wrap_svn_error_t_allocate);
  rb_define_method(SwigClassSvn_error_t.klass, "initialize", _wrap_new_svn_error_t, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "apr_err=", _wrap_svn_error_t_apr_err_set, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "apr_err", _wrap_svn_error_t_apr_err_get, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "message", _wrap_svn_error_t_message_get, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "child=", _wrap_svn_error_t_child_set, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "child", _wrap_svn_error_t_child_get, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "pool=", _wrap_svn_error_t_pool_set, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "pool", _wrap_svn_error_t_pool_get, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "file", _wrap_svn_error_t_file_get, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "line=", _wrap_svn_error_t_line_set, -1);
  rb_define_method(SwigClassSvn_error_t.klass, "line", _wrap_svn_error_t_line_get, -1);
  SwigClassSvn_error_t.mark = 0;
  SwigClassSvn_error_t.destroy = (void (*)(void *)) free_svn_error_t;
  SwigClassSvn_error_t.trackObjects = 0;
  rb_define_const(mCore, "Svn_node_none", SWIG_From_int((int)(svn_node_none)));
  rb_define_const(mCore, "Svn_node_file", SWIG_From_int((int)(svn_node_file)));
  rb_define_const(mCore, "Svn_node_dir", SWIG_From_int((int)(svn_node_dir)));
  rb_define_const(mCore, "Svn_node_unknown", SWIG_From_int((int)(svn_node_unknown)));
  rb_define_const(mCore, "Svn_node_symlink", SWIG_From_int((int)(svn_node_symlink)));
  rb_define_module_function(mCore, "svn_node_kind_to_word", _wrap_svn_node_kind_to_word, -1);
  rb_define_module_function(mCore, "svn_node_kind_from_word", _wrap_svn_node_kind_from_word, -1);
  rb_define_const(mCore, "Svn_tristate_false", SWIG_From_int((int)(svn_tristate_false)));
  rb_define_const(mCore, "Svn_tristate_true", SWIG_From_int((int)(svn_tristate_true)));
  rb_define_const(mCore, "Svn_tristate_unknown", SWIG_From_int((int)(svn_tristate_unknown)));
  rb_define_module_function(mCore, "svn_tristate__to_word", _wrap_svn_tristate__to_word, -1);
  rb_define_module_function(mCore, "svn_tristate__from_word", _wrap_svn_tristate__from_word, -1);
  rb_define_module_function(mCore, "svn_revnum_parse", _wrap_svn_revnum_parse, -1);
  rb_define_const(mCore, "SVN_REVNUM_T_FMT", SWIG_FromCharPtr("ld"));
  rb_define_const(mCore, "Svn_nonrecursive", SWIG_From_int((int)(svn_nonrecursive)));
  rb_define_const(mCore, "Svn_recursive", SWIG_From_int((int)(svn_recursive)));
  rb_define_const(mCore, "Svn_depth_unknown", SWIG_From_int((int)(svn_depth_unknown)));
  rb_define_const(mCore, "Svn_depth_exclude", SWIG_From_int((int)(svn_depth_exclude)));
  rb_define_const(mCore, "Svn_depth_empty", SWIG_From_int((int)(svn_depth_empty)));
  rb_define_const(mCore, "Svn_depth_files", SWIG_From_int((int)(svn_depth_files)));
  rb_define_const(mCore, "Svn_depth_immediates", SWIG_From_int((int)(svn_depth_immediates)));
  rb_define_const(mCore, "Svn_depth_infinity", SWIG_From_int((int)(svn_depth_infinity)));
  rb_define_module_function(mCore, "svn_depth_to_word", _wrap_svn_depth_to_word, -1);
  rb_define_module_function(mCore, "svn_depth_from_word", _wrap_svn_depth_from_word, -1);
  rb_define_const(mCore, "SVN_DIRENT_KIND", SWIG_From_int((int)(0x00001)));
  rb_define_const(mCore, "SVN_DIRENT_SIZE", SWIG_From_int((int)(0x00002)));
  rb_define_const(mCore, "SVN_DIRENT_HAS_PROPS", SWIG_From_int((int)(0x00004)));
  rb_define_const(mCore, "SVN_DIRENT_CREATED_REV", SWIG_From_int((int)(0x00008)));
  rb_define_const(mCore, "SVN_DIRENT_TIME", SWIG_From_int((int)(0x00010)));
  rb_define_const(mCore, "SVN_DIRENT_LAST_AUTHOR", SWIG_From_int((int)(0x00020)));
  
  SwigClassSvn_dirent_t.klass = rb_define_class_under(mCore, "Svn_dirent_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_dirent_t, (void *) &SwigClassSvn_dirent_t);
  rb_define_alloc_func(SwigClassSvn_dirent_t.klass, _wrap_svn_dirent_t_allocate);
  rb_define_method(SwigClassSvn_dirent_t.klass, "initialize", _wrap_new_svn_dirent_t, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "kind=", _wrap_svn_dirent_t_kind_set, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "kind", _wrap_svn_dirent_t_kind_get, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "size=", _wrap_svn_dirent_t_size_set, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "size", _wrap_svn_dirent_t_size_get, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "has_props=", _wrap_svn_dirent_t_has_props_set, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "has_props", _wrap_svn_dirent_t_has_props_get, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "created_rev=", _wrap_svn_dirent_t_created_rev_set, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "created_rev", _wrap_svn_dirent_t_created_rev_get, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "time=", _wrap_svn_dirent_t_time_set, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "time", _wrap_svn_dirent_t_time_get, -1);
  rb_define_method(SwigClassSvn_dirent_t.klass, "last_author", _wrap_svn_dirent_t_last_author_get, -1);
  SwigClassSvn_dirent_t.mark = 0;
  SwigClassSvn_dirent_t.destroy = (void (*)(void *)) free_svn_dirent_t;
  SwigClassSvn_dirent_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_dirent_dup", _wrap_svn_dirent_dup, -1);
  rb_define_module_function(mCore, "svn_dirent_create", _wrap_svn_dirent_create, -1);
  rb_define_const(mCore, "SVN_KEYWORD_MAX_LEN", SWIG_From_int((int)(255)));
  rb_define_const(mCore, "SVN_KEYWORD_REVISION_LONG", SWIG_FromCharPtr("LastChangedRevision"));
  rb_define_const(mCore, "SVN_KEYWORD_REVISION_SHORT", SWIG_FromCharPtr("Rev"));
  rb_define_const(mCore, "SVN_KEYWORD_REVISION_MEDIUM", SWIG_FromCharPtr("Revision"));
  rb_define_const(mCore, "SVN_KEYWORD_DATE_LONG", SWIG_FromCharPtr("LastChangedDate"));
  rb_define_const(mCore, "SVN_KEYWORD_DATE_SHORT", SWIG_FromCharPtr("Date"));
  rb_define_const(mCore, "SVN_KEYWORD_AUTHOR_LONG", SWIG_FromCharPtr("LastChangedBy"));
  rb_define_const(mCore, "SVN_KEYWORD_AUTHOR_SHORT", SWIG_FromCharPtr("Author"));
  rb_define_const(mCore, "SVN_KEYWORD_URL_LONG", SWIG_FromCharPtr("HeadURL"));
  rb_define_const(mCore, "SVN_KEYWORD_URL_SHORT", SWIG_FromCharPtr("URL"));
  rb_define_const(mCore, "SVN_KEYWORD_ID", SWIG_FromCharPtr("Id"));
  rb_define_const(mCore, "SVN_KEYWORD_HEADER", SWIG_FromCharPtr("Header"));
  
  SwigClassSvn_commit_info_t.klass = rb_define_class_under(mCore, "Svn_commit_info_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_commit_info_t, (void *) &SwigClassSvn_commit_info_t);
  rb_define_alloc_func(SwigClassSvn_commit_info_t.klass, _wrap_svn_commit_info_t_allocate);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "initialize", _wrap_new_svn_commit_info_t, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "revision=", _wrap_svn_commit_info_t_revision_set, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "revision", _wrap_svn_commit_info_t_revision_get, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "date=", _wrap_svn_commit_info_t_date_set, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "date", _wrap_svn_commit_info_t_date_get, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "author=", _wrap_svn_commit_info_t_author_set, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "author", _wrap_svn_commit_info_t_author_get, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "post_commit_err=", _wrap_svn_commit_info_t_post_commit_err_set, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "post_commit_err", _wrap_svn_commit_info_t_post_commit_err_get, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "repos_root=", _wrap_svn_commit_info_t_repos_root_set, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "repos_root", _wrap_svn_commit_info_t_repos_root_get, -1);
  rb_define_method(SwigClassSvn_commit_info_t.klass, "dup", _wrap_svn_commit_info_t_dup, -1);
  SwigClassSvn_commit_info_t.mark = 0;
  SwigClassSvn_commit_info_t.destroy = (void (*)(void *)) free_svn_commit_info_t;
  SwigClassSvn_commit_info_t.trackObjects = 0;
  
  SwigClassSvn_log_changed_path2_t.klass = rb_define_class_under(mCore, "Svn_log_changed_path2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_log_changed_path2_t, (void *) &SwigClassSvn_log_changed_path2_t);
  rb_define_alloc_func(SwigClassSvn_log_changed_path2_t.klass, _wrap_svn_log_changed_path2_t_allocate);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "initialize", _wrap_new_svn_log_changed_path2_t, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "action=", _wrap_svn_log_changed_path2_t_action_set, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "action", _wrap_svn_log_changed_path2_t_action_get, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "copyfrom_path=", _wrap_svn_log_changed_path2_t_copyfrom_path_set, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "copyfrom_path", _wrap_svn_log_changed_path2_t_copyfrom_path_get, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "copyfrom_rev=", _wrap_svn_log_changed_path2_t_copyfrom_rev_set, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "copyfrom_rev", _wrap_svn_log_changed_path2_t_copyfrom_rev_get, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "node_kind=", _wrap_svn_log_changed_path2_t_node_kind_set, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "node_kind", _wrap_svn_log_changed_path2_t_node_kind_get, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "text_modified=", _wrap_svn_log_changed_path2_t_text_modified_set, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "text_modified", _wrap_svn_log_changed_path2_t_text_modified_get, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "props_modified=", _wrap_svn_log_changed_path2_t_props_modified_set, -1);
  rb_define_method(SwigClassSvn_log_changed_path2_t.klass, "props_modified", _wrap_svn_log_changed_path2_t_props_modified_get, -1);
  SwigClassSvn_log_changed_path2_t.mark = 0;
  SwigClassSvn_log_changed_path2_t.destroy = (void (*)(void *)) free_svn_log_changed_path2_t;
  SwigClassSvn_log_changed_path2_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_log_changed_path2_create", _wrap_svn_log_changed_path2_create, -1);
  rb_define_module_function(mCore, "svn_log_changed_path2_dup", _wrap_svn_log_changed_path2_dup, -1);
  
  SwigClassSvn_log_changed_path_t.klass = rb_define_class_under(mCore, "Svn_log_changed_path_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_log_changed_path_t, (void *) &SwigClassSvn_log_changed_path_t);
  rb_define_alloc_func(SwigClassSvn_log_changed_path_t.klass, _wrap_svn_log_changed_path_t_allocate);
  rb_define_method(SwigClassSvn_log_changed_path_t.klass, "initialize", _wrap_new_svn_log_changed_path_t, -1);
  rb_define_method(SwigClassSvn_log_changed_path_t.klass, "action=", _wrap_svn_log_changed_path_t_action_set, -1);
  rb_define_method(SwigClassSvn_log_changed_path_t.klass, "action", _wrap_svn_log_changed_path_t_action_get, -1);
  rb_define_method(SwigClassSvn_log_changed_path_t.klass, "copyfrom_path", _wrap_svn_log_changed_path_t_copyfrom_path_get, -1);
  rb_define_method(SwigClassSvn_log_changed_path_t.klass, "copyfrom_rev=", _wrap_svn_log_changed_path_t_copyfrom_rev_set, -1);
  rb_define_method(SwigClassSvn_log_changed_path_t.klass, "copyfrom_rev", _wrap_svn_log_changed_path_t_copyfrom_rev_get, -1);
  SwigClassSvn_log_changed_path_t.mark = 0;
  SwigClassSvn_log_changed_path_t.destroy = (void (*)(void *)) free_svn_log_changed_path_t;
  SwigClassSvn_log_changed_path_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_log_changed_path_dup", _wrap_svn_log_changed_path_dup, -1);
  
  SwigClassSvn_log_entry_t.klass = rb_define_class_under(mCore, "Svn_log_entry_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_log_entry_t, (void *) &SwigClassSvn_log_entry_t);
  rb_define_alloc_func(SwigClassSvn_log_entry_t.klass, _wrap_svn_log_entry_t_allocate);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "initialize", _wrap_new_svn_log_entry_t, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "changed_paths=", _wrap_svn_log_entry_t_changed_paths_set, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "changed_paths", _wrap_svn_log_entry_t_changed_paths_get, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "revision=", _wrap_svn_log_entry_t_revision_set, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "revision", _wrap_svn_log_entry_t_revision_get, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "revprops=", _wrap_svn_log_entry_t_revprops_set, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "revprops", _wrap_svn_log_entry_t_revprops_get, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "has_children=", _wrap_svn_log_entry_t_has_children_set, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "has_children", _wrap_svn_log_entry_t_has_children_get, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "changed_paths2=", _wrap_svn_log_entry_t_changed_paths2_set, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "changed_paths2", _wrap_svn_log_entry_t_changed_paths2_get, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "non_inheritable=", _wrap_svn_log_entry_t_non_inheritable_set, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "non_inheritable", _wrap_svn_log_entry_t_non_inheritable_get, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "subtractive_merge=", _wrap_svn_log_entry_t_subtractive_merge_set, -1);
  rb_define_method(SwigClassSvn_log_entry_t.klass, "subtractive_merge", _wrap_svn_log_entry_t_subtractive_merge_get, -1);
  SwigClassSvn_log_entry_t.mark = 0;
  SwigClassSvn_log_entry_t.destroy = (void (*)(void *)) free_svn_log_entry_t;
  SwigClassSvn_log_entry_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_log_entry_create", _wrap_svn_log_entry_create, -1);
  rb_define_module_function(mCore, "svn_log_entry_dup", _wrap_svn_log_entry_dup, -1);
  rb_define_const(mCore, "SVN_STREAM_CHUNK_SIZE", SWIG_From_int((int)(102400)));
  rb_define_const(mCore, "SVN__STREAM_CHUNK_SIZE", SWIG_From_int((int)(16384)));
  rb_define_module_function(mCore, "svn_mime_type_validate", _wrap_svn_mime_type_validate, -1);
  rb_define_module_function(mCore, "svn_mime_type_is_binary", _wrap_svn_mime_type_is_binary, -1);
  
  SwigClassSvn_lock_t.klass = rb_define_class_under(mCore, "Svn_lock_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_lock_t, (void *) &SwigClassSvn_lock_t);
  rb_define_alloc_func(SwigClassSvn_lock_t.klass, _wrap_svn_lock_t_allocate);
  rb_define_method(SwigClassSvn_lock_t.klass, "initialize", _wrap_new_svn_lock_t, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "path=", _wrap_svn_lock_t_path_set, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "path", _wrap_svn_lock_t_path_get, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "token=", _wrap_svn_lock_t_token_set, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "token", _wrap_svn_lock_t_token_get, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "owner=", _wrap_svn_lock_t_owner_set, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "owner", _wrap_svn_lock_t_owner_get, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "comment=", _wrap_svn_lock_t_comment_set, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "comment", _wrap_svn_lock_t_comment_get, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "is_dav_comment=", _wrap_svn_lock_t_is_dav_comment_set, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "is_dav_comment", _wrap_svn_lock_t_is_dav_comment_get, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "creation_date=", _wrap_svn_lock_t_creation_date_set, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "creation_date", _wrap_svn_lock_t_creation_date_get, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "expiration_date=", _wrap_svn_lock_t_expiration_date_set, -1);
  rb_define_method(SwigClassSvn_lock_t.klass, "expiration_date", _wrap_svn_lock_t_expiration_date_get, -1);
  SwigClassSvn_lock_t.mark = 0;
  SwigClassSvn_lock_t.destroy = (void (*)(void *)) free_svn_lock_t;
  SwigClassSvn_lock_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_lock_create", _wrap_svn_lock_create, -1);
  rb_define_module_function(mCore, "svn_lock_dup", _wrap_svn_lock_dup, -1);
  rb_define_module_function(mCore, "svn_uuid_generate", _wrap_svn_uuid_generate, -1);
  
  SwigClassSvn_merge_range_t.klass = rb_define_class_under(mCore, "Svn_merge_range_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_merge_range_t, (void *) &SwigClassSvn_merge_range_t);
  rb_define_alloc_func(SwigClassSvn_merge_range_t.klass, _wrap_svn_merge_range_t_allocate);
  rb_define_method(SwigClassSvn_merge_range_t.klass, "initialize", _wrap_new_svn_merge_range_t, -1);
  rb_define_method(SwigClassSvn_merge_range_t.klass, "start=", _wrap_svn_merge_range_t_start_set, -1);
  rb_define_method(SwigClassSvn_merge_range_t.klass, "start", _wrap_svn_merge_range_t_start_get, -1);
  rb_define_method(SwigClassSvn_merge_range_t.klass, "end=", _wrap_svn_merge_range_t_end_set, -1);
  rb_define_method(SwigClassSvn_merge_range_t.klass, "end", _wrap_svn_merge_range_t_end_get, -1);
  rb_define_method(SwigClassSvn_merge_range_t.klass, "inheritable=", _wrap_svn_merge_range_t_inheritable_set, -1);
  rb_define_method(SwigClassSvn_merge_range_t.klass, "inheritable", _wrap_svn_merge_range_t_inheritable_get, -1);
  rb_define_method(SwigClassSvn_merge_range_t.klass, "dup", _wrap_svn_merge_range_t_dup, -1);
  SwigClassSvn_merge_range_t.mark = 0;
  SwigClassSvn_merge_range_t.destroy = (void (*)(void *)) free_svn_merge_range_t;
  SwigClassSvn_merge_range_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_merge_range_dup", _wrap_svn_merge_range_dup, -1);
  rb_define_module_function(mCore, "svn_merge_range_contains_rev", _wrap_svn_merge_range_contains_rev, -1);
  
  SwigClassSvn_location_segment_t.klass = rb_define_class_under(mCore, "Svn_location_segment_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_location_segment_t, (void *) &SwigClassSvn_location_segment_t);
  rb_define_alloc_func(SwigClassSvn_location_segment_t.klass, _wrap_svn_location_segment_t_allocate);
  rb_define_method(SwigClassSvn_location_segment_t.klass, "initialize", _wrap_new_svn_location_segment_t, -1);
  rb_define_method(SwigClassSvn_location_segment_t.klass, "range_start=", _wrap_svn_location_segment_t_range_start_set, -1);
  rb_define_method(SwigClassSvn_location_segment_t.klass, "range_start", _wrap_svn_location_segment_t_range_start_get, -1);
  rb_define_method(SwigClassSvn_location_segment_t.klass, "range_end=", _wrap_svn_location_segment_t_range_end_set, -1);
  rb_define_method(SwigClassSvn_location_segment_t.klass, "range_end", _wrap_svn_location_segment_t_range_end_get, -1);
  rb_define_method(SwigClassSvn_location_segment_t.klass, "path=", _wrap_svn_location_segment_t_path_set, -1);
  rb_define_method(SwigClassSvn_location_segment_t.klass, "path", _wrap_svn_location_segment_t_path_get, -1);
  SwigClassSvn_location_segment_t.mark = 0;
  SwigClassSvn_location_segment_t.destroy = (void (*)(void *)) free_svn_location_segment_t;
  SwigClassSvn_location_segment_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_location_segment_dup", _wrap_svn_location_segment_dup, -1);
  rb_define_module_function(mCore, "svn_log_invoke_entry_receiver", _wrap_svn_log_invoke_entry_receiver, -1);
  rb_define_module_function(mCore, "svn_log_invoke_message_receiver", _wrap_svn_log_invoke_message_receiver, -1);
  rb_define_module_function(mCore, "svn_commit_invoke_callback2", _wrap_svn_commit_invoke_callback2, -1);
  rb_define_module_function(mCore, "svn_commit_invoke_callback", _wrap_svn_commit_invoke_callback, -1);
  rb_define_module_function(mCore, "svn_cancel_invoke_func", _wrap_svn_cancel_invoke_func, -1);
  rb_define_module_function(mCore, "svn_location_invoke_segment_receiver", _wrap_svn_location_invoke_segment_receiver, -1);
  rb_define_const(mCore, "SVN_ALLOCATOR_RECOMMENDED_MAX_FREE", SWIG_From_int((int)((4096*1024))));
  rb_define_module_function(mCore, "svn_pool_create", _wrap_svn_pool_create, -1);
  rb_define_module_function(mCore, "svn_pool_create_allocator", _wrap_svn_pool_create_allocator, -1);
  rb_define_const(mCore, "SVN_VER_MAJOR", SWIG_From_int((int)(1)));
  rb_define_const(mCore, "SVN_VER_MINOR", SWIG_From_int((int)(11)));
  rb_define_const(mCore, "SVN_VER_PATCH", SWIG_From_int((int)(1)));
  rb_define_const(mCore, "SVN_VER_MICRO", SWIG_From_int((int)(1)));
  rb_define_const(mCore, "SVN_VER_LIBRARY", SWIG_From_int((int)(1)));
  rb_define_const(mCore, "SVN_VER_TAG", SWIG_FromCharPtr(" (r1850623)"));
  rb_define_const(mCore, "SVN_VER_NUMTAG", SWIG_FromCharPtr(""));
  rb_define_const(mCore, "SVN_VER_REVISION", SWIG_From_int((int)(1850623)));
  rb_define_const(mCore, "SVN_VER_NUM", SWIG_FromCharPtr("1.11.1"));
  rb_define_const(mCore, "SVN_VER_NUMBER", SWIG_FromCharPtr("1.11.1"));
  rb_define_const(mCore, "SVN_VERSION", SWIG_FromCharPtr("1.11.1 (r1850623)"));
  
  SwigClassSvn_version_t.klass = rb_define_class_under(mCore, "Svn_version_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_version_t, (void *) &SwigClassSvn_version_t);
  rb_define_alloc_func(SwigClassSvn_version_t.klass, _wrap_svn_version_t_allocate);
  rb_define_method(SwigClassSvn_version_t.klass, "initialize", _wrap_new_svn_version_t, -1);
  rb_define_method(SwigClassSvn_version_t.klass, "major=", _wrap_svn_version_t_major_set, -1);
  rb_define_method(SwigClassSvn_version_t.klass, "major", _wrap_svn_version_t_major_get, -1);
  rb_define_method(SwigClassSvn_version_t.klass, "minor=", _wrap_svn_version_t_minor_set, -1);
  rb_define_method(SwigClassSvn_version_t.klass, "minor", _wrap_svn_version_t_minor_get, -1);
  rb_define_method(SwigClassSvn_version_t.klass, "patch=", _wrap_svn_version_t_patch_set, -1);
  rb_define_method(SwigClassSvn_version_t.klass, "patch", _wrap_svn_version_t_patch_get, -1);
  rb_define_method(SwigClassSvn_version_t.klass, "tag=", _wrap_svn_version_t_tag_set, -1);
  rb_define_method(SwigClassSvn_version_t.klass, "tag", _wrap_svn_version_t_tag_get, -1);
  SwigClassSvn_version_t.mark = 0;
  SwigClassSvn_version_t.destroy = (void (*)(void *)) free_svn_version_t;
  SwigClassSvn_version_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_ver_compatible", _wrap_svn_ver_compatible, -1);
  rb_define_module_function(mCore, "svn_ver_equal", _wrap_svn_ver_equal, -1);
  
  SwigClassSvn_version_checklist_t.klass = rb_define_class_under(mCore, "Svn_version_checklist_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_version_checklist_t, (void *) &SwigClassSvn_version_checklist_t);
  rb_define_alloc_func(SwigClassSvn_version_checklist_t.klass, _wrap_svn_version_checklist_t_allocate);
  rb_define_method(SwigClassSvn_version_checklist_t.klass, "initialize", _wrap_new_svn_version_checklist_t, -1);
  rb_define_method(SwigClassSvn_version_checklist_t.klass, "label=", _wrap_svn_version_checklist_t_label_set, -1);
  rb_define_method(SwigClassSvn_version_checklist_t.klass, "label", _wrap_svn_version_checklist_t_label_get, -1);
  rb_define_method(SwigClassSvn_version_checklist_t.klass, "version_query=", _wrap_svn_version_checklist_t_version_query_set, -1);
  rb_define_method(SwigClassSvn_version_checklist_t.klass, "version_query", _wrap_svn_version_checklist_t_version_query_get, -1);
  SwigClassSvn_version_checklist_t.mark = 0;
  SwigClassSvn_version_checklist_t.destroy = (void (*)(void *)) free_svn_version_checklist_t;
  SwigClassSvn_version_checklist_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_ver_check_list2", _wrap_svn_ver_check_list2, -1);
  rb_define_module_function(mCore, "svn_ver_check_list", _wrap_svn_ver_check_list, -1);
  rb_define_module_function(mCore, "svn_subr_version", _wrap_svn_subr_version, -1);
  rb_define_module_function(mCore, "svn_version_extended", _wrap_svn_version_extended, -1);
  rb_define_module_function(mCore, "svn_version_ext_build_date", _wrap_svn_version_ext_build_date, -1);
  rb_define_module_function(mCore, "svn_version_ext_build_time", _wrap_svn_version_ext_build_time, -1);
  rb_define_module_function(mCore, "svn_version_ext_build_host", _wrap_svn_version_ext_build_host, -1);
  rb_define_module_function(mCore, "svn_version_ext_copyright", _wrap_svn_version_ext_copyright, -1);
  rb_define_module_function(mCore, "svn_version_ext_runtime_host", _wrap_svn_version_ext_runtime_host, -1);
  rb_define_module_function(mCore, "svn_version_ext_runtime_osname", _wrap_svn_version_ext_runtime_osname, -1);
  
  SwigClassSvn_version_ext_linked_lib_t.klass = rb_define_class_under(mCore, "Svn_version_ext_linked_lib_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_version_ext_linked_lib_t, (void *) &SwigClassSvn_version_ext_linked_lib_t);
  rb_define_alloc_func(SwigClassSvn_version_ext_linked_lib_t.klass, _wrap_svn_version_ext_linked_lib_t_allocate);
  rb_define_method(SwigClassSvn_version_ext_linked_lib_t.klass, "initialize", _wrap_new_svn_version_ext_linked_lib_t, -1);
  rb_define_method(SwigClassSvn_version_ext_linked_lib_t.klass, "name=", _wrap_svn_version_ext_linked_lib_t_name_set, -1);
  rb_define_method(SwigClassSvn_version_ext_linked_lib_t.klass, "name", _wrap_svn_version_ext_linked_lib_t_name_get, -1);
  rb_define_method(SwigClassSvn_version_ext_linked_lib_t.klass, "compiled_version=", _wrap_svn_version_ext_linked_lib_t_compiled_version_set, -1);
  rb_define_method(SwigClassSvn_version_ext_linked_lib_t.klass, "compiled_version", _wrap_svn_version_ext_linked_lib_t_compiled_version_get, -1);
  rb_define_method(SwigClassSvn_version_ext_linked_lib_t.klass, "runtime_version=", _wrap_svn_version_ext_linked_lib_t_runtime_version_set, -1);
  rb_define_method(SwigClassSvn_version_ext_linked_lib_t.klass, "runtime_version", _wrap_svn_version_ext_linked_lib_t_runtime_version_get, -1);
  SwigClassSvn_version_ext_linked_lib_t.mark = 0;
  SwigClassSvn_version_ext_linked_lib_t.destroy = (void (*)(void *)) free_svn_version_ext_linked_lib_t;
  SwigClassSvn_version_ext_linked_lib_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_version_ext_linked_libs", _wrap_svn_version_ext_linked_libs, -1);
  
  SwigClassSvn_version_ext_loaded_lib_t.klass = rb_define_class_under(mCore, "Svn_version_ext_loaded_lib_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_version_ext_loaded_lib_t, (void *) &SwigClassSvn_version_ext_loaded_lib_t);
  rb_define_alloc_func(SwigClassSvn_version_ext_loaded_lib_t.klass, _wrap_svn_version_ext_loaded_lib_t_allocate);
  rb_define_method(SwigClassSvn_version_ext_loaded_lib_t.klass, "initialize", _wrap_new_svn_version_ext_loaded_lib_t, -1);
  rb_define_method(SwigClassSvn_version_ext_loaded_lib_t.klass, "name=", _wrap_svn_version_ext_loaded_lib_t_name_set, -1);
  rb_define_method(SwigClassSvn_version_ext_loaded_lib_t.klass, "name", _wrap_svn_version_ext_loaded_lib_t_name_get, -1);
  rb_define_method(SwigClassSvn_version_ext_loaded_lib_t.klass, "version=", _wrap_svn_version_ext_loaded_lib_t_version_set, -1);
  rb_define_method(SwigClassSvn_version_ext_loaded_lib_t.klass, "version", _wrap_svn_version_ext_loaded_lib_t_version_get, -1);
  SwigClassSvn_version_ext_loaded_lib_t.mark = 0;
  SwigClassSvn_version_ext_loaded_lib_t.destroy = (void (*)(void *)) free_svn_version_ext_loaded_lib_t;
  SwigClassSvn_version_ext_loaded_lib_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_version_ext_loaded_libs", _wrap_svn_version_ext_loaded_libs, -1);
  rb_define_module_function(mCore, "svn_version_checklist_invoke_version_query", _wrap_svn_version_checklist_invoke_version_query, -1);
  rb_define_module_function(mCore, "svn_version_invoke_func", _wrap_svn_version_invoke_func, -1);
  rb_define_module_function(mCore, "svn_prop_dup", _wrap_svn_prop_dup, -1);
  rb_define_module_function(mCore, "svn_prop_array_dup", _wrap_svn_prop_array_dup, -1);
  
  SwigClassSvn_prop_inherited_item_t.klass = rb_define_class_under(mCore, "Svn_prop_inherited_item_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_prop_inherited_item_t, (void *) &SwigClassSvn_prop_inherited_item_t);
  rb_define_alloc_func(SwigClassSvn_prop_inherited_item_t.klass, _wrap_svn_prop_inherited_item_t_allocate);
  rb_define_method(SwigClassSvn_prop_inherited_item_t.klass, "initialize", _wrap_new_svn_prop_inherited_item_t, -1);
  rb_define_method(SwigClassSvn_prop_inherited_item_t.klass, "path_or_url=", _wrap_svn_prop_inherited_item_t_path_or_url_set, -1);
  rb_define_method(SwigClassSvn_prop_inherited_item_t.klass, "path_or_url", _wrap_svn_prop_inherited_item_t_path_or_url_get, -1);
  rb_define_method(SwigClassSvn_prop_inherited_item_t.klass, "prop_hash=", _wrap_svn_prop_inherited_item_t_prop_hash_set, -1);
  rb_define_method(SwigClassSvn_prop_inherited_item_t.klass, "prop_hash", _wrap_svn_prop_inherited_item_t_prop_hash_get, -1);
  SwigClassSvn_prop_inherited_item_t.mark = 0;
  SwigClassSvn_prop_inherited_item_t.destroy = (void (*)(void *)) free_svn_prop_inherited_item_t;
  SwigClassSvn_prop_inherited_item_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_prop_hash_to_array", _wrap_svn_prop_hash_to_array, -1);
  rb_define_module_function(mCore, "svn_prop_array_to_hash", _wrap_svn_prop_array_to_hash, -1);
  rb_define_module_function(mCore, "svn_prop_hash_dup", _wrap_svn_prop_hash_dup, -1);
  rb_define_module_function(mCore, "svn_prop_get_value", _wrap_svn_prop_get_value, -1);
  rb_define_const(mCore, "Svn_prop_entry_kind", SWIG_From_int((int)(svn_prop_entry_kind)));
  rb_define_const(mCore, "Svn_prop_wc_kind", SWIG_From_int((int)(svn_prop_wc_kind)));
  rb_define_const(mCore, "Svn_prop_regular_kind", SWIG_From_int((int)(svn_prop_regular_kind)));
  rb_define_module_function(mCore, "svn_property_kind2", _wrap_svn_property_kind2, -1);
  rb_define_module_function(mCore, "svn_property_kind", _wrap_svn_property_kind, -1);
  rb_define_module_function(mCore, "svn_prop_is_svn_prop", _wrap_svn_prop_is_svn_prop, -1);
  rb_define_module_function(mCore, "svn_prop_has_svn_prop", _wrap_svn_prop_has_svn_prop, -1);
  rb_define_module_function(mCore, "svn_prop_is_boolean", _wrap_svn_prop_is_boolean, -1);
  rb_define_module_function(mCore, "svn_prop_is_known_svn_rev_prop", _wrap_svn_prop_is_known_svn_rev_prop, -1);
  rb_define_module_function(mCore, "svn_prop_is_known_svn_node_prop", _wrap_svn_prop_is_known_svn_node_prop, -1);
  rb_define_module_function(mCore, "svn_prop_is_known_svn_file_prop", _wrap_svn_prop_is_known_svn_file_prop, -1);
  rb_define_module_function(mCore, "svn_prop_is_known_svn_dir_prop", _wrap_svn_prop_is_known_svn_dir_prop, -1);
  rb_define_module_function(mCore, "svn_prop_needs_translation", _wrap_svn_prop_needs_translation, -1);
  rb_define_module_function(mCore, "svn_categorize_props", _wrap_svn_categorize_props, -1);
  rb_define_module_function(mCore, "svn_prop_diffs", _wrap_svn_prop_diffs, -1);
  rb_define_module_function(mCore, "svn_prop_name_is_valid", _wrap_svn_prop_name_is_valid, -1);
  rb_define_const(mCore, "SVN_PROP_PREFIX", SWIG_FromCharPtr("svn:"));
  rb_define_const(mCore, "SVN_PROP_BOOLEAN_TRUE", SWIG_FromCharPtr("*"));
  rb_define_const(mCore, "SVN_PROP_MIME_TYPE", SWIG_FromCharPtr("svn:mime-type"));
  rb_define_const(mCore, "SVN_PROP_IGNORE", SWIG_FromCharPtr("svn:ignore"));
  rb_define_const(mCore, "SVN_PROP_EOL_STYLE", SWIG_FromCharPtr("svn:eol-style"));
  rb_define_const(mCore, "SVN_PROP_KEYWORDS", SWIG_FromCharPtr("svn:keywords"));
  rb_define_const(mCore, "SVN_PROP_EXECUTABLE", SWIG_FromCharPtr("svn:executable"));
  rb_define_const(mCore, "SVN_PROP_EXECUTABLE_VALUE", SWIG_FromCharPtr("*"));
  rb_define_const(mCore, "SVN_PROP_NEEDS_LOCK", SWIG_FromCharPtr("svn:needs-lock"));
  rb_define_const(mCore, "SVN_PROP_NEEDS_LOCK_VALUE", SWIG_FromCharPtr("*"));
  rb_define_const(mCore, "SVN_PROP_SPECIAL", SWIG_FromCharPtr("svn:special"));
  rb_define_const(mCore, "SVN_PROP_SPECIAL_VALUE", SWIG_FromCharPtr("*"));
  rb_define_const(mCore, "SVN_PROP_EXTERNALS", SWIG_FromCharPtr("svn:externals"));
  rb_define_const(mCore, "SVN_PROP_MERGEINFO", SWIG_FromCharPtr("svn:mergeinfo"));
  rb_define_const(mCore, "SVN_PROP_INHERITABLE_AUTO_PROPS", SWIG_FromCharPtr("svn:auto-props"));
  rb_define_const(mCore, "SVN_PROP_INHERITABLE_IGNORES", SWIG_FromCharPtr("svn:global-ignores"));
  rb_define_const(mCore, "SVN_PROP_TEXT_TIME", SWIG_FromCharPtr("svn:text-time"));
  rb_define_const(mCore, "SVN_PROP_OWNER", SWIG_FromCharPtr("svn:owner"));
  rb_define_const(mCore, "SVN_PROP_GROUP", SWIG_FromCharPtr("svn:group"));
  rb_define_const(mCore, "SVN_PROP_UNIX_MODE", SWIG_FromCharPtr("svn:unix-mode"));
  rb_define_const(mCore, "SVN_PROP_WC_PREFIX", SWIG_FromCharPtr("svn:wc:"));
  rb_define_const(mCore, "SVN_PROP_ENTRY_PREFIX", SWIG_FromCharPtr("svn:entry:"));
  rb_define_const(mCore, "SVN_PROP_ENTRY_COMMITTED_REV", SWIG_FromCharPtr("svn:entry:committed-rev"));
  rb_define_const(mCore, "SVN_PROP_ENTRY_COMMITTED_DATE", SWIG_FromCharPtr("svn:entry:committed-date"));
  rb_define_const(mCore, "SVN_PROP_ENTRY_LAST_AUTHOR", SWIG_FromCharPtr("svn:entry:last-author"));
  rb_define_const(mCore, "SVN_PROP_ENTRY_UUID", SWIG_FromCharPtr("svn:entry:uuid"));
  rb_define_const(mCore, "SVN_PROP_ENTRY_LOCK_TOKEN", SWIG_FromCharPtr("svn:entry:lock-token"));
  rb_define_const(mCore, "SVN_PROP_CUSTOM_PREFIX", SWIG_FromCharPtr("svn:custom:"));
  rb_define_const(mCore, "SVN_PROP_REVISION_AUTHOR", SWIG_FromCharPtr("svn:author"));
  rb_define_const(mCore, "SVN_PROP_REVISION_LOG", SWIG_FromCharPtr("svn:log"));
  rb_define_const(mCore, "SVN_PROP_REVISION_DATE", SWIG_FromCharPtr("svn:date"));
  rb_define_const(mCore, "SVN_PROP_REVISION_ORIG_DATE", SWIG_FromCharPtr("svn:original-date"));
  rb_define_const(mCore, "SVN_PROP_REVISION_AUTOVERSIONED", SWIG_FromCharPtr("svn:autoversioned"));
  rb_define_const(mCore, "SVNSYNC_PROP_PREFIX", SWIG_FromCharPtr("svn:sync-"));
  rb_define_const(mCore, "SVNSYNC_PROP_LOCK", SWIG_FromCharPtr("svn:sync-lock"));
  rb_define_const(mCore, "SVNSYNC_PROP_FROM_URL", SWIG_FromCharPtr("svn:sync-from-url"));
  rb_define_const(mCore, "SVNSYNC_PROP_FROM_UUID", SWIG_FromCharPtr("svn:sync-from-uuid"));
  rb_define_const(mCore, "SVNSYNC_PROP_LAST_MERGED_REV", SWIG_FromCharPtr("svn:sync-last-merged-rev"));
  rb_define_const(mCore, "SVNSYNC_PROP_CURRENTLY_COPYING", SWIG_FromCharPtr("svn:sync-currently-copying"));
  rb_define_const(mCore, "SVN_PROP_TXN_PREFIX", SWIG_FromCharPtr("svn:txn-"));
  rb_define_const(mCore, "SVN_PROP_TXN_CLIENT_COMPAT_VERSION", SWIG_FromCharPtr("svn:txn-client-compat-version"));
  rb_define_const(mCore, "SVN_PROP_TXN_USER_AGENT", SWIG_FromCharPtr("svn:txn-user-agent"));
  rb_define_const(mCore, "SVN_PROP_REVISION_PREFIX", SWIG_FromCharPtr("svn:revision-"));
  rb_define_const(mCore, "SVN_OPT_MAX_ALIASES", SWIG_From_int((int)(3)));
  rb_define_const(mCore, "SVN_OPT_MAX_OPTIONS", SWIG_From_int((int)(50)));
  rb_define_const(mCore, "SVN_OPT_MAX_PARAGRAPHS", SWIG_From_int((int)(100)));
  rb_define_const(mCore, "SVN_OPT_FIRST_LONGOPT_ID", SWIG_From_int((int)(256)));
  
  SwigClassSvn_opt_subcommand_desc3_t.klass = rb_define_class_under(mCore, "Svn_opt_subcommand_desc3_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_opt_subcommand_desc3_t, (void *) &SwigClassSvn_opt_subcommand_desc3_t);
  rb_define_alloc_func(SwigClassSvn_opt_subcommand_desc3_t.klass, _wrap_svn_opt_subcommand_desc3_t_allocate);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "initialize", _wrap_new_svn_opt_subcommand_desc3_t, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "name=", _wrap_svn_opt_subcommand_desc3_t_name_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "name", _wrap_svn_opt_subcommand_desc3_t_name_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "cmd_func=", _wrap_svn_opt_subcommand_desc3_t_cmd_func_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "cmd_func", _wrap_svn_opt_subcommand_desc3_t_cmd_func_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "aliases=", _wrap_svn_opt_subcommand_desc3_t_aliases_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "aliases", _wrap_svn_opt_subcommand_desc3_t_aliases_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "help=", _wrap_svn_opt_subcommand_desc3_t_help_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "help", _wrap_svn_opt_subcommand_desc3_t_help_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "valid_options=", _wrap_svn_opt_subcommand_desc3_t_valid_options_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "valid_options", _wrap_svn_opt_subcommand_desc3_t_valid_options_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t.klass, "desc_overrides", _wrap_svn_opt_subcommand_desc3_t_desc_overrides_get, -1);
  SwigClassSvn_opt_subcommand_desc3_t.mark = 0;
  SwigClassSvn_opt_subcommand_desc3_t.destroy = (void (*)(void *)) free_svn_opt_subcommand_desc3_t;
  SwigClassSvn_opt_subcommand_desc3_t.trackObjects = 0;
  
  SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.klass = rb_define_class_under(mCore, "Svn_opt_subcommand_desc3_t_desc_overrides", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_opt_subcommand_desc3_t_desc_overrides, (void *) &SwigClassSvn_opt_subcommand_desc3_t_desc_overrides);
  rb_define_alloc_func(SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.klass, _wrap_svn_opt_subcommand_desc3_t_desc_overrides_allocate);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.klass, "initialize", _wrap_new_svn_opt_subcommand_desc3_t_desc_overrides, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.klass, "optch=", _wrap_svn_opt_subcommand_desc3_t_desc_overrides_optch_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.klass, "optch", _wrap_svn_opt_subcommand_desc3_t_desc_overrides_optch_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.klass, "desc=", _wrap_svn_opt_subcommand_desc3_t_desc_overrides_desc_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.klass, "desc", _wrap_svn_opt_subcommand_desc3_t_desc_overrides_desc_get, -1);
  SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.mark = 0;
  SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.destroy = (void (*)(void *)) free_svn_opt_subcommand_desc3_t_desc_overrides;
  SwigClassSvn_opt_subcommand_desc3_t_desc_overrides.trackObjects = 0;
  
  SwigClassSvn_opt_subcommand_desc2_t.klass = rb_define_class_under(mCore, "Svn_opt_subcommand_desc2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_opt_subcommand_desc2_t, (void *) &SwigClassSvn_opt_subcommand_desc2_t);
  rb_define_alloc_func(SwigClassSvn_opt_subcommand_desc2_t.klass, _wrap_svn_opt_subcommand_desc2_t_allocate);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "initialize", _wrap_new_svn_opt_subcommand_desc2_t, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "name=", _wrap_svn_opt_subcommand_desc2_t_name_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "name", _wrap_svn_opt_subcommand_desc2_t_name_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "cmd_func=", _wrap_svn_opt_subcommand_desc2_t_cmd_func_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "cmd_func", _wrap_svn_opt_subcommand_desc2_t_cmd_func_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "aliases=", _wrap_svn_opt_subcommand_desc2_t_aliases_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "aliases", _wrap_svn_opt_subcommand_desc2_t_aliases_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "help=", _wrap_svn_opt_subcommand_desc2_t_help_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "help", _wrap_svn_opt_subcommand_desc2_t_help_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "valid_options=", _wrap_svn_opt_subcommand_desc2_t_valid_options_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "valid_options", _wrap_svn_opt_subcommand_desc2_t_valid_options_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t.klass, "desc_overrides", _wrap_svn_opt_subcommand_desc2_t_desc_overrides_get, -1);
  SwigClassSvn_opt_subcommand_desc2_t.mark = 0;
  SwigClassSvn_opt_subcommand_desc2_t.destroy = (void (*)(void *)) free_svn_opt_subcommand_desc2_t;
  SwigClassSvn_opt_subcommand_desc2_t.trackObjects = 0;
  
  SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.klass = rb_define_class_under(mCore, "Svn_opt_subcommand_desc2_t_desc_overrides", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_opt_subcommand_desc2_t_desc_overrides, (void *) &SwigClassSvn_opt_subcommand_desc2_t_desc_overrides);
  rb_define_alloc_func(SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.klass, _wrap_svn_opt_subcommand_desc2_t_desc_overrides_allocate);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.klass, "initialize", _wrap_new_svn_opt_subcommand_desc2_t_desc_overrides, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.klass, "optch=", _wrap_svn_opt_subcommand_desc2_t_desc_overrides_optch_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.klass, "optch", _wrap_svn_opt_subcommand_desc2_t_desc_overrides_optch_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.klass, "desc=", _wrap_svn_opt_subcommand_desc2_t_desc_overrides_desc_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.klass, "desc", _wrap_svn_opt_subcommand_desc2_t_desc_overrides_desc_get, -1);
  SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.mark = 0;
  SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.destroy = (void (*)(void *)) free_svn_opt_subcommand_desc2_t_desc_overrides;
  SwigClassSvn_opt_subcommand_desc2_t_desc_overrides.trackObjects = 0;
  
  SwigClassSvn_opt_subcommand_desc_t.klass = rb_define_class_under(mCore, "Svn_opt_subcommand_desc_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_opt_subcommand_desc_t, (void *) &SwigClassSvn_opt_subcommand_desc_t);
  rb_define_alloc_func(SwigClassSvn_opt_subcommand_desc_t.klass, _wrap_svn_opt_subcommand_desc_t_allocate);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "initialize", _wrap_new_svn_opt_subcommand_desc_t, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "name=", _wrap_svn_opt_subcommand_desc_t_name_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "name", _wrap_svn_opt_subcommand_desc_t_name_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "cmd_func=", _wrap_svn_opt_subcommand_desc_t_cmd_func_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "cmd_func", _wrap_svn_opt_subcommand_desc_t_cmd_func_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "aliases=", _wrap_svn_opt_subcommand_desc_t_aliases_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "aliases", _wrap_svn_opt_subcommand_desc_t_aliases_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "help=", _wrap_svn_opt_subcommand_desc_t_help_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "help", _wrap_svn_opt_subcommand_desc_t_help_get, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "valid_options=", _wrap_svn_opt_subcommand_desc_t_valid_options_set, -1);
  rb_define_method(SwigClassSvn_opt_subcommand_desc_t.klass, "valid_options", _wrap_svn_opt_subcommand_desc_t_valid_options_get, -1);
  SwigClassSvn_opt_subcommand_desc_t.mark = 0;
  SwigClassSvn_opt_subcommand_desc_t.destroy = (void (*)(void *)) free_svn_opt_subcommand_desc_t;
  SwigClassSvn_opt_subcommand_desc_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_opt_get_canonical_subcommand3", _wrap_svn_opt_get_canonical_subcommand3, -1);
  rb_define_module_function(mCore, "svn_opt_get_canonical_subcommand2", _wrap_svn_opt_get_canonical_subcommand2, -1);
  rb_define_module_function(mCore, "svn_opt_get_canonical_subcommand", _wrap_svn_opt_get_canonical_subcommand, -1);
  rb_define_module_function(mCore, "svn_opt_get_option_from_code3", _wrap_svn_opt_get_option_from_code3, -1);
  rb_define_module_function(mCore, "svn_opt_get_option_from_code2", _wrap_svn_opt_get_option_from_code2, -1);
  rb_define_module_function(mCore, "svn_opt_get_option_from_code", _wrap_svn_opt_get_option_from_code, -1);
  rb_define_module_function(mCore, "svn_opt_subcommand_takes_option4", _wrap_svn_opt_subcommand_takes_option4, -1);
  rb_define_module_function(mCore, "svn_opt_subcommand_takes_option3", _wrap_svn_opt_subcommand_takes_option3, -1);
  rb_define_module_function(mCore, "svn_opt_subcommand_takes_option2", _wrap_svn_opt_subcommand_takes_option2, -1);
  rb_define_module_function(mCore, "svn_opt_subcommand_takes_option", _wrap_svn_opt_subcommand_takes_option, -1);
  rb_define_module_function(mCore, "svn_opt_print_generic_help3", _wrap_svn_opt_print_generic_help3, -1);
  rb_define_module_function(mCore, "svn_opt_print_generic_help2", _wrap_svn_opt_print_generic_help2, -1);
  rb_define_module_function(mCore, "svn_opt_format_option", _wrap_svn_opt_format_option, -1);
  rb_define_module_function(mCore, "svn_opt_subcommand_help4", _wrap_svn_opt_subcommand_help4, -1);
  rb_define_module_function(mCore, "svn_opt_subcommand_help3", _wrap_svn_opt_subcommand_help3, -1);
  rb_define_module_function(mCore, "svn_opt_subcommand_help2", _wrap_svn_opt_subcommand_help2, -1);
  rb_define_module_function(mCore, "svn_opt_subcommand_help", _wrap_svn_opt_subcommand_help, -1);
  rb_define_const(mCore, "Svn_opt_revision_unspecified", SWIG_From_int((int)(svn_opt_revision_unspecified)));
  rb_define_const(mCore, "Svn_opt_revision_number", SWIG_From_int((int)(svn_opt_revision_number)));
  rb_define_const(mCore, "Svn_opt_revision_date", SWIG_From_int((int)(svn_opt_revision_date)));
  rb_define_const(mCore, "Svn_opt_revision_committed", SWIG_From_int((int)(svn_opt_revision_committed)));
  rb_define_const(mCore, "Svn_opt_revision_previous", SWIG_From_int((int)(svn_opt_revision_previous)));
  rb_define_const(mCore, "Svn_opt_revision_base", SWIG_From_int((int)(svn_opt_revision_base)));
  rb_define_const(mCore, "Svn_opt_revision_working", SWIG_From_int((int)(svn_opt_revision_working)));
  rb_define_const(mCore, "Svn_opt_revision_head", SWIG_From_int((int)(svn_opt_revision_head)));
  
  SwigClassSvn_opt_revision_value_t.klass = rb_define_class_under(mCore, "Svn_opt_revision_value_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_opt_revision_value_t, (void *) &SwigClassSvn_opt_revision_value_t);
  rb_define_alloc_func(SwigClassSvn_opt_revision_value_t.klass, _wrap_svn_opt_revision_value_t_allocate);
  rb_define_method(SwigClassSvn_opt_revision_value_t.klass, "initialize", _wrap_new_svn_opt_revision_value_t, -1);
  rb_define_method(SwigClassSvn_opt_revision_value_t.klass, "number=", _wrap_svn_opt_revision_value_t_number_set, -1);
  rb_define_method(SwigClassSvn_opt_revision_value_t.klass, "number", _wrap_svn_opt_revision_value_t_number_get, -1);
  rb_define_method(SwigClassSvn_opt_revision_value_t.klass, "date=", _wrap_svn_opt_revision_value_t_date_set, -1);
  rb_define_method(SwigClassSvn_opt_revision_value_t.klass, "date", _wrap_svn_opt_revision_value_t_date_get, -1);
  SwigClassSvn_opt_revision_value_t.mark = 0;
  SwigClassSvn_opt_revision_value_t.destroy = (void (*)(void *)) free_svn_opt_revision_value_t;
  SwigClassSvn_opt_revision_value_t.trackObjects = 0;
  
  SwigClassSvn_opt_revision_t.klass = rb_define_class_under(mCore, "Svn_opt_revision_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_opt_revision_t, (void *) &SwigClassSvn_opt_revision_t);
  rb_define_alloc_func(SwigClassSvn_opt_revision_t.klass, _wrap_svn_opt_revision_t_allocate);
  rb_define_method(SwigClassSvn_opt_revision_t.klass, "initialize", _wrap_new_svn_opt_revision_t, -1);
  rb_define_method(SwigClassSvn_opt_revision_t.klass, "kind=", _wrap_svn_opt_revision_t_kind_set, -1);
  rb_define_method(SwigClassSvn_opt_revision_t.klass, "kind", _wrap_svn_opt_revision_t_kind_get, -1);
  rb_define_method(SwigClassSvn_opt_revision_t.klass, "value=", _wrap_svn_opt_revision_t_value_set, -1);
  rb_define_method(SwigClassSvn_opt_revision_t.klass, "value", _wrap_svn_opt_revision_t_value_get, -1);
  SwigClassSvn_opt_revision_t.mark = 0;
  SwigClassSvn_opt_revision_t.destroy = (void (*)(void *)) free_svn_opt_revision_t;
  SwigClassSvn_opt_revision_t.trackObjects = 0;
  
  SwigClassSvn_opt_revision_range_t.klass = rb_define_class_under(mCore, "Svn_opt_revision_range_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_opt_revision_range_t, (void *) &SwigClassSvn_opt_revision_range_t);
  rb_define_alloc_func(SwigClassSvn_opt_revision_range_t.klass, _wrap_svn_opt_revision_range_t_allocate);
  rb_define_method(SwigClassSvn_opt_revision_range_t.klass, "initialize", _wrap_new_svn_opt_revision_range_t, -1);
  rb_define_method(SwigClassSvn_opt_revision_range_t.klass, "start=", _wrap_svn_opt_revision_range_t_start_set, -1);
  rb_define_method(SwigClassSvn_opt_revision_range_t.klass, "start", _wrap_svn_opt_revision_range_t_start_get, -1);
  rb_define_method(SwigClassSvn_opt_revision_range_t.klass, "end=", _wrap_svn_opt_revision_range_t_end_set, -1);
  rb_define_method(SwigClassSvn_opt_revision_range_t.klass, "end", _wrap_svn_opt_revision_range_t_end_get, -1);
  SwigClassSvn_opt_revision_range_t.mark = 0;
  SwigClassSvn_opt_revision_range_t.destroy = (void (*)(void *)) free_svn_opt_revision_range_t;
  SwigClassSvn_opt_revision_range_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_opt_parse_revision", _wrap_svn_opt_parse_revision, -1);
  rb_define_module_function(mCore, "svn_opt_parse_revision_to_range", _wrap_svn_opt_parse_revision_to_range, -1);
  rb_define_module_function(mCore, "svn_opt_resolve_revisions", _wrap_svn_opt_resolve_revisions, -1);
  rb_define_module_function(mCore, "svn_opt_parse_revprop", _wrap_svn_opt_parse_revprop, -1);
  rb_define_module_function(mCore, "svn_opt_push_implicit_dot_target", _wrap_svn_opt_push_implicit_dot_target, -1);
  rb_define_module_function(mCore, "svn_opt_parse_path", _wrap_svn_opt_parse_path, -1);
  rb_define_module_function(mCore, "svn_opt_print_help5", _wrap_svn_opt_print_help5, -1);
  rb_define_module_function(mCore, "svn_opt_print_help4", _wrap_svn_opt_print_help4, -1);
  rb_define_module_function(mCore, "svn_opt_print_help3", _wrap_svn_opt_print_help3, -1);
  rb_define_module_function(mCore, "svn_opt_print_help2", _wrap_svn_opt_print_help2, -1);
  rb_define_module_function(mCore, "svn_opt_print_help", _wrap_svn_opt_print_help, -1);
  rb_define_module_function(mCore, "svn_cmdline_init", _wrap_svn_cmdline_init, -1);
  rb_define_module_function(mCore, "svn_cmdline_create_auth_baton2", _wrap_svn_cmdline_create_auth_baton2, -1);
  rb_define_module_function(mCore, "svn_cmdline_create_auth_baton", _wrap_svn_cmdline_create_auth_baton, -1);
  
  SwigClassSvn_auth_provider_t.klass = rb_define_class_under(mCore, "Svn_auth_provider_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_provider_t, (void *) &SwigClassSvn_auth_provider_t);
  rb_define_alloc_func(SwigClassSvn_auth_provider_t.klass, _wrap_svn_auth_provider_t_allocate);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "initialize", _wrap_new_svn_auth_provider_t, -1);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "cred_kind=", _wrap_svn_auth_provider_t_cred_kind_set, -1);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "cred_kind", _wrap_svn_auth_provider_t_cred_kind_get, -1);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "first_credentials=", _wrap_svn_auth_provider_t_first_credentials_set, -1);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "first_credentials", _wrap_svn_auth_provider_t_first_credentials_get, -1);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "next_credentials=", _wrap_svn_auth_provider_t_next_credentials_set, -1);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "next_credentials", _wrap_svn_auth_provider_t_next_credentials_get, -1);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "save_credentials=", _wrap_svn_auth_provider_t_save_credentials_set, -1);
  rb_define_method(SwigClassSvn_auth_provider_t.klass, "save_credentials", _wrap_svn_auth_provider_t_save_credentials_get, -1);
  SwigClassSvn_auth_provider_t.mark = 0;
  SwigClassSvn_auth_provider_t.destroy = (void (*)(void *)) free_svn_auth_provider_t;
  SwigClassSvn_auth_provider_t.trackObjects = 0;
  
  SwigClassSvn_auth_provider_object_t.klass = rb_define_class_under(mCore, "Svn_auth_provider_object_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_provider_object_t, (void *) &SwigClassSvn_auth_provider_object_t);
  rb_define_alloc_func(SwigClassSvn_auth_provider_object_t.klass, _wrap_svn_auth_provider_object_t_allocate);
  rb_define_method(SwigClassSvn_auth_provider_object_t.klass, "initialize", _wrap_new_svn_auth_provider_object_t, -1);
  rb_define_method(SwigClassSvn_auth_provider_object_t.klass, "vtable=", _wrap_svn_auth_provider_object_t_vtable_set, -1);
  rb_define_method(SwigClassSvn_auth_provider_object_t.klass, "vtable", _wrap_svn_auth_provider_object_t_vtable_get, -1);
  rb_define_method(SwigClassSvn_auth_provider_object_t.klass, "provider_baton=", _wrap_svn_auth_provider_object_t_provider_baton_set, -1);
  rb_define_method(SwigClassSvn_auth_provider_object_t.klass, "provider_baton", _wrap_svn_auth_provider_object_t_provider_baton_get, -1);
  SwigClassSvn_auth_provider_object_t.mark = 0;
  SwigClassSvn_auth_provider_object_t.destroy = (void (*)(void *)) free_svn_auth_provider_object_t;
  SwigClassSvn_auth_provider_object_t.trackObjects = 0;
  rb_define_const(mCore, "SVN_AUTH_CRED_SIMPLE", SWIG_FromCharPtr("svn.simple"));
  
  SwigClassSvn_auth_cred_simple_t.klass = rb_define_class_under(mCore, "Svn_auth_cred_simple_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_cred_simple_t, (void *) &SwigClassSvn_auth_cred_simple_t);
  rb_define_alloc_func(SwigClassSvn_auth_cred_simple_t.klass, _wrap_svn_auth_cred_simple_t_allocate);
  rb_define_method(SwigClassSvn_auth_cred_simple_t.klass, "initialize", _wrap_new_svn_auth_cred_simple_t, -1);
  rb_define_method(SwigClassSvn_auth_cred_simple_t.klass, "username=", _wrap_svn_auth_cred_simple_t_username_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_simple_t.klass, "username", _wrap_svn_auth_cred_simple_t_username_get, -1);
  rb_define_method(SwigClassSvn_auth_cred_simple_t.klass, "password=", _wrap_svn_auth_cred_simple_t_password_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_simple_t.klass, "password", _wrap_svn_auth_cred_simple_t_password_get, -1);
  rb_define_method(SwigClassSvn_auth_cred_simple_t.klass, "may_save=", _wrap_svn_auth_cred_simple_t_may_save_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_simple_t.klass, "may_save", _wrap_svn_auth_cred_simple_t_may_save_get, -1);
  SwigClassSvn_auth_cred_simple_t.mark = 0;
  SwigClassSvn_auth_cred_simple_t.destroy = (void (*)(void *)) free_svn_auth_cred_simple_t;
  SwigClassSvn_auth_cred_simple_t.trackObjects = 0;
  rb_define_const(mCore, "SVN_AUTH_CRED_USERNAME", SWIG_FromCharPtr("svn.username"));
  
  SwigClassSvn_auth_cred_username_t.klass = rb_define_class_under(mCore, "Svn_auth_cred_username_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_cred_username_t, (void *) &SwigClassSvn_auth_cred_username_t);
  rb_define_alloc_func(SwigClassSvn_auth_cred_username_t.klass, _wrap_svn_auth_cred_username_t_allocate);
  rb_define_method(SwigClassSvn_auth_cred_username_t.klass, "initialize", _wrap_new_svn_auth_cred_username_t, -1);
  rb_define_method(SwigClassSvn_auth_cred_username_t.klass, "username=", _wrap_svn_auth_cred_username_t_username_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_username_t.klass, "username", _wrap_svn_auth_cred_username_t_username_get, -1);
  rb_define_method(SwigClassSvn_auth_cred_username_t.klass, "may_save=", _wrap_svn_auth_cred_username_t_may_save_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_username_t.klass, "may_save", _wrap_svn_auth_cred_username_t_may_save_get, -1);
  SwigClassSvn_auth_cred_username_t.mark = 0;
  SwigClassSvn_auth_cred_username_t.destroy = (void (*)(void *)) free_svn_auth_cred_username_t;
  SwigClassSvn_auth_cred_username_t.trackObjects = 0;
  rb_define_const(mCore, "SVN_AUTH_CRED_SSL_CLIENT_CERT", SWIG_FromCharPtr("svn.ssl.client-cert"));
  
  SwigClassSvn_auth_cred_ssl_client_cert_t.klass = rb_define_class_under(mCore, "Svn_auth_cred_ssl_client_cert_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t, (void *) &SwigClassSvn_auth_cred_ssl_client_cert_t);
  rb_define_alloc_func(SwigClassSvn_auth_cred_ssl_client_cert_t.klass, _wrap_svn_auth_cred_ssl_client_cert_t_allocate);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_t.klass, "initialize", _wrap_new_svn_auth_cred_ssl_client_cert_t, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_t.klass, "cert_file=", _wrap_svn_auth_cred_ssl_client_cert_t_cert_file_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_t.klass, "cert_file", _wrap_svn_auth_cred_ssl_client_cert_t_cert_file_get, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_t.klass, "may_save=", _wrap_svn_auth_cred_ssl_client_cert_t_may_save_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_t.klass, "may_save", _wrap_svn_auth_cred_ssl_client_cert_t_may_save_get, -1);
  SwigClassSvn_auth_cred_ssl_client_cert_t.mark = 0;
  SwigClassSvn_auth_cred_ssl_client_cert_t.destroy = (void (*)(void *)) free_svn_auth_cred_ssl_client_cert_t;
  SwigClassSvn_auth_cred_ssl_client_cert_t.trackObjects = 0;
  rb_define_const(mCore, "SVN_AUTH_CRED_SSL_CLIENT_CERT_PW", SWIG_FromCharPtr("svn.ssl.client-passphrase"));
  
  SwigClassSvn_auth_cred_ssl_client_cert_pw_t.klass = rb_define_class_under(mCore, "Svn_auth_cred_ssl_client_cert_pw_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t, (void *) &SwigClassSvn_auth_cred_ssl_client_cert_pw_t);
  rb_define_alloc_func(SwigClassSvn_auth_cred_ssl_client_cert_pw_t.klass, _wrap_svn_auth_cred_ssl_client_cert_pw_t_allocate);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_pw_t.klass, "initialize", _wrap_new_svn_auth_cred_ssl_client_cert_pw_t, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_pw_t.klass, "password=", _wrap_svn_auth_cred_ssl_client_cert_pw_t_password_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_pw_t.klass, "password", _wrap_svn_auth_cred_ssl_client_cert_pw_t_password_get, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_pw_t.klass, "may_save=", _wrap_svn_auth_cred_ssl_client_cert_pw_t_may_save_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_client_cert_pw_t.klass, "may_save", _wrap_svn_auth_cred_ssl_client_cert_pw_t_may_save_get, -1);
  SwigClassSvn_auth_cred_ssl_client_cert_pw_t.mark = 0;
  SwigClassSvn_auth_cred_ssl_client_cert_pw_t.destroy = (void (*)(void *)) free_svn_auth_cred_ssl_client_cert_pw_t;
  SwigClassSvn_auth_cred_ssl_client_cert_pw_t.trackObjects = 0;
  rb_define_const(mCore, "SVN_AUTH_CRED_SSL_SERVER_TRUST", SWIG_FromCharPtr("svn.ssl.server"));
  
  SwigClassSvn_auth_ssl_server_cert_info_t.klass = rb_define_class_under(mCore, "Svn_auth_ssl_server_cert_info_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_ssl_server_cert_info_t, (void *) &SwigClassSvn_auth_ssl_server_cert_info_t);
  rb_define_alloc_func(SwigClassSvn_auth_ssl_server_cert_info_t.klass, _wrap_svn_auth_ssl_server_cert_info_t_allocate);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "initialize", _wrap_new_svn_auth_ssl_server_cert_info_t, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "hostname=", _wrap_svn_auth_ssl_server_cert_info_t_hostname_set, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "hostname", _wrap_svn_auth_ssl_server_cert_info_t_hostname_get, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "fingerprint=", _wrap_svn_auth_ssl_server_cert_info_t_fingerprint_set, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "fingerprint", _wrap_svn_auth_ssl_server_cert_info_t_fingerprint_get, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "valid_from=", _wrap_svn_auth_ssl_server_cert_info_t_valid_from_set, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "valid_from", _wrap_svn_auth_ssl_server_cert_info_t_valid_from_get, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "valid_until=", _wrap_svn_auth_ssl_server_cert_info_t_valid_until_set, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "valid_until", _wrap_svn_auth_ssl_server_cert_info_t_valid_until_get, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "issuer_dname=", _wrap_svn_auth_ssl_server_cert_info_t_issuer_dname_set, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "issuer_dname", _wrap_svn_auth_ssl_server_cert_info_t_issuer_dname_get, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "ascii_cert=", _wrap_svn_auth_ssl_server_cert_info_t_ascii_cert_set, -1);
  rb_define_method(SwigClassSvn_auth_ssl_server_cert_info_t.klass, "ascii_cert", _wrap_svn_auth_ssl_server_cert_info_t_ascii_cert_get, -1);
  SwigClassSvn_auth_ssl_server_cert_info_t.mark = 0;
  SwigClassSvn_auth_ssl_server_cert_info_t.destroy = (void (*)(void *)) free_svn_auth_ssl_server_cert_info_t;
  SwigClassSvn_auth_ssl_server_cert_info_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_auth_ssl_server_cert_info_dup", _wrap_svn_auth_ssl_server_cert_info_dup, -1);
  
  SwigClassSvn_auth_cred_ssl_server_trust_t.klass = rb_define_class_under(mCore, "Svn_auth_cred_ssl_server_trust_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t, (void *) &SwigClassSvn_auth_cred_ssl_server_trust_t);
  rb_define_alloc_func(SwigClassSvn_auth_cred_ssl_server_trust_t.klass, _wrap_svn_auth_cred_ssl_server_trust_t_allocate);
  rb_define_method(SwigClassSvn_auth_cred_ssl_server_trust_t.klass, "initialize", _wrap_new_svn_auth_cred_ssl_server_trust_t, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_server_trust_t.klass, "may_save=", _wrap_svn_auth_cred_ssl_server_trust_t_may_save_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_server_trust_t.klass, "may_save", _wrap_svn_auth_cred_ssl_server_trust_t_may_save_get, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_server_trust_t.klass, "accepted_failures=", _wrap_svn_auth_cred_ssl_server_trust_t_accepted_failures_set, -1);
  rb_define_method(SwigClassSvn_auth_cred_ssl_server_trust_t.klass, "accepted_failures", _wrap_svn_auth_cred_ssl_server_trust_t_accepted_failures_get, -1);
  SwigClassSvn_auth_cred_ssl_server_trust_t.mark = 0;
  SwigClassSvn_auth_cred_ssl_server_trust_t.destroy = (void (*)(void *)) free_svn_auth_cred_ssl_server_trust_t;
  SwigClassSvn_auth_cred_ssl_server_trust_t.trackObjects = 0;
  rb_define_const(mCore, "SVN_AUTH_SSL_NOTYETVALID", SWIG_From_int((int)(0x00000001)));
  rb_define_const(mCore, "SVN_AUTH_SSL_EXPIRED", SWIG_From_int((int)(0x00000002)));
  rb_define_const(mCore, "SVN_AUTH_SSL_CNMISMATCH", SWIG_From_int((int)(0x00000004)));
  rb_define_const(mCore, "SVN_AUTH_SSL_UNKNOWNCA", SWIG_From_int((int)(0x00000008)));
  rb_define_const(mCore, "SVN_AUTH_SSL_OTHER", SWIG_From_int((int)(0x40000000)));
  rb_define_module_function(mCore, "svn_auth_set_parameter", _wrap_svn_auth_set_parameter, -1);
  rb_define_module_function(mCore, "svn_auth_get_parameter", _wrap_svn_auth_get_parameter, -1);
  rb_define_const(mCore, "SVN_AUTH_PARAM_PREFIX", SWIG_FromCharPtr("svn:auth:"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_DEFAULT_USERNAME", SWIG_FromCharPtr("svn:auth:username"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_DEFAULT_PASSWORD", SWIG_FromCharPtr("svn:auth:password"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_NON_INTERACTIVE", SWIG_FromCharPtr("svn:auth:non-interactive"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_DONT_STORE_PASSWORDS", SWIG_FromCharPtr("svn:auth:dont-store-passwords"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_STORE_PLAINTEXT_PASSWORDS", SWIG_FromCharPtr("svn:auth:store-plaintext-passwords"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_DONT_STORE_SSL_CLIENT_CERT_PP", SWIG_FromCharPtr("svn:auth:dont-store-ssl-client-cert-pp"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_STORE_SSL_CLIENT_CERT_PP_PLAINTEXT", SWIG_FromCharPtr("svn:auth:store-ssl-client-cert-pp-plaintext"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_NO_AUTH_CACHE", SWIG_FromCharPtr("svn:auth:no-auth-cache"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_SSL_SERVER_FAILURES", SWIG_FromCharPtr("svn:auth:ssl:failures"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_SSL_SERVER_CERT_INFO", SWIG_FromCharPtr("svn:auth:ssl:cert-info"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_CONFIG_CATEGORY_CONFIG", SWIG_FromCharPtr("svn:auth:config-category-config"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_CONFIG_CATEGORY_SERVERS", SWIG_FromCharPtr("svn:auth:config-category-servers"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_CONFIG", SWIG_FromCharPtr("svn:auth:config-category-servers"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_SERVER_GROUP", SWIG_FromCharPtr("svn:auth:server-group"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_CONFIG_DIR", SWIG_FromCharPtr("svn:auth:config-dir"));
  rb_define_module_function(mCore, "svn_auth_first_credentials", _wrap_svn_auth_first_credentials, -1);
  rb_define_module_function(mCore, "svn_auth_next_credentials", _wrap_svn_auth_next_credentials, -1);
  rb_define_module_function(mCore, "svn_auth_save_credentials", _wrap_svn_auth_save_credentials, -1);
  rb_define_module_function(mCore, "svn_auth_forget_credentials", _wrap_svn_auth_forget_credentials, -1);
  rb_define_module_function(mCore, "svn_auth_get_simple_prompt_provider", _wrap_svn_auth_get_simple_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_username_prompt_provider", _wrap_svn_auth_get_username_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_simple_provider2", _wrap_svn_auth_get_simple_provider2, -1);
  rb_define_module_function(mCore, "svn_auth_get_simple_provider", _wrap_svn_auth_get_simple_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_platform_specific_provider", _wrap_svn_auth_get_platform_specific_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_platform_specific_client_providers", _wrap_svn_auth_get_platform_specific_client_providers, -1);
  rb_define_const(mCore, "SVN_AUTH_PARAM_GNOME_KEYRING_UNLOCK_PROMPT_FUNC", SWIG_FromCharPtr("gnome-keyring-unlock-prompt-func"));
  rb_define_const(mCore, "SVN_AUTH_PARAM_GNOME_KEYRING_UNLOCK_PROMPT_BATON", SWIG_FromCharPtr("gnome-keyring-unlock-prompt-baton"));
  rb_define_module_function(mCore, "svn_auth_get_username_provider", _wrap_svn_auth_get_username_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_ssl_server_trust_file_provider", _wrap_svn_auth_get_ssl_server_trust_file_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_ssl_client_cert_file_provider", _wrap_svn_auth_get_ssl_client_cert_file_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_ssl_client_cert_pw_file_provider2", _wrap_svn_auth_get_ssl_client_cert_pw_file_provider2, -1);
  rb_define_module_function(mCore, "svn_auth_get_ssl_client_cert_pw_file_provider", _wrap_svn_auth_get_ssl_client_cert_pw_file_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_ssl_server_trust_prompt_provider", _wrap_svn_auth_get_ssl_server_trust_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_ssl_client_cert_prompt_provider", _wrap_svn_auth_get_ssl_client_cert_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_auth_get_ssl_client_cert_pw_prompt_provider", _wrap_svn_auth_get_ssl_client_cert_pw_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_auth_provider_invoke_first_credentials", _wrap_svn_auth_provider_invoke_first_credentials, -1);
  rb_define_module_function(mCore, "svn_auth_provider_invoke_next_credentials", _wrap_svn_auth_provider_invoke_next_credentials, -1);
  rb_define_module_function(mCore, "svn_auth_provider_invoke_save_credentials", _wrap_svn_auth_provider_invoke_save_credentials, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_simple_provider_func", _wrap_svn_auth_invoke_simple_provider_func, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_ssl_client_cert_pw_provider_func", _wrap_svn_auth_invoke_ssl_client_cert_pw_provider_func, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_simple_prompt_func", _wrap_svn_auth_invoke_simple_prompt_func, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_username_prompt_func", _wrap_svn_auth_invoke_username_prompt_func, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_ssl_server_trust_prompt_func", _wrap_svn_auth_invoke_ssl_server_trust_prompt_func, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_ssl_client_cert_prompt_func", _wrap_svn_auth_invoke_ssl_client_cert_prompt_func, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_ssl_client_cert_pw_prompt_func", _wrap_svn_auth_invoke_ssl_client_cert_pw_prompt_func, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_plaintext_prompt_func", _wrap_svn_auth_invoke_plaintext_prompt_func, -1);
  rb_define_module_function(mCore, "svn_auth_invoke_plaintext_passphrase_prompt_func", _wrap_svn_auth_invoke_plaintext_passphrase_prompt_func, -1);
  rb_define_const(mCore, "SVN_CONFIG_CATEGORY_SERVERS", SWIG_FromCharPtr("servers"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_GROUPS", SWIG_FromCharPtr("groups"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_GLOBAL", SWIG_FromCharPtr("global"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_PROXY_HOST", SWIG_FromCharPtr("http-proxy-host"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_PROXY_PORT", SWIG_FromCharPtr("http-proxy-port"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_PROXY_USERNAME", SWIG_FromCharPtr("http-proxy-username"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_PROXY_PASSWORD", SWIG_FromCharPtr("http-proxy-password"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_PROXY_EXCEPTIONS", SWIG_FromCharPtr("http-proxy-exceptions"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_TIMEOUT", SWIG_FromCharPtr("http-timeout"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_COMPRESSION", SWIG_FromCharPtr("http-compression"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_NEON_DEBUG_MASK", SWIG_FromCharPtr("neon-debug-mask"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_AUTH_TYPES", SWIG_FromCharPtr("http-auth-types"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SSL_AUTHORITY_FILES", SWIG_FromCharPtr("ssl-authority-files"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SSL_TRUST_DEFAULT_CA", SWIG_FromCharPtr("ssl-trust-default-ca"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SSL_CLIENT_CERT_FILE", SWIG_FromCharPtr("ssl-client-cert-file"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SSL_CLIENT_CERT_PASSWORD", SWIG_FromCharPtr("ssl-client-cert-password"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SSL_PKCS11_PROVIDER", SWIG_FromCharPtr("ssl-pkcs11-provider"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_LIBRARY", SWIG_FromCharPtr("http-library"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_STORE_PASSWORDS", SWIG_FromCharPtr("store-passwords"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_STORE_PLAINTEXT_PASSWORDS", SWIG_FromCharPtr("store-plaintext-passwords"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_STORE_AUTH_CREDS", SWIG_FromCharPtr("store-auth-creds"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_STORE_SSL_CLIENT_CERT_PP", SWIG_FromCharPtr("store-ssl-client-cert-pp"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_STORE_SSL_CLIENT_CERT_PP_PLAINTEXT", SWIG_FromCharPtr("store-ssl-client-cert-pp-plaintext"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_USERNAME", SWIG_FromCharPtr("username"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_BULK_UPDATES", SWIG_FromCharPtr("http-bulk-updates"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_MAX_CONNECTIONS", SWIG_FromCharPtr("http-max-connections"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HTTP_CHUNKED_REQUESTS", SWIG_FromCharPtr("http-chunked-requests"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SERF_LOG_COMPONENTS", SWIG_FromCharPtr("serf-log-components"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SERF_LOG_LEVEL", SWIG_FromCharPtr("serf-log-level"));
  rb_define_const(mCore, "SVN_CONFIG_CATEGORY_CONFIG", SWIG_FromCharPtr("config"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_AUTH", SWIG_FromCharPtr("auth"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_PASSWORD_STORES", SWIG_FromCharPtr("password-stores"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_KWALLET_WALLET", SWIG_FromCharPtr("kwallet-wallet"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_KWALLET_SVN_APPLICATION_NAME_WITH_PID", SWIG_FromCharPtr("kwallet-svn-application-name-with-pid"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SSL_CLIENT_CERT_FILE_PROMPT", SWIG_FromCharPtr("ssl-client-cert-file-prompt"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_HELPERS", SWIG_FromCharPtr("helpers"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_EDITOR_CMD", SWIG_FromCharPtr("editor-cmd"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_DIFF_CMD", SWIG_FromCharPtr("diff-cmd"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_DIFF_EXTENSIONS", SWIG_FromCharPtr("diff-extensions"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_DIFF3_CMD", SWIG_FromCharPtr("diff3-cmd"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_DIFF3_HAS_PROGRAM_ARG", SWIG_FromCharPtr("diff3-has-program-arg"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_MERGE_TOOL_CMD", SWIG_FromCharPtr("merge-tool-cmd"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_MISCELLANY", SWIG_FromCharPtr("miscellany"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_GLOBAL_IGNORES", SWIG_FromCharPtr("global-ignores"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_LOG_ENCODING", SWIG_FromCharPtr("log-encoding"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_USE_COMMIT_TIMES", SWIG_FromCharPtr("use-commit-times"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_TEMPLATE_ROOT", SWIG_FromCharPtr("template-root"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_ENABLE_AUTO_PROPS", SWIG_FromCharPtr("enable-auto-props"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_ENABLE_MAGIC_FILE", SWIG_FromCharPtr("enable-magic-file"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_NO_UNLOCK", SWIG_FromCharPtr("no-unlock"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_MIMETYPES_FILE", SWIG_FromCharPtr("mime-types-file"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_PRESERVED_CF_EXTS", SWIG_FromCharPtr("preserved-conflict-file-exts"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_INTERACTIVE_CONFLICTS", SWIG_FromCharPtr("interactive-conflicts"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_MEMORY_CACHE_SIZE", SWIG_FromCharPtr("memory-cache-size"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_DIFF_IGNORE_CONTENT_TYPE", SWIG_FromCharPtr("diff-ignore-content-type"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_TUNNELS", SWIG_FromCharPtr("tunnels"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_AUTO_PROPS", SWIG_FromCharPtr("auto-props"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_WORKING_COPY", SWIG_FromCharPtr("working-copy"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SQLITE_EXCLUSIVE", SWIG_FromCharPtr("exclusive-locking"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SQLITE_EXCLUSIVE_CLIENTS", SWIG_FromCharPtr("exclusive-locking-clients"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_SQLITE_BUSY_TIMEOUT", SWIG_FromCharPtr("busy-timeout"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_GENERAL", SWIG_FromCharPtr("general"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_ANON_ACCESS", SWIG_FromCharPtr("anon-access"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_AUTH_ACCESS", SWIG_FromCharPtr("auth-access"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_PASSWORD_DB", SWIG_FromCharPtr("password-db"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_REALM", SWIG_FromCharPtr("realm"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_AUTHZ_DB", SWIG_FromCharPtr("authz-db"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_GROUPS_DB", SWIG_FromCharPtr("groups-db"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_FORCE_USERNAME_CASE", SWIG_FromCharPtr("force-username-case"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_HOOKS_ENV", SWIG_FromCharPtr("hooks-env"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_SASL", SWIG_FromCharPtr("sasl"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_USE_SASL", SWIG_FromCharPtr("use-sasl"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_MIN_SSF", SWIG_FromCharPtr("min-encryption"));
  rb_define_const(mCore, "SVN_CONFIG_OPTION_MAX_SSF", SWIG_FromCharPtr("max-encryption"));
  rb_define_const(mCore, "SVN_CONFIG_SECTION_USERS", SWIG_FromCharPtr("users"));
  rb_define_const(mCore, "SVN_CONFIG__DEFAULT_GLOBAL_IGNORES_LINE_1", SWIG_FromCharPtr("*.o *.lo *.la *.al .libs *.so *.so.[0-9]* *.a *.pyc *.pyo __pycache__"));
  rb_define_const(mCore, "SVN_CONFIG__DEFAULT_GLOBAL_IGNORES_LINE_2", SWIG_FromCharPtr("*.rej *~ #*# .#* .*.swp .DS_Store [Tt]humbs.db"));
  rb_define_const(mCore, "SVN_CONFIG_DEFAULT_GLOBAL_IGNORES", SWIG_FromCharPtr("*.o *.lo *.la *.al .libs *.so *.so.[0-9]* *.a *.pyc *.pyo __pycache__ *.rej *~ #*# .#* .*.swp .DS_Store [Tt]humbs.db"));
  rb_define_const(mCore, "SVN_CONFIG_TRUE", SWIG_FromCharPtr("TRUE"));
  rb_define_const(mCore, "SVN_CONFIG_FALSE", SWIG_FromCharPtr("FALSE"));
  rb_define_const(mCore, "SVN_CONFIG_ASK", SWIG_FromCharPtr("ASK"));
  rb_define_const(mCore, "SVN_CONFIG_DEFAULT_OPTION_STORE_PASSWORDS", SWIG_From_int((int)(1)));
  rb_define_const(mCore, "SVN_CONFIG_DEFAULT_OPTION_STORE_PLAINTEXT_PASSWORDS", SWIG_FromCharPtr("ASK"));
  rb_define_const(mCore, "SVN_CONFIG_DEFAULT_OPTION_STORE_AUTH_CREDS", SWIG_From_int((int)(1)));
  rb_define_const(mCore, "SVN_CONFIG_DEFAULT_OPTION_STORE_SSL_CLIENT_CERT_PP", SWIG_From_int((int)(1)));
  rb_define_const(mCore, "SVN_CONFIG_DEFAULT_OPTION_STORE_SSL_CLIENT_CERT_PP_PLAINTEXT", SWIG_FromCharPtr("ASK"));
  rb_define_const(mCore, "SVN_CONFIG_DEFAULT_OPTION_HTTP_MAX_CONNECTIONS", SWIG_From_int((int)(4)));
  rb_define_module_function(mCore, "svn_config_get_config", _wrap_svn_config_get_config, -1);
  rb_define_module_function(mCore, "svn_config_create2", _wrap_svn_config_create2, -1);
  rb_define_module_function(mCore, "svn_config_create", _wrap_svn_config_create, -1);
  rb_define_module_function(mCore, "svn_config_read3", _wrap_svn_config_read3, -1);
  rb_define_module_function(mCore, "svn_config_read2", _wrap_svn_config_read2, -1);
  rb_define_module_function(mCore, "svn_config_read", _wrap_svn_config_read, -1);
  rb_define_module_function(mCore, "svn_config_parse", _wrap_svn_config_parse, -1);
  rb_define_module_function(mCore, "svn_config_merge", _wrap_svn_config_merge, -1);
  rb_define_module_function(mCore, "svn_config_get", _wrap_svn_config_get, -1);
  rb_define_module_function(mCore, "svn_config_set", _wrap_svn_config_set, -1);
  rb_define_module_function(mCore, "svn_config_get_bool", _wrap_svn_config_get_bool, -1);
  rb_define_module_function(mCore, "svn_config_set_bool", _wrap_svn_config_set_bool, -1);
  rb_define_module_function(mCore, "svn_config_get_int64", _wrap_svn_config_get_int64, -1);
  rb_define_module_function(mCore, "svn_config_set_int64", _wrap_svn_config_set_int64, -1);
  rb_define_module_function(mCore, "svn_config_get_yes_no_ask", _wrap_svn_config_get_yes_no_ask, -1);
  rb_define_module_function(mCore, "svn_config_get_tristate", _wrap_svn_config_get_tristate, -1);
  rb_define_module_function(mCore, "svn_config_enumerate_sections", _wrap_svn_config_enumerate_sections, -1);
  rb_define_module_function(mCore, "svn_config_enumerate_sections2", _wrap_svn_config_enumerate_sections2, -1);
  rb_define_module_function(mCore, "svn_config_enumerate", _wrap_svn_config_enumerate, -1);
  rb_define_module_function(mCore, "svn_config_enumerate2", _wrap_svn_config_enumerate2, -1);
  rb_define_module_function(mCore, "svn_config_has_section", _wrap_svn_config_has_section, -1);
  rb_define_module_function(mCore, "svn_config_find_group", _wrap_svn_config_find_group, -1);
  rb_define_module_function(mCore, "svn_config_get_server_setting", _wrap_svn_config_get_server_setting, -1);
  rb_define_module_function(mCore, "svn_config_get_server_setting_int", _wrap_svn_config_get_server_setting_int, -1);
  rb_define_module_function(mCore, "svn_config_get_server_setting_bool", _wrap_svn_config_get_server_setting_bool, -1);
  rb_define_module_function(mCore, "svn_config_ensure", _wrap_svn_config_ensure, -1);
  rb_define_const(mCore, "SVN_CONFIG_REALMSTRING_KEY", SWIG_FromCharPtr("svn:realmstring"));
  rb_define_const(mCore, "SVN_CONFIG_AUTHN_USERNAME_KEY", SWIG_FromCharPtr("username"));
  rb_define_const(mCore, "SVN_CONFIG_AUTHN_PASSWORD_KEY", SWIG_FromCharPtr("password"));
  rb_define_const(mCore, "SVN_CONFIG_AUTHN_PASSPHRASE_KEY", SWIG_FromCharPtr("passphrase"));
  rb_define_const(mCore, "SVN_CONFIG_AUTHN_PASSTYPE_KEY", SWIG_FromCharPtr("passtype"));
  rb_define_const(mCore, "SVN_CONFIG_AUTHN_ASCII_CERT_KEY", SWIG_FromCharPtr("ascii_cert"));
  rb_define_const(mCore, "SVN_CONFIG_AUTHN_FAILURES_KEY", SWIG_FromCharPtr("failures"));
  rb_define_module_function(mCore, "svn_config_read_auth_data", _wrap_svn_config_read_auth_data, -1);
  rb_define_module_function(mCore, "svn_config_write_auth_data", _wrap_svn_config_write_auth_data, -1);
  rb_define_module_function(mCore, "svn_config_walk_auth_data", _wrap_svn_config_walk_auth_data, -1);
  rb_define_module_function(mCore, "svn_config_get_user_config_path", _wrap_svn_config_get_user_config_path, -1);
  rb_define_module_function(mCore, "svn_config_dup", _wrap_svn_config_dup, -1);
  rb_define_module_function(mCore, "svn_config_copy_config", _wrap_svn_config_copy_config, -1);
  rb_define_module_function(mCore, "svn_config_invoke_section_enumerator", _wrap_svn_config_invoke_section_enumerator, -1);
  rb_define_module_function(mCore, "svn_config_invoke_section_enumerator2", _wrap_svn_config_invoke_section_enumerator2, -1);
  rb_define_module_function(mCore, "svn_config_invoke_enumerator", _wrap_svn_config_invoke_enumerator, -1);
  rb_define_module_function(mCore, "svn_config_invoke_enumerator2", _wrap_svn_config_invoke_enumerator2, -1);
  rb_define_module_function(mCore, "svn_config_invoke_auth_walk_func", _wrap_svn_config_invoke_auth_walk_func, -1);
  rb_define_module_function(mCore, "svn_utf_initialize2", _wrap_svn_utf_initialize2, -1);
  rb_define_module_function(mCore, "svn_utf_initialize", _wrap_svn_utf_initialize, -1);
  rb_define_module_function(mCore, "svn_utf_stringbuf_to_utf8", _wrap_svn_utf_stringbuf_to_utf8, -1);
  rb_define_module_function(mCore, "svn_utf_string_to_utf8", _wrap_svn_utf_string_to_utf8, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_to_utf8", _wrap_svn_utf_cstring_to_utf8, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_to_utf8_ex2", _wrap_svn_utf_cstring_to_utf8_ex2, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_to_utf8_ex", _wrap_svn_utf_cstring_to_utf8_ex, -1);
  rb_define_module_function(mCore, "svn_utf_stringbuf_from_utf8", _wrap_svn_utf_stringbuf_from_utf8, -1);
  rb_define_module_function(mCore, "svn_utf_string_from_utf8", _wrap_svn_utf_string_from_utf8, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_from_utf8", _wrap_svn_utf_cstring_from_utf8, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_from_utf8_ex2", _wrap_svn_utf_cstring_from_utf8_ex2, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_from_utf8_ex", _wrap_svn_utf_cstring_from_utf8_ex, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_from_utf8_fuzzy", _wrap_svn_utf_cstring_from_utf8_fuzzy, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_from_utf8_stringbuf", _wrap_svn_utf_cstring_from_utf8_stringbuf, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_from_utf8_string", _wrap_svn_utf_cstring_from_utf8_string, -1);
  rb_define_module_function(mCore, "svn_utf_cstring_utf8_width", _wrap_svn_utf_cstring_utf8_width, -1);
  rb_define_module_function(mCore, "svn_nls_init", _wrap_svn_nls_init, -1);
  rb_define_module_function(mCore, "svn_path_internal_style", _wrap_svn_path_internal_style, -1);
  rb_define_module_function(mCore, "svn_path_local_style", _wrap_svn_path_local_style, -1);
  rb_define_module_function(mCore, "svn_path_splitext", _wrap_svn_path_splitext, -1);
  rb_define_module_function(mCore, "svn_path_is_empty", _wrap_svn_path_is_empty, -1);
  rb_define_module_function(mCore, "svn_dirent_is_root", _wrap_svn_dirent_is_root, -1);
  rb_define_module_function(mCore, "svn_path_canonicalize", _wrap_svn_path_canonicalize, -1);
  rb_define_module_function(mCore, "svn_path_is_canonical", _wrap_svn_path_is_canonical, -1);
  rb_define_module_function(mCore, "svn_path_compare_paths", _wrap_svn_path_compare_paths, -1);
  rb_define_module_function(mCore, "svn_path_get_longest_ancestor", _wrap_svn_path_get_longest_ancestor, -1);
  rb_define_module_function(mCore, "svn_path_is_dotpath_present", _wrap_svn_path_is_dotpath_present, -1);
  rb_define_module_function(mCore, "svn_path_is_url", _wrap_svn_path_is_url, -1);
  rb_define_module_function(mCore, "svn_path_is_uri_safe", _wrap_svn_path_is_uri_safe, -1);
  rb_define_module_function(mCore, "svn_path_url_add_component2", _wrap_svn_path_url_add_component2, -1);
  rb_define_module_function(mCore, "svn_path_is_repos_relative_url", _wrap_svn_path_is_repos_relative_url, -1);
  rb_define_module_function(mCore, "svn_path_resolve_repos_relative_url", _wrap_svn_path_resolve_repos_relative_url, -1);
  rb_define_module_function(mCore, "svn_path_illegal_path_escape", _wrap_svn_path_illegal_path_escape, -1);
  rb_define_module_function(mCore, "svn_dirent_internal_style", _wrap_svn_dirent_internal_style, -1);
  rb_define_module_function(mCore, "svn_dirent_local_style", _wrap_svn_dirent_local_style, -1);
  rb_define_module_function(mCore, "svn_relpath__internal_style", _wrap_svn_relpath__internal_style, -1);
  rb_define_module_function(mCore, "svn_dirent_join", _wrap_svn_dirent_join, -1);
  rb_define_module_function(mCore, "svn_relpath_join", _wrap_svn_relpath_join, -1);
  rb_define_module_function(mCore, "svn_dirent_basename", _wrap_svn_dirent_basename, -1);
  rb_define_module_function(mCore, "svn_dirent_dirname", _wrap_svn_dirent_dirname, -1);
  rb_define_module_function(mCore, "svn_dirent_split", _wrap_svn_dirent_split, -1);
  rb_define_module_function(mCore, "svn_relpath_split", _wrap_svn_relpath_split, -1);
  rb_define_module_function(mCore, "svn_relpath_basename", _wrap_svn_relpath_basename, -1);
  rb_define_module_function(mCore, "svn_relpath_dirname", _wrap_svn_relpath_dirname, -1);
  rb_define_module_function(mCore, "svn_relpath_prefix", _wrap_svn_relpath_prefix, -1);
  rb_define_module_function(mCore, "svn_uri_split", _wrap_svn_uri_split, -1);
  rb_define_module_function(mCore, "svn_uri_basename", _wrap_svn_uri_basename, -1);
  rb_define_module_function(mCore, "svn_uri_dirname", _wrap_svn_uri_dirname, -1);
  rb_define_module_function(mCore, "svn_dirent_is_absolute", _wrap_svn_dirent_is_absolute, -1);
  rb_define_module_function(mCore, "svn_uri_is_root", _wrap_svn_uri_is_root, -1);
  rb_define_module_function(mCore, "svn_dirent_canonicalize", _wrap_svn_dirent_canonicalize, -1);
  rb_define_module_function(mCore, "svn_relpath_canonicalize", _wrap_svn_relpath_canonicalize, -1);
  rb_define_module_function(mCore, "svn_uri_canonicalize", _wrap_svn_uri_canonicalize, -1);
  rb_define_module_function(mCore, "svn_dirent_is_canonical", _wrap_svn_dirent_is_canonical, -1);
  rb_define_module_function(mCore, "svn_relpath_is_canonical", _wrap_svn_relpath_is_canonical, -1);
  rb_define_module_function(mCore, "svn_uri_is_canonical", _wrap_svn_uri_is_canonical, -1);
  rb_define_module_function(mCore, "svn_dirent_get_longest_ancestor", _wrap_svn_dirent_get_longest_ancestor, -1);
  rb_define_module_function(mCore, "svn_relpath_get_longest_ancestor", _wrap_svn_relpath_get_longest_ancestor, -1);
  rb_define_module_function(mCore, "svn_uri_get_longest_ancestor", _wrap_svn_uri_get_longest_ancestor, -1);
  rb_define_module_function(mCore, "svn_dirent_get_absolute", _wrap_svn_dirent_get_absolute, -1);
  rb_define_module_function(mCore, "svn_dirent_is_child", _wrap_svn_dirent_is_child, -1);
  rb_define_module_function(mCore, "svn_dirent_is_ancestor", _wrap_svn_dirent_is_ancestor, -1);
  rb_define_module_function(mCore, "svn_uri__is_ancestor", _wrap_svn_uri__is_ancestor, -1);
  rb_define_module_function(mCore, "svn_dirent_skip_ancestor", _wrap_svn_dirent_skip_ancestor, -1);
  rb_define_module_function(mCore, "svn_relpath_skip_ancestor", _wrap_svn_relpath_skip_ancestor, -1);
  rb_define_module_function(mCore, "svn_uri_skip_ancestor", _wrap_svn_uri_skip_ancestor, -1);
  rb_define_module_function(mCore, "svn_uri_get_dirent_from_file_url", _wrap_svn_uri_get_dirent_from_file_url, -1);
  rb_define_module_function(mCore, "svn_uri_get_file_url_from_dirent", _wrap_svn_uri_get_file_url_from_dirent, -1);
  rb_define_const(mCore, "SVN_MERGEINFO_NONINHERITABLE_STR", SWIG_FromCharPtr("*"));
  rb_define_module_function(mCore, "svn_mergeinfo_parse", _wrap_svn_mergeinfo_parse, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_diff2", _wrap_svn_mergeinfo_diff2, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_diff", _wrap_svn_mergeinfo_diff, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_merge2", _wrap_svn_mergeinfo_merge2, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_catalog_merge", _wrap_svn_mergeinfo_catalog_merge, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_remove", _wrap_svn_mergeinfo_remove, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_remove2", _wrap_svn_mergeinfo_remove2, -1);
  rb_define_module_function(mCore, "svn_rangelist_diff", _wrap_svn_rangelist_diff, -1);
  rb_define_module_function(mCore, "svn_rangelist_merge2", _wrap_svn_rangelist_merge2, -1);
  rb_define_module_function(mCore, "svn_rangelist_remove", _wrap_svn_rangelist_remove, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_intersect2", _wrap_svn_mergeinfo_intersect2, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_intersect", _wrap_svn_mergeinfo_intersect, -1);
  rb_define_module_function(mCore, "svn_rangelist_intersect", _wrap_svn_rangelist_intersect, -1);
  rb_define_module_function(mCore, "svn_rangelist_to_string", _wrap_svn_rangelist_to_string, -1);
  rb_define_module_function(mCore, "svn_rangelist_inheritable2", _wrap_svn_rangelist_inheritable2, -1);
  rb_define_module_function(mCore, "svn_rangelist_inheritable", _wrap_svn_rangelist_inheritable, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_inheritable2", _wrap_svn_mergeinfo_inheritable2, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_inheritable", _wrap_svn_mergeinfo_inheritable, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_to_string", _wrap_svn_mergeinfo_to_string, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_catalog_dup", _wrap_svn_mergeinfo_catalog_dup, -1);
  rb_define_module_function(mCore, "svn_mergeinfo_dup", _wrap_svn_mergeinfo_dup, -1);
  rb_define_module_function(mCore, "svn_rangelist_dup", _wrap_svn_rangelist_dup, -1);
  rb_define_const(mCore, "Svn_mergeinfo_explicit", SWIG_From_int((int)(svn_mergeinfo_explicit)));
  rb_define_const(mCore, "Svn_mergeinfo_inherited", SWIG_From_int((int)(svn_mergeinfo_inherited)));
  rb_define_const(mCore, "Svn_mergeinfo_nearest_ancestor", SWIG_From_int((int)(svn_mergeinfo_nearest_ancestor)));
  rb_define_module_function(mCore, "svn_inheritance_to_word", _wrap_svn_inheritance_to_word, -1);
  rb_define_module_function(mCore, "svn_inheritance_from_word", _wrap_svn_inheritance_from_word, -1);
  rb_define_const(mCore, "Svn_io_file_del_none", SWIG_From_int((int)(svn_io_file_del_none)));
  rb_define_const(mCore, "Svn_io_file_del_on_close", SWIG_From_int((int)(svn_io_file_del_on_close)));
  rb_define_const(mCore, "Svn_io_file_del_on_pool_cleanup", SWIG_From_int((int)(svn_io_file_del_on_pool_cleanup)));
  
  SwigClassSvn_io_dirent2_t.klass = rb_define_class_under(mCore, "Svn_io_dirent2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_io_dirent2_t, (void *) &SwigClassSvn_io_dirent2_t);
  rb_define_alloc_func(SwigClassSvn_io_dirent2_t.klass, _wrap_svn_io_dirent2_t_allocate);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "initialize", _wrap_new_svn_io_dirent2_t, -1);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "kind=", _wrap_svn_io_dirent2_t_kind_set, -1);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "kind", _wrap_svn_io_dirent2_t_kind_get, -1);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "special=", _wrap_svn_io_dirent2_t_special_set, -1);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "special", _wrap_svn_io_dirent2_t_special_get, -1);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "filesize=", _wrap_svn_io_dirent2_t_filesize_set, -1);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "filesize", _wrap_svn_io_dirent2_t_filesize_get, -1);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "mtime=", _wrap_svn_io_dirent2_t_mtime_set, -1);
  rb_define_method(SwigClassSvn_io_dirent2_t.klass, "mtime", _wrap_svn_io_dirent2_t_mtime_get, -1);
  SwigClassSvn_io_dirent2_t.mark = 0;
  SwigClassSvn_io_dirent2_t.destroy = (void (*)(void *)) free_svn_io_dirent2_t;
  SwigClassSvn_io_dirent2_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_io_dirent2_create", _wrap_svn_io_dirent2_create, -1);
  rb_define_module_function(mCore, "svn_io_dirent2_dup", _wrap_svn_io_dirent2_dup, -1);
  
  SwigClassSvn_io_dirent_t.klass = rb_define_class_under(mCore, "Svn_io_dirent_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_io_dirent_t, (void *) &SwigClassSvn_io_dirent_t);
  rb_define_alloc_func(SwigClassSvn_io_dirent_t.klass, _wrap_svn_io_dirent_t_allocate);
  rb_define_method(SwigClassSvn_io_dirent_t.klass, "initialize", _wrap_new_svn_io_dirent_t, -1);
  rb_define_method(SwigClassSvn_io_dirent_t.klass, "kind=", _wrap_svn_io_dirent_t_kind_set, -1);
  rb_define_method(SwigClassSvn_io_dirent_t.klass, "kind", _wrap_svn_io_dirent_t_kind_get, -1);
  rb_define_method(SwigClassSvn_io_dirent_t.klass, "special=", _wrap_svn_io_dirent_t_special_set, -1);
  rb_define_method(SwigClassSvn_io_dirent_t.klass, "special", _wrap_svn_io_dirent_t_special_get, -1);
  SwigClassSvn_io_dirent_t.mark = 0;
  SwigClassSvn_io_dirent_t.destroy = (void (*)(void *)) free_svn_io_dirent_t;
  SwigClassSvn_io_dirent_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_io_open_uniquely_named", _wrap_svn_io_open_uniquely_named, -1);
  rb_define_module_function(mCore, "svn_io_open_unique_file3", _wrap_svn_io_open_unique_file3, -1);
  rb_define_module_function(mCore, "svn_io_open_unique_file2", _wrap_svn_io_open_unique_file2, -1);
  rb_define_module_function(mCore, "svn_io_open_unique_file", _wrap_svn_io_open_unique_file, -1);
  rb_define_module_function(mCore, "svn_io_copy_perms", _wrap_svn_io_copy_perms, -1);
  rb_define_module_function(mCore, "svn_io_sleep_for_timestamps", _wrap_svn_io_sleep_for_timestamps, -1);
  rb_define_module_function(mCore, "svn_io_filesizes_three_different_p", _wrap_svn_io_filesizes_three_different_p, -1);
  rb_define_module_function(mCore, "svn_io_file_checksum2", _wrap_svn_io_file_checksum2, -1);
  rb_define_module_function(mCore, "svn_io_file_checksum", _wrap_svn_io_file_checksum, -1);
  rb_define_module_function(mCore, "svn_io_files_contents_same_p", _wrap_svn_io_files_contents_same_p, -1);
  rb_define_module_function(mCore, "svn_io_files_contents_three_same_p", _wrap_svn_io_files_contents_three_same_p, -1);
  rb_define_module_function(mCore, "svn_io_file_create_bytes", _wrap_svn_io_file_create_bytes, -1);
  rb_define_module_function(mCore, "svn_io_file_create_empty", _wrap_svn_io_file_create_empty, -1);
  rb_define_module_function(mCore, "svn_io_lock_open_file", _wrap_svn_io_lock_open_file, -1);
  rb_define_module_function(mCore, "svn_io_unlock_open_file", _wrap_svn_io_unlock_open_file, -1);
  rb_define_module_function(mCore, "svn_stream_set_read2", _wrap_svn_stream_set_read2, -1);
  rb_define_module_function(mCore, "svn_stream_set_skip", _wrap_svn_stream_set_skip, -1);
  rb_define_module_function(mCore, "svn_stream_set_mark", _wrap_svn_stream_set_mark, -1);
  rb_define_module_function(mCore, "svn_stream_set_seek", _wrap_svn_stream_set_seek, -1);
  rb_define_module_function(mCore, "svn_stream_set_data_available", _wrap_svn_stream_set_data_available, -1);
  rb_define_module_function(mCore, "svn_stream_set_readline", _wrap_svn_stream_set_readline, -1);
  rb_define_module_function(mCore, "svn_stream_empty", _wrap_svn_stream_empty, -1);
  rb_define_module_function(mCore, "svn_stream_disown", _wrap_svn_stream_disown, -1);
  rb_define_module_function(mCore, "svn_stream_open_readonly", _wrap_svn_stream_open_readonly, -1);
  rb_define_module_function(mCore, "svn_stream_open_writable", _wrap_svn_stream_open_writable, -1);
  rb_define_module_function(mCore, "svn_stream_open_unique", _wrap_svn_stream_open_unique, -1);
  rb_define_module_function(mCore, "svn_stream_from_aprfile2", _wrap_svn_stream_from_aprfile2, -1);
  rb_define_module_function(mCore, "svn_stream_from_aprfile", _wrap_svn_stream_from_aprfile, -1);
  rb_define_module_function(mCore, "svn_stream_for_stdin2", _wrap_svn_stream_for_stdin2, -1);
  rb_define_module_function(mCore, "svn_stream_for_stdin", _wrap_svn_stream_for_stdin, -1);
  rb_define_module_function(mCore, "svn_stream_for_stderr", _wrap_svn_stream_for_stderr, -1);
  rb_define_module_function(mCore, "svn_stream_for_stdout", _wrap_svn_stream_for_stdout, -1);
  rb_define_module_function(mCore, "svn_stringbuf_from_stream", _wrap_svn_stringbuf_from_stream, -1);
  rb_define_module_function(mCore, "svn_stream_from_stringbuf", _wrap_svn_stream_from_stringbuf, -1);
  rb_define_module_function(mCore, "svn_stream_from_string", _wrap_svn_stream_from_string, -1);
  rb_define_module_function(mCore, "svn_stream_buffered", _wrap_svn_stream_buffered, -1);
  rb_define_module_function(mCore, "svn_stream_compressed", _wrap_svn_stream_compressed, -1);
  rb_define_module_function(mCore, "svn_stream_checksummed2", _wrap_svn_stream_checksummed2, -1);
  rb_define_module_function(mCore, "svn_stream_contents_checksum", _wrap_svn_stream_contents_checksum, -1);
  rb_define_module_function(mCore, "svn_stream_read_full", _wrap_svn_stream_read_full, -1);
  rb_define_module_function(mCore, "svn_stream_supports_partial_read", _wrap_svn_stream_supports_partial_read, -1);
  rb_define_module_function(mCore, "svn_stream_read2", _wrap_svn_stream_read2, -1);
  rb_define_module_function(mCore, "svn_stream_read", _wrap_svn_stream_read, -1);
  rb_define_module_function(mCore, "svn_stream_skip", _wrap_svn_stream_skip, -1);
  rb_define_module_function(mCore, "svn_stream_write", _wrap_svn_stream_write, -1);
  rb_define_module_function(mCore, "svn_stream_close", _wrap_svn_stream_close, -1);
  rb_define_module_function(mCore, "svn_stream_reset", _wrap_svn_stream_reset, -1);
  rb_define_module_function(mCore, "svn_stream_supports_mark", _wrap_svn_stream_supports_mark, -1);
  rb_define_module_function(mCore, "svn_stream_supports_reset", _wrap_svn_stream_supports_reset, -1);
  rb_define_module_function(mCore, "svn_stream_mark", _wrap_svn_stream_mark, -1);
  rb_define_module_function(mCore, "svn_stream_seek", _wrap_svn_stream_seek, -1);
  rb_define_module_function(mCore, "svn_stream_data_available", _wrap_svn_stream_data_available, -1);
  rb_define_module_function(mCore, "svn_stream_tee", _wrap_svn_stream_tee, -1);
  rb_define_module_function(mCore, "svn_stream_puts", _wrap_svn_stream_puts, -1);
  rb_define_module_function(mCore, "svn_stream_readline", _wrap_svn_stream_readline, -1);
  rb_define_module_function(mCore, "svn_stream_copy3", _wrap_svn_stream_copy3, -1);
  rb_define_module_function(mCore, "svn_stream_copy2", _wrap_svn_stream_copy2, -1);
  rb_define_module_function(mCore, "svn_stream_copy", _wrap_svn_stream_copy, -1);
  rb_define_module_function(mCore, "svn_stream_contents_same2", _wrap_svn_stream_contents_same2, -1);
  rb_define_module_function(mCore, "svn_stream_contents_same", _wrap_svn_stream_contents_same, -1);
  rb_define_module_function(mCore, "svn_string_from_stream2", _wrap_svn_string_from_stream2, -1);
  rb_define_module_function(mCore, "svn_string_from_stream", _wrap_svn_string_from_stream, -1);
  rb_define_module_function(mCore, "svn_stream_lazyopen_create", _wrap_svn_stream_lazyopen_create, -1);
  rb_define_module_function(mCore, "svn_stringbuf_from_file2", _wrap_svn_stringbuf_from_file2, -1);
  rb_define_module_function(mCore, "svn_stringbuf_from_file", _wrap_svn_stringbuf_from_file, -1);
  rb_define_module_function(mCore, "svn_stringbuf_from_aprfile", _wrap_svn_stringbuf_from_aprfile, -1);
  rb_define_module_function(mCore, "svn_io_remove_file2", _wrap_svn_io_remove_file2, -1);
  rb_define_module_function(mCore, "svn_io_remove_dir2", _wrap_svn_io_remove_dir2, -1);
  rb_define_module_function(mCore, "svn_io_get_dirents3", _wrap_svn_io_get_dirents3, -1);
  rb_define_module_function(mCore, "svn_io_stat_dirent2", _wrap_svn_io_stat_dirent2, -1);
  rb_define_module_function(mCore, "svn_io_stat_dirent", _wrap_svn_io_stat_dirent, -1);
  rb_define_module_function(mCore, "svn_io_dir_walk2", _wrap_svn_io_dir_walk2, -1);
  rb_define_module_function(mCore, "svn_io_start_cmd3", _wrap_svn_io_start_cmd3, -1);
  rb_define_module_function(mCore, "svn_io_start_cmd2", _wrap_svn_io_start_cmd2, -1);
  rb_define_module_function(mCore, "svn_io_run_diff2", _wrap_svn_io_run_diff2, -1);
  rb_define_module_function(mCore, "svn_io_run_diff3_3", _wrap_svn_io_run_diff3_3, -1);
  rb_define_module_function(mCore, "svn_io_parse_mimetypes_file", _wrap_svn_io_parse_mimetypes_file, -1);
  rb_define_module_function(mCore, "svn_io_detect_mimetype2", _wrap_svn_io_detect_mimetype2, -1);
  rb_define_module_function(mCore, "svn_io_detect_mimetype", _wrap_svn_io_detect_mimetype, -1);
  rb_define_module_function(mCore, "svn_io_is_binary_data", _wrap_svn_io_is_binary_data, -1);
  rb_define_module_function(mCore, "svn_io_file_putc", _wrap_svn_io_file_putc, -1);
  rb_define_module_function(mCore, "svn_io_file_size_get", _wrap_svn_io_file_size_get, -1);
  rb_define_module_function(mCore, "svn_io_file_get_offset", _wrap_svn_io_file_get_offset, -1);
  rb_define_module_function(mCore, "svn_io_file_read_full2", _wrap_svn_io_file_read_full2, -1);
  rb_define_module_function(mCore, "svn_io_file_aligned_seek", _wrap_svn_io_file_aligned_seek, -1);
  rb_define_module_function(mCore, "svn_io_file_flush", _wrap_svn_io_file_flush, -1);
  rb_define_module_function(mCore, "svn_io_write_atomic2", _wrap_svn_io_write_atomic2, -1);
  rb_define_module_function(mCore, "svn_io_write_atomic", _wrap_svn_io_write_atomic, -1);
  rb_define_module_function(mCore, "svn_io_write_unique", _wrap_svn_io_write_unique, -1);
  rb_define_module_function(mCore, "svn_io_file_trunc", _wrap_svn_io_file_trunc, -1);
  rb_define_module_function(mCore, "svn_io_file_rename2", _wrap_svn_io_file_rename2, -1);
  rb_define_module_function(mCore, "svn_io_dir_close", _wrap_svn_io_dir_close, -1);
  rb_define_module_function(mCore, "svn_io_file_name_get", _wrap_svn_io_file_name_get, -1);
  rb_define_module_function(mCore, "svn_io_file_readline", _wrap_svn_io_file_readline, -1);
  rb_define_module_function(mCore, "svn_read_invoke_fn", _wrap_svn_read_invoke_fn, -1);
  rb_define_module_function(mCore, "svn_stream_invoke_skip_fn", _wrap_svn_stream_invoke_skip_fn, -1);
  rb_define_module_function(mCore, "svn_write_invoke_fn", _wrap_svn_write_invoke_fn, -1);
  rb_define_module_function(mCore, "svn_close_invoke_fn", _wrap_svn_close_invoke_fn, -1);
  rb_define_module_function(mCore, "svn_stream_invoke_mark_fn", _wrap_svn_stream_invoke_mark_fn, -1);
  rb_define_module_function(mCore, "svn_stream_invoke_seek_fn", _wrap_svn_stream_invoke_seek_fn, -1);
  rb_define_module_function(mCore, "svn_stream_invoke_data_available_fn", _wrap_svn_stream_invoke_data_available_fn, -1);
  rb_define_module_function(mCore, "svn_stream_invoke_readline_fn", _wrap_svn_stream_invoke_readline_fn, -1);
  rb_define_module_function(mCore, "svn_stream_invoke_lazyopen_func", _wrap_svn_stream_invoke_lazyopen_func, -1);
  rb_define_module_function(mCore, "svn_io_invoke_walk_func", _wrap_svn_io_invoke_walk_func, -1);
  rb_define_const(mCore, "Svn_checksum_md5", SWIG_From_int((int)(svn_checksum_md5)));
  rb_define_const(mCore, "Svn_checksum_sha1", SWIG_From_int((int)(svn_checksum_sha1)));
  rb_define_const(mCore, "Svn_checksum_fnv1a_32", SWIG_From_int((int)(svn_checksum_fnv1a_32)));
  rb_define_const(mCore, "Svn_checksum_fnv1a_32x4", SWIG_From_int((int)(svn_checksum_fnv1a_32x4)));
  
  SwigClassSvn_checksum_t.klass = rb_define_class_under(mCore, "Svn_checksum_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_checksum_t, (void *) &SwigClassSvn_checksum_t);
  rb_define_alloc_func(SwigClassSvn_checksum_t.klass, _wrap_svn_checksum_t_allocate);
  rb_define_method(SwigClassSvn_checksum_t.klass, "initialize", _wrap_new_svn_checksum_t, -1);
  rb_define_method(SwigClassSvn_checksum_t.klass, "digest=", _wrap_svn_checksum_t_digest_set, -1);
  rb_define_method(SwigClassSvn_checksum_t.klass, "digest", _wrap_svn_checksum_t_digest_get, -1);
  rb_define_method(SwigClassSvn_checksum_t.klass, "kind=", _wrap_svn_checksum_t_kind_set, -1);
  rb_define_method(SwigClassSvn_checksum_t.klass, "kind", _wrap_svn_checksum_t_kind_get, -1);
  SwigClassSvn_checksum_t.mark = 0;
  SwigClassSvn_checksum_t.destroy = (void (*)(void *)) free_svn_checksum_t;
  SwigClassSvn_checksum_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_checksum_create", _wrap_svn_checksum_create, -1);
  rb_define_module_function(mCore, "svn_checksum_clear", _wrap_svn_checksum_clear, -1);
  rb_define_module_function(mCore, "svn_checksum_match", _wrap_svn_checksum_match, -1);
  rb_define_module_function(mCore, "svn_checksum_dup", _wrap_svn_checksum_dup, -1);
  rb_define_module_function(mCore, "svn_checksum_to_cstring_display", _wrap_svn_checksum_to_cstring_display, -1);
  rb_define_module_function(mCore, "svn_checksum_to_cstring", _wrap_svn_checksum_to_cstring, -1);
  rb_define_module_function(mCore, "svn_checksum_serialize", _wrap_svn_checksum_serialize, -1);
  rb_define_module_function(mCore, "svn_checksum_deserialize", _wrap_svn_checksum_deserialize, -1);
  rb_define_module_function(mCore, "svn_checksum_parse_hex", _wrap_svn_checksum_parse_hex, -1);
  rb_define_module_function(mCore, "svn_checksum", _wrap_svn_checksum, -1);
  rb_define_module_function(mCore, "svn_checksum_empty_checksum", _wrap_svn_checksum_empty_checksum, -1);
  rb_define_module_function(mCore, "svn_checksum_ctx_create", _wrap_svn_checksum_ctx_create, -1);
  rb_define_module_function(mCore, "svn_checksum_ctx_reset", _wrap_svn_checksum_ctx_reset, -1);
  rb_define_module_function(mCore, "svn_checksum_update", _wrap_svn_checksum_update, -1);
  rb_define_module_function(mCore, "svn_checksum_final", _wrap_svn_checksum_final, -1);
  rb_define_module_function(mCore, "svn_checksum_size", _wrap_svn_checksum_size, -1);
  rb_define_module_function(mCore, "svn_checksum_is_empty_checksum", _wrap_svn_checksum_is_empty_checksum, -1);
  rb_define_module_function(mCore, "svn_checksum_mismatch_err", _wrap_svn_checksum_mismatch_err, -1);
  
  SwigClassSvn_cache_config_t.klass = rb_define_class_under(mCore, "Svn_cache_config_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_cache_config_t, (void *) &SwigClassSvn_cache_config_t);
  rb_define_alloc_func(SwigClassSvn_cache_config_t.klass, _wrap_svn_cache_config_t_allocate);
  rb_define_method(SwigClassSvn_cache_config_t.klass, "initialize", _wrap_new_svn_cache_config_t, -1);
  rb_define_method(SwigClassSvn_cache_config_t.klass, "cache_size=", _wrap_svn_cache_config_t_cache_size_set, -1);
  rb_define_method(SwigClassSvn_cache_config_t.klass, "cache_size", _wrap_svn_cache_config_t_cache_size_get, -1);
  rb_define_method(SwigClassSvn_cache_config_t.klass, "file_handle_count=", _wrap_svn_cache_config_t_file_handle_count_set, -1);
  rb_define_method(SwigClassSvn_cache_config_t.klass, "file_handle_count", _wrap_svn_cache_config_t_file_handle_count_get, -1);
  rb_define_method(SwigClassSvn_cache_config_t.klass, "single_threaded=", _wrap_svn_cache_config_t_single_threaded_set, -1);
  rb_define_method(SwigClassSvn_cache_config_t.klass, "single_threaded", _wrap_svn_cache_config_t_single_threaded_get, -1);
  SwigClassSvn_cache_config_t.mark = 0;
  SwigClassSvn_cache_config_t.destroy = (void (*)(void *)) free_svn_cache_config_t;
  SwigClassSvn_cache_config_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_cache_config_get", _wrap_svn_cache_config_get, -1);
  rb_define_module_function(mCore, "svn_cache_config_set", _wrap_svn_cache_config_set, -1);
  rb_define_module_function(mCore, "svn_auth_set_gnome_keyring_unlock_prompt_func", _wrap_svn_auth_set_gnome_keyring_unlock_prompt_func, -1);
  rb_define_module_function(mCore, "svn_md5_empty_string_digest", _wrap_svn_md5_empty_string_digest, -1);
  rb_define_module_function(mCore, "svn_md5_digest_to_cstring_display", _wrap_svn_md5_digest_to_cstring_display, -1);
  rb_define_module_function(mCore, "svn_md5_digest_to_cstring", _wrap_svn_md5_digest_to_cstring, -1);
  rb_define_module_function(mCore, "svn_md5_digests_match", _wrap_svn_md5_digests_match, -1);
  
  svn_swig_rb_initialize();
  
  rb_define_const(mCore, "SVN_ALLOCATOR_MAX_FREE_UNLIMITED",
    UINT2NUM(APR_ALLOCATOR_MAX_FREE_UNLIMITED));
  
  
  SwigClassApr_pool_wrapper_t.klass = rb_define_class_under(mCore, "Apr_pool_wrapper_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_apr_pool_wrapper_t, (void *) &SwigClassApr_pool_wrapper_t);
  rb_define_alloc_func(SwigClassApr_pool_wrapper_t.klass, _wrap_apr_pool_wrapper_t_allocate);
  rb_define_method(SwigClassApr_pool_wrapper_t.klass, "initialize", _wrap_new_apr_pool_wrapper_t, -1);
  rb_define_singleton_method(SwigClassApr_pool_wrapper_t.klass, "destroy", _wrap_apr_pool_wrapper_t_destroy, -1);
  rb_define_singleton_method(SwigClassApr_pool_wrapper_t.klass, "set_default_max_free_size", _wrap_apr_pool_wrapper_t_set_default_max_free_size, -1);
  rb_define_method(SwigClassApr_pool_wrapper_t.klass, "set_max_free_size", _wrap_apr_pool_wrapper_t_set_max_free_size, -1);
  rb_define_method(SwigClassApr_pool_wrapper_t.klass, "_destroy", _wrap_apr_pool_wrapper_t__destroy, -1);
  SwigClassApr_pool_wrapper_t.mark = 0;
  SwigClassApr_pool_wrapper_t.destroy = (void (*)(void *)) free_apr_pool_wrapper_t;
  SwigClassApr_pool_wrapper_t.trackObjects = 0;
  
  SwigClassSvn_stream_t.klass = rb_define_class_under(mCore, "Svn_stream_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_stream_t, (void *) &SwigClassSvn_stream_t);
  rb_define_alloc_func(SwigClassSvn_stream_t.klass, _wrap_svn_stream_t_allocate);
  rb_define_method(SwigClassSvn_stream_t.klass, "initialize", _wrap_new_svn_stream_t, -1);
  SwigClassSvn_stream_t.mark = 0;
  SwigClassSvn_stream_t.destroy = (void (*)(void *)) free_svn_stream_t;
  SwigClassSvn_stream_t.trackObjects = 0;
  
  SwigClassSvn_auth_baton_t.klass = rb_define_class_under(mCore, "Svn_auth_baton_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_auth_baton_t, (void *) &SwigClassSvn_auth_baton_t);
  rb_define_alloc_func(SwigClassSvn_auth_baton_t.klass, _wrap_svn_auth_baton_t_allocate);
  rb_define_method(SwigClassSvn_auth_baton_t.klass, "initialize", _wrap_new_svn_auth_baton_t, -1);
  SwigClassSvn_auth_baton_t.mark = 0;
  SwigClassSvn_auth_baton_t.destroy = (void (*)(void *)) free_svn_auth_baton_t;
  SwigClassSvn_auth_baton_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_diff_version", _wrap_svn_diff_version, -1);
  rb_define_const(mCore, "Svn_diff_datasource_original", SWIG_From_int((int)(svn_diff_datasource_original)));
  rb_define_const(mCore, "Svn_diff_datasource_modified", SWIG_From_int((int)(svn_diff_datasource_modified)));
  rb_define_const(mCore, "Svn_diff_datasource_latest", SWIG_From_int((int)(svn_diff_datasource_latest)));
  rb_define_const(mCore, "Svn_diff_datasource_ancestor", SWIG_From_int((int)(svn_diff_datasource_ancestor)));
  
  SwigClassSvn_diff_fns2_t.klass = rb_define_class_under(mCore, "Svn_diff_fns2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_diff_fns2_t, (void *) &SwigClassSvn_diff_fns2_t);
  rb_define_alloc_func(SwigClassSvn_diff_fns2_t.klass, _wrap_svn_diff_fns2_t_allocate);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "initialize", _wrap_new_svn_diff_fns2_t, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "datasources_open=", _wrap_svn_diff_fns2_t_datasources_open_set, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "datasources_open", _wrap_svn_diff_fns2_t_datasources_open_get, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "datasource_close=", _wrap_svn_diff_fns2_t_datasource_close_set, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "datasource_close", _wrap_svn_diff_fns2_t_datasource_close_get, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "datasource_get_next_token=", _wrap_svn_diff_fns2_t_datasource_get_next_token_set, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "datasource_get_next_token", _wrap_svn_diff_fns2_t_datasource_get_next_token_get, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "token_compare=", _wrap_svn_diff_fns2_t_token_compare_set, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "token_compare", _wrap_svn_diff_fns2_t_token_compare_get, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "token_discard=", _wrap_svn_diff_fns2_t_token_discard_set, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "token_discard", _wrap_svn_diff_fns2_t_token_discard_get, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "token_discard_all=", _wrap_svn_diff_fns2_t_token_discard_all_set, -1);
  rb_define_method(SwigClassSvn_diff_fns2_t.klass, "token_discard_all", _wrap_svn_diff_fns2_t_token_discard_all_get, -1);
  SwigClassSvn_diff_fns2_t.mark = 0;
  SwigClassSvn_diff_fns2_t.destroy = (void (*)(void *)) free_svn_diff_fns2_t;
  SwigClassSvn_diff_fns2_t.trackObjects = 0;
  
  SwigClassSvn_diff_fns_t.klass = rb_define_class_under(mCore, "Svn_diff_fns_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_diff_fns_t, (void *) &SwigClassSvn_diff_fns_t);
  rb_define_alloc_func(SwigClassSvn_diff_fns_t.klass, _wrap_svn_diff_fns_t_allocate);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "initialize", _wrap_new_svn_diff_fns_t, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "datasource_open=", _wrap_svn_diff_fns_t_datasource_open_set, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "datasource_open", _wrap_svn_diff_fns_t_datasource_open_get, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "datasource_close=", _wrap_svn_diff_fns_t_datasource_close_set, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "datasource_close", _wrap_svn_diff_fns_t_datasource_close_get, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "datasource_get_next_token=", _wrap_svn_diff_fns_t_datasource_get_next_token_set, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "datasource_get_next_token", _wrap_svn_diff_fns_t_datasource_get_next_token_get, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "token_compare=", _wrap_svn_diff_fns_t_token_compare_set, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "token_compare", _wrap_svn_diff_fns_t_token_compare_get, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "token_discard=", _wrap_svn_diff_fns_t_token_discard_set, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "token_discard", _wrap_svn_diff_fns_t_token_discard_get, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "token_discard_all=", _wrap_svn_diff_fns_t_token_discard_all_set, -1);
  rb_define_method(SwigClassSvn_diff_fns_t.klass, "token_discard_all", _wrap_svn_diff_fns_t_token_discard_all_get, -1);
  SwigClassSvn_diff_fns_t.mark = 0;
  SwigClassSvn_diff_fns_t.destroy = (void (*)(void *)) free_svn_diff_fns_t;
  SwigClassSvn_diff_fns_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_diff_diff_2", _wrap_svn_diff_diff_2, -1);
  rb_define_module_function(mCore, "svn_diff_diff", _wrap_svn_diff_diff, -1);
  rb_define_module_function(mCore, "svn_diff_diff3_2", _wrap_svn_diff_diff3_2, -1);
  rb_define_module_function(mCore, "svn_diff_diff3", _wrap_svn_diff_diff3, -1);
  rb_define_module_function(mCore, "svn_diff_diff4_2", _wrap_svn_diff_diff4_2, -1);
  rb_define_module_function(mCore, "svn_diff_diff4", _wrap_svn_diff_diff4, -1);
  rb_define_module_function(mCore, "svn_diff_contains_conflicts", _wrap_svn_diff_contains_conflicts, -1);
  rb_define_module_function(mCore, "svn_diff_contains_diffs", _wrap_svn_diff_contains_diffs, -1);
  
  SwigClassSvn_diff_output_fns_t.klass = rb_define_class_under(mCore, "Svn_diff_output_fns_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_diff_output_fns_t, (void *) &SwigClassSvn_diff_output_fns_t);
  rb_define_alloc_func(SwigClassSvn_diff_output_fns_t.klass, _wrap_svn_diff_output_fns_t_allocate);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "initialize", _wrap_new_svn_diff_output_fns_t, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_common=", _wrap_svn_diff_output_fns_t_output_common_set, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_common", _wrap_svn_diff_output_fns_t_output_common_get, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_diff_modified=", _wrap_svn_diff_output_fns_t_output_diff_modified_set, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_diff_modified", _wrap_svn_diff_output_fns_t_output_diff_modified_get, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_diff_latest=", _wrap_svn_diff_output_fns_t_output_diff_latest_set, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_diff_latest", _wrap_svn_diff_output_fns_t_output_diff_latest_get, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_diff_common=", _wrap_svn_diff_output_fns_t_output_diff_common_set, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_diff_common", _wrap_svn_diff_output_fns_t_output_diff_common_get, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_conflict=", _wrap_svn_diff_output_fns_t_output_conflict_set, -1);
  rb_define_method(SwigClassSvn_diff_output_fns_t.klass, "output_conflict", _wrap_svn_diff_output_fns_t_output_conflict_get, -1);
  SwigClassSvn_diff_output_fns_t.mark = 0;
  SwigClassSvn_diff_output_fns_t.destroy = (void (*)(void *)) free_svn_diff_output_fns_t;
  SwigClassSvn_diff_output_fns_t.trackObjects = 0;
  rb_define_const(mCore, "Svn_diff_conflict_display_modified_latest", SWIG_From_int((int)(svn_diff_conflict_display_modified_latest)));
  rb_define_const(mCore, "Svn_diff_conflict_display_resolved_modified_latest", SWIG_From_int((int)(svn_diff_conflict_display_resolved_modified_latest)));
  rb_define_const(mCore, "Svn_diff_conflict_display_modified_original_latest", SWIG_From_int((int)(svn_diff_conflict_display_modified_original_latest)));
  rb_define_const(mCore, "Svn_diff_conflict_display_modified", SWIG_From_int((int)(svn_diff_conflict_display_modified)));
  rb_define_const(mCore, "Svn_diff_conflict_display_latest", SWIG_From_int((int)(svn_diff_conflict_display_latest)));
  rb_define_const(mCore, "Svn_diff_conflict_display_only_conflicts", SWIG_From_int((int)(svn_diff_conflict_display_only_conflicts)));
  rb_define_module_function(mCore, "svn_diff_output2", _wrap_svn_diff_output2, -1);
  rb_define_module_function(mCore, "svn_diff_output", _wrap_svn_diff_output, -1);
  rb_define_const(mCore, "Svn_diff_file_ignore_space_none", SWIG_From_int((int)(svn_diff_file_ignore_space_none)));
  rb_define_const(mCore, "Svn_diff_file_ignore_space_change", SWIG_From_int((int)(svn_diff_file_ignore_space_change)));
  rb_define_const(mCore, "Svn_diff_file_ignore_space_all", SWIG_From_int((int)(svn_diff_file_ignore_space_all)));
  
  SwigClassSvn_diff_file_options_t.klass = rb_define_class_under(mCore, "Svn_diff_file_options_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_diff_file_options_t, (void *) &SwigClassSvn_diff_file_options_t);
  rb_define_alloc_func(SwigClassSvn_diff_file_options_t.klass, _wrap_svn_diff_file_options_t_allocate);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "initialize", _wrap_new_svn_diff_file_options_t, -1);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "ignore_space=", _wrap_svn_diff_file_options_t_ignore_space_set, -1);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "ignore_space", _wrap_svn_diff_file_options_t_ignore_space_get, -1);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "ignore_eol_style=", _wrap_svn_diff_file_options_t_ignore_eol_style_set, -1);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "ignore_eol_style", _wrap_svn_diff_file_options_t_ignore_eol_style_get, -1);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "show_c_function=", _wrap_svn_diff_file_options_t_show_c_function_set, -1);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "show_c_function", _wrap_svn_diff_file_options_t_show_c_function_get, -1);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "context_size=", _wrap_svn_diff_file_options_t_context_size_set, -1);
  rb_define_method(SwigClassSvn_diff_file_options_t.klass, "context_size", _wrap_svn_diff_file_options_t_context_size_get, -1);
  SwigClassSvn_diff_file_options_t.mark = 0;
  SwigClassSvn_diff_file_options_t.destroy = (void (*)(void *)) free_svn_diff_file_options_t;
  SwigClassSvn_diff_file_options_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_diff_file_options_parse", _wrap_svn_diff_file_options_parse, -1);
  rb_define_module_function(mCore, "svn_diff_file_diff_2", _wrap_svn_diff_file_diff_2, -1);
  rb_define_module_function(mCore, "svn_diff_file_diff", _wrap_svn_diff_file_diff, -1);
  rb_define_module_function(mCore, "svn_diff_file_diff3_2", _wrap_svn_diff_file_diff3_2, -1);
  rb_define_module_function(mCore, "svn_diff_file_diff3", _wrap_svn_diff_file_diff3, -1);
  rb_define_module_function(mCore, "svn_diff_file_diff4_2", _wrap_svn_diff_file_diff4_2, -1);
  rb_define_module_function(mCore, "svn_diff_file_diff4", _wrap_svn_diff_file_diff4, -1);
  rb_define_module_function(mCore, "svn_diff_file_output_unified4", _wrap_svn_diff_file_output_unified4, -1);
  rb_define_module_function(mCore, "svn_diff_file_output_unified3", _wrap_svn_diff_file_output_unified3, -1);
  rb_define_module_function(mCore, "svn_diff_file_output_unified2", _wrap_svn_diff_file_output_unified2, -1);
  rb_define_module_function(mCore, "svn_diff_file_output_unified", _wrap_svn_diff_file_output_unified, -1);
  rb_define_module_function(mCore, "svn_diff_file_output_merge3", _wrap_svn_diff_file_output_merge3, -1);
  rb_define_module_function(mCore, "svn_diff_file_output_merge2", _wrap_svn_diff_file_output_merge2, -1);
  rb_define_module_function(mCore, "svn_diff_file_output_merge", _wrap_svn_diff_file_output_merge, -1);
  rb_define_module_function(mCore, "svn_diff_output_binary", _wrap_svn_diff_output_binary, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_diff", _wrap_svn_diff_mem_string_diff, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_diff3", _wrap_svn_diff_mem_string_diff3, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_diff4", _wrap_svn_diff_mem_string_diff4, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_output_unified3", _wrap_svn_diff_mem_string_output_unified3, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_output_unified2", _wrap_svn_diff_mem_string_output_unified2, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_output_unified", _wrap_svn_diff_mem_string_output_unified, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_output_merge3", _wrap_svn_diff_mem_string_output_merge3, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_output_merge2", _wrap_svn_diff_mem_string_output_merge2, -1);
  rb_define_module_function(mCore, "svn_diff_mem_string_output_merge", _wrap_svn_diff_mem_string_output_merge, -1);
  rb_define_const(mCore, "Svn_diff_op_unchanged", SWIG_From_int((int)(svn_diff_op_unchanged)));
  rb_define_const(mCore, "Svn_diff_op_added", SWIG_From_int((int)(svn_diff_op_added)));
  rb_define_const(mCore, "Svn_diff_op_deleted", SWIG_From_int((int)(svn_diff_op_deleted)));
  rb_define_const(mCore, "Svn_diff_op_copied", SWIG_From_int((int)(svn_diff_op_copied)));
  rb_define_const(mCore, "Svn_diff_op_moved", SWIG_From_int((int)(svn_diff_op_moved)));
  rb_define_const(mCore, "Svn_diff_op_modified", SWIG_From_int((int)(svn_diff_op_modified)));
  rb_define_module_function(mCore, "svn_diff_hunk_readline_diff_text", _wrap_svn_diff_hunk_readline_diff_text, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_readline_original_text", _wrap_svn_diff_hunk_readline_original_text, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_readline_modified_text", _wrap_svn_diff_hunk_readline_modified_text, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_reset_diff_text", _wrap_svn_diff_hunk_reset_diff_text, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_reset_original_text", _wrap_svn_diff_hunk_reset_original_text, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_reset_modified_text", _wrap_svn_diff_hunk_reset_modified_text, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_get_original_start", _wrap_svn_diff_hunk_get_original_start, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_get_original_length", _wrap_svn_diff_hunk_get_original_length, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_get_modified_start", _wrap_svn_diff_hunk_get_modified_start, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_get_modified_length", _wrap_svn_diff_hunk_get_modified_length, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_get_leading_context", _wrap_svn_diff_hunk_get_leading_context, -1);
  rb_define_module_function(mCore, "svn_diff_hunk_get_trailing_context", _wrap_svn_diff_hunk_get_trailing_context, -1);
  
  SwigClassSvn_prop_patch_t.klass = rb_define_class_under(mCore, "Svn_prop_patch_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_prop_patch_t, (void *) &SwigClassSvn_prop_patch_t);
  rb_define_alloc_func(SwigClassSvn_prop_patch_t.klass, _wrap_svn_prop_patch_t_allocate);
  rb_define_method(SwigClassSvn_prop_patch_t.klass, "initialize", _wrap_new_svn_prop_patch_t, -1);
  rb_define_method(SwigClassSvn_prop_patch_t.klass, "name=", _wrap_svn_prop_patch_t_name_set, -1);
  rb_define_method(SwigClassSvn_prop_patch_t.klass, "name", _wrap_svn_prop_patch_t_name_get, -1);
  rb_define_method(SwigClassSvn_prop_patch_t.klass, "operation=", _wrap_svn_prop_patch_t_operation_set, -1);
  rb_define_method(SwigClassSvn_prop_patch_t.klass, "operation", _wrap_svn_prop_patch_t_operation_get, -1);
  rb_define_method(SwigClassSvn_prop_patch_t.klass, "hunks=", _wrap_svn_prop_patch_t_hunks_set, -1);
  rb_define_method(SwigClassSvn_prop_patch_t.klass, "hunks", _wrap_svn_prop_patch_t_hunks_get, -1);
  SwigClassSvn_prop_patch_t.mark = 0;
  SwigClassSvn_prop_patch_t.destroy = (void (*)(void *)) free_svn_prop_patch_t;
  SwigClassSvn_prop_patch_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_diff_get_binary_diff_original_stream", _wrap_svn_diff_get_binary_diff_original_stream, -1);
  rb_define_module_function(mCore, "svn_diff_get_binary_diff_result_stream", _wrap_svn_diff_get_binary_diff_result_stream, -1);
  
  SwigClassSvn_patch_t.klass = rb_define_class_under(mCore, "Svn_patch_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_patch_t, (void *) &SwigClassSvn_patch_t);
  rb_define_alloc_func(SwigClassSvn_patch_t.klass, _wrap_svn_patch_t_allocate);
  rb_define_method(SwigClassSvn_patch_t.klass, "initialize", _wrap_new_svn_patch_t, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "old_filename=", _wrap_svn_patch_t_old_filename_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "old_filename", _wrap_svn_patch_t_old_filename_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "new_filename=", _wrap_svn_patch_t_new_filename_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "new_filename", _wrap_svn_patch_t_new_filename_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "hunks=", _wrap_svn_patch_t_hunks_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "hunks", _wrap_svn_patch_t_hunks_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "prop_patches=", _wrap_svn_patch_t_prop_patches_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "prop_patches", _wrap_svn_patch_t_prop_patches_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "operation=", _wrap_svn_patch_t_operation_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "operation", _wrap_svn_patch_t_operation_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "reverse=", _wrap_svn_patch_t_reverse_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "reverse", _wrap_svn_patch_t_reverse_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "mergeinfo=", _wrap_svn_patch_t_mergeinfo_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "mergeinfo", _wrap_svn_patch_t_mergeinfo_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "reverse_mergeinfo=", _wrap_svn_patch_t_reverse_mergeinfo_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "reverse_mergeinfo", _wrap_svn_patch_t_reverse_mergeinfo_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "binary_patch=", _wrap_svn_patch_t_binary_patch_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "binary_patch", _wrap_svn_patch_t_binary_patch_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "old_executable_bit=", _wrap_svn_patch_t_old_executable_bit_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "old_executable_bit", _wrap_svn_patch_t_old_executable_bit_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "new_executable_bit=", _wrap_svn_patch_t_new_executable_bit_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "new_executable_bit", _wrap_svn_patch_t_new_executable_bit_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "old_symlink_bit=", _wrap_svn_patch_t_old_symlink_bit_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "old_symlink_bit", _wrap_svn_patch_t_old_symlink_bit_get, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "new_symlink_bit=", _wrap_svn_patch_t_new_symlink_bit_set, -1);
  rb_define_method(SwigClassSvn_patch_t.klass, "new_symlink_bit", _wrap_svn_patch_t_new_symlink_bit_get, -1);
  SwigClassSvn_patch_t.mark = 0;
  SwigClassSvn_patch_t.destroy = (void (*)(void *)) free_svn_patch_t;
  SwigClassSvn_patch_t.trackObjects = 0;
  rb_define_module_function(mCore, "svn_diff_open_patch_file", _wrap_svn_diff_open_patch_file, -1);
  rb_define_module_function(mCore, "svn_diff_parse_next_patch", _wrap_svn_diff_parse_next_patch, -1);
  rb_define_module_function(mCore, "svn_diff_close_patch_file", _wrap_svn_diff_close_patch_file, -1);
  rb_define_module_function(mCore, "svn_diff_fns2_invoke_datasources_open", _wrap_svn_diff_fns2_invoke_datasources_open, -1);
  rb_define_module_function(mCore, "svn_diff_fns2_invoke_datasource_close", _wrap_svn_diff_fns2_invoke_datasource_close, -1);
  rb_define_module_function(mCore, "svn_diff_fns2_invoke_datasource_get_next_token", _wrap_svn_diff_fns2_invoke_datasource_get_next_token, -1);
  rb_define_module_function(mCore, "svn_diff_fns2_invoke_token_compare", _wrap_svn_diff_fns2_invoke_token_compare, -1);
  rb_define_module_function(mCore, "svn_diff_fns2_invoke_token_discard", _wrap_svn_diff_fns2_invoke_token_discard, -1);
  rb_define_module_function(mCore, "svn_diff_fns2_invoke_token_discard_all", _wrap_svn_diff_fns2_invoke_token_discard_all, -1);
  rb_define_module_function(mCore, "svn_diff_fns_invoke_datasource_open", _wrap_svn_diff_fns_invoke_datasource_open, -1);
  rb_define_module_function(mCore, "svn_diff_fns_invoke_datasource_close", _wrap_svn_diff_fns_invoke_datasource_close, -1);
  rb_define_module_function(mCore, "svn_diff_fns_invoke_datasource_get_next_token", _wrap_svn_diff_fns_invoke_datasource_get_next_token, -1);
  rb_define_module_function(mCore, "svn_diff_fns_invoke_token_compare", _wrap_svn_diff_fns_invoke_token_compare, -1);
  rb_define_module_function(mCore, "svn_diff_fns_invoke_token_discard", _wrap_svn_diff_fns_invoke_token_discard, -1);
  rb_define_module_function(mCore, "svn_diff_fns_invoke_token_discard_all", _wrap_svn_diff_fns_invoke_token_discard_all, -1);
  rb_define_module_function(mCore, "svn_diff_output_fns_invoke_output_common", _wrap_svn_diff_output_fns_invoke_output_common, -1);
  rb_define_module_function(mCore, "svn_diff_output_fns_invoke_output_diff_modified", _wrap_svn_diff_output_fns_invoke_output_diff_modified, -1);
  rb_define_module_function(mCore, "svn_diff_output_fns_invoke_output_diff_latest", _wrap_svn_diff_output_fns_invoke_output_diff_latest, -1);
  rb_define_module_function(mCore, "svn_diff_output_fns_invoke_output_diff_common", _wrap_svn_diff_output_fns_invoke_output_diff_common, -1);
  rb_define_module_function(mCore, "svn_diff_output_fns_invoke_output_conflict", _wrap_svn_diff_output_fns_invoke_output_conflict, -1);
  rb_define_module_function(mCore, "svn_default_charset", _wrap_svn_default_charset, -1);
  rb_define_module_function(mCore, "svn_locale_charset", _wrap_svn_locale_charset, -1);
  rb_define_module_function(mCore, "svn_swig_rb_auth_get_simple_prompt_provider", _wrap_svn_swig_rb_auth_get_simple_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_swig_rb_auth_get_ssl_client_cert_prompt_provider", _wrap_svn_swig_rb_auth_get_ssl_client_cert_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_swig_rb_auth_get_ssl_client_cert_pw_prompt_provider", _wrap_svn_swig_rb_auth_get_ssl_client_cert_pw_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_swig_rb_auth_get_ssl_server_trust_prompt_provider", _wrap_svn_swig_rb_auth_get_ssl_server_trust_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_swig_rb_auth_get_username_prompt_provider", _wrap_svn_swig_rb_auth_get_username_prompt_provider, -1);
  rb_define_module_function(mCore, "svn_swig_mergeinfo_merge", _wrap_svn_swig_mergeinfo_merge, -1);
  rb_define_module_function(mCore, "svn_swig_mergeinfo_sort", _wrap_svn_swig_mergeinfo_sort, -1);
  rb_define_module_function(mCore, "svn_swig_rangelist_merge", _wrap_svn_swig_rangelist_merge, -1);
  rb_define_module_function(mCore, "svn_swig_rangelist_reverse", _wrap_svn_swig_rangelist_reverse, -1);
}

